! Grey Gordon and Qiu Shi, 2014
! Note: for the RBC model, k and z are exactly as in the paper. For the Arellano model, think of k as "b" and z as "y"
module mod_data
    use mod_matlab
    use mod_discreteAR1
    use mod_util
    implicit none

    type(timer) :: time_max, time_quad, time_tot

    real(8) :: maxTime 
    logical :: useStdOut
    logical :: dodiscreteshocks, dogridsearch, docontmono, docontmono_but_discrete,  domaxexact, doexploit2state, dotranspose2state
    integer :: shocktype
    integer :: maxmethod 
    real(8) :: lambda = -1d0, lambda_alt = -1d0  

    real(8) :: override_cover = -1d0 ! this value is no longer used, but keep it for legacy code
    real(8), protected :: crra
    real(8) :: override_beta = -1d0 ! like override_cover above
    real(8) :: beta
    real(8), parameter :: NOT_INITIALIZED_GUESS = -sqrt(sqrt(huge(0d0)))

    integer, protected :: nK, nZ

    real(8) :: nonmonofound = 0d0

    integer, parameter :: nquad = 1001

    ! RBC specific parameters
    real(8), parameter :: deprec = .025d0, alpha= .36d0
    real(8), protected :: kSS, cSS, vSS
    integer, protected :: iKeq0 

    ! DEF specific parameters 
    integer :: defaultcost_type = 1 ! For replicating Arellano, set defaultcost_type = 0. For smoother function, use defaultcost_type = 1
    real(8), parameter :: rstar = .017d0
    real(8) :: thresh = .969d0 ! only used if defaultcost_type==0
    real(8) :: theta = .282d0
    real(8) :: def_std = .025d0, def_ar1 = .945d0

    ! Large arrays
    real(8), allocatable :: gK(:),gZ(:),Fzpcz(:,:),c(:,:),kp(:,:),V(:,:),work(:,:),work2(:,:),EEE(:,:),dsn(:,:,:),q(:,:),d(:,:),CV(:,:),Vnd(:,:),Vd(:,:),cnd(:,:),QUADz(:,:),QUADw(:,:) !,QUADqz(:,:,:)

    integer, allocatable :: kpi(:,:)

contains

    subroutine saveplots()
        implicit none
        integer, allocatable :: plot_zi(:)
        real(8), allocatable :: x(:,:)
        character(50) :: dir
        integer :: iZ
        
        allocate(x(nK,nZ))
        do iZ = 1,nZ
            x(:,iZ) = q(:,iZ)*gK - gZ(iZ)
        end do

        allocate(plot_zi(3))
        if (nZ==3) then
            plot_zi = [1,2,3]
        else
            call bsearch(plot_zi(1),.93d0,gZ)
            call bsearch(plot_zi(2),1d0,gZ)
            call bsearch(plot_zi(3),1.02d0,gZ)
        end if
        ! plot_zi = [1,nZ/2,nZ]

        if (dogridsearch) then
            dir = 'pics_grid/'
        elseif (docontmono) then
            dir = 'pics_cont/'
        end if

        call gnuplot(gK,c(:,plot_zi),title='c',saveas=trim(dir)//'c.eps')
        call gnuplot(gK,kp(:,plot_zi),title='kp',saveas=trim(dir)//'kp.eps')
        call gnuplot(gK,V(:,plot_zi),title='V',saveas=trim(dir)//'V.eps')
        call gnuplot(gK,q(:,plot_zi),title='q',saveas=trim(dir)//'q.eps')
        call gnuplot(gK,CV(:,plot_zi),title='CV',saveas=trim(dir)//'CV.eps')
        call gnuplot(gK,x(:,plot_zi),title='x',saveas=trim(dir)//'x.eps')
        call gnuplot(gK,EEE(:,plot_zi),title='EEE',saveas=trim(dir)//'EEE.eps')
        call gnuplot(gK,dsn(:,plot_zi,1),title='Distribution',saveas=trim(dir)//'dsn.eps')
    end subroutine 
    
    subroutine setRBCparam()
        implicit none
        crra = 2d0
        beta = .99d0
        kSS = ((1d0/beta - 1d0 + deprec)/alpha)**(1d0/(alpha-1d0))
        cSS = kSS**alpha - deprec*kSS
        vSS = cSS**(1d0-crra)/(1d0-crra)/(1d0-beta)
    end subroutine

    subroutine setDEFparam()
        implicit none
        crra = 2d0
        if (override_beta==-1d0) then
            beta = .95282d0
        else
            beta = override_beta
        end if
        iKeq0 = 7*nK/10
    end subroutine

    subroutine allocatedata(nKin,nZin)
        implicit none
        integer, intent(in) :: nKin, nZin

        nK = nKin
        nZ = nZin

        if (allocated(gK)) deallocate(gK); allocate(gK(nK))
        if (allocated(gZ)) deallocate(gZ); allocate(gZ(nZ))
        if (allocated(Fzpcz)) deallocate(Fzpcz); allocate(Fzpcz(nZ,nZ))
        if (allocated(c)) deallocate(c); allocate(c(nK,nZ))
        if (allocated(V)) deallocate(V); allocate(V(nK,nZ))
        if (allocated(kp)) deallocate(kp); allocate(kp(nK,nZ))
        if (allocated(kpi)) deallocate(kpi); allocate(kpi(nK,nZ))
        if (allocated(work)) deallocate(work); allocate(work(nK,nZ))
        if (allocated(work2)) deallocate(work2); allocate(work2(nK,nZ))
        if (allocated(EEE)) deallocate(EEE); allocate(EEE(nK,nZ))
        if (allocated(dsn)) deallocate(dsn); allocate(dsn(nK,nZ,2))
        if (allocated(q)) deallocate(q); allocate(q(nK,nZ))
        if (allocated(d)) deallocate(d); allocate(d(nK,nZ))
        if (allocated(CV)) deallocate(CV); allocate(CV(nK,nZ))
        if (allocated(Vnd)) deallocate(Vnd); allocate(Vnd(nK,nZ))
        if (allocated(Vd)) deallocate(Vd); allocate(Vd(nK,nZ))
        if (allocated(cnd)) deallocate(cnd); allocate(cnd(nK,nZ))
        if (.not. dodiscreteshocks) then
            if (allocated(QUADz)) deallocate(QUADz); allocate(QUADz(nquad,nZ))
            if (allocated(QUADw)) deallocate(QUADw); allocate(QUADw(nquad,nZ))
        end if

    end subroutine

!     subroutine deallocatedata
!         implicit none
! 
!         deallocate(gK,gZ,Fzpcz,c,V,kp,kpi,work,work2,EEE,dsn,q,d,CV,Vnd,Vd,cnd)
!         if (.not. dodiscreteshocks) deallocate(QUADz,QUADw) ! QUADqz
! 
!     end subroutine

    elemental function ydef(y) result(ydefout)
        implicit none
        real(8), intent(in) :: y
        real(8) :: ydefout
        
        select case (defaultcost_type) 
            case (0)
                ! Same as Arellano
                ydefout = min(y,thresh) ! cutoff -- in the paper, she says relative to mean y, but she does not factor into account that the process is not mean 1.
!                 if (y>thresh) then
!                     ydefout = thresh
!                 else
!                     ydefout = y
!                 end if
                
            case (1)
                ! Smooth default cost from CE, gives a flatter price function
                ydefout = y - (-.18819d0 * y + .24558d0 * y **2) 
            case (2)
                ! Clausen and Strub default cost
                ydefout = y - .001d0
            case default
                ydefout = huge(0d0)
        end select

    end function 

    ! Computes x := u'(x)
    elemental subroutine dPU(x)
        implicit none
        real(8), intent(inout) :: x
        ! CRRA of 2 has u' = c**(-gamma) so u' = 1d0/c**2
        if (crra==2d0) then
            x = x**(-2d0)
        else
            x = x**(-crra)
        end if
    end subroutine 

    ! Computes x := u'^-1(x)
    elemental subroutine dPUinv(x)
        implicit none
        real(8), intent(inout) :: x
        ! u' = 1d0/c**2 <=> c = u'^{-1/2} 
        if (crra==2d0) then
            x = sqrt(1d0/x)
        else
            x = x**(-1d0/crra)
        end if
    end subroutine

    ! Computes r = z*alpha*k**(alpha-1)
    elemental function mpk(z,k) result(r)
        implicit none
        real(8), intent(in) :: z,k
        real(8) :: r

        r = z*alpha*k**(alpha-1d0)

    end function


end module 

module mod_routines
    implicit none
contains

    subroutine vfiRBC(iter_out,cnt,time_out,overtime,opt)
        use mod_binarymono
        use mod_csv
        use mod_data
        implicit none
        real(8), intent(out) :: cnt, time_out ! average count per iteration and total time 
        integer, intent(in) :: opt(2)
        logical, intent(out) :: overtime
        integer, intent(out) :: iter_out
        ! local
        real(8) :: sup, time
        integer :: onecnt,iZ,iK,iter
        real(8), allocatable :: cash(:,:)
        real(8), allocatable :: cntmap(:,:)
        integer, allocatable :: searchmap(:,:,:)

!         real(8), allocatable :: V_trans(:,:)
!         integer, allocatable :: kpi_trans(:,:)
!         allocate(V_trans(nZ,nK))
!         allocate(kpi_trans(nZ,nK))

        allocate(cash(nK,nZ))
        allocate(cntmap(nK,nZ))
        allocate(searchmap(nK,nK,nZ))

        cntmap = 0d0
        
        overtime = .false.

        ! Precomputation
        do iZ = 1,nZ
            cash(:,iZ) = gZ(iZ)*gK**alpha + gK*(1d0-deprec)
        end do

        ! Make initial guess
        V = vSS
        cnt = 0d0

        do iter = 1,10000

            work = V ! save old V for supnorm check

            ! Compute the continuation utility
            call time_quad%start()
            call mygemm(CV,V,Fzpcz,nK,nZ,nZ,nZ,alpha=beta)
            call time_quad%stop()

            onecnt = 0

            if (doexploit2state) then

                ! Exploit monotonicity in both k and z simultaneously
                if (iter==100) then
                    searchmap = 0
                end if

                call time_max%start()
                call sub_bmono_2s_disc(V,kpi,fmulti,nK,monotone=opt(1),concave=opt(2),dotranspose=dotranspose2state)
                call time_max%stop()

                if (iter==100) then
                    do iK = 1,nK
                        do iZ = 1,nZ
                            searchmap(kpi(iK,iZ),iK,iZ) = 2
                        end do
                    end do
                end if

                ! call sub_bmono_2s_disc(V_trans,kpi_trans,fmulti_trans,nK)
                ! V = transpose(V_trans)
                ! kpi = transpose(kpi_trans)
                ! print*,'doing transpose version'

                ! Check if have exceeded time limit
                call toc(time)
                if (time>maxTime) then
                    overtime = .true.
                    return ! do not modify cnt or time_out
                end if

            else

                ! Exploit monotonicity in just k while looping over z

                do iZ = 1,nZ

                    ! In the first version of the paper submitted to QE, used general routine with f(iK,iKp) etc. 
                    ! Presumably this is due to the compiler being better able to optimize sub_bmono_cont than sub_bmono_disc. If there is no concavity (meaning sub_bmono_cont may be used), then sub_bmono_cont is recommended.
                    call time_max%start()
                    call sub_bmono_disc(V(:,iZ),kpi(:,iZ),f,nK,opt(1),opt(2))
                    ! call sub_bmono_cont(V(:,iZ),c(:,iZ),kp(:,iZ),B=gK,z=cash(:,iZ),Bp=gK,W=CV(:,iZ),x=gK,monotone=opt(1),concave=opt(2),iscontinuous=.false.,crra=crra)
                    call time_max%stop()

                    ! Check if have exceeded time limit
                    call toc(time)
                    if (time>maxTime) then
                        overtime = .true.
                        return ! do not modify cnt or time_out
                    end if
                end do

            end if
  
            cnt = (onecnt/real(nZ,8) + cnt*(iter-1))/iter

            call toc(time_out)
            iter_out = iter

            sup = maxval(abs(V-work))

            if (mod(iter,100)==0 .and. useStdOut) write(*,*) 'iter: ',iter,'supnorm V',sup
            
            if (sup<1d-6*abs(vSS)) exit

        end do

        cntmap = cntmap / iter ! average across iterations

        ! Recover c and kp
        do iZ = 1,nZ
            kp(:,iZ) = gK(kpi(:,iZ))
        end do
        c = cash - kp

        ! Check the number of non-monotonicities in kpi
        nonmonofound = nonmonofound + real(checknonmono(kpi),8)

        if (doexploit2state) then

            lambda = -huge(0d0)
            do iZ = 2,nZ-1
                ! $g(n_1,j) - g(1,j) +1 \le \lambda (g(n_1,j+1) - g(1,j-1)+1)$
                ! divide by term on RHS to get ... \le lambda and take max over LHS
                lambda = max(lambda,(kpi(nK,iZ)-kpi(1,iZ)+1d0)/(kpi(nK,iZ+1)-kpi(1,iZ-1)+1))
            end do

            lambda_alt = -huge(0d0)
            do iK = 2,nK-1
                lambda_alt = max(lambda_alt,(kpi(iK,nZ)-kpi(iK,1)+1d0)/(kpi(iK+1,nZ)-kpi(iK-1,1)+1))
            end do

        end if

        if (doexploit2state .and. opt(2)==0 .and. opt(1)==2) then

            ! Binary monotonicity, 2 state, brute concavity
            if (nK==25 .and. nZ==25) then
                call usedir('./results/')
!                 call csvwrite('cntmap',cntmap)
                call csvwrite('searchmap25',searchmap)
            elseif (nK==100 .and. nZ==25) then
                call usedir('./results/')
                call csvwrite('searchmap100',searchmap)
                call csvwrite('lambda100',lambda)
                call csvwrite('lambda_alt100',lambda_alt)
            end if
        end if

    contains

        function f(iK,iKp)
            implicit none
            integer, intent(in) :: iK,iKp
            real(8) :: ctmp
            real(8) :: f

            onecnt = onecnt + 1
            cntmap(iK,iZ) = cntmap(iK,iZ) + 1d0

            ctmp = cash(iK,iZ)-gK(iKp)
            if (ctmp>0d0) then
                if (crra==2d0) then
                    f = -1d0/ctmp + CV(iKp,iZ)
                else
                    f = ctmp**(1d0-crra)/(1d0-crra) + CV(iKp,iZ)
                end if
            else
                f = -1d100
            end if

        end function 

        function fmulti(iK,iZ,iKp) result(f)
            implicit none
            integer, intent(in) :: iK,iZ,iKp
            real(8) :: ctmp
            real(8) :: f

            onecnt = onecnt + 1
            cntmap(iK,iZ) = cntmap(iK,iZ) + 1d0
            if (iter==100) then
                searchmap(iKp,iK,iZ) = 1
            end if

            ctmp = cash(iK,iZ)-gK(iKp)
            if (ctmp>0d0) then
                if (crra==2d0) then
                    f = -1d0/ctmp + CV(iKp,iZ)
                else
                    f = ctmp**(1d0-crra)/(1d0-crra) + CV(iKp,iZ)
                end if
            else
                f = -1d100
            end if

        end function 
        function fmulti_trans(iZ,iK,iKp) result(f)
            implicit none
            integer, intent(in) :: iK,iZ,iKp
            real(8) :: ctmp
            real(8) :: f

            onecnt = onecnt + 1
            cntmap(iK,iZ) = cntmap(iK,iZ) + 1d0

            ctmp = cash(iK,iZ)-gK(iKp)
            if (ctmp>0d0) then
                if (crra==2d0) then
                    f = -1d0/ctmp + CV(iKp,iZ)
                else
                    f = ctmp**(1d0-crra)/(1d0-crra) + CV(iKp,iZ)
                end if
            else
                f = -1d100
            end if


        end function 

    end subroutine 

    ! Counts the number of non-monotonicities found in the matrix x
    integer function checknonmono(x)
        implicit none
        integer, intent(in) :: x(:,:)
        integer :: i, j
        integer :: cntnoni, cntnonj

        cntnoni = 0
        cntnonj = 0

        do j = 1,size(X,2)
            do i = 1,size(X,1)-1
                if (x(i,j)>x(i+1,j)) cntnoni = cntnoni + 1
            end do
        end do
        do j = 1,size(X,2)-1
            do i = 1,size(X,1)
                if (x(i,j)>x(i,j+1)) cntnonj = cntnonj + 1
            end do
        end do

        checknonmono = cntnoni + cntnonj

!         print*,'Non monotonicities found in i ',cntnoni,' found in j ',cntnonj,' magn i ',maxmagi,' magn j ',maxmagj

    end function 

    ! Compute the conditional expectation giving CV and q for a given k' and z (where the z is given by iZ)
    subroutine compute_contE_scalar_DEF(CVval,qval,kpval,iZ,dCVval)
        use mod_data
        implicit none
        real(8), intent(out) :: CVval, qval
        real(8), intent(in) :: kpval
        integer, intent(in) :: iZ
        real(8), intent(out), optional :: dCVval
        ! local
        real(8), allocatable :: Vnd_int(:),Vnd_quad(:),Vd_quad(:),cnd_quad(:),cnd_int(:)

        ! Interpolate Vnd at kp
        allocate(Vnd_int(nZ))
        Vnd_int = interp1q(gK,Vnd,kpval)

        ! Now, interpolate in z' dimension
        allocate(Vnd_quad(nquad),Vd_quad(nquad))
        Vnd_quad = interp1q(gZ,Vnd_int,QUADz(:,iZ)) ! matmul(QUADqz(:,:,iZ),Vnd_int)
        Vd_quad = interp1q(gZ,Vd(nK,:),QUADz(:,iZ)) ! matmul(QUADqz(:,:,iZ),Vd(nK,:))

        ! Now sum up values at quadrature points to get CV and q
        CVval = beta*dot_product(max(Vnd_quad,Vd_quad),QUADw(:,iZ))
        qval = sum(QUADw(:,iZ),Vnd_quad>=Vd_quad)/(1d0+rstar)

        if (present(dCVval)) then
            ! beta * \int_{nodef} u'(cnd) dF(z'|z) 
            allocate(cnd_int(nZ))
            allocate(cnd_quad(nquad))
            cnd_int = interp1q(gK,cnd,kpval)
            cnd_quad = interp1q(gZ,cnd_int,QUADz(:,iZ))
            dCVval = beta*sum(cnd_quad**(-crra)*QUADw(:,iZ),mask=Vnd_quad>=Vd_quad)
        end if

    end subroutine
    subroutine compute_contE_DEF(CVd) ! ,dq,dCV)
        use mod_data
        use mod_matlab
        implicit none
        real(8), intent(out), optional :: CVd(:)
        ! dq(:,:),dCV(:,:)
        ! local
        real(8), allocatable :: quad_x(:),quad_w(:),quad_pdf(:),quad_Vnd(:,:),quad_Vd(:),quad_V(:,:),quad_repay(:,:)
        integer :: iZ,iquad
            
        allocate(quad_x(nquad),quad_w(nquad),quad_Vnd(nK,nquad),quad_Vd(nquad),quad_V(nK,nquad),quad_pdf(nquad),quad_repay(nK,nquad)) 

        ! Numerically integrate 
        do iZ = 1,nZ

! print*,'shape(QUADqz(:,:,iZ))',shape(QUADqz(:,:,iZ)),'shape(transpose(Vnd))',shape(transpose(Vnd))
            ! quad_Vnd = matmul(QUADqz(:,:,iZ),transpose(Vnd)) ! nQ x nK = (nQ x nZ) * (nK x nZ)'
            ! quad_Vd = matmul(QUADqz(:,:,iZ),Vd(nK,:)) ! nQ x 1 = (nQ x nZ) * (nZ x 1)
            ! call mygemm(quad_Vnd,QUADqz(:,:,iZ),Vnd,nquad,nZ,nK,nZ,tB='T') 
            ! call mygemm(quad_Vd,QUADqz(:,:,iZ),Vd(nK,:),nquad,nZ,nZ,1)


            call sub_interp1q(quad_Vnd,gZ,Vnd,QUADz(:,iZ),dim=2) ! interpolates along the second dimension
            ! call sub_interp1q(quad_Vd,gZ,Vd,QUADz(:,iZ),dim=2) ! interpolates along the second dimension
            quad_Vd = interp1q(gZ,Vd(1,:),QUADz(:,iZ)) 
    
            quad_repay = 1d0
            quad_V = quad_Vnd
            do iquad = 1,nquad
                where (quad_Vnd(:,iquad)<quad_Vd(iquad)) 
                    quad_repay(:,iquad) = 0d0
                    quad_V(:,iquad) = quad_Vd(iquad)
                endwhere
            end do

!             do iK = 1,nK
!                 quad_V(:,iK) = max(quad_Vnd(:,iK),quad_Vd)
!                 where (quad_Vnd(:,iK)>=quad_Vd) quad_repay(:,iK) = 1d0
!             end do
            
            ! CV(:,iZ) = matmul(QUADw(:,iZ),quad_V) ! 1 x nK
            ! CV(:,iZ) = CV(:,iZ) * beta
            ! q(:,iZ) = matmul(QUADw(:,iZ),quad_repay) ! 1 x nK
            ! q(:,iZ) = q(:,iZ) / (1d0+rstar)
            call mygemm(CV(:,iZ),quad_V,QUADw(:,iZ),nK,nquad,nquad,1,alpha=beta)
            call mygemm(q(:,iZ),quad_repay,QUADw(:,iZ),nK,nquad,nquad,1,alpha=1d0/(1d0+rstar))
            
            if (present(CVd)) CVd(iZ) = beta*dot_product(theta*quad_Vnd(iKeq0,:) + (1d0-theta)*quad_Vd,QUADw(:,iZ))

! 
!             ! Compute the numerical derivatives of q by interpolating Vnd and Vd
!             if (present(dq) .or. present(dCV)) then
!                 ! Interpolate Vnd(k,z) to get Vnd(k+h,z)
!                 do iK = 1,nK
!                     quad_Vnd_plus(:,iK) = interp1q(gK,quad_Vnd,gK(iK)+h)
!                 end do
! 
!                 ! Recover from this the default decisions 
!                 if (present(dq)) then
!                     quad_repay_plus = 0d0
!                     do iK = 1,nK
!                         where (quad_Vnd_plus(:,iK) > quad_Vd) quad_repay_plus(:,iK) = 1d0
!                     end do
!                     call mygemm(dq(:,iZ),QUADw(:,iZ),quad_repay_plus,1,nquad,nquad,nK,alpha=1d0/(1d0+rstar))
!                     dq(:,iZ) = (dq(:,iZ) - q(:,iZ))/h
!                 end if
!                 if (present(dCV)) then
!                     do iK = 1,nK
!                         quad_V_plus(:,iK) = max(quad_Vnd_plus(:,iK),quad_Vd)
!                     end do
!                     call mygemm(dCV(:,iZ),QUADw(:,iZ),quad_V_plus,1,nquad,nquad,nK,alpha=beta)
!                     dCV(:,iZ) = (dCV(:,iZ) - CV(:,iZ))/h
!                 end if
!             end if

        end do

    end subroutine

    subroutine vfiDEF(iter_out,cnt,time_out,overtime,opt,tol,guess,soln)
        use mod_quad
        use mod_util
        use mod_matlab
        use mod_binarymono
        use mod_data
        implicit none
        real(8), allocatable, intent(out), optional :: soln(:,:,:)
        real(8), allocatable, intent(in), optional :: guess(:,:,:)
        real(8), intent(out) :: cnt, time_out ! average count per iteration and run time over cnt
        integer, intent(in) :: opt(2)
        real(8), intent(in) :: tol
        logical, intent(out) :: overtime
        integer, intent(out) :: iter_out ! vfi iteration
        ! local
        real(8) :: sup, time
        integer :: onecnt,iK,iZ,iter,ilo,itmpvec(1),tmpcnt
        real(8), allocatable :: x(:,:),dx(:),dWx(:),CVd(:)

        ! Start the overall timer
        call time_tot%start()

        allocate(x(nK,nZ),dx(nK),dWx(nK),CVd(nZ))
        
        overtime = .false.


        ! Make initial guess
        Vnd = 0d0
        Vd = 0d0
        V = 0d0
        d = 0d0
        cnt = 0d0

        ! Override if applicable
        if (present(guess)) then
            if (allocated(guess)) then
                ! Special case that is important in estimation is where the grids haven't changed.
                if (size(guess,1)==nK .and. size(guess,2)==nZ) then
                    ! This may be inferior in terms of run counts but avoids interp
                    Vnd = guess(:,:,3)
                    Vd = guess(:,:,4)
                    q = guess(:,:,5)
                    V = max(Vnd,Vd)
                    d = 1d0
                    where (Vnd>=Vd) d = 0d0
                ! Otherwise, have to interpolate
                else
                    Vnd = interp2q(guess(:,1,1),guess(1,:,2),guess(:,:,3),gK,gZ) ! old gK,gZ grid to new
                    Vd = interp2q(guess(:,1,1),guess(1,:,2),guess(:,:,4),gK,gZ) ! old gK,gZ grid to new
                    q = interp2q(guess(:,1,1),guess(1,:,2),guess(:,:,5),gK,gZ) ! old gK,gZ grid to new
                    V = max(Vnd,Vd)
                    d = 1d0
                    where (Vnd>=Vd) d = 0d0
               end if
            end if
        end if

        ! Get quadrature rules (this is low cost)
        ! call sub_simpsonquad(quad_x,quad_w,log(gZ(1)),log(gZ(nZ)))

        do iter = 1,500
        
            work = V

            ! Compute bond prices and continuation utility
            ! q = matmul(1d0-d,Fzpcz)/(1d0+rstar)
            ! call mygemm(dCV,c**(-crra),Fzpcz,nK,nZ,nZ,nZ,alpha=beta)

            ! If doing continuous approximation for y, need to interpolate Vnd and Vd in y dimension
            call time_quad%start()
            if (dodiscreteshocks) then

                call mygemm(CV,V,Fzpcz,nK,nZ,nZ,nZ,alpha=beta)
                call mygemm(q,d,Fzpcz,nK,nZ,nZ,nZ,alpha=1d0/(1d0+rstar)) ! qbar * E d
                q = 1d0/(1d0+rstar) - q ! qbar - qbar * E d = qbar * E(1-d)
!                 q = matmul(1d0-d,Fzpcz) / (1d0+rstar)

                ! Check that q is increasing 
                do iZ = 1,nZ
                    do iK = 1,nK-1
                        if (q(iK,iZ)>1d-6 + q(iK+1,iZ)) then
                            write(0,*) 'WARNING: q is not increasing and should be'
                            stop
                        end if
                    end do
                end do

                CVd = beta*matmul(theta*V(iKeq0,:) + (1d0-theta)*Vd(nK,:),Fzpcz)

            else

                ! Compute q and CV (both module variables), as well as CVd 
                call compute_contE_DEF(CVd=CVd)

            end if
            call time_quad%stop()

!             call gnuplot(gK,CV(:,[1,nZ/2,nZ]),title='CV',saveas='pics/CV_iter'//trim(conv_i_to_s(iter))//'.eps')
!             call gnuplot(gK,q(:,[1,nZ/2,nZ]),title='q',saveas='pics/q_iter'//trim(conv_i_to_s(iter))//'.eps')
!             if (iter==10) stop

            ! Solve the Vnd problem
            onecnt = 0

            !!! Construct expenditures less endowment (now, c = b - x)
            do iZ = 1,nZ
                x(:,iZ) = q(:,iZ)*gK - gZ(iZ) 
            end do

            if (dogridsearch) then

                do iZ = 1,nZ

                        ! NOTE: To anyone reading this code, sub_bmono_cont is somewhat faster (maybe 10% faster) than
                        ! using sub_bmono_disc because there is no need for function calls (as we restrict the problem's
                        ! form). For that reason, it should generally be preferred. However, here we go with the slower
                        ! one to be consistent across the RBC and DEF versions.

                        ! Grid search
                        call time_max%start()
                        call sub_bmono_disc(Vnd(:,iZ),kpi(:,iZ),f,nK,opt(1),opt(2),evalcnt=tmpcnt)
                        ! call sub_bmono_cont(Vnd(:,iZ),cnd(:,iZ),kp(:,iZ),B=gK,z=gK,Bp=gK,W=CV(:,iZ),x=x(:,iZ),monotone=opt(1),concave=opt(2),iscontinuous=.false.,crra=crra,evalcnt=tmpcnt)
                        call time_max%stop()

                        onecnt = onecnt + tmpcnt

                end do

            else

                !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iZ,ilo,tmpcnt,itmpvec) REDUCTION(+:onecnt)
                do iZ = 1,nZ

                    ! CONT 
                    call time_max%start()
                    call sub_bmono_cont(Vnd(:,iZ),cnd(:,iZ),kp(:,iZ),B=gK,z=gK,Bp=gK,W=CV(:,iZ),x=x(:,iZ),&
                                        monotone=opt(1),concave=opt(2),&
                                        iscontinuous=.not. docontmono_but_discrete,& ! NOTE: will use this subroutine but solve discrete version (for fair comparisons in Section 6) if docontmono_but_discrete
                                        crra=crra,domaxexact=domaxexact,evalcnt=tmpcnt)
                    call time_max%stop()

                    onecnt = onecnt + tmpcnt

                end do
                !$OMP END PARALLEL DO

            end if

            ! Check if have exceeded time limit
            call toc(time)
            overtime = time>maxTime

            if (overtime) return

            cnt = (onecnt/real(nZ,8) + cnt*(iter-1))/iter
            call toc(time_out)
            iter_out = iter

            ! Solve the Vd problem
            ! Vd(nK,:) = ydef(gZ)**(1d0-crra)/(1d0-crra) + beta*matmul(theta*V(iKeq0,:) + (1d0-theta)*Vd(nK,:),Fzpcz)
            Vd(nK,:) = ydef(gZ)**(1d0-crra)/(1d0-crra) + CVd
            do iK = 1,nK-1 
                Vd(iK,:) = Vd(nK,:)
            end do
             
            ! Compute V and d
            where (Vnd>=Vd)
                d = 0d0
                V = Vnd
            elsewhere
                d = 1d0
                V = Vd
            endwhere
! if (iter==3) stop 'checking plots'

            if (.not. dogridsearch) then
                c = cnd
                do iZ = 1,nZ
                    where (d(:,iZ)==1d0) c(:,iZ) = ydef(gZ(iZ))
! write(*,'(i4,A)',advance='no') iZ,' : '
                    call errorcheck(iZ) ! check k' policy (not applicable for grid search until the end
                end do
            end if

            sup = maxval(abs(V-work))

            if (mod(iter,1)==0 .and. useStdOut) write(*,*) 'iter: ',iter,'supnorm V',sup
            
            ! If doing EEE check, use tight tolerance here because q is updated an extra time
            if (sup<tol*maxval(abs(V))) exit

        end do

        if (dogridsearch) then
            do iZ = 1,nZ

                ! Get capital choice (non index) conditional on repayment 
                kp(:,iZ) = gK(kpi(:,iZ)) 

                ! Back out consumption conditional on repayment
                cnd(:,iZ) = gZ(iZ) - q(kpi(:,iZ),iZ)*kp(:,iZ) + gK

                ! Recover the overall consumption policy
                where (d(:,iZ)==1d0)
                    c(:,iZ) = ydef(gZ(iZ))
                elsewhere
                    c(:,iZ) = cnd(:,iZ)
                endwhere
            end do

            ! Check the number of non-monotonicities in kpi
            nonmonofound = nonmonofound + real(checknonmono(kpi),8)

        end if

        do iZ = 1,nZ
            call errorcheck(iZ)
        end do

        ! Save
        if (present(soln)) then
            allocate(soln(nK,nZ,9))
            soln(:,1,1) = gK
            soln(1,:,2) = gZ
            soln(:,:,3) = Vnd
            soln(:,:,4) = Vd
            soln(:,:,5) = q
            soln(:,:,6) = kp ! this is "kpnd"
            soln(:,:,7) = cnd
            soln(:,:,8:9) = NOT_INITIALIZED_GUESS ! will store the invariant distribution
        end if

        ! This time_tot%stop captures the time from just one run, not the average
        call time_tot%stop()

    contains

        subroutine errorcheck(iZ)
            implicit none
            integer, intent(in) :: iZ
            ! local
            integer :: iK
            if (any(d(:,iZ)==0d0 .and. (kp(:,iZ)<gK(1) - 1d-10 .or. kp(:,iZ)>gK(nK) + 1d-10))) then
                do iK = 1,nK
                    if (d(iK,iZ)==0d0 .and. (kp(iK,iZ)<gK(1) .or. kp(iK,iZ)>gK(nK))) then
                        write(*,*) 'Bad kp value: kp(iK,iZ) ',kp(iK,iZ),' gK(1) ',gK(1),' gK(nK) ',gK(nK)
                    end if
                end do
                ! call saveplots()
                ! write(0,*) 'WARNING: bad kp value. This may happen with GEGM when certain candidate points are rejected. See the paper.' !The variant is nK,nZ',nK,nZ,'docontmono',docontmono,'doaccgegm',doaccgegm
            end if
! write(*,*)
        end subroutine 

        function f(iK,iKp)
            implicit none
            integer, intent(in) :: iK,iKp
            real(8) :: f
!             real(8) :: ctmp 

            ! Note: z is output, k is bond holdings, and k' is next period bond holdings
            ! ctmp = gZ(iZ) - q(iKp,iZ)*gK(iKp) + gK(iK)
            ! ctmp = gK(iK) - x(iKp,iZ) ! x = -y + q(b',y)*b'
            if (gK(iK)>x(iKp,iZ)) then
                if (crra==2d0) then
                    f = -1d0/(gK(iK)-x(iKp,iZ))+ CV(iKp,iZ) !  beta*dot_product(Vtran(:,iKp),Fzpcz(:,iZ))
                else
                    f = (gK(iK)-x(iKp,iZ))**(1d0-crra)/(1d0-crra) + CV(iKp,iZ) !beta*dot_product(Vtran(:,iKp),Fzpcz(:,iZ))
                end if
            else
                f = -1d100
            end if

        end function 

    end subroutine 

    subroutine computeEEEDEF(avgEEE,maxEEE,extrastats)
        use mod_data
        use mod_stata
        use mod_spline
        implicit none
        real(8), intent(out) :: avgEEE,maxEEE
        real(8), intent(out), allocatable :: extrastats(:)
        real(8), allocatable :: dx(:,:),dCV(:,:),r(:,:),tmpmat(:,:)
        logical, allocatable :: isvalid(:,:)
        real(8) :: qval,dCVval,dxval,qval_left,qval_right,CVval,CVval_left,CVval_right,xval_left,xval_right,xval,h_CV,h_x,tmp
        integer :: iK,iZ
        type(stata_wkspace) :: s
        character(len=100) :: dir
        real(8), parameter :: deriv_agree_tol = 1d-3

        ! First order condition is
        ! u' * (dq(b')b'/db') = beta * E (dV'/db') and so
        ! u' * (dq/db'*b' + q(b')) = beta * E(dV'/db'). Hence, the EEE is 
        ! log10(abs(1-u'^{-1}(beta*E(dV'/db')/(dq/db'*b' + q)/c)
        !

        allocate(dx(nK,nZ),isvalid(nK,nZ),dCV(nK,nZ),r(nK,nZ))

        ! Stepsize for numerical derivative
        h_x = 1d-3 ! was 1d-4 until 7/29/16 ! to figure out what was reasonable here (given that in truth it is always going to be a step function for any finite quadrature rule, I adjusted h_x and examined dx_fcn_of_kp.eps to see if it looked monotone (which it should be since x in the relevant range looks very nearly convex, cf. x.eps, and k' is monotone). I found h_x = 1d-3 was too large, so I used this value, despite it being more of a secant.
        h_CV = 1d-5

        dCV = 0d0
        dx = 0d0
        r = STATA_MISSING
! 
        ! Use the c and kp policy functions in mod_data
        isvalid = .false.
        work = 1d0 ! put in some positive number so u'^-1 does not return NaNs
        do iZ = 1,nZ

            do iK = 1,nK

                if (d(iK,iZ)==0d0) then
                    
                    isvalid(iK,iZ) = .true.

                    ! Compute q and CV at k' +/- h
                    if (dodiscreteshocks) then

                        CVval = beta*dot_product(Fzpcz(:,iZ),max(interp1q(gK,Vnd,kp(iK,iZ)),Vd(nK,:))) ! beta*E(max(Vnd,Vd))
                        CVval_right = beta*dot_product(Fzpcz(:,iZ),max(interp1q(gK,Vnd,kp(iK,iZ)+h_CV),Vd(nK,:))) ! beta*E(max(Vnd,Vd))
                        CVval_left = beta*dot_product(Fzpcz(:,iZ),max(interp1q(gK,Vnd,kp(iK,iZ)-h_CV),Vd(nK,:))) ! beta*E(max(Vnd,Vd))

                        qval = sum(Fzpcz(:,iZ),mask=interp1q(gK,Vnd,kp(iK,iZ))>=Vd(nK,:))/(1d0+rstar) ! q at kp
                        qval_right = sum(Fzpcz(:,iZ),mask=interp1q(gK,Vnd,kp(iK,iZ)+h_x)>=Vd(nK,:))/(1d0+rstar) ! q at kp+h
                        qval_left = sum(Fzpcz(:,iZ),mask=interp1q(gK,Vnd,kp(iK,iZ)-h_x)>=Vd(nK,:))/(1d0+rstar) ! q at kp-h

                    else
                        
                        call compute_contE_scalar_DEF(CVval,qval,kp(iK,iZ),iZ)
                        call compute_contE_scalar_DEF(CVval_right,tmp,kp(iK,iZ)+h_CV,iZ) ! actually returns CV and q at kp+h
                        call compute_contE_scalar_DEF(CVval_left,tmp,kp(iK,iZ)-h_CV,iZ) ! actually returns CV and q at kp-h
                        call compute_contE_scalar_DEF(tmp,qval_right,kp(iK,iZ)+h_x,iZ) ! actually returns CV and q at kp+h
                        call compute_contE_scalar_DEF(tmp,qval_left,kp(iK,iZ)-h_x,iZ) ! actually returns CV and q at kp-h
                    
                    end if

                    ! Compute numerical derivatives and see if the functions appear to be differentiable
                    xval = qval*kp(iK,iZ) - gZ(iZ)
                    xval_right = qval_right*(kp(iK,iZ)+h_x) - gZ(iZ) ! x at kp+h  
                    xval_left = qval_left*(kp(iK,iZ)-h_x) - gZ(iZ) ! x at kp-h
                    dxval = (xval_right - xval_left)/(2d0*h_x)

                    dCVval = (CVval_right - CVval_left)/(2d0*h_CV)

                    ! Check whether left and right derivatives disagree from this much. If so, make invalid
                    if (abs((xval-xval_left)/h_x - dxval)>deriv_agree_tol .or. &
                        abs((xval_right-xval)/h_x - dxval)>deriv_agree_tol) then
                        isvalid(iK,iZ) = .false.
                    end if

                    ! Check whether at binding constraint
                    if (kp(iK,iZ)>=gK(nK) .or. kp(iK,iZ)<=gK(1)) isvalid(iK,iZ) = .false.

                    ! Save some values
                    dCV(iK,iZ) = dCVval
                    dx(iK,iZ) = dxval
                    r(iK,iZ) = 1d0/qval-1d0 - rstar

                    if (isvalid(iK,iZ) .and. (dCVval<=0d0 .or. dxval<=0d0)) then
                        isvalid(iK,iZ) = .false. 
                        write(0,*) 'WARNING: in EEE computation, dCVval or dxval <= 0d0. Excluding from sample. dCVval and dxval are ',dCVval,dxval
                    else
                        work(iK,iZ) = dCVval/dxval
                    end if
                    
                end if 
                
            end do
        end do

        call dPUinv(work) ! u'^{-1} E[...]

        ! Set EEE to invalid by default and put in actual value where well defined
        EEE = STATA_INVALID
        where (isvalid) EEE = log10(epsilon(0d0)+abs(1d0-work/c))

        call stata_gen_assumedsize(s,EEE,'EEE',size(EEE))
        call stata_gen_assumedsize(s,kp,'kp',size(kp))
        call stata_gen_assumedsize(s,dx,'dx',size(dx))
        call stata_gen_assumedsize(s,r,'spread',size(r))
        call stata_gen_assumedsize(s,spread(gZ,1,nK),'output',nK*nZ)

        allocate(tmpmat(nK,nZ))
        tmpmat = 1d0-d ! condition on not repaying (for EEE stats, they are implicitly conditioned on isvalid because non-isvalids have STATA_INVALID or STATA_MISSING
        call stata_setweight_assumedsize(s,tmpmat,nK*nZ)

        if (useStdOut) then
            print*,'Weighted only conditional on not defaulting'
            call s%summarize()
        end if

        ! Set weights
        tmpmat = dsn(:,:,1) * (1d0-d)
        tmpmat = tmpmat/sum(tmpmat)
        call stata_setweight_assumedsize(s,tmpmat,nK*nZ)

        if (useStdOut) then
            print*,'Weighted by invariant distribution conditional on not defaulting'
            call s%summarize()
        end if

        ! call s%hist('EEE')

        avgEEE = sum(EEE*dsn(:,:,1),mask=isvalid)/sum(dsn(:,:,1),mask=isvalid)
        maxEEE = maxval(EEE,mask=isvalid)

        if (useStdOut) then
            print*,'50,75,90,95,99,99.9 pctiles of EEE ',s%pctile('EEE',[50d0,75d0,90d0,95d0,99d0,99.9d0])
        end if

        allocate(extrastats(8))
        extrastats(1:6) = s%pctile('EEE',[50d0,75d0,90d0,95d0,99d0,99.9d0])
        extrastats(7) = s%stdev('spread') ! spread volatility 
        extrastats(8) = s%pwcorr('spread','output') ! spread cyclicality


        where (.not. isvalid) EEE = 0d0 ! for plotting


        if (dogridsearch) then
            dir = 'pics_grid/'
        elseif (docontmono) then
            dir = 'pics_cont/'
        end if
        
        if (useStdOut) then
            call gnuplot(gK,dx(:,[1,(nZ+1)/2,nZ]),title='Derivative of expenditure functions',saveas=trim(dir)//'dx.eps')
            call gnuplot(gK,dCV(:,[1,(nZ+1)/2,nZ]),title='Derivative of continuation utility',saveas=trim(dir)//'dCV.eps')
            call gnuplot(kp(:,(nZ+1)/2),dx(:,(nZ+1)/2),title='Derivative of x with optimal policy on horizontal axis',saveas=trim(dir)//'dx_fcn_of_kp.eps')

            call s%scatter('kp','dx',trim(dir)//'scatter_kp_vs_dx.eps')
        end if

!         do iZ = 1,nZ
!             do iK = 1,nK
!                 if (isvalid(iK,iZ) .and. EEE(iK,iZ)>=.9d0*maxEEE + .1d0*avgEEE) then
!                     print*,'EEE(iK,iZ)',EEE(iK,iZ)
!                     print*,'kp(iK,iZ)',kp(iK,iZ)
!                     call bsearch(ilo,kp(iK,iZ),gK)
!                     print*,'gK(ilo:ilo+1) @ (iK,iZ)',gK(ilo:ilo+1)
!                     print*,'CV slope @ (iK,iZ)',(CV(ilo+1,iZ)-CV(ilo,iZ))/(gK(ilo+1)-gK(ilo))
!                     print*,'dCV(iK,iZ)',dCV(iK,iZ)
!                     print*,'x slope @ (iK,iZ)',(x(ilo+1,iZ)-x(ilo,iZ))/(gK(ilo+1)-gK(ilo))
!                     print*,'dx(iK,iZ)',dx(iK,iZ)
!                 
!                 end if
!             end do
!         end do
! 


    end subroutine 


    subroutine computeEEERBC(avgEEE,maxEEE)
        use mod_data
        implicit none
        real(8), intent(out) :: avgEEE,maxEEE
        real(8) :: kpval,tmp(nZ)
        integer :: iK,iZ,iZp

        ! Use the c and kp policy functions in mod_data

        do iZ = 1,nZ
            do iK = 1,nK
                kpval = kp(iK,iZ)
                do iZp = 1,nZ
                    tmp(iZp) = c(kpi(iK,iZ),iZp)
                end do
                call dPU(tmp) ! u'(c)
                work(iK,iZ) = dot_product(Fzpcz(:,iZ),beta*tmp*(1d0+mpk(gZ,kpval)-deprec)) ! E[beta*u'(c')*(1+r'-delta)|z]
                
            end do
        end do

        call dPUinv(work) ! u'^{-1} E[...]
        EEE = log10(abs(1d0-work/c)) 

        avgEEE = sum(EEE*dsn(:,:,1))
        maxEEE = maxval(EEE)

    end subroutine 

    subroutine invarRBC
        use mod_data
        implicit none
        real(8) :: sup
        integer :: iZ,iK,iter

        dsn = 0d0

        dsn(nK/2,nZ/2,1) = 1d0  

        do iter = 1,100000

            work = 0d0

            do iZ = 1,nZ
                do iK = 1,nK
                    work(kpi(iK,iZ),iZ) = work(kpi(iK,iZ),iZ) + dsn(iK,iZ,1)
                end do
            end do

            work2 = dsn(:,:,1)

            dsn(:,:,1) = matmul(work,transpose(Fzpcz))

            sup = maxval(abs(dsn(:,:,1)-work2))

            if (mod(iter,25000)==0) print*,'DSN: iter ',iter,' sup ',sup

            if (sup<1d-10) exit


        end do

    end subroutine 

    subroutine invarDEF(guess,soln)
        use mod_matlab
        use mod_data
        implicit none
        real(8), allocatable, intent(inout), optional :: soln(:,:,:)
        real(8), allocatable, intent(in), optional :: guess(:,:,:)
        real(8), allocatable :: dsn2(:,:,:),mlomat(:,:)
        integer, allocatable :: ilomat(:,:)
        real(8) :: sup,mlo
        integer :: iZ,iK,iter,ilo


        allocate(dsn2(nK,nZ,2))
        allocate(mlomat(nK,nZ))
        allocate(ilomat(nK,nZ))

        dsn = 0d0
        dsn(nK,(nZ+1)/2,1) = 1d0  
        if (present(guess)) then
            if (allocated(guess)) then
                if (.not. any(guess(:,:,8:9)==NOT_INITIALIZED_GUESS)) then
                    if (all(shape(guess(:,:,8:9))==shape(dsn))) dsn = guess(:,:,8:9)
                end if
            end if
        end if

        do iZ = 1,nZ
            do iK = 1,nK
                call noexinterp1qlocweight(ilomat(iK,iZ),mlomat(iK,iZ),kp(iK,iZ),gK)
            end do
        end do

        do iter = 1,merge(10000,1000,theta>0d0)

            dsn2 = 0d0

            ! h = 0
            do iZ = 1,nZ
                do iK = 1,nK
                    if (d(iK,iZ)==1d0) then
                        dsn2(iKeq0,iZ,2) = dsn2(iKeq0,iZ,2) + dsn(iK,iZ,1)
                    else
                        ! dsn2(kpi(iK,iZ),iZ,1) = dsn2(kpi(iK,iZ),iZ,1) + dsn(iK,iZ,1)
!                         call noexinterp1qlocweight(ilo,mlo,kp(iK,iZ),gK)
                        ilo = ilomat(iK,iZ)
                        mlo = mlomat(iK,iZ)
                        dsn2(ilo:ilo+1,iZ,1) = dsn2(ilo:ilo+1,iZ,1) + [mlo,1d0-mlo]*dsn(iK,iZ,1)
                    end if
                end do
            end do

            ! h = 1
            do iZ = 1,nZ
                dsn2(iKeq0,iZ,1) = dsn2(iKeq0,iZ,1) + theta*dsn(iKeq0,iZ,2)
                dsn2(iKeq0,iZ,2) = dsn2(iKeq0,iZ,2) + (1d0-theta)*dsn(iKeq0,iZ,2)
            end do

            work = dsn(:,:,1)
            work2 = dsn(:,:,2)

            ! dsn(:,:,1) = matmul(dsn2(:,:,1),transpose(Fzpcz))
            ! dsn(iKeq0,:,2) = matmul(dsn2(iKeq0,:,2),transpose(Fzpcz))
            call mygemm(dsn(:,:,1),dsn2(:,:,1),Fzpcz,nK,nZ,nZ,nZ,tb='T')
            dsn(iKeq0,:,2) = matmul(Fzpcz,dsn2(iKeq0,:,2))  ! b*A' = (A*b')'

            sup = max(maxval(abs(dsn(:,:,1)-work)),maxval(abs(dsn(:,:,2)-work2)))

            if (mod(iter,1000)==0) print*,'DSN: iter ',iter,' sup ',sup, 'sum ',sum(dsn)

            if (sup<1d-10) exit

        end do

        if ((sum(dsn)-1d0)>1d-8) write(0,*) 'WARNING: invarDEF: invariant dsn sums to ',sum(dsn)

        if (present(soln)) then
            if (all(shape(soln(:,:,8:9))==shape(dsn))) soln(:,:,8:9) = dsn
        end if

    end subroutine 


end module mod_routines

module mod_main
    implicit none
contains
    subroutine sub_main(model,dims,opt,maxTime_in,useStdOut_in,shocktype_in,maxmethod_in,doextras,guess,soln,loopfortime_in,estimating_in,converged_out)
        use mod_quad
!         use mod_myprint
        use mod_data
        use mod_routines
        implicit none
        real(8), allocatable, intent(out), optional :: soln(:,:,:)
        real(8), allocatable, intent(in), optional :: guess(:,:,:)
        real(8), allocatable :: quad_x(:),quad_w(:),extrastats(:)
        integer, intent(in) :: dims(2),opt,model,maxTime_in,shocktype_in,maxmethod_in
        logical, intent(in) :: useStdOut_in,doextras
        logical, intent(in), optional :: loopfortime_in,estimating_in
        logical, intent(out), optional :: converged_out
        integer :: unitno,i,vfi_iter,iZ
        real(8) :: cnt,minutes,stats(7),avgEEE,maxEEE,vfi_time,minqZ,maxqZ
        logical :: overtime,doRBC,loopfortime,estimating
        character(50) :: arg
        character(50*9) :: arglist
        real(8) :: vfi_total_time, vfi_avg_time, tol, total_time(3), avg_time(3)
        integer :: repeat_iter


        estimating = .false.
        if (present(estimating_in)) estimating = estimating_in

        tol = merge(1d-5,1d-7,estimating)

        if (.not. estimating) then
            call execute_command_line('export VECLIB_MAXIMUM_THREADS=1') ! If using Mac, single threaded BLAS via the Accelerate framework can be done by setting the environment variable VECLIB_MAXIMUM_THREADS to 1'
            call execute_command_line('export OMP_NUM_THREADS=1')
        end if

        maxTime = real(maxTime_in,8)
        shocktype = shocktype_in
        maxmethod = maxmethod_in
        useStdOut = useStdOut_in

        loopfortime = .true.
        if (present(loopfortime_in)) loopfortime = loopfortime_in
        if (estimating) loopfortime = .false.

        ! RBC = 0; Arellano = 1
        select case (model)
            case (0:1) 
                doRBC = 0==model
            case default
                stop 'ERROR: invalid model'
        end select

        ! Shocks
        select case (shocktype)
            case (0:1)
                dodiscreteshocks = 0==shocktype
                if (.not. dodiscreteshocks .and. doRBC) stop 'ERROR: RBC has not been programmed for continuous shocks'
            case default
                stop 'ERROR: invalid shocktype'
        end select

        ! Maximization method
        select case (maxmethod)
            case (0:5)
                dogridsearch = maxmethod==0 .or. maxmethod==3 .or. maxmethod==4
                docontmono = maxmethod==1 .or. maxmethod==2 .or. maxmethod==5
                docontmono_but_discrete = maxmethod==5
                domaxexact = maxmethod==1
                doexploit2state = maxmethod==3 .or. maxmethod==4
                dotranspose2state = maxmethod==4
                if (.not. dogridsearch .and. doRBC) stop 'ERROR: RBC has only been programmed to handle grid search'
            case default
                stop 'sub_main: case not found'
        end select

        ! Allocate space
        call allocatedata(dims(1),dims(2))

        if (useStdOut) then
            write(*,'(A,A12)') '                              model: ',merge('RBC','DEF',doRBC)
            write(*,'(A,L12)') '                    discrete shocks: ',dodiscreteshocks
            write(*,'(A,L12)') '                        grid search: ',dogridsearch
            write(*,'(A,L12)') '                         cont. mono: ',docontmono
            write(*,'(A,L12)') ' cont. mono subroutine but discrete: ',docontmono
            write(*,'(A,L12)') '                        max exactly: ',domaxexact
            write(*,'(A,L12)') '                     exploit 2 mono: ',doexploit2state
            write(*,'(A,L12)') '                   transpose 2 mono: ',dotranspose2state
            write(*,'(A,I12)') '                                 nK: ',nK
            write(*,'(A,I12)') '                                 nZ: ',nZ
        end if

        ! Set params and grids
        if (doRBC) then

            call setRBCparam()

            gK = linspace(.8d0*kSS,1.2d0*kSS,nK)
            call sub_tauchen(gZ,Fzpcz,.95d0,.007d0,nZ,3d0)
            gZ = exp(gZ)
            Fzpcz = transpose(Fzpcz)

        else

            call setDEFparam()

            ! Note: Arellano baseline has nZ = 21 and nK = 200
            gK(1:iKeq0) = linspace(-.35d0,0d0,iKeq0)
            gK(iKeq0:nK) = linspace(0d0,.15d0,nK-iKeq0+1)

            if (dodiscreteshocks) then

                if (.not. estimating) then
                    ! +/- 3stdev corresponds to Arellano's AER grid (but she uses Tauchen and Hussey not Tauchen)
                    call sub_tauchen(gZ,Fzpcz,def_ar1,def_std,nZ,3d0)
                    Fzpcz = transpose(Fzpcz)
                else
                    call sub_rouwenhorst(gZ,Fzpcz,def_ar1,def_std,nZ)
                    Fzpcz = transpose(Fzpcz)
                end if

            else

                ! Use large coverage so don't need to extrapolate too much
                ! With linear interpolation, suppose the quadrature rule says we need to evaluate at z' with weight w.
                ! Then linear interpolation will place weights mlo and 1-mlo on the two closest nodes. So we can interpret this as
                ! one huge transition matrix. However, note that it is not proper, as in the case of extrapolating, it will have
                ! negative values. Rather, it is  by nZ where the columns give the current iZ
                ! and the rows give the mixture of quadrature weights and 
                call sub_tauchen(gZ,Fzpcz,def_ar1,def_std,nZ,4d0) ! Fzpcz is only used in the discrete case for computing the invariant distribution
                ! Fzpcz = -1d0

                allocate(quad_x(nquad))
                allocate(quad_w(nquad))

                call legendrequad(quad_x,quad_w,-1d0,1d0)

                do iZ = 1,nZ

                    ! Generate the quadrature points
                    ! call chebyquad(quad_x,quad_w,def_ar1*gZ(iZ)-5d0*def_std,def_ar1*gZ(iZ)+5d0*def_std)
                    ! call legendrequad(quad_x,quad_w,def_ar1*gZ(iZ)-5d0*def_std,def_ar1*gZ(iZ)+5d0*def_std)

                    ! Legendre quadrature : integrate \int f(x) dx from [-1,1]
                    ! Adapt to minqZ maxqZ
                    minqZ = def_ar1*gZ(iZ)-4d0*def_std
                    maxqZ = def_ar1*gZ(iZ)+4d0*def_std

                    QUADz(:,iZ) = minqZ + .5d0*(maxqZ-minqZ)*(quad_x+1d0)
                    QUADw(:,iZ) = .5d0*(maxqZ-minqZ)*quad_w

                    ! Calculate the normal density and apply it to the quadrature weights
                    QUADw(:,iZ) = normpdf(QUADz(:,iZ),def_ar1*gZ(iZ),def_std)*QUADw(:,iZ) !/sum(normpdf(quad_x,def_ar1*gZ(iZ),def_std)) ! ensure probability distribution over nodes
                    QUADz(:,iZ) = exp(QUADz(:,iZ))

                end do

            end if

            gZ = exp(gZ)
            
        end if


        ! Solve, running for at least 2 minutes and averaging times when not meeting that
        vfi_total_time = 0. ! minutes
        total_time = 0d0 
        repeat_iter = 0
        ! Run for 
        do 

            ! Start the timer
            call tic()

            if (doRBC) then
                select case (opt)
                    ! Monotonicity is the first argument, concavity is the second
                    case (1); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[0,0])
                    case (2); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[0,1])
                    case (3); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[0,2])
                    case (4); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[1,0])
                    case (5); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[1,1])
                    case (6); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[1,2])
                    case (7); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[2,0])
                    case (8); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[2,1])
                    case (9); call vfiRBC(vfi_iter,cnt,vfi_time,overtime,[2,2])
                    case default; STOP 'ERROR: invalid opt'
                end select
            else
                select case (opt)
                    ! Monotonicity is the first argument, concavity is the second
                    case (1); call vfiDEF(vfi_iter,cnt,vfi_time,overtime,[0,0],tol,guess,soln)
                    case (4); call vfiDEF(vfi_iter,cnt,vfi_time,overtime,[1,0],tol,guess,soln)
                    case (7); call vfiDEF(vfi_iter,cnt,vfi_time,overtime,[2,0],tol,guess,soln)
                    case default; STOP 'ERROR: invalid opt'
                end select
            end if
            
            if (present(converged_out)) converged_out = vfi_iter<500

            vfi_total_time = vfi_total_time + vfi_time  ! initialized to 0
            total_time = total_time + [time_max%elapsed(),time_quad%elapsed(),time_tot%elapsed()]
            repeat_iter = repeat_iter + 1               ! initialized to 0 
            vfi_avg_time = vfi_total_time / repeat_iter
            avg_time = total_time / repeat_iter

            ! stop looping if more than 2 minutes or not trying to loop for time
            if (overtime .or. vfi_total_time>2d0 .or. .not. loopfortime) then
                vfi_time = vfi_avg_time
                exit 
            end if

        end do

        if (.not. overtime) then

            call toc(minutes)

            if (doextras .and. nK<=100000) then

                if (doRBC) then

                    ! Compute the invariant dsn
                    call invarRBC()

                    ! Compute Euler errors
                    call computeEEERBC(avgEEE,maxEEE)

                else

                    ! Compute the invariant dsn
                    call invarDEF(guess,soln)
                    if (estimating) return ! If doing the estimation, should be done here

                    ! Compute Euler errors
                    call computeEEEDEF(avgEEE,maxEEE,extrastats)

                end if
            else
                avgEEE = -1d100
                maxEEE = -1d100
            end if

        else
            
            write(unitno,'(e24.16,A)',advance='no') cnt,','
            stats = [cnt,0d0,0d0,0d0,1d0,real(vfi_iter,8),vfi_time]

            avgEEE = -1d100
            maxEEE = -1d100

        end if

        if (useStdOut) then
            write(*,*) 'Average count    :',cnt
            write(*,*) 'Minutes          :',minutes
            write(*,*) 'Average EEE      :',avgEEE
            write(*,*) 'Maximum EEE      :',maxEEE
            write(*,*) 'VFI iter cnt     :',vfi_iter
            write(*,*) 'Time maximizing  :',time_max%elapsed()
            write(*,*) 'Time integrating :',time_quad%elapsed()
            write(*,*) 'Time total       :',time_tot%elapsed()
        end if

        ! Output the results to a csv
        arglist = ''
        write(arg,*) model; arglist = trim(arglist)//trim(adjustl(arg))//'_'
        write(arg,*) dims(1); arglist = trim(arglist)//trim(adjustl(arg))//'_'
        write(arg,*) dims(2); arglist = trim(arglist)//trim(adjustl(arg))//'_'
        write(arg,*) opt; arglist = trim(arglist)//trim(adjustl(arg))//'_'
        write(arg,*) shocktype; arglist = trim(arglist)//trim(adjustl(arg))//'_'
        write(arg,*) maxmethod; arglist = trim(arglist)//trim(adjustl(arg))//'_'
        write(arg,*) maxTime_in; arglist = trim(arglist)//trim(adjustl(arg))

        open(newunit=unitno,file='results/stats_'//trim(arglist)//'.csv',status='replace')

        ! Replace _ with ,
        do while (scan(arglist,'_')/=0)
            i = scan(arglist,'_')
            arglist(i:i) = ','
        end do
        arglist = trim(arglist)//','

        write(6,'(A)',advance='no') trim(arglist)

        do repeat_iter = 1,2
            if (repeat_iter==2) unitno=6
            write(unitno,'(e25.17e3,A)',advance='no') cnt,','
            write(unitno,'(e25.17e3,A)',advance='no') minutes,','
            write(unitno,'(e25.17e3,A)',advance='no') avgEEE,','
            write(unitno,'(e25.17e3,A)',advance='no') maxEEE,','
            write(unitno,'(e25.17e3,A)',advance='no') merge(1d0,0d0,overtime),','
            write(unitno,'(e25.17e3,A)',advance='no') real(vfi_iter,8),','
            write(unitno,'(e25.17e3,A)',advance='no') avg_time(1),','
            write(unitno,'(e25.17e3,A)',advance='no') avg_time(2),','
            write(unitno,'(e25.17e3,A)',advance='no') avg_time(3),','
            write(unitno,'(e25.17e3,A)',advance='no') nonmonofound
            if (allocated(extrastats)) then
                do i = 1,size(extrastats)
                    write(unitno,'(A,e25.17e3)',advance='no') ',',extrastats(i)
                end do
            end if

            if (doexploit2state) then
                write(unitno,'(A,e25.17e3)',advance='no') ',',lambda
                write(unitno,'(A,e25.17e3)',advance='no') ',',lambda_alt
            end if

            write(unitno,'(A)') ' ' 

            if (repeat_iter==1) then 
                close(unitno) 
            end if

        end do

        ! Plot errors
        if (useStdOut) call saveplots()

!         call deallocatedata

    end subroutine
end module
