! Grey Gordon, 2012
! 
! For God so loved the world, that He gave His only begotten Son, 
! that whoever believes in Him shall not perish, but have eternal life.
!
module mod_minNdim
    use mod_types
    implicit none

    ! Options for minNdim 
    type minNdim_opts
        ! Commonly used ones
        integer :: printmod = 1, maxiter = 10000
        real(rt) :: xabstoler = 1e-6_rt, xreltoler = 1e-6_rt
        real(rt) :: yabstoler = 1e-10_rt, yreltoler = 1e-10_rt
        integer :: dbglvl = 0
        ! Others (maybe only a few algorithms use them)
        ! For these, -1 means there is no default specified
        real(rt) :: fcnerr = 0._rt  ! used by MINPACK routines for computing approximate jacobian
        integer :: algorithm = 1
        integer :: poolsize = -1
        integer :: sleeptime = -1 
        integer :: displaybestx = 1 ! 1==means true (the default)
        real(rt) :: mutationprob = .001_rt
        integer :: reproducible = 0 ! random default
    end type 

    private :: slave

contains

    subroutine disp_opts(o)
        implicit none
        type(minNdim_opts), intent(in) :: o
        write(*,*) 'printmod',o%printmod
        write(*,*) 'maxiter',o%maxiter
        write(*,*) 'xabstoler',o%xabstoler
        write(*,*) 'xreltoler',o%xreltoler
        write(*,*) 'yabstoler',o%yabstoler
        write(*,*) 'yreltoler',o%yreltoler
        write(*,*) 'dbglvl',o%dbglvl
        write(*,*) 'poolsize',o%poolsize
        write(*,*) 'sleeptime',o%sleeptime
        write(*,*) 'displaybestx',o%displaybestx
        write(*,*) 'mutationprob',o%mutationprob
        write(*,*) 'reproducible',o%reproducible
        write(*,*) 'algorithm',o%algorithm
        write(*,*) 'fcnerr',o%fcnerr
    end subroutine

    ! Here, options are specified via a string such as "xtoler=1d6,printmod=10000"
    function optimset(str) result(opts)
        use mod_util
        implicit none
        character(len=*), intent(in) :: str
        type(minNdim_opts) :: opts
        ! local
        character(len=len(str)) :: s,arg,val
        type(minNdim_opts) :: def
        integer :: nargs,i

        nargs = count_delim_arg(str,",")

        do i = 1,nargs

            ! Get the substring
            s = ''
            s = trim(get_delim_arg(str,",",i))

            ! Process to get what is left and right of the equal sign
            arg = ''
            val = ''
            arg = get_delim_arg(s,"=",1)
            val = get_delim_arg(s,"=",2)

            if     (trim(arg)=="xabstoler")     then; read(val,'(d30.16)',err=999) opts%xabstoler 
            elseif (trim(arg)=="xreltoler")     then; read(val,'(d30.16)',err=999) opts%xreltoler 
            elseif (trim(arg)=="yabstoler")     then; read(val,'(d30.16)',err=999) opts%yabstoler 
            elseif (trim(arg)=="yreltoler")     then; read(val,'(d30.16)',err=999) opts%yreltoler 
            elseif (trim(arg)=="mutationprob")  then; read(val,'(d30.16)',err=999) opts%mutationprob
            elseif (trim(arg)=="fcnerr")        then; read(val,'(d30.16)',err=999) opts%fcnerr
            elseif (trim(arg)=="printmod")      then; read(val,'(i25)',   err=999) opts%printmod
            elseif (trim(arg)=="maxiter")       then; read(val,'(i25)',   err=999) opts%maxiter
            elseif (trim(arg)=="poolsize")      then; read(val,'(i25)',   err=999) opts%poolsize
            elseif (trim(arg)=="sleeptime")     then; read(val,'(i25)',   err=999) opts%sleeptime
            elseif (trim(arg)=="displaybestx")  then; read(val,'(i25)',   err=999) opts%displaybestx
            elseif (trim(arg)=="dbglvl")        then; read(val,'(i25)',   err=999) opts%dbglvl
            elseif (trim(arg)=="reproducible")  then; read(val,'(i25)',   err=999) opts%reproducible
            elseif (trim(arg)=="algorithm")     then; read(val,'(i25)',   err=999) opts%algorithm
            else
                write(0,*) 'WARNING: optimset: did not find option corresponding to argument ',&
                           trim(arg),'  with value ',trim(val),' and overall input ',trim(str)
            end if

        end do

        return

999     write(0,*) 'ERROR: optimset: error encountered processing ',trim(str),', with arg ',trim(arg),' and val ',trim(val),' falling back to default arguments'
        opts = def
    

    end function

    ! Levenberg-Marquardt with a finite-difference Jacobian
    ! 
    ! options specific to LM
    ! -> fcnerr : determines the step-size in computing the gradient. stepsize is (roughly?) 10**(fcnerr/2-8). So fcnerr=16 is 100%, fcnerr=14 is 10%, fcnerr=12 is 1%, .... If too small, will just use machine epsilon
    ! 
    ! Interface to the MINPACK routine lmdif
    subroutine sub_levenbergmarquardt(xstar,f,m,opts)
        implicit none
        interface 
            function f(x,m) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                integer, intent(in) :: m
                real(rt) :: f(m)
            end function
        end interface
        real(rt), intent(in) :: xstar(:)
        integer, intent(in) :: m
        type(minNdim_opts), optional :: opts
        ! local
        type(minNdim_opts) :: o
        !!! NOTE: keep as double precision because required by MINPACK
        real(8), allocatable :: fvec(:),wa(:)
        integer, allocatable :: iwa(:)
        real(8) :: epsfcn,factor,ftol,gtol,xtol
        integer :: n,info,lwa,maxfev,mode,nfev,nprint,neval,iter

        if (present(opts)) o = opts

        neval = 0
        iter = 0
        n = size(xstar)
        lwa = m*n+5*n+m

        allocate(fvec(m))
        allocate(wa(lwa))
        allocate(iwa(n))

        info = 0
        if (n .le. 0 .or. m .lt. n .or. lwa .lt. m*n + 5*n + m) then
            write(0,*) 'WARNING: bad input param'
            return
        end if

        maxfev = o%maxiter*(n + 1)
        ftol = o%yreltoler
        xtol = o%xreltoler
        gtol = 0.d0
        epsfcn = o%fcnerr ! used to construct appropriate step-length (higher means error, and uses bigger step). if < epsilon(0), uses epsilon(0)
        mode = 1 ! internally scaled
        nprint = max(0,o%printmod)
        factor = 100d0

        call lmdif(sub_f,m,n,xstar,fvec,ftol,xtol,gtol,maxfev,epsfcn,wa(1),&
                   mode,factor,nprint,info,nfev,wa(m+5*n+1),m,iwa,& 
                   wa(n+1),wa(2*n+1),wa(3*n+1),wa(4*n+1),wa(5*n+1))

        ! call lmdif1(sub_f,m,n,xstar,fvec,tol,info,iwa,wa,lwa)

        select case (info)
            case(0); write(*,'(A)') 'LM: improper input parameters.'
            case(1); write(*,'(A)') 'LM: both actual and predicted relative reductions in the sum of squares are at most ftol.'
            case(2); write(*,'(A)') 'LM: relative error between two consecutive iterates is at most xtol.'
            case(3); write(*,'(A)') 'LM: conditions for info = 1 and info = 2 both hold.'
            case(4); write(*,'(A)') 'LM: the cosine of the angle between fvec and any column of the jacobian is at most gtol in absolute value.'
            case(5); write(*,'(A)') 'LM: number of calls to fcn has reached or exceeded maxfev.'
            case(6); write(*,'(A)') 'LM: ftol is too small. no further reduction in the sum of squares is possible.'
            case(7); write(*,'(A)') 'LM: xtol is too small. no further improvement in the approximate solution x is possible.'
            case(8); write(*,'(A)') 'LM: gtol is too small. fvec is orthogonal to the columns of the jacobian to machine precision.'
        end select

    contains 

        subroutine sub_f(m,n,x,fvec,iflag)
            use mod_util, only: printx
            implicit none
            integer :: m,n,iflag
            real(8) :: x(n),fvec(m)

            if (iflag==0) then
                write(*,'(6A,2(A,i5))')'LM: x = ',printx(x),' f(x) = ',printx(fvec),' SSRS = ',printx([norm2(fvec)]),' iter ',iter,' fevals ',neval
                iter = iter + nprint
            else
                fvec = f(x,m)
                neval = neval + 1
            end if

        end subroutine 

    end subroutine 

    ! Optimizes only along a given dimension
    subroutine sub_opt_along_dim(xstar,f,a,b,dim,opts,ngridsearch)
        use mod_min1dim
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a,b
        integer, intent(in) :: dim
        type(minNdim_opts), optional :: opts
        integer, intent(in), optional :: ngridsearch
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local
        integer :: nx, ngs, flag
        real(rt), allocatable :: x(:)
        real(rt) :: x1
        type(minNdim_opts) :: o

        ngs = 100
        if (present(ngridsearch)) ngs = ngridsearch

        if (present(opts)) o = opts


        nx = size(xstar)

        if (dim<0 .or. dim>nx) stop 'ERROR: sub_opt_along_dim: invalid dim'

        allocate(x(nx))

        x = xstar

        print*,'sub_opt_along_dim: initial guess: ',xstar(dim)

        x1 = xstar(dim) 
        call sub_gridsearch1dim_nogrid(x1,g,a,b,ngs,printmod=o%printmod)
        call sub_simplex1dim(x1,g,e=.75d0*(b-a)/(ngs-1),xtoler=o%xabstoler,maxiter=o%maxiter,printmod=o%printmod,flag=flag) 
        ! Note: don't use brent as current setup doesn't use initial information for x1

        xstar(dim) = x1

        print*,'sub_opt_along_dim: final guess: ',xstar(dim)

    contains

        function g(x1) 
            use mod_types
            implicit none
            real(rt), intent(in) :: x1
            real(rt) :: g

            x(dim) = x1
            g = f(x)

        end function 


    end subroutine

    ! Controlled random search --- serial version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. 
    subroutine sub_CRS(xstar,f,a,b,opts)
        use mod_util
        use mod_matlab
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        type(minNdim_opts), optional :: opts
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable :: X(:,:),y(:) 
        real(rt), allocatable :: R(:),g(:),smx(:,:),tmpy(:)
        integer, allocatable :: tmpI(:)
        real(rt) :: tmp, yR
        integer :: nx, N, M, ix, iter, i, ism, itmp(1)
        logical :: dostop
        type(minNdim_opts) :: o

        if (present(opts)) o = opts
    

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_CRS: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_CRS: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_CRS: a>=b at least once'

        ! Check that input x is in a,b for all i
        if (any(xstar<a) .or. any(xstar>b)) stop 'ERROR: sub_CRS: x is not in a,b for all dimensions'

        ! M is the number to simultaneously try to get rid of
        M = 1

        ! N is number of trial points (the pool of points)
        N = max(5*M, ((100+nx)/M+1)*M, nx+2+M)
        if (o%poolsize/=-1) then
            N = max(N,o%poolsize)
        end if

        if (N-M<=nx+1) stop 'ERROR: sub_CRS: too few N points set relative to M'

        ! Generate N random points uniformly from [a,b]
        allocate(X(nx,N))
        allocate(y(N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    
        allocate(tmpy(N),tmpI(N))

        ! Generate initial pool of points
        call randstate%save()
        call randstate%randomize()
        X(:,1) = xstar
        do i = 2,N
            do ix = 1,nx
                call random_number(tmp)
                X(ix,i) = tmp*a(ix) + (1e0_rt-tmp)*b(ix)
            end do
        end do
        call randstate%revert()

        ! Evaluate at each of the N points in parallel
        if (o%printmod>0 .and. o%printmod==1) write(*,'(A,i4,A)') 'CRS: Solving for initial pool of ',N,' points'
        do i = 1,N

            if (o%printmod>=1 .and. mod(i,o%printmod)==0 .and. o%displaybestx==1) &
                write(*,'(A,i4,A,i4,A,A)') 'CRS: solving for point ',i,' of ',N,' x = ',printx(X(:,i))
            if (o%printmod>=1 .and. mod(i,o%printmod)==0 .and. o%displaybestx==0) &
                write(*,'(A,i4,A,i4)')     'CRS: solving for point ',i,' of ',N

            ! Evaluate the point
            y(i) = f(X(:,i))

        end do

        call randstate%save()
        call randstate%randomize()
        call init_random_seed(N) ! in case f(.) has reset the seed, set a new one
        call randstate%revert()

        if (o%dbglvl>1) print*,'done initial point distribution, proceeding to main loop'

        do iter = 1,o%maxiter

            ! Sort the current points (ascending)
            call sort2(tmpy,tmpI,y) 
            X = X(:,tmpI) ! sort X
            y = tmpy ! sort y

            ! Display progress
            if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                if (o%displaybestx==1) &
                    write(*,'(A,i8,2(A,e14.6),2A)') 'CRS: iter',iter,'  best ',y(1),'  worst ',y(N),'  xbest = ',printx(X(:,1))
                if (o%displaybestx==0) &
                    write(*,'(A,i8,2(A,e14.6))')    'CRS: iter',iter,'  best ',y(1),'  worst ',y(N)
            end if
 
            ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
            xstar = X(:,1)

            ! Check stopping criteria
            dostop = abs(y(1)-y(N))<o%yabstoler .or. sqrt(sum((x(:,1)-x(:,N))**2))<o%xabstoler

            if (dostop .or. iter==o%maxiter) then

                if (dostop) then
                    if (o%printmod==1) write(*,'(A)') 'CRS: convergence criteria satisfied'
                else
                    if (o%printmod==1) write(*,'(A)') 'CRS: max iterations reached'
                end if

                return

            end if

            ! Form a random simplex excluding current worst point 
            do ism = 1,nx+1
                
                ! CRS with local mutation where simplex always includes best point
                if (ism==1) then
                    smx(:,ism) = x(:,1) 
                else
                    smx(:,ism) = x(:,randi(1,N-1))
                end if

            end do

            ! Compute the centroid
            g = sum(smx,2)/(nx+1)

            ! Reflect the worst point
            R = 2e0_rt*g - X(:,N) ! pure reflection

            ! If goes through boundary, then bound at boundary
            R = max(R,a)
            R = min(R,b)

            ! To make more local, can 50% of the time combine with local search which is convex combination with best point
            call randstate%save()
            call randstate%randomize()
            call random_number(tmp)
            if (tmp<=.5e0_rt) then
                do ix = 1,nx
                    call random_number(tmp)
                    R(ix) = tmp*x(ix,1) + (1e0_rt-tmp)*R(ix) ! convex combination, so in [a,b]
                end do
            end if
            call randstate%revert()
            
            ! Compute the reflected point
            yR = f(R)

            ! Replace the worst point if found a better one
            if (yR<maxval(y)) then
                itmp = maxloc(y) 
                y(itmp(1)) = yR
                x(:,itmp(1)) = R
            end if

        end do

    end subroutine

    subroutine check_guess_in_range(flagout,x,a,b)
        implicit none
        real(rt), intent(in) :: x(:),a(:),b(:)
        integer, intent(out) :: flagout
        ! local
        integer :: i,n,flag

        n = size(x)
        if (size(a)/=n) stop 'ERROR: check_guess_in_range: a and x do not have same length'
        if (size(b)/=n) stop 'ERROR: check_guess_in_range: b and x do not have same length'

        flag = 0
        do i = 1,n 
        if (x(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: x<a for element',i; flag = merge(-1,-999,flag==0); end if
        if (x(i)>b(i)) then; write(0,*) 'WARNING: check_guess_in_range: x>b for element',i; flag = merge(-2,-999,flag==0); end if
        if (b(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: b<a for element',i; flag = merge(-3,-999,flag==0); end if
        end do
        flagout = flag

    end subroutine 

    ! Controlled random search --- MPI version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. 
    !
    ! If e_in is present, then random draws will not be taken uniformly from [a,b], but rather
    ! from [xstar-e,xstar+e] (component-wise). Combined with a small poolsize (which can be set 
    ! to 0 which will then be overridden), this makes the routine effectively like a parallel 
    ! version of Nelder-Mead.
    subroutine sub_parCRS_MPI(xstar,f,a,b,e_in,opts)
        use mod_util
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), intent(in), optional :: e_in(:)
        type(minNdim_opts), intent(in), optional :: opts
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable, target :: X(:,:),y(:),slavexlist(:,:),slaveylist(:) ! NOTE: these have target attribute and will be pointed at
        logical, allocatable :: finishedtask(:),slaveeverassignedlist(:)
        real(rt), allocatable :: R(:),g(:),smx(:,:),tmpy(:)
        integer, allocatable :: tmpI(:), fevaldsn(:),requestlist(:)
        real(rt) :: tmp
        integer :: nx, N, M, ix, iter, i, ism, who, itmp(1), flag, minPoolSize, effPoolSize, effIter
        type(minNdim_opts) :: o

        if (present(opts)) o = opts

        if (mympi_nproc()<=1) then
            write(0,*) 'WARNING: parCRS MPI version was called with only one rank, calling serial version instead'
            call sub_CRS(xstar,f,a,b,o)
            return
        end if

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_parCRS2LM: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_parCRS2LM: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_parCRS2LM: a>=b at least once'

        ! Check that input x is in a,b for all i
        call check_guess_in_range(flag,xstar,a,b)
        if (flag/=0) stop 'ERROR: sub_parCRS2LM: x is not in a,b or b<a for some dimension'

        ! M is the number to simultaneously try to get rid of
        M = mympi_nproc()-1

        ! N is number of trial points (the pool of points)
        ! N = max(2*M, ((100+nx)/M+1)*M, nx+2+M)
        N = max(((100+nx)/M+1)*M, nx+2+M)
        ! N = ((100+nx)/M+1)*M ! roughly evenly divide 100+nx points across the M ranks
        if (o%poolsize/=-1) N = max(nx+2+M,o%poolsize)

        minPoolSize = max(nx+1,N-mympi_nproc(),M)

        if (N-M<=nx+1) call mympi_stop('ERROR: sub_parCRS2LM: too few N points set relative to M')

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave(f,nx)
            call mympi_bcast(xstar,0) ! receive the best point from 
            call mympi_barrier()
            return
        end if

        ! Generate N random points uniformly from [a,b]
        allocate(finishedtask(mympi_nproc()-1))
        allocate(X(nx,N))
        allocate(y(N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    
        allocate(tmpy(N),tmpI(N))
        allocate(fevaldsn(mympi_nproc()-1))
        allocate(requestlist(mympi_nproc()-1))
        allocate(slavexlist(nx,mympi_nproc()-1))
        allocate(slaveylist(mympi_nproc()-1))
        allocate(slaveeverassignedlist(mympi_nproc()-1))

        fevaldsn = 0

        finishedtask = .true.
        slaveeverassignedlist = .false.

        ! Evaluate at each of the N points in parallel
        effPoolSize = 0
        write(*,'(A,i8,A,i8)') 'PARCRS: Solving for initial pool of ',N,' points; min pool size is ',minPoolSize
        do i = 1,N

            ! Wait for a free worker
            if (o%dbglvl>0) print*,'MASTER: WAITING FOR FREE WORKER'
            who = wait_for_free_worker()

            ! If they had a previously assigned point, use its value
            if (slaveeverassignedlist(who)) then
                ! Add to the pool
                effPoolSize = effPoolSize + 1
                X(:,effPoolsize) = slavexlist(:,who)
                Y(effPoolsize) = slaveylist(who)

                if (isnan(y(effPoolSize))) then
                    write(0,*) 'WARNING: sub_parCRS_MPI: function returned a NaN value for x :',printx(X(:,effPoolSize)),' replacing with huge(0d0)'
                    y(effPoolSize) = huge(0d0)
                end if

                if (o%printmod>0 .and. mod(effPoolSize,o%printmod)==0) then
                    itmp = minloc(Y(1:effPoolSize)) 
                    write(*,'(A,i8,2(A,ES14.6),2A)') 'PARCRS: pool ',effPoolSize,' trial ',y(effPoolSize),' best ',y(itmp(1)),'  xbest = ',printx(X(:,itmp(1)))
                end if
            end if

            ! Generate a random point and assign it
            call randstate%save()
            call randstate%randomize()
            slaveeverassignedlist(who) = .true.
            if (i==1) then
                slavexlist(:,who) = xstar
            else
                do ix = 1,nx
                    if (.not. present(e_in)) then
                        call random_number(tmp)
                        slavexlist(ix,who) = tmp*a(ix) + (1e0_rt-tmp)*b(ix)
                    else
                        call random_number(tmp)
                        tmp = 2._rt*tmp-1
                        slavexlist(ix,who) = min(b(ix),max(a(ix),xstar(ix)+tmp*e_in(ix)))
                    end if
                end do
            end if
            call randstate%revert()
            call assign_xy_to_worker(slavexlist(:,who),slaveylist(who),who)

            ! For statistical purposes, collect who was assigned that feval
            fevaldsn(who) = fevaldsn(who) + 1

        end do

        if (o%dbglvl>1) print*,'waiting for everyone to finish'

        ! Wait for everyone to finish
        ! do while (any(.not. finishedtask))
        do while (effPoolsize<minPoolSize) 

            do who = 1,mympi_nproc()-1
                if (.not. finishedtask(who)) then
                    finishedtask(who) = mympi_test(requestlist(who)) ! requestlist is not valid after mympi_test returns true
                    if (finishedtask(who)) then
                        ! Add to the pool
                        effPoolSize = effPoolSize + 1
                        X(:,effPoolsize) = slavexlist(:,who)
                        Y(effPoolsize) = slaveylist(who)
                        if (isnan(y(effPoolSize))) then
                            write(0,*) 'WARNING: sub_parCRS_MPI: function returned a NaN value for x :',printx(X(:,effPoolSize)),' replacing with huge(0d0)'
                            y(effPoolSize) = huge(0d0)
                        end if
                    end if
                end if
            end do

        end do

        write(*,'(A,i8,A)') 'PARCRS: Achieved pool size of ',effPoolSize,' proceeding to main loop, pool size will be increased in main loop'

        if (o%dbglvl>1) print*,'done initial point distribution, proceeding to main loop'

        effIter = 0

        do iter = 1,o%maxiter

            ! Wait for a slave. If there were previously assigned a task, then record its value.
            who = wait_for_free_worker()
!             print*,'iter',iter,'who',who

            ! If they had a previously assigned point, use its value
            if (slaveeverassignedlist(who)) then

                if (isnan(slaveylist(who))) then
                    write(0,*) 'WARNING: sub_parCRS_MPI: function returned a NaN value for x :',printx(slavexlist(:,who)),' replacing with huge(0d0)'
                    slaveylist(who) = huge(0d0)
                end if
                
                if (effPoolSize<N) then
                    ! If not at desired pool size yet, just add it to the pile
                    effPoolSize = effPoolSize + 1
                    X(:,effPoolsize) = slavexlist(:,who)
                    Y(effPoolsize) = slaveylist(who)

                else

                    ! Otherwise, replace the worst point if found a better one
                    itmp = maxloc(y(1:effPoolSize)) 
                    if (slaveylist(who)<y(itmp(1))) then
                        X(:,itmp(1)) = slavexlist(:,who)
                        Y(itmp(1)) = slaveylist(who)
                    end if

                end if

            end if

            effIter = merge(effIter+1,effIter,effPoolSize==N)

            ! After this point, "who" will be assigned an x
            slaveeverassignedlist(who) = .true.

            ! Sort the current points (ascending)
            call sort2(tmpy(1:effPoolSize),tmpI(1:effPoolSize),y(1:effPoolSize)) 
            X(:,1:effPoolSize) = X(:,tmpI(1:effPoolSize)) ! sort X
            y(1:effPoolSize) = tmpy(1:effPoolSize) ! sort y

            ! Display progress
            if (iter==1 .or. (o%printmod>0 .and. mod(iter,o%printmod)==0)) then
                
                if (effPoolSize/=N .or. effIter==0) then
                    write(*,'(A,i8,2(A,ES14.6),2A)') 'PARCRS: pool ',effPoolSize,' worst ',y(effPoolSize),' best ',y(1),'  xbest = ',printx(X(:,1))
!                     write(*,'(A,*(i4))') 'fevaldsn',fevaldsn
                else
                    write(*,'(A,i8,2(A,ES14.6),2A)') 'PARCRS: iter ',effIter,' worst ',y(effPoolSize),' best ',y(1),'  xbest = ',printx(X(:,1))
                end if

            end if
 
            ! Check stopping criteria
            if (effPoolSize==N .and. abs(y(1)-y(effPoolSize))<o%yabstoler .or. iter==o%maxiter) then

                ! In case exit either through maxiter being reached or satisfying exit criteria, save xstar
                xstar = X(:,1)

                if (abs(y(1)-y(effPoolSize))<o%yabstoler) then
                    write(*,'(A)') 'PARCRS: convergence criteria satisfied'
                else
                    write(*,'(A)') 'PARCRS: max iterations reached'
                end if

                write(*,'(*(A,i5))') 'PARCRS: rank eval. count; min: ',minval(fevaldsn),' max: ',maxval(fevaldsn),' avg: ',sum(fevaldsn)/size(fevaldsn)

                ! Shutdown slave
                call shutdown_slaves()

                ! Distribute xbest
                call mympi_bcast(xstar,0)

                ! Wait for everyone to finish 
                call mympi_barrier()

                return

            end if

            ! Form a random simplex excluding current worst point 
            call randstate%save()
            call randstate%randomize()
            do ism = 1,nx+1
                
                ! CRS with local mutation where simplex always includes best point
                if (ism==1) then
                    smx(:,ism) = x(:,1) 
                else
                    smx(:,ism) = x(:,randi(1,effPoolSize-1))
                end if

            end do

            ! Compute the centroid
            g = sum(smx,2)/(nx+1)

            ! Reflect the worst point
            R = 2._rt*g - X(:,effPoolSize) ! pure reflection

            ! If goes through boundary, then bound at boundary
            R = min(max(R,a),b)

            ! To make more local, can 50% of the time combine with local search which is convex combination with best point
            call random_number(tmp)
            if (tmp<=.5e0_rt) then
                do ix = 1,nx
                    call random_number(tmp)
                    R(ix) = tmp*x(ix,1) + (1e0_rt-tmp)*R(ix) ! convex combination, so in [a,b]
                end do
            end if
            call randstate%revert()
            
            ! Request that they solve for x and store it in slavey
            slavexlist(:,who) = R
            call assign_xy_to_worker(slavexlist(:,who),slaveylist(who),who)

            fevaldsn(who) = fevaldsn(who) + 1

        end do

    contains

        ! This routine sends the value of x to a slave "who", who will then store the value in y. 
        subroutine assign_xy_to_worker(x,y,who)
            use iso_c_binding, only: c_loc,c_ptr,c_f_pointer ! GETS THE ACTUALY ADDRESS OF Y WHICH WILL PERSIST AFTER CALL TO ASSIGN ENDS
            implicit none
            real(rt), intent(in), target :: x(:)
            real(rt), intent(out), target :: y
            integer, intent(in) :: who ! who you want assigned
            ! local
            type(c_ptr) :: yptr_c ! pointer in C
            real(rt), pointer :: yptr_f ! pointer in Fortran 
            integer :: flag,ierr

            ! Assign it
            flag = 1
            finishedtask(who) = .false.

            call mympi_send(flag,who) ! blocking send (tell them to proceed to execute)

            call mympi_send(x,who) ! assign an x value (wait for it to complete)

            ! In order to be able to do this assignment where the memory location associated with y will be stored even after 
            ! the subroutine is exited, the safest/most reliable way I have found is to send MPI the actual memory address.
            ! This can be accomplished by getting the memory address of y using the F2003 intrinsic "c_loc" and assigning that
            ! to a C pointer. Then, using c_f_pointer, copy that to a Fortran pointer. Then pass that Fortran poitner to MPI. 
            !
            ! This may be a hack (it seems like completely legal Fortran though), but it simplifies bookkeeping substantially.
            ! I.e., it becomes "set it and forget it."

            yptr_c = c_loc(y) ! get the actual address of y using "c_loc" and have yptr_c point to that
            call c_f_pointer(yptr_c,yptr_f) ! get a fortran pointer to that same address
            call MPI_Irecv(yptr_f, 1, MPI_DOUBLE_PRECISION, who, who, MPI_COMM_WORLD, requestlist(who), ierr)

        end subroutine 

        ! a routine to shut down all slaves
        subroutine shutdown_slaves() 
            implicit none
            ! local
            integer :: i
            logical :: ft

            ! Cycle through and shutdown
            do i = 1,mympi_nproc()-1

                ! Wait for rank to finish if not finished task
                if (.not. finishedtask(i)) then

                    ! Infinite loop, wait for them to finish 
                    do 

                        ! Check whether i has finished
                        ft = mympi_test(requestlist(i))
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            exit
                        end if

                        ! If sleeptime, then sleep
                        if (o%sleeptime>=1) then

                            call sleep(o%sleeptime)

                        else

                            ! Do some busy work so as to not thrash the system
                            call short_sleep()

                        end if

                    end do

                end if
                    
                ! Shut down, sending flag = -1 and xstar
                call mympi_send(-1,i) 

            end do

        end subroutine


        ! a routine returning the index of a free slave
        ! if no worker is available, who is -1
        function wait_for_free_worker() result(who)
            implicit none
            integer :: who
            ! local
            integer :: i
            logical :: ft

            do 

                ! Infinite loop
                do i = 1,mympi_nproc()-1

                    if (finishedtask(i)) then

                        who = i
                        return  

                    else

                        ! Check whether i has finished
                        if (o%dbglvl>1) print*,'MASTER: TESTING MESSAGE'
                        if (o%dbglvl>2) print*,'MASTER: finishedtask',finishedtask
                        if (o%dbglvl>2) print*,'MASTER: REQUESTLIST',REQUESTLIST
                        if (o%dbglvl>2) print*,'MASTER: i: ',i

                        ft = mympi_test(requestlist(i))
                        if (o%dbglvl>1) print*,'MASTER: FINISHED TEST'
                        if (ft) then

                            finishedtask(i) = .true. ! show that they are now free


                            who = i
                            if (o%dbglvl>1) print*,'MASTER: RECEIVED MESSAGE'
                            return
                        end if

                    end if

                end do

                ! If sleeptime, then sleep
                if (o%sleeptime>=1) then

                    call sleep(o%sleeptime)

                else

                    ! Do some busy work so as to not thrash the system
                    call short_sleep()

                end if

            end do

        end function


    end subroutine

    ! Do something according to the value of flag(this_image())[1]
    !  1 := evaluate slavex and store in slavey
    ! -1 := exit the slave routine
    subroutine slave(f,nx)
        use mod_mympi
        implicit none
        integer, intent(in) :: nx
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local
        integer :: flag
        real(rt) :: slavey
        real(rt), allocatable :: slavex(:)

        allocate(slavex(nx))

        do 

            ! Wait for a flag 
            call mympi_recv(flag,0)

            ! If flag==-1, then shutdown
            if (flag==-1) then
                return
            end if

            ! If still here, wait for an x
            call mympi_recv(slavex,0)

            ! If flag==1, execute f(x) and store on master.
            if (flag==1) then

                ! Compute the result
                slavey = f(slavex)

                ! Send the result to master tagged by mympi_iproc()
                call mympi_send(slavey,0,mympi_iproc()) 

            end if

        end do

    end subroutine

    ! Generate a random integer in a,...,b
    function randi(a,b) result(i)
        implicit none
        integer, intent(in) :: a,b
        integer :: i
        real(rt) :: tmp

        ! Generate a U[a,b+1-epsilon]
        ! Then floor to get a in 
        call random_number(tmp)
        tmp = tmp*(b+1-a) + a ! U[a,b+1]
        tmp = min(tmp,b+.99999999e0_rt)
        i = floor(tmp)

        if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'

    end function


!     ! Noisy Levenberg-Marquardt with MPI
!     !
!     ! solve min_{x \in R^{nx}} \sum^{nf}_{j=1} f^j(x)
!     !
!     ! This is a routine I made up (pdf is in noisyLM folder).
!     ! Compute a linear projection of the function at hand via OLS and random points to obtain an 
!     ! approximate Jacobian. Then this Jacobian is used in the usual Levenberg-Marquardt procedure.
!     ! 
!     subroutine sub_parnoisyLM(x,f,nf,stdev,ndraws,opts)
!         use mod_mympi
!         use mod_matlab
!         implicit none
!         integer, intent(in) :: nf
!         real(rt), dimension(:), intent(inout) :: x
!         interface 
!             ! f is vector valued taking x (a length nx vector) to a length nf vector.
!             function f(x,nf) 
!                 implicit none
!                 integer, intent(in) :: nf
!                 real(rt), dimension(:), intent(in) :: x
!                 real(rt), dimension(nf) :: f
!             end function
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: stdev
!         integer, intent(in), optional :: ndraws
!         type(minNdim_opts), optional :: opts
!         !local
!         type(minNdim_opts) :: o
!         real(rt), allocatable :: stdev_(:), xtrial(:), e_(:,:), Ymattranspose(:,:), Ymat(:,:), Xmat(:,:), J(:,:), fx(:), gL(:), s(:), eyex(:,:), ftrial(:), xtrials(:,:), ftrials(:,:)
!         integer, allocatable :: request(:)
!         logical, allocatable :: isFree(:)
!         integer :: iter, nx, itmp(1), nL, iL, idraw, ndraws_, iproc
!         logical :: haveRootDoIter1
!         real(rt) :: gradSSRSnorm,stepnorm,lambda,SSRSx,tmpSSRS,lambdabest
! 
!         if (mympi_nproc()==1) then
!             write(*,*) 'WARNING: If only using one node, call sub_noisyLM instead'
!             write(*,*) 'WARNING: Calling it for you with default parameters'
!             call sub_noisyLM(x,f,nf)
!             return
!         end if
! 
!         nx = size(x)
!         allocate(fx(nf))
! 
!     if (mympi_iproc()==0) then
! 
!         if (present(opts)) o = opts
! 
!         ndraws_ = 100 ! 10*nx
!         if (present(ndraws)) ndraws_ = ndraws
! 
!         if (ndraws_<mympi_nproc()-1) then
!             ! Increase ndraws so don't waste computation
!             write(*,*) 'WARNING: Increasing ndraws to not waste nodes'
!             ndraws_ = mympi_nproc()-1 
!         end if
!             
!         ! Set nL so that node nodes are wasted and that at least 12 points are evaluated
!         nL = max(12,mympi_nproc()-1)
! 
!         allocate(s(nx))
!         allocate(gL(nL))
!         allocate(stdev_(nx))
!         allocate(xtrial(nx))
!         allocate(ftrial(nf))
!         allocate(xtrials(nx,nL))
!         allocate(ftrials(nf,nL))
!         allocate(e_(nx,ndraws_))
!         allocate(Ymat(ndraws_,nf))
!         allocate(Ymattranspose(nf,ndraws_))
!         allocate(Xmat(ndraws_,nx))
!         allocate(eyex(nx,nx))
!         allocate(isFree(1:mympi_nproc()-1))
!         allocate(request(1:mympi_nproc()-1))
! 
!         isFree = .true.
!         eyex = eye(nx)
! 
!         if (nL==12) then
!             gL(1:12) = [0e0_rt,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1e0_rt,1d1,1d2,1d3,1d4]
!         else
!             gL(1) = 0e0_rt
!             gL(2:nL) = logspace(-6e0_rt,4e0_rt,nL-1)
!         end if
! 
!         ! Give each process a random seed based on the clock
!         call init_random_seed(mympi_iproc())
! 
!         !Set default values
!         stdev_ = 1d-3
! 
!         !Overwrite if applicable
!         if (present(stdev)) stdev_ = stdev
! 
!         ! Determine whether the root node should compute f on the first iteration
!         ! This is done purely to avoid wasting resources because the first iteration requires 
!         ! ndraws_+1 function evaluations whereas every subsequent iteration requires ndraws_.
!         haveRootDoIter1 = ndraws_ == mympi_nproc()-1
! 
!         ! Main loop
!         do iter = 1,o%maxiter
! 
!             ! Compute draws
!             do idraw = 1,ndraws_
!                 e_(:,idraw) = stdev_*randn(nx)
!             end do
! 
!             ! Compute function values
!             do idraw = 1,merge(ndraws_+1,ndraws_,iter==1)
! 
!                 if (idraw==ndraws_+1) then
!                     ! Special case: first iteration, need to compute f of initial guess
!                     xtrial = x
!                 else
!                     ! Get xtrial
!                     xtrial = x + e_(:,idraw)
!                 end if
! 
!                 ! Compute the trial point
!                 if (idraw==ndraws_+1 .and. haveRootDoIter1) then
! 
!                     fx = f(xtrial,nf)
! 
!                 else
! 
!                     ! If there are no free nodes, wait for one to be free
!     ! write(*,'(A)',advance='no') '|'
!                     do while (.not. any(isFree))
! 
!                         ! Test each node to see if free
!                         do iproc = 1,mympi_nproc()-1
!                             
!                             ! Check whether finished
!                             isFree(iproc) = mympi_test(request(iproc))
! 
!                         end do
!                         
!                     end do
! 
!                     ! Use the free node
!                     if (any(isFree)) then
! 
!                         ! If there is a free node, get it
!                         iproc = find1(isFree,'first')
!                     
!                     else
!                         
!                         call mympi_stop('ERROR: could not find free node')
!                         
!                     end if
! 
!                     ! Send free node xtrial
!                     call mympi_send(xtrial,iproc)
! 
!                     ! Request the value of f(xtrial,nf) in Ymat(idraw,:)
!                     if (idraw==ndraws_+1) then
!                         ! print*,mympi_iproc(),': non-blocking receive of fx from ',iproc
!                         call mympi_irecv(request(iproc),fx,iproc)
!                     else
!                         ! print*,mympi_iproc(),': non-blocking receive of Ymattran from ',iproc
!                         call mympi_irecv(request(iproc),Ymattranspose(:,idraw),iproc)
!                     end if
! 
!                     ! Mark as not free 
!                     isFree(iproc) = .false.
! 
!                 end if
! 
!             end do
! 
!             ! Force all requests to complete before continuing
!             ! print*,mympi_iproc(),': waiting for all processes to complete'
!             call mympi_waitall(pack(request,.not. isFree))
!             isFree = .true.
! 
!             ! If first iteration, save SSRSx
!             if (iter==1) then
!                 SSRSx = dot_product(fx,fx)
!             end if
! 
!             ! Evaluate f at the trial points
!             Ymat = transpose(Ymattranspose)
!             Xmat = transpose(e_)
!             J = transpose(mldivide(matmul(transpose(Xmat),Xmat),matmul(transpose(Xmat),Ymat)))
! 
!             gradSSRSnorm = sqrt(dot_product(matmul(transpose(J),fx),matmul(transpose(J),fx)))
! 
!             ! For lambda on the line search, compute x+s and f(x+s)
!             do iL = 1,nL
! 
!                 lambda = gL(iL)
! 
!                 ! s = - mldivide(matmul(transpose(J),J) + lambda*eyex, matmul(transpose(J),fx))
!                 s = - mldivide(matmul(transpose(J),J) + lambda*diag(diag(matmul(transpose(J),J))), matmul(transpose(J),fx))
! 
!                 xtrials(:,iL) = x + s
! 
!                 ! If there are no free nodes, wait for one to be free
!                 do while (.not. any(isFree))
! 
!                     ! Test each node to see if free
!                     do iproc = 1,mympi_nproc()-1
!                         
!                         ! Check whether finished
!                         isFree(iproc) = mympi_test(request(iproc))
! 
!                     end do
!                     
!                 end do
! 
!                 ! If there is a free node, get it
!                 iproc = find1(isFree,'first')
! 
!                 ! Send the free node xtrial
!                 call mympi_send(xtrials(:,iL),iproc)
! 
!                 ! Request the value of f(xtrial,nf) in ftrials(:,iL)
!                 call mympi_irecv(request(iproc),ftrials(:,iL),iproc)
! 
!                 ! Mark as not free 
!                 isFree(iproc) = .false.
! 
!                 if (iL==nL) then
! 
!                     ! Wait for everyone to finish
!                     call mympi_waitall(pack(request,.not. isFree))
!                     isFree = .true.
! 
!                 end if
! 
!             end do 
! 
!             itmp = minloc(sum(ftrials**2,dim=1))
!             tmpSSRS = sum(ftrials(:,itmp(1))**2)
! 
!             if (tmpSSRS<SSRSx) then
! 
!                 ! Best line search is better than current best
!                 lambdabest = gL(itmp(1))
!                 s = xtrials(:,itmp(1)) - x
!                 stepnorm = sqrt(dot_product(s,s))
!                 x = xtrials(:,itmp(1))
!                 fx = ftrials(:,itmp(1))
!                 SSRSx = tmpSSRS
! 
!             else
! 
!                 ! Keep old value
!                 lambdabest = -1e0_rt
!                 stepnorm = 0e0_rt
! 
!             end if
! 
!             if (o%printmod>0 .and. mod(iter,o%printmod)==0) write(*,'(A,i8,4(A,d12.5))') ' NLM: iter',iter, ' step norm ',stepnorm,' SSRS ',SSRSx,' SSRS grad norm', gradSSRSnorm,' lambda ', lambdabest
! 
!             ! Check convergence criteria
!             if (stepnorm<o%xtoler .and.  gradSSRSnorm<o%xtoler) exit
!         
!         end do
! 
!         ! Tell every node to stop by sending them -huge(0e0_rt)
!         do iproc = 1,mympi_nproc()-1
!             xtrial = -huge(0e0_rt)
!             call mympi_send(xtrial,iproc)
!         end do
! 
!     else
!         
!         ! If not the root node, enter an infinite loop where you wait to receive an x value. Compute f(x) when you do.
!         do 
! 
!             ! Receive from root
!             ! print*,mympi_iproc(),': receiving x from root'
!             call mympi_recv(x,0) 
! 
!             ! Test whether x is less than -1d100, signaling it is time to stop the loop
!             if (all(x<-1d100)) exit
! 
!             ! Evaluate f
!             fx = f(x,nf)
! 
!             ! Send the value back to root
!             ! print*,mympi_iproc(),': sending fx to root'
!             call mympi_send(fx,0)
! 
!         end do
! 
! 
!     end if
! 
!         ! Send x from root to every node
!         call mympi_bcast(x,0)
! 
!     end subroutine 
! 
!     
!     ! Noisy Levenberg-Marquardt
!     !
!     ! solve min_{x \in R^{nx}} \sum^{nf}_{j=1} f^j(x)
!     !
!     ! This is a routine I made up (pdf is in noisyLM folder).
!     ! Compute a linear projection of the function at hand via OLS and random points to obtain an 
!     ! approximate Jacobian. Then this Jacobian is used in the usual Levenberg-Marquardt procedure.
!     ! 
!     subroutine sub_noisyLM(x,f,nf,stdev,ndraws,opts)
!         use mod_mympi
!         use mod_matlab
!         implicit none
!         integer, intent(in) :: nf
!         real(rt), dimension(:), intent(inout) :: x
!         interface 
!             ! f is vector valued taking x (a length nx vector) to a length nf vector.
!             function f(x,nf) 
!                 implicit none
!                 integer, intent(in) :: nf
!                 real(rt), dimension(:), intent(in) :: x
!                 real(rt), dimension(nf) :: f
!             end function
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: stdev
!         integer, intent(in), optional :: ndraws
!         type(minNdim_opts), optional :: opts
!         !local
!         type(minNdim_opts) :: o
!         real(rt), allocatable :: stdev_(:), xtrial(:), e_(:,:), Ymat(:,:), Xmat(:,:), J(:,:), fx(:), gL(:), s(:), eyex(:,:), ftrial(:), fxbest(:), xbest(:)
!         integer :: iter, nx, nL, iL, idraw, ndraws_
!         real(rt) :: gradSSRSnorm,stepnorm,lambda,SSRSx,tmpSSRS,lambdabest,SSRSbest
! 
!         if (present(opts)) o = opts
! 
!         nx = size(x)
! 
!         ndraws_ = 100 
!         if (present(ndraws)) ndraws_ = ndraws
! 
!         nL = 12
! 
!         allocate(s(nx))
!         allocate(xbest(nx))
!         allocate(fx(nf))
!         allocate(fxbest(nf))
!         allocate(gL(nL))
!         allocate(stdev_(nx))
!         allocate(xtrial(nx))
!         allocate(ftrial(nf))
!         allocate(e_(nx,ndraws_))
!         allocate(Ymat(ndraws_,nf))
!         allocate(Xmat(ndraws_,nx))
!         allocate(eyex(nx,nx))
! 
!         eyex = eye(nx)
! 
!         if (nL==12) then
!             gL(1:12) = [0e0_rt,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1e0_rt,1d1,1d2,1d3,1d4]
!         else
!             gL(1) = 0e0_rt
!             gL(2:nL) = logspace(-6e0_rt,4e0_rt,nL-1)
!         end if
! 
!         ! Give each process a random seed based on the clock
!         call init_random_seed(mympi_iproc())
! 
!         !Set default values
!         stdev_ = 1d-3
! 
!         !Overwrite if applicable
!         if (present(stdev)) stdev_ = stdev
! 
! 
!         ! Initialize
!         fx = f(x,nf) 
!         SSRSx = dot_product(fx,fx)
! 
!         ! Main loop
!         do iter = 1,o%maxiter
! 
!             do idraw = 1,ndraws_
! 
!                 ! Compute draws
!                 e_(:,idraw) = stdev_*randn(nx)
! 
!                 xtrial = x + e_(:,idraw)
! 
!                 ! Evaluate f at the trial points
!                 Ymat(idraw,:) = f(xtrial,nf)
! 
!             end do
! 
!             Xmat = transpose(e_)
!             J = transpose(mldivide(matmul(transpose(Xmat),Xmat),matmul(transpose(Xmat),Ymat)))
! 
!             gradSSRSnorm = sqrt(dot_product(matmul(transpose(J),fx),matmul(transpose(J),fx)))
! 
!             ! Do a line search
!             ! Initialize values to current
!             xbest = x
!             fxbest = fx
!             SSRSbest = SSRSx
!             lambdabest = -1e0_rt ! keeping current value 
!             stepnorm = 0e0_rt
!             do iL = 1,nL
!                 lambda = gL(iL)
! 
!                 ! s = - mldivide(matmul(transpose(J),J) + lambda*eyex, matmul(transpose(J),fx))
!                 s = - mldivide(matmul(transpose(J),J) + lambda*diag(diag(matmul(transpose(J),J))), matmul(transpose(J),fx))
! 
!                 xtrial = x + s
!                 ftrial = f(xtrial,nf)
!                 tmpSSRS = dot_product(ftrial,ftrial)
! 
!                 if (tmpSSRS<SSRSbest) then
!                     xbest = xtrial
!                     fxbest = ftrial
!                     SSRSbest = tmpSSRS
!                     lambdabest = lambda
!                     stepnorm = sqrt(dot_product(s,s))
!                 end if
! 
!             end do
! 
!             x = xbest
!             fx = fxbest
!             SSRSx = SSRSbest
! 
!             if (o%printmod>0 .and. mod(iter,o%printmod)==0) write(*,'(A,i8,4(A,d12.5))') ' NLM: iter',iter, ' step norm ',stepnorm,' SSRS ',SSRSx,' SSRS grad norm', gradSSRSnorm,' lambda ', lambdabest
! 
!             ! Check convergence criteria
!             if (stepnorm<o%xtoler .and.  gradSSRSnorm<o%xtoler) exit
!         
!         end do
! 
!     end subroutine 


    ! MPI implementation of random search
    ! Beginning with an initial point x, draws N proposals x+eps where eps is N(0,factor^2*e^2)
    ! where N is the number of MPI ranks, e is given by the user, and factor is internally set.
    ! The factor begins at 1e0_rt. If during an iteration an improvement is found, then factor = 
    ! factor*expand. O/w, factor = factor*shrink. The routine ends when factor*e<xtoler*abs(x) uniformly.
    ! 
    ! NOTE: this sets the random seed to distinct values for each node based on the clock. 
    subroutine sub_parrandom(x,f,e_in,maxiter_in,printmod_in,xtoler_in,shrink_in,expand_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(rt), intent(in), optional :: xtoler_in, shrink_in, expand_in
        !local
        real(rt), allocatable :: e(:), ftrials(:), xtrial(:) !, xtmp(:)
        real(rt) :: xtoler, fx, expand, shrink, factor
        integer :: maxiter, printmod, iter, nproc, nx, itmp(1) !, iproc

        nx = size(x)

        ! Give each process a random seed based on the clock
        call init_random_seed(mympi_iproc())

        !Set default values
        allocate(e(nx))
        e = max(.1e0_rt*abs(x),10e0_rt*epsilon(0e0_rt)) !the size of neighborhood about each point which is a 
                                   !hypersphere stretched to edge length e in each dimension
        maxiter = 100000
        printmod = 1
        xtoler = 1e-6_rt
        shrink =  .9e0_rt 
        expand = 1.1e0_rt

        !Overwrite if applicable
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(shrink_in)) shrink = shrink_in
        if (present(expand_in)) expand = expand_in

        ! Get MPI info
        nproc = mympi_nproc()
        allocate(ftrials(nproc))
        allocate(xtrial(nx))
        ! allocate(xtmp(nx))

        ! Initialize
        fx = huge(0e0_rt) ! set large so replace for sure
        factor = 1e0_rt

        ! Main loop
        do iter = 1,maxiter

            ! Compute a trial value
            xtrial = x + randn(nx)*e*factor
!             do iproc = 1,nproc
!                 if (iproc - 1 == mympi_iproc()) then
!                     xtrial = x + randn(nx)*e*factor
!                 else
!                     ! So that xtrial is definitely different across all nodes, call many randn calls
!                     xtmp = randn(nx)
!                 end if
!             end do

            ! Special case: first iteration, master node, compute value for x
            if (iter==1 .and. mympi_iproc()==0) then
                xtrial = x
            end if

            ! Compute the function value, but only for your rank
            ftrials(mympi_iproc()+1) = f(xtrial)

            ! Gather the function values only
            call mympi_allgatherv(ftrials)

            ! Select the best, check whether better than the current by a significant margin.
            if (minval(ftrials)<fx - (10e0_rt*epsilon(0e0_rt))*abs(fx)) then
                itmp = minloc(ftrials)
                fx = ftrials(itmp(1))
                
                ! If you are the best rank, send xtrial to everyone else. O/w, recieve.
                call mympi_bcast(xtrial,itmp(1)-1)
                
                ! Now all ranks have the copy of the best x
                x = xtrial
                factor = factor*expand
            else
                factor = factor*shrink
            end if

            if (mympi_iproc()==0 .and. printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i10,A,d12.5,A,d12.5)') ' PR: iter',iter, ' f',fx,' factor', factor

            ! Check stopping criteria
            if (all(factor*e<xtoler*abs(x))) exit
        
        end do

        if (mympi_iproc()==0 .and. printmod>0 .and. printmod==1) then
            write(*,*) 'PR: final fx', fx, 'final x',x
        end if

    end subroutine 

    ! This code is a random-start simplex. Random draws are taken from [a,b] for each
    ! node. Each node then optimizes starting from that point. The controls specified
    ! by e_in and opt are passed along to the simplex. 
    !
    ! This is a ridiculously parallel minization. The optimization never stops, so it 
    ! is up to the user to keep track of the best (which is displayed to stdout after
    ! each iteration).
    ! 
    subroutine sub_parmin_MPI(x,f,a,b,opt,e_in)
        use mod_util
        use mod_matlab
        use mod_mympi
        implicit none
        real(rt), intent(inout) :: x(:)
        real(rt), intent(in) :: a(:), b(:)
        type(minNdim_opts), intent(in), optional :: opt
        real(rt), intent(in), optional :: e_in(:)
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        ! local
        real(rt), allocatable :: xtmp(:)
        type(minNdim_opts) :: o
        type(type_seed) :: randstate
        integer :: flag,i,iter

        allocate(xtmp(size(x)))

        ! Set default parameters and overwrite if optional argument is present
        if (present(opt)) o = opt

        call check_guess_in_range(flag,x,a,b)
        if (flag/=0) write(0,*) 'WARNING: x not in a,b for all dimensions'
        if (any(b<a)) stop 'ERROR: b<a in at least one dimension'

    

        iter = 0 

        ! BEGIN INFINITE LOOP
        ! Never exit (this way all ranks keep working forever) 
        do 

            iter = iter + 1
            
            call randstate%save()
            call randstate%randomize()

            ! Generate initial points (here, burn through points to ensure each node gets a different x)
            do i = 0,mympi_nproc()-1

                ! Generate a point in [a,b]
                call random_number(xtmp)
                xtmp = a + (b-a)*xtmp

                ! Use as starting point if i is for your node
                if (i==mympi_iproc()) x = xtmp

            end do 

            call randstate%revert()

            write(*,*) 'PARMIN: iter: ',iter,' rank ',mympi_iproc(),' xbeforeNM = ',printx(x)
            ! Call the minimization routine in parallel
            if (present(e_in)) then
                call sub_simplex(x,f,e_in=e_in,maxiter_in=o%maxiter,printmod_in=o%printmod,xtoler_in=o%xabstoler)
            else
                xtmp = .1e0_rt*(b-a)
                call sub_simplex(x,f,e_in=xtmp,maxiter_in=o%maxiter,printmod_in=o%printmod,xtoler_in=o%xabstoler)
            end if

            write(*,*) 'PARMIN: iter: ',iter,' rank ',mympi_iproc(),' xafterNM = ',printx(x)

        end do

    end subroutine 

    
    ! Same as the above, but with an OpenMP implementation
    ! NOTE:
    ! If e_in is present, the results will not be reproducible (even calling random_seed directly before the call)
    ! because the draws are from xtrial+-e_in.
    ! However, if e_in is not present, then the results will be reproducible since all the trials are taken randomly
    ! from a,b and the same trials are taken each time. So, instaed of using e_in, would could call with a=x-e,b=x+e
    ! repeatedly and have something similar, but reproducible.
    !
    ! If seed is present, then it is used to set the seed for random_number upon entry and is replaced with the
    ! state of the seed after sub_parmin_OMP has completed.
    ! 
    ! If randunitid is present, the random numbers are read from the file associated with randunitid
    ! 
    ! If randvals is present, then the random numbers are taken from randvals (as if they were read in from file). 
    ! Hence, randvals must be sufficient for the computation. The variable i_randvals is required in this case
    ! and tells the routine where to begin reading randvals. It is modified on exit to return the last location.
    !
    ! THIS CODE HAS BEEN WORKED EXTENSIVELY TO PRODUCE EXACTLY THE SAME RESULTS THREADED AS NON-THREADED (PROVIDED
    ! THE FUNCTION F DOES NOT HAVE ANY OMP ISSUES). THE ORDER IN WHICH ITERATIONS ARE EXECUTED AND THE TRIALS
    ! ARE EXACTLY THE SAME AS IN THE SERIAL CODE.
    subroutine sub_parmin_OMP(x,f,a,b,e_in,opt,seed,randunitid,randvals,i_randvals)
        use mod_matlab
        use omp_lib
        use mod_util
        implicit none
        real(rt), intent(inout) :: x(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), optional, intent(in) :: e_in(:)
        type(minNdim_opts), intent(in), optional :: opt
        integer, intent(inout), optional :: seed(:)
        integer, intent(in), optional :: randunitid
        real(rt), intent(in), optional :: randvals(:)
        integer, intent(inout), optional :: i_randvals
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        ! local
        integer :: i,iter,tid,priviter,nthreads
        integer :: whosturn1,whosturn2
        integer :: whosturn1priv,whosturn2priv
        type(minNdim_opts) :: o
        real(rt), allocatable :: xtrial(:)
        real(rt) :: fx,fxtrial

        if (present(opt)) o = opt

        if (o%poolsize==-1) then
            write(0,*) 'WARNING: sub_parmin_OMP: poolsize should be specified. Setting to 1000.'
            o%poolsize = 1000
        end if

        if (any(b<a)) stop 'ERROR: sub_parmin: b is less than a'
        if (any(x<a .or. x>b)) stop 'ERROR: sub_parmin: x is not in [a,b]'

        iter = 0

        fx = f(x)

        if (present(seed)) then
            call random_seed(size=i)
            if (size(seed)/=i) stop 'ERROR: seed is not the correct length' 
            call random_seed(put=seed)
        end if

!         call dbgmsg(1,'about to enter parallel section')
    
        ! For strict reproducibility, need an ordered construct to handle cases of indifference. 

        !$OMP PARALLEL DEFAULT(NONE) SHARED(x,fx,o,a,b,whosturn1,whosturn2,randunitid,iter,nthreads,e_in,randvals,i_randvals) &
        !$OMP PRIVATE(xtrial,fxtrial,i,tid,priviter,whosturn1priv,whosturn2priv) 

        if (allocated(xtrial)) deallocate(xtrial)
        allocate(xtrial(size(x)))

        ! Get thread info
        tid = omp_get_thread_num()
        nthreads = omp_get_num_threads()

        ! Let thread 0 begin
        whosturn1 = 0
        whosturn2 = 0

        !$OMP BARRIER

!         print*,'tid',tid,'nthreads',nthreads,'whosturn1',whosturn1,'whosturn2',whosturn2

        ! Set counters to 0
        priviter = 0

        do i = 1,o%poolsize

            if (tid/=mod(i-1,nthreads)) cycle
    
            ! Wait for your turn to begin executiong
            do 
                !$OMP FLUSH (whosturn1)
                ! !$OMP ATOMIC READ
                whosturn1priv = whosturn1
                ! !$OMP END ATOMIC
                if (tid==mod(whosturn1priv,nthreads)) exit
                call spin(priviter)
            end do
            
            !$OMP CRITICAL (TURN1)

            ! Generate a new trial point
            if (present(randvals)) then
                xtrial = randvals(i_randvals:i_randvals+size(xtrial)-1)
                i_randvals = i_randvals + size(xtrial)
            elseif (present(randunitid)) then
                read(randunitid,*) xtrial
            else
                call random_number(xtrial)
            end if

            ! Generate a random point in [a,b]
            if (present(e_in)) then
                ! Draw uniformly about current best +/- e
                xtrial = x + (2e0_rt*xtrial - 1e0_rt)*e_in
                xtrial = max(a,min(b,xtrial)) ! bound
            else
                ! Draw uniformly from bounds
                xtrial = a + xtrial*(b-a)
            end if

            ! To ensure descent, make sure that do simplex from x
            if (tid==0 .and. priviter==0) xtrial = x 

            priviter = priviter + 1

            ! Free the next worker
            whosturn1 = whosturn1 + 1

            !$OMP END CRITICAL (TURN1)
            
            ! Call the minimization routine in parallel
            call sub_simplex(xtrial,f,a_in=a,b_in=b,e_in=(b-a)*.5e0_rt,xtoler_in=o%xabstoler,maxiter_in=o%maxiter,printmod_in=-1)

            ! Get value of best point found, xtrial
            fxtrial = f(xtrial)

            ! Wait for your turn to compare
            do
                !$OMP FLUSH (whosturn2)
                ! !$OMP ATOMIC READ
                whosturn2priv = whosturn2
                ! !$OMP END ATOMIC
                if (tid==mod(whosturn2priv,nthreads)) exit ! now its your turn
                call spin(priviter)
            end do

            !$OMP CRITICAL (TURN2)
            iter = iter + 1

            if (mod(iter-1,nthreads)/=tid) then
                write(0,*) 'WARNING: sub_parmin_OMP: execution of iterations was not ordered properly, there is a bug!'
            end if

            ! Compare against current best
            if (fxtrial<fx) then
                x = xtrial
                fx = fxtrial
            end if
            if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                if (o%displaybestx==1) then
                    write(*,'(A,i4,2(A,d16.8),2A)') 'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial,' best x ',printx(x)
                else
                    write(*,'(A,i4,2(A,d16.8))')    'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial
                end if
            end if

            ! Free the next worker
            whosturn2 = whosturn2 + 1

            !$OMP END CRITICAL (TURN2)

        end do

        ! don't exit the parallel region til everyone is finished (I think this is not necessary)
        !$OMP BARRIER 

        deallocate(xtrial)

        !$OMP END PARALLEL 

        ! Output the seed if requested
        if (present(seed)) then
            call random_seed(get=seed)
        end if

!     contains
!         subroutine dbgmsg(priority,msg)
!             implicit none
!             character(len=*), intent(in) :: msg
!             integer, intent(in) :: priority
!             if (priority>=o%dbglvl)  print*,'DEBUG: ',priority,msg
!         end subroutine 

    end subroutine 

    subroutine spin(i)
        implicit none
        integer :: i,j
        real(rt) :: tmp

        tmp = .12e0_rt + i

        do j = 0,1000
            tmp = cos(sin(tmp))
            if (tmp==2e0_rt) then
                write(0,*) 'This should not happen'
                exit
            end if
        end do
    end subroutine 

    ! Same as _OMP but with all openmp directives removed.
    subroutine sub_parmin_NOPAR(x,f,a,b,e_in,opt,seed,randunitid)
        use mod_matlab
        use omp_lib
        use mod_util
        implicit none
        real(rt), intent(inout) :: x(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), optional, intent(in) :: e_in(:)
        type(minNdim_opts), intent(in), optional :: opt
        integer, intent(inout), optional :: seed(:)
        integer, intent(in), optional :: randunitid
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        ! local
        integer :: i,iter,tid,priviter
        type(minNdim_opts) :: o
        real(rt), allocatable :: xtrial(:)
        real(rt) :: fx,fxtrial

        allocate(xtrial(size(x)))

        if (present(opt)) o = opt

        if (o%poolsize==-1) then
            write(0,*) 'WARNING: sub_parmin_OMP: poolsize should be specified. Setting to 1000.'
            o%poolsize = 1000
        end if

        if (any(b<a)) stop 'ERROR: sub_parmin: b is less than a'
        if (any(x<a .or. x>b)) stop 'ERROR: sub_parmin: x is not in [a,b]'

        iter = 0
        priviter = 0

        fx = f(x)

        if (present(seed)) then
            call random_seed(size=i)
            if (size(seed)/=i) stop 'ERROR: seed is not the correct length' 
            call random_seed(put=seed)
        end if
    
        ! For strict reproducibility, need an ordered construct to handle cases of indifference. 

        do i = 1,o%poolsize

            tid = omp_get_thread_num()

            if (present(randunitid)) then
                read(randunitid,*) xtrial
            else
                call random_number(xtrial)
            end if

            ! Generate a random point in [a,b]
            if (present(e_in)) then
                ! Draw uniformly about current best +/- e
                xtrial = x + (2e0_rt*xtrial - 1e0_rt)*e_in
                xtrial = max(a,min(b,xtrial)) ! bound
            else
                ! Draw uniformly from bounds
                xtrial = a + xtrial*(b-a)
            end if

            ! To ensure descent, make sure that do simplex from x
            if (tid==0 .and. priviter==0) then
                xtrial = x 
                priviter = 1
            end if
            
            ! Call the minimization routine in parallel
            call sub_simplex(xtrial,f,a_in=a,b_in=b,e_in=(b-a)*.5e0_rt,xtoler_in=o%xabstoler,maxiter_in=o%maxiter,printmod_in=-1)

            ! Call two more times to compare against initial x. This way, the routine never does worse than the input guess.
            fxtrial = f(xtrial)

            if (fxtrial<fx) then
                x = xtrial
                fx = fxtrial
            end if
            iter = iter + 1
            if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                if (o%displaybestx==1) then
                    write(*,'(A,i4,2(A,d16.8),2A)') 'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial,' best x ',printx(x)
                else
                    write(*,'(A,i4,2(A,d16.8))')    'PARMINOMP: iter ',iter,' best f ',fx,' trial f ',fxtrial
                end if
            end if

        end do

        ! Output the seed if requested
        if (present(seed)) then
            call random_seed(get=seed)
        end if

    end subroutine 

    ! Maximizes using a genetic algorithm
    ! Note that this routine is different from the rest because
    ! (1) the search space is integers
    ! (2) it maximizes (as this is more natural in thinking about fitness)
    !
    ! By default, searches in binary.
    ! 
    ! The stopping criterion is measured in terms of the standard deviations of the fitness
    ! in the population. When this falls below toler, execution is terminated and the 
    ! best point is returned.
    ! 
    ! The crossover method can be selected through the cross option
    ! cross             meaning
    ! 0                 single crossover point
    ! 1 (default)       double crossover point
    ! ... others not yet implemented
    subroutine sub_ga(x,f,npop_in,nelite_in,pmute_in,cross_in,toler_in,maxiter_in,printmod_in,&
                        maxfevals_in,openmp_in,mpi_in)
        use mod_matlab
        implicit none
        integer, dimension(:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                integer, dimension(:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: toler_in,pmute_in
        integer, intent(in), optional :: maxiter_in,printmod_in,maxfevals_in
        integer, intent(in), optional :: npop_in,nelite_in,cross_in
        logical, intent(in), optional :: openmp_in,mpi_in
        !local
        !controls
        integer :: maxiter,printmod,maxfevals,cross
        integer :: npop ! number of members in the population
        integer :: nelite ! number of elite members to carry over unchanged 
        logical :: openmp,mpi

        real(rt) :: pmute ! pmute/nx is the probability of a particular gene mutating
        real(rt), parameter :: pcross = .65e0_rt ! probability of two parents crossing
        
        !tools
        integer, dimension(:,:), allocatable :: pop ! population members
        integer, dimension(:,:), allocatable :: parents ! new population members
        integer, dimension(:,:), allocatable :: elite ! elite member
        real(rt), dimension(:), allocatable :: fit ! fitness of population members
        real(rt), dimension(:), allocatable :: probdraw ! used for drawing survivors
        real(rt), dimension(:), allocatable :: fit_elite
        real(rt) :: tmp,toler
        integer :: ipop,ix,nx,iter,itmp,itmp2,itmp3,fevals,ivec1(1)

        nx = size(x)        

        !Set defaults
        maxiter = 1000
        npop = 32
        nelite = 1
        pmute = 1e0_rt
        cross = 1
        printmod = 1
        toler = 1e-9_rt
        openmp = .false.
        mpi = .false.
        
        ! Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(maxfevals_in)) maxfevals = maxfevals_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(toler_in)) toler = toler_in 
        if (present(openmp_in)) openmp = openmp_in 
        if (present(mpi_in)) mpi = mpi_in  
        if (present(npop_in)) npop = npop_in
        if (present(cross_in)) cross = cross_in
        if (present(nelite_in)) nelite = nelite_in
        if (present(pmute_in)) pmute = pmute_in

        ! Set other defaults
        if (.not. present(maxfevals_in)) maxfevals = 1000*npop

        ! Check inputs
        if (mod(npop,2)/=0) STOP 'npop must be even'
        if (nelite>1) STOP 'nelite must less than or equal 1 or need to reprogram'

        ! Allocate space
        allocate(pop(nx,npop),parents(nx,npop),fit(npop),probdraw(npop))
        if (nelite>0) allocate(elite(nx,nelite),fit_elite(nelite))

        if (printmod>=0) write(*,'(A)') 'GA:________________________________________________________________________'            
        if (printmod>=0) write(*,'(A)') 'GA: Beginning Genetic Algorithm'            
        if (printmod>=0) write(*,'(A)') 'GA:  iter          best         worst           avg          stdev   fevals'

        !Create initial population by random draws except for input chromosone x
        pop(:,1) = x ! 
        do ipop = 2,npop
            do ix = 1,nx
                call random_number(tmp)
                pop(ix,ipop) = merge(1,0,tmp>.5e0_rt) ! binary
            end do
        end do

        fevals = 0
        if (nelite>0) fit_elite = -huge(0e0_rt)
        
        if (printmod==0) write(*,*) 'GA: Evaluating initial population: '

        ! Iterate
        do iter = 1,maxiter

            ! Evaluate the fitness of the population
            if (.not. openmp .and. .not. mpi) then
                do ipop = 1,npop
                    fit(ipop) = f(pop(:,ipop))
                end do
            elseif (openmp) then
                !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(ipop)
                do ipop = 1,npop
                    fit(ipop) = f(pop(:,ipop))
                end do
                !$OMP END PARALLEL DO
            elseif (mpi) then
                STOP 'have not programmed mpi yet'
            end if
            fevals = fevals + npop


            ! Check that fitness is all positive
            if (any(fit<0e0_rt)) STOP 'ERROR: GA: fitness is negative but should be positive'

            ! Drop the worst performing completely and add back in elite
            if (nelite==1) then
                if (maxval(fit)>fit_elite(1)) then
                    ivec1 = maxloc(fit)
                    elite(:,1) = pop(:,ivec1(1))
                    fit_elite(1) = fit(ivec1(1))
                else
                    ivec1 = minloc(fit)
                    pop(:,ivec1(1)) = elite(:,1)
                    fit(ivec1(1)) = fit_elite(1)
                end if
            end if

            ! Display some info
            if (printmod==0) call disp('GA: pop fitness: ',fit)
            if (printmod==0 .or. (printmod>0 .and. mod(iter,printmod)==0)) &
                write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,A,e12.5,A,i6)') &
                'GA:',iter,'  ', maxval(fit),'  ', minval(fit),'  ', mean(fit),'   ',std(fit),'   ',fevals

            ! Check stopping criteria
            if (iter==maxiter .or. fevals>=maxfevals .or. std(fit)<=toler) then
                print*,'iter',iter,'maxiter',maxiter
                print*,'fevals',fevals,'maxfevals',maxfevals
                print*,'std(fit)',std(fit),'toler',toler
                print*,'fit',fit
                exit
            end if

            ! Generate a new population 
            probdraw = cumsum(fit)/sum(fit)

            ! Roulette selection
            !!! Draw parents according to cumsum(fit)/sum(fit)
            do ipop = 1,npop
                call random_number(tmp)
                itmp = find1(tmp<=probdraw,'first')
                parents(:,ipop) = pop(:,ipop)
            end do

            ! Crossover the parents
            select case (cross)
                case (0)
                    !!! One-point crossover
                    do ipop = 1,npop-1,2

                        call random_number(tmp)
                        if (tmp>pcross) then 

                            ! Don't cross
                            pop(:,ipop) = parents(:,ipop)
                            pop(:,ipop+1) = parents(:,ipop+1)

                        else

                            !!! Draw one crossover 
                            call random_number(tmp)
                            itmp = nint(1e0_rt + tmp*(nx-1)) ! scale to 1,nx and round to nearest int

                            ! Now have a crossover point, so do it
                            pop(1:itmp,ipop) = parents(1:itmp,ipop)
                            pop(itmp+1:nx,ipop) = parents(itmp+1:nx,ipop+1)

                            pop(1:itmp,ipop+1) = parents(1:itmp,ipop+1)
                            pop(itmp+1:nx,ipop+1) = parents(itmp+1:nx,ipop)

                        end if

                    end do
                case (1)
                    !!! Two-point crossover
                    do ipop = 1,npop-1,2

                        call random_number(tmp)
                        if (tmp>pcross) then 

                            ! Don't cross
                            pop(:,ipop) = parents(:,ipop)
                            pop(:,ipop+1) = parents(:,ipop+1)

                        else

                            !!! Two-point crossover. Draw two crossover points
                            call random_number(tmp)
                            itmp = nint(1e0_rt + tmp*(nx-1)) ! scale to 1,nx and round to nearest int
                            call random_number(tmp)
                            itmp2 = nint(1e0_rt + tmp*(nx-1)) ! ditto

                            itmp3 = itmp
                            itmp = min(itmp3,itmp2)
                            itmp2 = max(itmp3,itmp2)

                            ! Now have two points for cross over. They may be equal.
                            pop(1:itmp,ipop) = parents(1:itmp,ipop)
                            pop(itmp+1:itmp2,ipop) = parents(itmp+1:itmp2,ipop+1)
                            pop(itmp2+1:nx,ipop) = parents(itmp2+1:nx,ipop)

                            pop(1:itmp,ipop+1) = parents(1:itmp,ipop+1)
                            pop(itmp+1:itmp2,ipop+1) = parents(itmp+1:itmp2,ipop)
                            pop(itmp2+1:nx,ipop+1) = parents(itmp2+1:nx,ipop+1)

                        end if

                    end do

!                 case (2)
!                     !!! WARNING: this is not actually uniform crossover
!                     !!! Uniform crossover
!                     do ipop = 1,npop-1,2
!                         do ix = 1,nx
!                             !!! W.p. 1/2, swap the gene
!                             call random_number(tmp)
! 
!                             if (tmp>.5e0_rt) then
!                                 pop(ix,ipop) = parents(ix,ipop)
!                                 pop(ix,ipop+1) = parents(ix,ipop+1)
!                             else
!                                 pop(ix,ipop) = parents(ix,ipop+1)
!                                 pop(ix,ipop+1) = parents(ix,ipop)
!                             end if
! 
!                         end do
!                     end do
                case default
                    STOP 'ERROR: crossover case not found'
            end select

            ! Mutate the population
            do ipop = 1,npop
                do ix = 1,nx
                    call random_number(tmp)
                    if (tmp<pmute) pop(ix,ipop) = merge(1,0,pop(ix,ipop)==0) ! mutate
                end do
            end do

        end do

        ! Return the best point and exit
        ivec1 = maxloc(fit)
        x = pop(:,ivec1(1))

!         if (printmod>0 .AND. printmod==1) then
!             write(*,*) 'GA:Final minimum x: '
!             call disp(x)            
!             write(*,*) 'GA:Final minimum f: '
!             write(*,*) 'GA:',minval(f_simplex)
!         end if
        if (printmod>=0) write(*,'(A)') 'GA:________________________________________________________________________'           

    end subroutine sub_ga


!     ! sub_sequentialmin
!     ! Cycles through each dimension and optimizes it before moving onto the next dimension
!     ! Good for problems where one-dimension may be more important than others or the problem
!     ! is mostly independent. This is definitely *not* guaranteed to find a global min or even
!     ! a local min because doing so may require changing x in more than one direction. 
!     ! However, it can still be useful!
!     subroutine sub_sequentialmin(x,f,e_in,shrink_in,toler_in,xtoler_in,maxiter_in,printmod_in)
!         use mod_min1dim
!         use mod_matlab
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: e_in
!         real(rt), intent(in), optional :: toler_in, xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(rt), intent(in), optional :: shrink_in
!         !local
!         real(rt) :: toler, xtoler, shrink, lastf, enorm
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e
!         integer :: n,i,iter,flag
!         
!         n = size(x)        
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1e0_rt*abs(x),.01e0_rt) !inital step size
!         shrink = .5e0_rt
!         xtoler = 1d-12
!         toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(shrink_in)) shrink = shrink_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
! 
!         if (printmod>0) then
!             write(*,'(A)') 'SM:__________________________________________________________________'            
!             write(*,'(A)') 'SM:Beginning Sequential Minimization Method  '        
!             write(*,'(A)') 'SM:    iter   dim   f(x)   enorm                                  '
!         end if 
! 
!         ! Do the search
!         do iter = 1,maxiter
! 
!             enorm = sqrt(sum(e**2))
!            
!             ! Exit if xtoler exceeded 
!             if (enorm < xtoler) exit 
! 
!             ! Search in each direction
!             do i = 1,n
!                 
!                 ! 
!                 call sub_simplex1dim(x(i),fonedim,e(i),1d-6,10,-1,flag)
! 
!                 ! Update status 
!                 if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
!                     write(*,'(A,2i5,2f13.7)') 'SM:   ',iter,i,lastf,enorm
! 
!             end do
!             
!             ! So it will eventually converge, shrink e
!             e = shrink*e
! 
!         end do
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'SM: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'SM: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'SM:__________________________________________________________________'            
!         end if
! 
!     contains
!         
!         function fonedim(xi)
!             implicit none
!             real(rt), intent(in) :: xi
!             real(rt) :: fonedim
! 
!             x(i) = xi
!             lastf = f(x) ! Used for stats
!             fonedim = lastf
! 
!         end function 
! 
!     end subroutine sub_sequentialmin


    ! Stochastic Gradient Descent
    ! 
    ! This minimizes a function that can only be measured noisily.
    ! That is, min_x f(x) where f(x) = E(g(x,Z)) and z is a measurement error.
    ! Is essential a noisy version of gradient descent.
    !
    subroutine sub_stochGradDescent(x,f,e_in,a_in,b_in,lambda_in,scaling_in,xtoler_in,maxiter_in,printmod_in)
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: xtoler_in, scaling_in, lambda_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        real(rt) :: xtoler, f0
        integer :: maxiter,printmod
        real(rt), dimension(1:size(x)) :: e, a, b, grad, scaling_init, scaling
        real(rt) :: snorm, lambda, lambda_init, tmp
        integer :: n,iter,fevals
        
        n = size(x)        

        ! Set defaults and overwrite if present
        e = max(.1e0_rt*abs(x),.01e0_rt); if (present(e_in)) e = e_in
        a = -huge(0e0_rt);             if (present(a_in)) a = a_in
        b = huge(0e0_rt);              if (present(b_in)) b = b_in
        maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;              if (present(xtoler_in)) xtoler = xtoler_in
        printmod = 1;               if (present(printmod_in)) printmod = printmod_in
        lambda_init = 1e0_rt;          if (present(lambda_in)) lambda_init = lambda_in
        scaling_init = 1e0_rt;         if (present(scaling_in)) scaling_init = scaling_in

        ! Init
        fevals = 0

        if (printmod>0) then
            write(*,'(A)') 'SGD:__________________________________________________________________'            
            write(*,'(A)') 'SGD:Beginning Stochastic Gradient Descent Method       '            
            write(*,'(A)') 'SGD:    iter  fevals         f(x)        snorm'
        end if 

    
        do iter = 1,maxiter
            
            ! Compute the gradient estimate
            scaling = scaling_init*real(iter,rt)**(-1e0_rt/3e0_rt)
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            fevals = fevals + n + 1
            ! Compute the new point: use line search with a maximum step of lambda
            tmp = lambda_init/real(iter,rt)
            call sub_lineSearch(x,lambda,-grad,f,f0,grad,lambda_init_in=tmp,maxiter_in=10)
            ! x = x - lambda*grad

            ! Compute the norm of the step size
            snorm = lambda*sqrt(sum(grad**2))

            ! Update status 
            if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
                write(*,'(A,i5,i8,2f13.7)') 'SGD:   ',iter, fevals, f0, snorm

            ! Exit if converged
            if (snorm<xtoler) exit

        end do 

        if (printmod>0) then
            write(*,'(A)',advance='no') 'SGD: Final x:'
            write(*,'(99d16.6)',advance='no') x
            write(*,*)
            write(*,'(A)') 'SGD:__________________________________________________________________'            
        end if

        
    end subroutine 


!     ! Smolyak search
!     ! 
!     ! Smolyak search computes a smolyak collocation projection of the function f on a hypercube
!     ! defined by x +- e. Subsequently, the projection (which is an infinitely differentiable
!     ! function) is optimized using derivative based methods. After the projection is optimized,
!     ! the hypercube is recentered around the new x. The projection is applied again on the 
!     ! hypercube x +- e. If little improvement is made, the hypercube is shrunk: e = shrink*e and
!     ! the process begins again (terminated when the L2 norm of e falls below xtoler).
!     ! 
!     ! The philosophy behind this is that if the underlying function is "roughly smooth" --- 
!     ! i.e. almost differentiable but maybe not because of noise --- that we can get a close
!     ! approximation by the smolyak function and then optimize that. The smolyak projection (verse
!     ! a tensor product approx) is used because it is suitable for high dimension problems.
!     ! 
!     ! This is totally heuristic --- although I believe it could be formalized. I believe it is
!     ! my creation, but it is a fairly obvious.
!     !
!     ! The user may choose e, shrink, and also "mu", which is the degree of the smolyak 
!     ! approximation used. mu should generally be at least 2, because mu==1 has no crossterms
!     ! in the approximation. Also, for higher order approximations, there may be a lot of local
!     ! minima. Generally, the default value mu==2 is recommended. a,b, if present, are bounds
!     ! on the problem that are enforced in both in the smolyak approximation construction and 
!     ! the inner maximization (which is the simplex method currently).
!     ! 
!     subroutine sub_smolyaksearch(x,f,e_in,a_in,b_in,shrink_in,mu_in,xtoler_in,maxiter_in,printmod_in)
!         use mod_matlab
!         use mod_smolyak
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
!         real(rt), intent(in), optional :: xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(rt), intent(in), optional :: shrink_in
!         integer, intent(in), optional :: mu_in
!         !local
!         type(smolyakFcn) :: sApprox ! Smolyak approximation
!         real(rt) :: xtoler
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e, xoldbest, xbest, a, b, hcubelb, hcubeub
!         real(rt) :: shrink, enorm, snorm, fbest, foldbest
!         integer :: n,i,iter,fevals,mu,inner, tmpi1(1)
!         real(rt), allocatable :: coll(:,:),fcoll(:)
!         
!         n = size(x)        
! 
!         ! Set defaults and overwrite if present
!         mu = 2;                     if (present(mu_in)) mu = mu_in
!         e = max(.1e0_rt*abs(x),.01e0_rt); if (present(e_in)) e = e_in
!         a = -huge(0e0_rt);             if (present(a_in)) a = a_in
!         b = huge(0e0_rt);              if (present(b_in)) b = b_in
!         shrink = .5e0_rt;              if (present(shrink_in)) shrink = shrink_in
!         maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
!         xtoler = 1d-6;              if (present(xtoler_in)) xtoler = xtoler_in
!         printmod = 1;               if (present(printmod_in)) printmod = printmod_in
! 
!         ! Check inputs and try to recover when appropriate
!         if (any(e<0e0_rt)) then
!             print*,'WARNING: e has some negative, replacing'
!             e = abs(e)
!         end if
!         if (any(e==0e0_rt)) STOP 'SS: ERROR: e is equal to 0'
! 
!         ! Init
!         fevals = 0
! 
!         if (printmod>0) then
!             write(*,'(A)') 'SS:__________________________________________________________________'            
!             write(*,'(A)') 'SS:Beginning Smolyak Projection Search Method       '            
!             write(*,'(A)') 'SS:    iter  fevals    avg.fcoll        fbest        enorm        snorm'
!         end if 
! 
!         xoldbest = x
!         foldbest = f(x) ! inefficient
!         fevals = fevals + 1
! 
!         ! Do the smolyak search
!         OUTER: do iter = 1,maxiter
! 
!             ! Exit if xtoler exceeded 
!             enorm = sqrt(sum(e**2))
!             if (enorm < xtoler) exit 
! 
!             ! Inner loop: updating the hypercube center successively 
!             do inner = 1,20
! 
!                 ! Get the smolyak grid
!                 hcubelb = max(xoldbest-e,a)
!                 hcubeub = min(xoldbest+e,b)
!                 call smolyakapprox_step1(coll,sApprox,n,mu,hcubelb,hcubeub)
! 
!                 ! Evaluate the function at each collocation point
!                 if (allocated(fcoll)) deallocate(fcoll)
!                 allocate(fcoll(size(coll,1)))
!                 do i = 1,size(fcoll)
!                     fcoll(i) = f(coll(i,:))
!                     fevals = fevals + 1
!                 end do
! 
!                 ! Construct the smolyak approximation
!                 call smolyakapprox_step2(fcoll,sApprox)
! 
!                 ! Minimize the approximation within the hypercube 
!                 !!! Start from every collocation point and find the best of the approx
!                 do i = 1,size(fcoll)
!                     x = coll(i,:)                 
!                     call sub_simplex(x,fapprox,xtoler_in=1d-11,printmod_in=-1)
!                     if (i==1 .or. fapprox(x)<fbest) then
!                         xbest = x
!                         fbest = fapprox(x)
!                     end if
!                 end do
! 
!                 ! This is inefficient to do, but for info sake and ease, do it here 
!                 ! (could save this value since will be a collocation point next time
!                 fbest = f(xbest)
!                 fevals = fevals + 1
! 
!                 ! To ensure we always descend according to the *real* function values rather
!                 ! than the *approximate* function values, we can do this: 
!                 ! If this new "xbest" is worse than one of the collocation points we checked,
!                 ! replace it. 
!                 ! Note that with this, smolyak is at worst a grid search in several important
!                 ! dimensions.
!                 tmpi1 = minloc(fcoll)
!                 if (fcoll(tmpi1(1))<fbest) then
!                     xbest = coll(tmpi1(1),:) 
!                     fbest = fcoll(tmpi1(1))
!                 end if
! 
!                 snorm = sqrt(sum((xbest-xoldbest)**2))
! 
!                 xoldbest = xbest
!                 foldbest = fbest
! 
!                 ! Update status 
!                 if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
!                     write(*,'(A,i5,i8,4f13.7)') 'SS:   ',iter, fevals, mean(fcoll), fbest, enorm, snorm
! 
!                 ! Exit inner loop if best x has not changed much
!                 ! if (xnorm<maxval(abs(xbest))*(1e0_rt + 1d-3)) exit
!                 if (snorm<1d2*xtoler) exit
! 
!             end do 
!                 
!             ! Shrink the hypercube dimensions
!             e = shrink*e
! 
!         end do OUTER
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'SS: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'SS: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'SS:__________________________________________________________________'            
!         end if
! 
!     contains 
!         
!         function fapprox(x)
!             implicit none
!             real(rt), intent(in) :: x(1:)
!             real(rt) :: fapprox
!             ! local
!             real(rt) :: fvec(1),xmat(1,size(x))
! 
!             if (any(x<hcubelb) .or. any(x>hcubeub)) then
!                 fapprox = 1d100
!                 return
!             end if
! 
!             xmat(1,:) = x
! 
!             ! Evaluate the smolyak approximation at x
!             call smolyakapprox_step3(fvec,sApprox,xmat)
!             fapprox = fvec(1)
! 
!         end function
!     
!     end subroutine sub_smolyaksearch


!     ! Compass search
!     ! 
!     ! This is a comparison method that searches in each coordinate direction successfully,
!     ! shrinking the step size only when no improvement is found. It can be awfully slow,
!     ! but it is also reliable --- more so than the simplex algorithm. In my testing on the 
!     ! rosenbrock function, SA got stuck far from the global optimium while CS was in the 
!     ! right area (but without a precise estimate).
!     ! 
!     ! The step sizes in each direction are initially given by "e". At each step, the new e
!     ! is e*shrink. The routine ceases when the L2 norm of e falls below xtoler or maxiter
!     ! is reached.
!     ! 
!     subroutine sub_compass(x,f,e_in,shrink_in,toler_in,xtoler_in,maxiter_in,printmod_in,usecache_in)
!         use mod_matlab
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         procedure(minN_f_template) :: f
!         procedure(minN_f_template_w_cache), pointer :: g
!         type(minN_f_cache) :: cache
!         real(rt), dimension(1:size(x)), intent(in), optional :: e_in
!         real(rt), intent(in), optional :: toler_in, xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(rt), intent(in), optional :: shrink_in
!         logical, intent(in), optional :: usecache_in
!         !local
!         real(rt) :: toler, xtoler
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e,xbest,xtrial
!         real(rt) :: shrink,fbest,ftrial,avgftrial,enorm, fpos, fneg
!         integer :: n,i,iter,nimprove,ntrial,fevals
!         logical :: usecache
!         
!         n = size(x)        
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1e0_rt*abs(x),.01e0_rt) !inital step size
!         shrink = .5e0_rt
!         xtoler = 1d-12
!         toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
!         usecache = .true.
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(shrink_in)) shrink = shrink_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
!         if (present(usecache_in)) usecache = usecache_in
! 
!         ! Point g to either use f with a cache or without one
!         if (usecache) then
!             ! Create a cache
!             call init_cache(cache,size(x),1 + 2*n*maxiter)
!             g => f_with_cache
!         else
!             ! Proceed with a cache
!             g => f_without_cache 
!         end if
! 
!         ! Init
!         fevals = 0
! 
!         if (printmod>0) then
!             write(*,'(A)') 'CS:__________________________________________________________________'            
!             write(*,'(A)') 'CS:Beginning Compass Search Method       '            
!             write(*,'(A)') 'CS:    iter  fevals #better   avg.ftrial        fbest        xsize'
!         end if 
! 
!         ! Compute the function value at the initial point
!         xbest = x
!         fbest = g(x,f,cache)
! 
!         ! Do the compass search
!         do iter = 1,maxiter
! 
!             ! Search in each direction
!             nimprove = 0
!             avgftrial = 0e0_rt
!             ntrial = 0
!             do i = 1,n
! 
!                 ! Try positive direction
!                 xtrial = xbest
!                 xtrial(i) = xtrial(i) + e(i)
!                 fpos = g(xtrial,f,cache)
!                 avgftrial = avgftrial + fpos
!                 ntrial = ntrial + 1
! 
!                 ! Try negative direction
!                 xtrial = xbest
!                 xtrial(i) = xtrial(i) - e(i)
!                 fneg = g(xtrial,f,cache)
!                 avgftrial = avgftrial + fneg
!                 ntrial = ntrial + 1
!                 
!                 ! Get the best of the two directions
!                 if (fpos<fneg) then
!                     xtrial = xbest
!                     xtrial(i) = xtrial(i) + e(i)
!                     ftrial = fpos
!                 else
!                     xtrial = xbest
!                     xtrial(i) = xtrial(i) - e(i)
!                     ftrial = fneg
!                 end if
! 
!                 ! Store if better 
!                 if (ftrial<fbest) then
!                     xbest = xtrial
!                     fbest = ftrial
!                     nimprove = nimprove+1
! !                 else
! !                     ! Shrink that direction
! !                     e(i) = shrink*e(i)
!                 end if    
! 
!             end do
! 
!             ! If didn't find a better, shrink
!             if (nimprove==0) then
!                 e = shrink*e
!             end if
! 
!             ! Compute L2 norm of compass size e
!             enorm = sqrt(sum(e**2))
! 
!             ! Update status 
!             if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0 .or. enorm<xtoler)) &
!               write(*,'(A,i5,2i8,3f13.7)') 'CS:   ',iter, fevals, nimprove, avgftrial/real(ntrial,8), fbest, enorm
!            
!             ! Exit if xtoler exceeded 
!             if (enorm < xtoler) exit 
! 
!         end do
! 
!         x = xbest
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'CS: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'CS: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'CS:__________________________________________________________________'            
!         end if
! 
!         if (usecache) then
!             ! Free cache
!             call free_cache(cache)
!         end if
! 
!     end subroutine sub_compass


    ! sub_lineSearch
    ! 
    ! This is a utility routine that, given a function, initial point (x), and direction (s),
    ! minimizes the function along the direction.
    !
    subroutine sub_lineSearch(x,lambda,s,f,fx,gradx,lambda_init_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_min1dim
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        real(rt), dimension(1:), intent(in) :: s, gradx ! direction and gradient at initial point x
        real(rt), intent(in) :: fx ! f at initial point x
        real(rt), intent(out) :: lambda ! step size chosen
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: toler_in, xtoler_in, lambda_init_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, gp0, g0, flambda, glambda, bestf, bestlambda, lambda_init
        real(rt), parameter :: sig = 2e0_rt, eps = 1e-4_rt

        n = size(x)        

        ! Set defaults and overwrite if present
        maxiter = 40; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = -1;   if (present(printmod_in)) printmod = printmod_in
        lambda_init = 1e0_rt;  if (present(lambda_init_in)) lambda_init = lambda_init_in 

        ! We can use an inexact line search, the Armijo rule, to try to get a good value quickly
        ! Get the total derivative of f at x. This is the derivative of g(lambda) at lambda = 0  
        g0 = fx
        gp0 = dot_product(gradx,s)

        ! Begin with lambda = 1e0_rt, and creep it down until find an acceptable
        lambda = lambda_init 
        bestf = huge(0e0_rt)
        bestlambda = 1e0_rt
        do iter = 1,maxiter
            flambda = f(x+lambda*s)
            glambda = g0 + lambda*eps*gp0
            if (flambda<=glambda) then
                x = x+lambda*s
                return
            else
                lambda = lambda/sig
            end if
            if (flambda<bestf) then
                bestlambda = lambda
            end if
        end do
        
        lambda = bestlambda
        x = x + lambda*s

    end subroutine sub_lineSearch

    ! NOTE: THIS HAS NOT BEEN TESTED MUCH
    ! Broyden-Fletcher-Goldfarb-Shanno (BFGS) method
    ! 
    ! This uses line search with a Hessian estimate.
    ! 
    subroutine sub_BFGS(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in,hess_out)
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(rt), intent(out), optional :: hess_out(size(x),size(x))
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, f0, snorm, gnorm, lambda
        real(rt), allocatable :: x0(:),s(:), hess(:,:), scaling(:), grad(:), grae0_rt(:), y(:)
        
        n = size(x)        
        allocate(x0(n),s(n),grad(n),hess(n,n),grae0_rt(n),y(n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1e0_rt;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'BFGS:__________________________________________________________'            
            write(*,'(A)') 'BFGS: Beginning Broyden-Fletcher-Goldfarb-Shanno Method        '          
            write(*,'(A)') 'BFGS: Computing initial function value and gradient            '         
        end if

        ! Set the initial hessian to do steepest descent
        hess = eye(N)

        ! Compute the initial gradient
        x0 = x
        call sub_numGradHess(x0,f,scaling,grae0_rt,f0=f0)
        gnorm = sum(grae0_rt)**2
        
        if (printmod>0) then
            write(*,'(A)') 'BFGS:  iter          f0       snorm       gnorm      lambda    '
        end if

        ! Do the gradient descent method
        do iter = 1,maxiter
         
            ! Compute the direction using the hessian estimate and the gradient
            s = -mldivide(hess,grae0_rt)

            ! Compute the step via line search and the new x
            x = x0
            call sub_lineSearch(x,lambda,s,f,f0,grae0_rt) ! x overwritten
            s = x-x0 ! Now s is the step (not the direction)
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,5e12.5)') 'BFGS:',iter,f0,snorm,gnorm,lambda

            ! Check for convergence
            if (snorm<xtoler*(1e0_rt+sum(x0**2))) then
                if (printmod>0) then
                    if (gnorm<toler*(1e0_rt+abs(f0))) then
                        write(*,'(A)') 'BFGS: Converged to critical point: gradient is nearly 0'
                    else
                        write(*,'(A)') 'BFGS: Converged to non optimal point: gradient is not nearly 0, but step is small'
                    end if
                end if
                exit
            end if

            ! Update the gradient and hessian
            !!! Compute the new gradient
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            gnorm = sum(grad**2)
            !!! Update the hessian
            y = grad-grae0_rt
            if (abs(dot_product(y,s))>1d-10 .and. sum(y**2)/dot_product(y,s)<1d10) then
                hess = hess - matmul(vec2mat(matmul(hess,s)),transpose(vec2mat(matmul(s,hess))))/&
                              dot_product(s,matmul(hess,s)) + matmul(vec2mat(y),transpose(vec2mat(y)))/dot_product(y,s)
            else
                ! hess = hess
            end if

            !!! Update the gradient
            grae0_rt = grad

            ! Update
            x0 = x

            if (iter==maxiter) then
                write(*,'(A)') 'BFGS: Did not converge before max iter reached'
            end if

        end do

        if (present(hess_out)) hess_out = hess

        if (printmod>0) write(*,'(A)') 'BFGS:__________________________________________________________'

    end subroutine sub_BFGS


    ! NOTE: THIS HAS NOT BEEN TESTED MUCH
    ! Gradient descent (GD)
    ! 
    ! Gradient descent method employing a line search
    ! 
    subroutine sub_gradDescent(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, f0, snorm, gnorm, lambda
        real(rt), allocatable :: x0(:),s(:), hess(:,:), grad(:), scaling(:)
        
        n = size(x)        
        allocate(x0(n),s(n),grad(n),hess(n,n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1e0_rt;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'GS:________________________________________________________'            
            write(*,'(A)') 'GS:Beginning Gradient Descent Method       '          
            write(*,'(A)') 'GS:iter      f0       snorm      gnorm     '
        end if

        ! Do the gradient descent method
        do iter = 1,maxiter
         
            ! Compute the derivative information 
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            gnorm = sum(grad**2)

! print*,'x is ',x

            ! Compute the step via line search
            x0 = x
            grad = -grad
            call sub_lineSearch(x0,lambda,grad,f,f0,grad) ! x0 overwritten
            grad = -grad
            s = x-x0
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,4e12.5)') 'GS:',iter,f0,snorm,gnorm

            ! Check for convergence
            if (snorm<xtoler*(1e0_rt+sum(x**2))) then
                if (gnorm<toler*(1e0_rt+abs(f0))) then
                    print*,'GS: Converged to critical point: gradient is nearly 0'
                else
                    print*,'GS: Converged to non optimal point: gradient is not nearly 0, but step is small'
                end if
                exit
            end if

            ! Update
            x = x0

            if (iter==maxiter) then
                print*,'GS: Did not converge before max iter reached'
            end if

        end do

        if (printmod>0) write(*,'(A)') 'GS:________________________________________________________'

    end subroutine sub_gradDescent


    ! Newton's method in N dimensions (NN)
    ! 
    ! Pure newton's method, calculating the hessian, without linesearch.
    ! 
    subroutine sub_newtonNdim(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_matlab
        use mod_deriv
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(rt) :: toler, xtoler, f0, snorm, gnorm
        real(rt), allocatable :: s(:), hess(:,:), grad(:), scaling(:)
        
        n = size(x)        
        allocate(s(n),grad(n),hess(n,n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1e-6_rt;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1e-6_rt;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1e0_rt;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'NN:________________________________________________________'            
            write(*,'(A)') 'NN:Beginning Newton''s Method                     '          
            write(*,'(A)') 'NN:  iter          f0       snorm       gnorm     '
        end if

        ! Do Newton's method
        do iter = 1,maxiter
         
            ! Compute the derivative information 
            call sub_numGradHess(x,f,scaling,grad,hess,f0)
            gnorm = sum(grad**2)

            ! Compute the step
            s = -mldivide(hess,grad)
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,4e12.5)') 'NN:',iter,f0,snorm,gnorm

            ! Check for convergence
            if (snorm<xtoler*(1e0_rt+sum(x**2))) then
                if (gnorm<toler*(1e0_rt+abs(f0))) then
                    if (sum(sum(hess,1))<0e0_rt) then
                        print*,'NN: Converged to min'
                    else
                        print*,'NN: Converged to max'
                    end if
                else
                    print*,'NN: Converged to non optimal point: gradient is not nearly 0, but step is small'
                end if
                exit
            end if

            ! Update
            x = x + s

            if (iter==maxiter) then
                print*,'NN: Did not converge before max iter reached'
            end if

        end do

        if (printmod>0) write(*,'(A)') 'NN:________________________________________________________'

    end subroutine sub_newtonNdim


    !Nelder-Mead algorithm
    ! 
    ! I should change that so it just treats anywhere outside of [a,b] as infinite.
    !
    ! This code essentially follows the algorithm in wikipedia
    ! However, I try a small modification where I accept the reflection point if it is better
    ! than the worst point, not the second worse point. This does not affect the volume of the 
    ! simplex, but gets around one pathology I encountered with a non-continuous function.
    ! 
    ! There are two routines here: 
    !   sub_simplex: full featured but slow. Ideal for flexibility, debugging, and calibration. 
    !   sub_simplex_core: not as many options and info, but faster (and so suitable for core routines)
    ! I recently added a new option: "randomize". If this is selected, the initial simplex is 
    ! drawn randomly with x_j = x + randn*e so that is e is the standard deviation of the shock
    !
    subroutine sub_simplex_core(x,f,e,xtoler,maxiter,printmod)
        use mod_sort
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in) :: e
        real(rt), intent(in) :: xtoler
        integer, intent(in) :: maxiter,printmod
        !local
        !controls
        real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt, gamma = .5e0_rt, eta = .5e0_rt
        !tools
        real(rt), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(rt), dimension(1:size(x)+1) :: f_simplex
        real(rt), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(rt) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(rt) :: dist
        integer :: n, i, iter
        logical :: doNotTryReflect

        n = size(x)        

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        do i = 2,n+1
            x_simplex(i,:) = x
            x_simplex(i,i-1) = x(i-1) + e(i-1)            
        end do

        !Evaluate simplex at corners
        do i = 1,n+1
            f_simplex(i) = f(x_simplex(i,:))
        end do

        if (printmod>0) write(*,'(A)') 'NM:iter     min(f)        max(f)        simplex size'

        !Main routine
        doNotTryReflect = .false.
        do iter = 1,maxiter

            ! Reorder the simplex from smallest to largest (ascending sort)
            !!! Use insertion sort because almost surely this will be a very small list
            call isort(f_simplex,ranking) 
            x_simplex = x_simplex(ranking,:)
 
            ! Output
            if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') &
                'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.e0_rt
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do

            ! Check for convergence or end of loop
            if (dist<xtoler .or. iter==maxiter) then
                if (printmod>0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') &
                    'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist
                ! Finished
                x = x_simplex(1,:)
                return
            end if

            !Try to replace the worst point, n+1, by another better point

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n,:),1)/real(n,rt) 

            !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
            x_reflect = x_gravity + alpha*(x_gravity-x_simplex(n+1,:))
            if (.not. doNotTryReflect) then
                f_reflect = f(x_reflect)
            else
                f_reflect = huge(0e0_rt)
                doNotTryReflect = .false.
            end if
            
            !Determine what to do with reflected point
            !if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n)) then ! The old algorithm had this
            if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1)) then 
                ! NOTE: next time, reflect will not satisfy this criterion---could optimize by avoiding it.
                ! If this is true, then next time reflect is tried, it will be the old point.
                ! We already know this is worse -- so don't waste an feval!
                if (f_reflect>=f_simplex(n)) then
                    doNotTryReflect = .true.
                end if 

                !Reflect: If f(x1)>f(xr)>f(xn) then replace and continue
                x_simplex(n+1,:) = x_reflect
                f_simplex(n+1) = f_reflect

            elseif (f_reflect>=f_simplex(n)) then

                ! If f_reflect is worst point, generate a new point by contraction
                x_contract = x_gravity + gamma*(x_simplex(n+1,:)-x_gravity)
                f_contract = f(x_contract)

                if (f_contract<f_simplex(n+1)) then

                    ! Contract: found a better point, so use it
                    x_simplex(n+1,:) = x_contract
                    f_simplex(n+1) = f_contract
                
                else
                
                    ! Contract all: no better point found, contract the entire simplex toward the best point
                    do i = 2,n+1                        
                        x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                        f_simplex(i) = f(x_simplex(i,:))
                    end do

                end if 

            elseif (f_reflect<=f_simplex(1)) then
                
                ! If f_reflect is best point, then try expanding in this direction for further improve
                x_expand = x_reflect + beta*(x_reflect-x_gravity)
                f_expand = f(x_expand)

                if (f_expand<f_reflect) then
                    ! Expand
                    x_simplex(n+1,:) = x_expand
                    f_simplex(n+1) = f_expand
                else
                    ! Reflect
                    x_simplex(n+1,:) = x_reflect
                    f_simplex(n+1) = f_reflect
                end if
            else 
                !Case not found
                STOP 'simplex: case not found'
            end if

        end do

    end subroutine sub_simplex_core
    subroutine sub_simplex(x,f,e_in,a_in,b_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in,uid_in)
        use mod_util
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(:), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: toler_in, xtoler_in, contract_in
        integer, intent(in), optional :: maxiter_in,printmod_in,uid_in
        logical, intent(in), optional :: randomize_in
        !local
        !controls
        real(rt) :: toler, xtoler
        integer :: maxiter,printmod
        real(rt), dimension(1:size(x)) :: e
        real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt
        real(rt) :: gamma, eta
        
        !tools
        real(rt), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(rt), dimension(1:size(x)+1) :: f_simplex, tmp
        real(rt), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(rt) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(rt) :: dist
        integer :: n, i, iter, fevals, uid, flag
        character(len=15) :: action
        logical :: is_constrained, doNotTryReflect, randomize

        n = size(x)        

        ! Check the sizes conform
        if (present(e_in)) then 
            if (size(e_in)/=n) STOP 'ERROR: sub_simplex: e_in does not conform to x'
        end if
        if (present(a_in)) then 
            if (size(a_in)/=n) STOP 'ERROR: sub_simplex: a_in does not conform to x'
        end if
        if (present(b_in)) then 
            if (size(b_in)/=n) STOP 'ERROR: sub_simplex: b_in does not conform to x'
        end if

        !Set defaults
        maxiter = 1000
        e = max(.1e0_rt*abs(x),.0001e0_rt) !inital step size
        xtoler = 1e-12_rt
        toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        randomize = .false.
        gamma = .5e0_rt ! contraction param
        eta = .5e0_rt ! contraction param      
        uid = 6 ! stdout

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(randomize_in)) randomize = randomize_in
        if (present(uid_in)) uid = uid_in
        if (present(contract_in)) then
            gamma = (1e0_rt-contract_in)
            eta = (1e0_rt-contract_in)
            if (contract_in<=0e0_rt .or. contract_in>=1e0_rt) then
                STOP 'ERROR: sub_simplex: contract_in invalid'
            end if
        end if

        ! Check that x is inside a,b if present
        if (present(a_in) .and. present(b_in)) then
            call check_guess_in_range(flag,x,a_in,b_in)
            if (flag/=0) write(0,*) 'WARNING: x was not in range, so bounding and proceeding'
            x = min(max(a_in,x),b_in)
        end if

        if (printmod==1) then
            write(uid,'(A)') 'NM:'//repeat('_',120) 
            write(uid,'(A)') 'NM:Beginning Simplex Method (Nelder-Mead)'            
        end if

        is_constrained = present(a_in) .and. present(b_in)

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        if (.not. randomize) then
            do i = 2,n+1
                x_simplex(i,:) = x
                x_simplex(i,i-1) = x(i-1) + e(i-1)            
            end do
        else
            do i = 2,n+1
                x_simplex(i,:) = x + randn(n)*e
            end do
        end if

        !Evaluate simplex at corners
        do i = 1,n+1
            if (.not. is_constrained) then
                f_simplex(i) = f(x_simplex(i,:))
            else
                if (any(x_simplex(i,:)<a_in .or. x_simplex(i,:)>b_in)) then
                    f_simplex(i) = huge(0e0_rt)
                else
                    f_simplex(i) = f(x_simplex(i,:))
                end if
            end if
        end do
        fevals = n+1
    
        if (printmod==1) then
            write(uid,'(A)') 'NM: Initial simplex: '
            do i = 1,n+1
                write(uid,'(A,i4,A,A)') 'NM: i ',i,' x ',printx(x_simplex(i,:))
            end do
            write(uid,'(A)') 'NM: Initial simplex values: '
            write(uid,'(A,A)') 'NM: ',printx(f_simplex)
        end if

        dist = -1e0_rt

        !Main routine
!         if (printmod>0) write(uid,'(A)') 'NM:iter     min(f)        max(f)        simplex size  fevals   action              bestx'
        action = ' Init'
        doNotTryReflect = .false.
        do iter = 1,maxiter

            !Reorder the simplex from smallest to largest (ascending sort)
            call sort2(tmp,ranking,f_simplex)
            f_simplex = f_simplex(ranking)
            x_simplex = x_simplex(ranking,:)
            
! print*,'f_simplex',f_simplex
! print*,'x_simplex'
! call disp(x_simplex)

            if (printmod>0 .AND. mod(iter,printmod)==0) then
                write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                'NM:',iter,' best ', f_simplex(1),' worst ', f_simplex(n+1),' s.size ', dist, action, printx(x_simplex(1,:))
            end if

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.e0_rt
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do
            if (dist<xtoler) then
                if (printmod>0) &
                    write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                    'NM:',iter,' best ', minval(f_simplex),' worst ', maxval(f_simplex),&
                    ' s.size ', dist, action, printx(x_simplex(1,:))
                if (printmod==1) write(uid,'(A)') 'NM:Simplex finished: simplex size less than xtoler'
                exit
            end if
            if (f_simplex(1)<toler) then
                if (printmod>0) &
                    write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                    'NM:',iter,' best ', minval(f_simplex),' worst ', maxval(f_simplex),&
                    ' s.size ', dist, action, printx(x_simplex(1,:))
                if (printmod==1) write(uid,'(A)') 'NM:Simplex finished: f is less than toler'
                exit
            end if            

            !Try to replace the worst pt, N+1, by another better point

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n,:),1)/real(n,rt) 

            !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
            x_reflect = x_gravity + alpha*(x_gravity-x_simplex(n+1,:))
! print*,'x_reflect', x_reflect            
            if (.not. doNotTryReflect) then
                if (.not. is_constrained) then
                    f_reflect = f(x_reflect)
                else
                    if (any(x_reflect<a_in .or. x_reflect>b_in)) then
                        f_reflect = huge(0e0_rt)
                    else
                        f_reflect = f(x_reflect)
                    end if
                end if
! print*,'f_reflect', f_reflect            
                fevals = fevals+1
            else
                f_reflect = huge(0e0_rt)
                doNotTryReflect = .false.
            end if
            
            !Determine what to do with reflected point
            !if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n)) then ! The old algorithm had this
            if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1)) then 
                ! NOTE: next time, reflect will not satisfy this criterion---could optimize by avoiding it.
                ! If this is true, then next time reflect is tried, it will be the old point.
                ! We already know this is worse -- so don't waste an feval!
                if (f_reflect>=f_simplex(n)) then
                    doNotTryReflect = .true.
                end if 
                action = ' Reflected'
                !If f(x1)>f(xr)>f(xn) then replace and continue
                x_simplex(n+1,:) = x_reflect
                f_simplex(n+1) = f_reflect

            elseif (f_reflect>=f_simplex(n)) then
                !If f_reflect is worst point, generate a new point by contraction
                x_contract = x_gravity + gamma*(x_simplex(n+1,:)-x_gravity)
                if (.NOT. is_constrained) then
                    f_contract = f(x_contract)
                else
                    if (any(x_contract<a_in .or. x_contract>b_in)) then
                        f_contract = huge(0e0_rt)
                    else
                        f_contract = f(x_contract)
                    end if
                end if
                fevals = fevals+1
                if (f_contract<f_simplex(n+1)) then
                    action = ' Contracted'
                    !If found a better point, then use it
                    x_simplex(n+1,:) = x_contract
                    f_simplex(n+1) = f_contract
                else
                    !If no better point found, contract the entire simplex toward the best point
                    action = ' Contracted All'
                    do i = 2,n+1                        
                        x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                        !f_simplex(i) = f(x_simplex(i,:))
                        if (.not. is_constrained) then
                            f_simplex(i) = f(x_simplex(i,:))
                        else
                            if (any(x_simplex(i,:)<a_in .or. x_simplex(i,:)>b_in)) then
                                f_simplex(i) = huge(0e0_rt)
                            else
                                f_simplex(i) = f(x_simplex(i,:))
                            end if
                        end if
                        fevals = fevals+1
                    end do
                end if 
            elseif (f_reflect<=f_simplex(1)) then
                
                !If f_reflect is best point, then try expanding in this direction for further improve
                x_expand = x_reflect + beta*(x_reflect-x_gravity)
                !f_expand = f(x_expand)
                if (.not. is_constrained) then
                    f_expand = f(x_expand)
                else
                    if (any(x_expand<a_in .or. x_expand>b_in)) then
                        f_expand = huge(0e0_rt)
                    else
                        f_expand = f(x_expand)
                    end if
                end if
                fevals = fevals+1
                if (f_expand<f_reflect) then
                    action = ' Expanded'
                    x_simplex(n+1,:) = x_expand
                    f_simplex(n+1) = f_expand
                else
                    action = ' Reflected'
                    x_simplex(n+1,:) = x_reflect
                    f_simplex(n+1) = f_reflect
                end if
            else 
                !Case not found
                print*,'f_reflect:',f_reflect
                print*,'f_simplex:',f_simplex
                STOP 'simplex: case not found'
            end if

        end do
 
        !Set out
        x = x_simplex(minloc(f_simplex,1),:)

        if (printmod==1) then
            write(uid,'(A)') 'NM: Final simplex: '
            do i = 1,n+1
                write(uid,'(A,i4,A,99(e12.5))') 'NM: i ',i,' x ',x_simplex(i,:)
            end do
            write(uid,'(A)') 'NM: Final simplex values: '
            write(uid,'(A,99(e12.5))') 'NM: ',f_simplex
            write(uid,'(A)') 'NM:Final minimum x: '
            write(uid,'(A,99(e12.5))') 'NM: ',x
            write(uid,'(A)') 'NM:Final minimum f: '
            write(uid,'(A,e12.5)') 'NM:',minval(f_simplex)
            write(uid,'(A)') 'NM:'//repeat('_',120)
        end if

    end subroutine sub_simplex

    ! OpenMP version of simplex
!     subroutine sub_parallelsimplex_OMP(x,f,e_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in)
!         use mod_matlab
!         use mod_util
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(:), intent(in), optional :: e_in
!         real(rt), intent(in), optional :: toler_in, xtoler_in, contract_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         logical, intent(in), optional :: randomize_in
!         !local
!         !controls
!         real(rt) :: toler, xtoler
!         integer :: maxiter,printmod
!         real(rt), dimension(1:size(x)) :: e
!         real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt
!         real(rt) :: gamma, eta
!         !tools
!         real(rt), dimension(:,:), allocatable :: x_simplex
!         real(rt), dimension(:), allocatable :: f_simplex, tmp
!         integer, dimension(:), allocatable :: ranking
!         logical, dimension(:), allocatable :: didImprove
!         real(rt), dimension(:), allocatable :: x_gravity,x_reflect,x_expand,x_contract
!         real(rt) :: f_reflect,f_expand,f_contract
!         real(rt) :: dist
!         integer :: n, i, iter, fevals, ip, np, ilo, ihi
!         logical :: randomize
!         
! 
!         n = size(x)        
! 
!         allocate(x_simplex(n+1,n))
!         allocate(f_simplex(n+1))
!         allocate(tmp(n+1))
!         allocate(ranking(n+1))
!         allocate(didImprove(n+1))
!         allocate(x_gravity(n))
!         allocate(x_reflect(n))
!         allocate(x_expand(n))
!         allocate(x_contract(n))
! 
!         ! Get number of threads
!         np = 0
!         !$OMP PARALLEL REDUCTION(+:np)
!         np = np + 1
!         !$OMP END PARALLEL 
!         np = min(n,np)
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1e0_rt*abs(x),.0001e0_rt) !inital step size
!         xtoler = 1d-12
!         toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
!         randomize = .false.
!         gamma = .5e0_rt ! contraction param
!         eta = .5e0_rt ! contraction param      
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
!         if (present(randomize_in)) randomize = randomize_in
!         if (present(contract_in)) then
!             gamma = (1e0_rt-contract_in)
!             eta = (1e0_rt-contract_in)
!             if (contract_in<=0e0_rt .or. contract_in>=1e0_rt) then
!                 STOP 'ERROR: sub_simplex: contract_in invalid'
!             end if
!         end if
! 
!         if (randomize) STOP 'ERROR: sub_parallelsimplex: not sure randomization is okay across multiple processes'
! 
!         printmod = merge(printmod,-1,mympi_iproc()==0) 
! 
!         !Create initial simplex and evaluate at simplex corners
!         x_simplex(1,:) = x 
!         if (.not. randomize) then
!             do i = 2,n+1
!                 x_simplex(i,:) = x
!                 x_simplex(i,i-1) = x(i-1) + e(i-1)            
!             end do
!         else
!             do i = 2,n+1
!                 x_simplex(i,:) = x + randn(n)*e
!             end do
!         end if
! 
!         !Evaluate simplex at corners
!         !$OMP PARALLEL DO DEFAULT(SHARED)
!         do i = 1,n+1
!             f_simplex(i) = f(x_simplex(i,:))
!         end do
!         !$OMP PARALLEL DO
! 
!         fevals = n+1
!     
!         if (printmod>0 .AND. printmod==1) then
!             write(*,'(A)') 'PNM:Initial simplex: '
!             do i = 1,n+1
!                 write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
!             end do
!             write(*,'(A)') 'PNM: Initial simplex values: '
!             write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
!         end if
! 
!         dist = -1e0_rt
! 
!         !Main routine
!         if (printmod>0) write(*,'(A)') 'PNM:iter     min(f)        max(f)        simplex size  '            
!         do iter = 1,maxiter
! 
!             !Reorder the simplex from smallest to largest (ascending sort)
!             call sort2(tmp,ranking,f_simplex)
!             f_simplex = f_simplex(ranking); x_simplex = x_simplex(ranking,:)
! 
!             if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') 'PNM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist,' xbest = ',printx(x_simplex(1,:))
! 
!             !Check whether length of simplex sides from best point are less than xtoler
!             dist = 0.e0_rt
!             do i = 2,n+1
!                 dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
!             end do
!             if (dist<xtoler) then
!                 if (printmod>0) &
!                     write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
!                 if (printmod>0) &
!                     write(*,'(A)') 'PNM:Simplex finished: simplex size less than xtoler'
!                 exit
!             end if
!             if (f_simplex(1)<toler) then
!                 if (printmod>0) &
!                     write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
!                 if (printmod>0) &
!                     write(*,'(A)') 'PNM:Simplex finished: f is less than toler'
!                 exit
!             end if            
! 
!             !Calc center of gravity of n other pts
!             x_gravity = sum(x_simplex(1:n+1-np,:),1)/real(n+1-np,8) 
! 
!             !!!!! PARALLEL ALGORITHM !!!!!!
!             ! Try to replace the worst P points, N-P,N-P+1,...,N+1, with other points.
!             !$OMP PARALLEL DO
!             do ip = n-np,n+1
! 
!                 !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
!                 x_reflect(:,ip) = x_gravity + alpha*(x_gravity-x_simplex(ip,:))
!                 f_reflect(ip) = f(x_reflect)
! 
!             end do
!             !$OMP END PARALLEL DO
! 
!             ! For each of the reflected points, see if they improved and which ones to keep
!             ! Sort the new points in with the simplex, and see how many of the old ones are no longer in the simplex.
!             do ip = n-np,n+1
!                 if (f_simplex()
!             end do
!             call sort2()
!             call sort2(tmp,ranking,f_simplex)
! 
!             do ip = n-np,n+1
!                 
!                 !Determine what to do with reflected point
!                 if (f_reflect(ip)<=f_simplex(1)) then
!                     
!                     !If f_reflect is best point, then try expanding in this direction for further improve
!                     x_expand = x_reflect(:,ip) + beta*(x_reflect(:,ip)-x_gravity); f_expand = f(x_expand)
! 
!                     if (f_expand<f_reflect(ip)) then
!                         x_simplex(ip,:) = x_expand;  f_simplex(ip) = f_expand
!                     else
!                         x_simplex(ip,:) = x_reflect(:,ip); f_simplex(ip) = f_reflect(ip)
!                     end if
! 
!                     didImprove(ip) = .true.
! 
!                 elseif (f_simplex(1)<f_reflect(ip) .and. f_reflect(ip)<f_simplex(n+1-np)) then 
! 
!                     x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect(ip)
!                     didImprove(ip) = .true.
! 
!                 elseif (f_reflect(ip)>=f_simplex(n+1-np)) then
! 
!                     !If f_reflect is worst point, generate a new point by contraction
!                     if (f_reflect(ip)>=f_simplex(ip)) then
!                         x_reflect = x_simplex(ip,:); f_reflect(ip) = f_simplex(ip)
!                     end if
! 
!                     x_contract = x_gravity + gamma*(x_reflect-x_gravity)
!                     f_contract = f(x_contract)
! 
!                     if (f_contract<f_reflect(ip)) then
!                         !If found a better point, then use it
!                         x_simplex(ip,:) = x_contract; f_simplex(ip) = f_contract
!                         didImprove(ip) = .true.
!                     else
!                         ! These are not actually the reflection points, but the original point
!                         x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect(ip)
!                         didImprove(ip) = .false.
!                     end if 
!                 else 
!                     !Case not found
!                     print*,'f_reflect(ip):',f_reflect(ip)
!                     print*,'f_simplex:',f_simplex
!                     STOP 'simplex: case not found'
!                 end if
!             end do
! 
!             ! Now every process has updated a simplex point. Use all gather to collect them
!             ilo = n+1-np+1
!             ihi = n+1
!             call mympi_allgatherv(f_simplex(ilo:ihi))
!             x_simplex_transpose = transpose(x_simplex)
!             call mympi_allgatherv(x_simplex_transpose(:,ilo:ihi))
!             x_simplex = transpose(x_simplex_transpose)
!             call mympi_allgatherv(didImprove(ilo:ihi))
! 
!             if (.not. any(didImprove(ilo:ihi))) then
! 
!                 ! Contract all
!                 do i = 2,n+1
!                     x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
!                 end do
! 
!                 ! Evaluate at the new simplex 
!                 do i = 1+mympi_a(n),1+mympi_b(n)
!                     f_simplex(i) = f(x_simplex(i,:))
!                 end do
!                 call mympi_allgatherv(f_simplex(2:n+1))
! 
!             end if
! 
!         end do
!  
!         !Set out
!         x = x_simplex(minloc(f_simplex,1),:)
! 
!         if (printmod>0 .AND. printmod==1) then
!             write(*,'(A)') 'PNM: Final simplex: '
!             do i = 1,n+1
!                 write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
!             end do
!             write(*,'(A)') 'PNM: Final simplex values: '
!             write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
!             write(*,'(A)') 'PNM:Final minimum x: '
!             write(*,'(A,99(e12.5))') 'PNM: ',x
!             write(*,'(A)') 'PNM:Final minimum f: '
!             write(*,'(A,e12.5)') 'PNM:',minval(f_simplex)
!         end if
! 
!     end subroutine sub_parallelsimplex_OMP

    ! NOTE: currently this tries to replace the worst P points. However, this does not scale well to a large number of processors
    ! Instead, what could be done is a line search. When a reflection is called for, try reflecting at P different lengths.
    ! When a contraction is called for, try contracting at P different lengths. When a contract all is called for, do a round-robin 
    ! distributing N/P but checking multiple lengths.
    ! 
    subroutine sub_parallelsimplex(x,f,e_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in)
        use mod_mympi
        use mod_matlab
        use mod_util
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in
        real(rt), intent(in), optional :: toler_in, xtoler_in, contract_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: randomize_in
        !local
        !controls
        real(rt) :: toler, xtoler
        integer :: maxiter,printmod
        real(rt), dimension(1:size(x)) :: e
        real(rt), parameter :: alpha = 1.e0_rt, beta = 1.e0_rt
        real(rt) :: gamma, eta
        !tools
        real(rt), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(rt), dimension(1:size(x)+1) :: f_simplex, tmp
        real(rt), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(rt) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(rt) :: dist
        integer :: n, i, iter, fevals, ip, np, ilo, ihi
        logical :: randomize
        logical, dimension(1:size(x)+1) :: didImprove
        real(rt), dimension(1:size(x),1:size(x)+1) :: x_simplex_transpose

        n = size(x)        

        ! Get number of nodes
        np = min(n,mympi_nproc())

        !Set defaults
        maxiter = 1000
        e = max(.1e0_rt*abs(x),.0001e0_rt) !inital step size
        xtoler = 1e-10_rt
        toler = -huge(0.e0_rt) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        randomize = .false.
        gamma = .5e0_rt ! contraction param
        eta = .5e0_rt ! contraction param      

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(randomize_in)) randomize = randomize_in
        if (present(contract_in)) then
            gamma = (1e0_rt-contract_in)
            eta = (1e0_rt-contract_in)
            if (contract_in<=0e0_rt .or. contract_in>=1e0_rt) then
                STOP 'ERROR: sub_simplex: contract_in invalid'
            end if
        end if

        if (randomize) STOP 'ERROR: sub_parallelsimplex: not sure randomization is okay across multiple processes'

        printmod = merge(printmod,-1,mympi_iproc()==0) 

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        if (.not. randomize) then
            do i = 2,n+1
                x_simplex(i,:) = x
                x_simplex(i,i-1) = x(i-1) + e(i-1)            
            end do
        else
            do i = 2,n+1
                x_simplex(i,:) = x + randn(n)*e
            end do
        end if

        !Evaluate simplex at corners
        do i = mympi_a(n+1),mympi_b(n+1)
            f_simplex(i) = f(x_simplex(i,:))
        end do
        call mympi_allgatherv(f_simplex)

        fevals = n+1
    
        if (printmod>0 .AND. printmod==1) then
            write(*,'(A)') 'PNM:Initial simplex: '
            do i = 1,n+1
                write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
            end do
            write(*,'(A)') 'PNM: Initial simplex values: '
            write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
        end if

        dist = -1e0_rt

        !Main routine
        if (printmod>0) write(*,'(A)') 'PNM:iter     min(f)        max(f)        simplex size  '            
        do iter = 1,maxiter

            !Reorder the simplex from smallest to largest (ascending sort)
            call sort2(tmp,ranking,f_simplex)
            f_simplex = f_simplex(ranking); x_simplex = x_simplex(ranking,:)

            if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,2A)') &
                'PNM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist,' xbest = ',printx(x_simplex(1,:))

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.e0_rt
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do
            if (dist<xtoler) then
                if (printmod>0) &
                    write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
                if (printmod>0) &
                    write(*,'(A)') 'PNM:Simplex finished: simplex size less than xtoler'
                exit
            end if
            if (f_simplex(1)<toler) then
                if (printmod>0) &
                    write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
                if (printmod>0) &
                    write(*,'(A)') 'PNM:Simplex finished: f is less than toler'
                exit
            end if            

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n+1-np,:),1)/real(n+1-np,rt) 

            !!!!! PARALLEL ALGORITHM !!!!!!
            ! Try to replace the worst P points, N-P,N-P+1,...,N+1, with other points.
            do ip = n+1-np+ mympi_a(np),n+1-np+ mympi_b(np)

                !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
                x_reflect = x_gravity + alpha*(x_gravity-x_simplex(ip,:))
                f_reflect = f(x_reflect)
                
                !Determine what to do with reflected point
                if (f_reflect<=f_simplex(1)) then
                    
                    !If f_reflect is best point, then try expanding in this direction for further improve
                    x_expand = x_reflect + beta*(x_reflect-x_gravity); f_expand = f(x_expand)

                    if (f_expand<f_reflect) then
                        x_simplex(ip,:) = x_expand;  f_simplex(ip) = f_expand
                    else
                        x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect
                    end if

                    didImprove(ip) = .true.

                elseif (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1-np)) then 

                    x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect
                    didImprove(ip) = .true.

                elseif (f_reflect>=f_simplex(n+1-np)) then

                    !If f_reflect is worst point, generate a new point by contraction
                    if (f_reflect>=f_simplex(ip)) then
                        x_reflect = x_simplex(ip,:); f_reflect = f_simplex(ip)
                    end if

                    x_contract = x_gravity + gamma*(x_reflect-x_gravity)
                    f_contract = f(x_contract)

                    if (f_contract<f_reflect) then
                        !If found a better point, then use it
                        x_simplex(ip,:) = x_contract; f_simplex(ip) = f_contract
                        didImprove(ip) = .true.
                    else
                        ! These are not actually the reflection points, but the original point
                        x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect  
                        didImprove(ip) = .false.
                    end if 
                else 
                    !Case not found
                    print*,'f_reflect:',f_reflect
                    print*,'f_simplex:',f_simplex
                    STOP 'simplex: case not found'
                end if

            end do

            ! Now every process has updated a simplex point. Use all gather to collect them
            ilo = n+1-np+1
            ihi = n+1
            call mympi_allgatherv(f_simplex(ilo:ihi))
            x_simplex_transpose = transpose(x_simplex)
            call mympi_allgatherv(x_simplex_transpose(:,ilo:ihi))
            x_simplex = transpose(x_simplex_transpose)
            call mympi_allgatherv(didImprove(ilo:ihi))

            if (.not. any(didImprove(ilo:ihi))) then

                ! Contract all
                do i = 2,n+1
                    x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                end do

                ! Evaluate at the new simplex 
                do i = 1+mympi_a(n),1+mympi_b(n)
                    f_simplex(i) = f(x_simplex(i,:))
                end do
                call mympi_allgatherv(f_simplex(2:n+1))

            end if

        end do
 
        !Set out
        x = x_simplex(minloc(f_simplex,1),:)

        if (printmod>0 .AND. printmod==1) then
            write(*,'(A)') 'PNM: Final simplex: '
            do i = 1,n+1
                write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
            end do
            write(*,'(A)') 'PNM: Final simplex values: '
            write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
            write(*,'(A)') 'PNM:Final minimum x: '
            write(*,'(A,99(e12.5))') 'PNM: ',x
            write(*,'(A)') 'PNM:Final minimum f: '
            write(*,'(A,e12.5)') 'PNM:',minval(f_simplex)
        end if

    end subroutine sub_parallelsimplex

    !Shrinking hypercube
    ! Starting with an initial guess and a hypercube centered on that guess, 
    ! this draws a random point from within the hypercube.
    ! If the value is better (less) than the current value, the hypercube edges are doubled
    ! If the value is worse (greater), the hypercube edges are decreased by a small amount
    ! 
    ! I originally got this idea from the web. It is discussed in http://linneus20.ethz.ch:8080/4_4_2.html
    ! which looks like it was copied from somewhere. If you know where the original article is from, please
    ! let me know and I'll credit appropriately.
    subroutine sub_shrinkingcube(x,f,e_in,maxiter_in,printmod_in,xtoler_in,shrinkFactor_in)
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(rt), intent(in), optional :: xtoler_in, shrinkFactor_in
        !local
        integer :: maxiter, printmod, iter
        real(rt), dimension(1:size(x)) :: e, bestx, draw
        real(rt) :: xtoler, fx, bestf, shrinkFactor
        integer :: seedSize,seed(100)

        ! Get the current state of the seed and save it to not interfere with a main program
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_shrinkingcube: must increase seed size'
        call random_seed(get=seed(1:seedSize))        

        !Set default values
        e = .02e0_rt*max(abs(x),1e-5_rt) !the size of neighborhood about each point which is a 
                                   !hypersphere stretched to edge length e in each dimension
        maxiter = 1000
        printmod = 1
        xtoler = 1e-8_rt
        shrinkFactor = .99e0_rt !Default suggested on website

        !Overwrite if applicable
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(shrinkFactor_in)) shrinkFactor = shrinkFactor_in

        if (printmod>0) write(*,'(A)') 'SH: ________________________________________________________'            
        if (printmod>0) write(*,'(A)') 'SH: Beginning Shrinking Hypercube'            

        ! Initialize
        fx = f(x)
        bestf = fx
        bestx = x        

        ! 
        do iter = 1,maxiter

            if (printmod>0) write(*,'(A,i10,A,d12.5,A,d12.5,A,d12.5)') &
                'SH: iter',iter, ' f',fx,' bestf',bestf, ' cubeVol', product(e)

            ! If a cube edge is less than xtoler stop
            if (any(e<xtoler)) exit
        
            ! Take random draw from within cube
            call random_number(draw)
            draw = draw*2.e0_rt-1.e0_rt !now uniform in [-.5,.5]
            x = bestx + e*draw

            ! Evalute current point
            fx = f(x)

            ! If best save it and expand, otherwise shrink it
            if (fx<bestf) then
                bestf = fx
                bestx = x
                e = e*2.e0_rt
            else
                e = e*shrinkFactor
            end if

        end do

        x = bestx

        if (printmod>0 .AND. printmod==1) write(*,'(A)') 'SH: Final minimum x: '
        if (printmod>0 .AND. printmod==1) write(*,*) 'SH:', x !call disp(x)            
        if (printmod>0 .AND. printmod==1) write(*,'(A)') 'SH: Final minimum f: '
        if (printmod>0 .AND. printmod==1) write(*,'(A,d12.5)') 'SH:',bestf
        if (printmod>0) write(*,'(A)') 'SH:________________________________________________________'

    end subroutine sub_shrinkingcube

    ! Simulated Annealing (SA)
    ! 
    ! Algorithm description: 
    ! Begins at x and draws random normal innovations to generate x'. x' is accepted whenever f(x')<f(x)
    ! Even when f(x')>f(x), it still accepted according to the probability that U < exp((fx-f(x'))/T).
    ! The temperature T "cools" slowly trending to 0. The innovations are drawn with a decreasing variance.
    ! 
    ! The temperature begins at T0 and cools with T(k) = T0/(1+log(k)).
    ! The variance begins at Var(0)=e**2 and decreases with Var(k) = Var(0)/(1+log(k)).
    !
    ! Global convergence is guaranteed for T0 "large enough". Unfortunately, this is not really helpful.
    ! Part of the program output is an acceptance rate. One may try to tune T0 and e so that this is not much
    ! more than 30% but also not 0%. Also, the "currf" and "bestf" should not be too far apart
    ! 
    ! Optional input:
    ! e: initial standard deviation of the shocks to x.
    ! T0: initial temperature.
    ! a: lower bound on parameter space
    ! b: upper bound on parameter space
    ! maxiter: self-explanatory
    ! printmod: self-explanatory
    ! 
    subroutine sub_simanneal(x,f,e_in,a_in,b_in,T0_in,maxiter_in,printmod_in,aggressive_in)
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: T0_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: aggressive_in
        !local
        real(rt), parameter :: fracEps = .02e0_rt
        integer :: maxiter, printmod
        real(rt) :: T,T0
        real(rt), dimension(1:size(x)) :: a, b, e, y, xbest
        real(rt) :: fx,fy,fxbest
        integer :: seedSize,seed(100)
        integer :: n, iter, naccept, nreject
        real(rt) :: randnDraw(size(x)), randuDraw
        logical :: aggressive

        n = size(x)

        ! Check seedSize is large enough
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_simanneal: must increase seed size'

        !Set defaults
        maxiter = 10000
        printmod = 1
        a = -huge(0.e0_rt) !the lowerbound on the state space
        b = huge(0.e0_rt)  !the upperbound on the state space
        e = .01e0_rt*max(abs(x),1e-4_rt) !the size of neighborhood about each point which is a hypersphere stretched to diameter e in each dimension
        T0 = 1e0_rt !temperature scaling  (setting T0 sufficiently ensures a.s. convergence)
        aggressive = .false.

        !Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(a_in)) a = a_in
        if (present(b_in)) b = b_in
        if (present(e_in)) e = e_in
        if (present(T0_in)) T0 = T0_in
        if (present(aggressive_in)) aggressive = aggressive_in

        ! Check inputs
        if (any(x<a) .or. any(x>b)) STOP 'sub_simanneal: x was not initialized to be between a and b!'

        ! Evaluate f at x
        fx = f(x)

        ! Initialize best values
        fxbest = fx
        xbest = x

        naccept = 0
        nreject = 0

        if (printmod>0) then
            write(*,'(A)') 'SA: ________________________________________________________'            
            write(*,'(A)') 'SA: Beginning Simulated Annealing Method '            
        end if
        do iter = 1,maxiter

!             ! Save the seed state, draw a random normal (size n) and random uniform 
!             !!! Get the current state of the seed and save it to not interfere with a main program
!             call random_seed(get=seed(1:seedSize))
            !!! Draw the variables
            randnDraw = randn(n)
            call random_number(randuDraw)
!             !!! Return the state of the seed to what it was before sim annealing was called
!             call random_seed(put=seed(1:seedSize))

            ! Set the current temperature
            T = T0/(1e0_rt+log(real(iter,rt))) ! T0 to 0
! T = T0/(1e0_rt+real(iter-1,8)) ! T0 to 0

            ! Draw a newpoint within the neighborhood of current x
            !!! I saw here http://stats.stackexchange.com/questions/27923/should-the-proposal-distribution-in-simulated-annealing-depend-on-the-temperatur
            !!! that can draw from this variance decreasing in T and will still get convergence (for large enough T0)
            if (.not. aggressive) then
                y = x + e*randnDraw*sqrt(T/T0) ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
            else
                y = x + e*randnDraw*T/T0 ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
            end if


            ! Bound y so it remains in the parameter space
            y = min(max(y,a),b)

            ! Evaluate f at y 
            fy = f(y)

            ! Determine accept reject
            if (fy<fx) then 
                ! In this variant we always accept better moves
                x = y
                fx = fy
                naccept = naccept + 1
            else
                if (randuDraw<exp((fx-fy)/T)) then
                    ! Move from x to y
                    x = y
                    fx = fy
                    naccept = naccept + 1
                else
                    ! Keep current x
                    nreject = nreject + 1
                end if
            end if

            ! Try
            if (printmod>0 .and. mod(iter,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5,A,d12.5,A,d12.5,A,d12.5)') &
                    'SA: iter ',iter,' ftrial',fy,' fcurr',fx,' fbest',fxbest,' Temp', T,&
                    ' Accept%', 100.e0_rt*real(naccept,8)/real(naccept+nreject,8)
            end if

            ! Save best f
            if (fx<fxbest) then 
                fxbest = fx
                xbest = x
            end if

        end do

        ! Return the best function value
        if (printmod>0) then
            if (fxbest/=fx) then
                write(*,'(A,d12.5,A)') 'SA: best value ever encountered, ',fxbest,', being returned, it is /= fx'
            end if
        end if
        x = xbest
        fx = fxbest

    end subroutine sub_simanneal

    ! Parallel Simulated Annealing (PSA)
    ! 
    ! All the argumetns are the same as the regular simulated annealing. 
    ! The algorithm is the same except that at each iteration, P proposals are drawn at the same time.
    ! The best of the proposals is taken. 
    ! 
    subroutine sub_parallelsimanneal(x,f,e_in,a_in,b_in,T0_in,maxiter_in,printmod_in,aggressive_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(rt), intent(in), optional :: T0_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: aggressive_in
        !local
        real(rt), parameter :: fracEps = .02e0_rt
        integer :: maxiter, printmod, n, iter, naccept, nreject, np, ip, itmp(1), seedSize, seed(100)
        real(rt) :: fx,fy,fxbest,T,T0,randuDraw
        real(rt), dimension(1:size(x)) :: a, b, e, y, xbest
        real(rt), allocatable :: z(:,:),fz(:)
        logical :: aggressive

        n = size(x)
        np = mympi_nproc()

        allocate(z(n,np),fz(np))

        ! Check seedSize is large enough
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_simanneal: must increase seed size'

        !Set defaults
        maxiter = 10000
        printmod = 1
        a = -huge(0.e0_rt) !the lowerbound on the state space
        b = huge(0.e0_rt)  !the upperbound on the state space
        e = .01e0_rt*max(abs(x),1e-4_rt) !the size of neighborhood about each point which is a hypersphere stretched to diameter e in each dimension
        T0 = 1e0_rt !temperature scaling  (setting T0 sufficiently ensures a.s. convergence)
        aggressive = .false.

        !Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(a_in)) a = a_in
        if (present(b_in)) b = b_in
        if (present(e_in)) e = e_in
        if (present(T0_in)) T0 = T0_in
        if (present(aggressive_in)) aggressive = aggressive_in

        printmod = merge(printmod,-1,mympi_iproc()==0)

        ! Check inputs
        if (any(x<a) .or. any(x>b)) STOP 'sub_simanneal: x was not initialized to be between a and b!'

        ! Initialize best values
        fxbest = huge(0e0_rt)
        xbest = 0e0_rt

        naccept = 0
        nreject = 0

        if (printmod>0) then
            write(*,'(A)') 'PSA: ________________________________________________________'            
            write(*,'(A)') 'PSA: Beginning Parallel Simulated Annealing Method '            
        end if
        do iter = 1,maxiter

            ! Set the current temperature
            T = T0/(1e0_rt+log(real(iter,rt))) ! T0 to 0

            !!! Draw the variables
            ! Draw a newpoint within the neighborhood of current x
            !!! I saw here http://stats.stackexchange.com/questions/27923/should-the-proposal-distribution-in-simulated-annealing-depend-on-the-temperatur
            !!! that can draw from this variance decreasing in T and will still get convergence (for large enough T0)
            do ip = 1,np

                if (iter==1 .and. ip==1) then
                    z(:,ip) = x
                else
                    do while (.true.)

                        if (.not. aggressive) then
                            z(:,ip) = x + e*randn(n)*sqrt(T/T0) ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
                        else
                            z(:,ip) = x + e*randn(n)*T/T0 ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
                        end if

                        ! Use rejection sampling to get the new point
                        if (all(z(:,ip)>=a) .and. all(z(:,ip)<=b)) exit

                    end do
                end if

            end do

            ! Evaluate f in parallel and share
            do ip = mympi_a(np),mympi_b(np)
                fz(ip) = f(z(:,ip))
            end do

            call mympi_allgatherv(fz)

            ! Take the best
            fy = minval(fz)
            itmp = minloc(fz)
            y = z(:,itmp(1))

            ! Determine accept reject
            if (iter==1) then
                x = y
                fx = fy
            elseif (fy<fx) then 
                ! In this variant we always accept better moves
                x = y
                fx = fy
                naccept = naccept + 1
            else
                call random_number(randuDraw)
                if (randuDraw<exp((fx-fy)/T)) then
                    ! Move from x to y
                    x = y
                    fx = fy
                    naccept = naccept + 1
                else
                    ! Keep current x
                    nreject = nreject + 1
                end if
            end if

            ! Try
            if (printmod>0 .and. mod(iter,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5,A,d12.5,A,d12.5,A,d12.5)') &
                    'PSA: iter ',iter,' ftrial',fy,' fcurr',fx,' fbest',fxbest,' Temp', T,&
                    ' Accept%', 100.e0_rt*real(naccept,8)/real(naccept+nreject,8)
            end if

            ! Save best f
            if (fx<fxbest) then 
                fxbest = fx
                xbest = x
            end if

        end do

        ! Return the best function value
        if (printmod>0) then
            if (fxbest/=fx) then
                write(*,'(A,d12.5,A)') 'PSA: best value ever encountered, ',fxbest,', being returned, it is /= fx'
            end if
        end if
        x = xbest
        fx = fxbest

    end subroutine sub_parallelsimanneal

    ! Grid search (GS)
    ! 
    ! Performs a brute force grid search. The tensor product grid is formed from the variables "grids" and "n".
    ! The grid in the ith dimension is grids(1:n(i),i). 
    subroutine sub_gridsearch(x,f,n,grids,printmod_in)
        use mod_matlab
        implicit none
        real(rt), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                use mod_types
                implicit none
                real(rt), dimension(1:), intent(in) :: x
                real(rt) :: f
            end function f
        end interface
        real(rt), dimension(:,:), intent(in) :: grids
        integer, dimension(:), intent(in) :: n
        integer, intent(in), optional :: printmod_in
        !local
        real(rt), dimension(size(x)) :: xbest
        real(rt) :: fxbest,fx
        integer :: printmod, d, j, i
        integer, dimension(size(x)) :: ivec

        ! Get dimension of the problem
        d = size(x)

        ! Check 
        if (d/=size(n)) STOP 'ERROR: sub_gridsearch: n has wrong size'
        if (d/=size(grids,2)) STOP 'ERROR: sub_gridsearch: grids has wrong size 2'
        if (maxval(n)>size(grids,1)) STOP 'ERROR: sub_gridsearch: grids has wrong size 1'

        ! Set defaults
        printmod = 1

        ! Overwrite if present
        if (present(printmod_in)) printmod = printmod_in

        ! Check inputs
!         if (any(b<a)) STOP 'sub_gridsearch: ERROR: b<a' 
        if (any(n<=0)) STOP 'sub_gridsearch: ERROR: n<=0'

        if (printmod>0) then
            write(*,'(A)') 'GS: ________________________________________________________'            
            write(*,'(A)') 'GS: Beginning Grid Search'            
        end if

        ! Do the grid search
        fxbest = huge(0e0_rt)
        ivec(:) = 1
        do i = 1,product(n)

            ! Get x
            do j = 1,d
                x(j) = grids(ivec(j),j)
            end do

            ! Evaluate 
            fx = f(x)

            ! Store the best
            if (fx<fxbest) then
                xbest = x
                fxbest = fx
            end if

            ! Try
            if (printmod>0 .and. mod(i,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5)') 'GS: iter ',i,' f',fx,' fbest',fxbest
            end if

            ! Increment ivec
            do j = 1,d
                if (mod(ivec(j),n(j))==0) then
                    ivec(j) = 1 
                else
                    ivec(j) = ivec(j)+1
                    exit
                end if
            end do
        end do
    
        ! Return the best
        x = xbest

        ! Return the best function value
        if (printmod>0) then
            write(*,'(A)') 'GS: finished.'
        end if

    end subroutine sub_gridsearch


!     ! Grid search (GS)
!     ! 
!     ! Performs a basic grid search over the domain [a,b] with n giving the points in each dimension
!     !
!     subroutine sub_gridsearch(x,f,a,b,n,printmod_in)
!         use mod_matlab
!         implicit none
!         real(rt), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(rt), dimension(1:), intent(in) :: x
!                 real(rt) :: f
!             end function f
!         end interface
!         real(rt), dimension(1:size(x)), intent(in) :: a,b
!         integer, dimension(1:size(x)), intent(in) :: n
!         integer, intent(in), optional :: printmod_in
!         !local
!         real(rt), dimension(size(x)) :: xbest
!         real(rt) :: fxbest,fx
!         integer :: printmod, d, j, i
!         integer, dimension(size(x)) :: ivec
! 
!         ! Get dimension of the problem
!         d = size(x)
! 
!         ! Set defaults
!         printmod = 1
! 
!         ! Overwrite if present
!         if (present(printmod_in)) printmod = printmod_in
! 
!         ! Check inputs
!         if (any(b<a)) STOP 'sub_gridsearch: ERROR: b<a' 
!         if (any(n<=0)) STOP 'sub_gridsearch: ERROR: n<=0'
! 
!         if (printmod>0) then
!             write(*,'(A)') 'GS: ________________________________________________________'            
!             write(*,'(A)') 'GS: Beginning Grid Search'            
!         end if
! 
!         ! Do the grid search
!         fxbest = huge(0e0_rt)
!         ivec(:) = 1
!         do i = 1,product(n)
! 
! !             print*,'ivec',ivec
! 
!             ! Get x
!             do j = 1,d
!                 x(j) = a(j) + real(ivec(j)-1,8)/real(n(j)-1,8)*(b(j)-a(j))
!             end do
! 
!             ! Evaluate 
!             fx = f(x)
! 
!             ! Store the best
!             if (fx<fxbest) then
!                 xbest = x
!                 fxbest = fx
!             end if
! 
!             ! Try
!             if (printmod>0 .and. mod(i,printmod)==0) then
!                 write(*,'(A,i6,A,d12.5,A,d12.5)') 'GS: iter ',i,' f',fx,' fbest',fxbest
!             end if
! 
!             ! Increment ivec
!             do j = 1,d
!                 if (mod(ivec(j),n(j))==0) then
!                     ivec(j) = 1 
!                 else
!                     ivec(j) = ivec(j)+1
!                     exit
!                 end if
!             end do
!         end do
!     
!         ! Return the best
!         x = xbest
! 
!         ! Return the best function value
!         if (printmod>0) then
!             write(*,'(A)') 'GS: finished.'
!         end if
! 
!     end subroutine sub_gridsearch

    ! Genetic algorithm --- MPI version
    ! 
    ! attempts to solve max f(x) s.t. x in {0,1}^n 
    ! 
    ! It generates N random uniform points.
    ! It takes the worst point and tries to get rid of it by reflecting it
    ! (like in Nelder-Mead) through a simplex of points. This simplex is chosen 
    ! by randomly selecting n+1 points (whose function values are known) with 
    ! replacement, giving a total of n+1 points. 
    !
    ! The centroid of the simplex is then used for reflecting the worst point. 
    ! If there reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. The parallel design here,
    ! in contrast to CRS2LM, is more advantageous when the functions time to completion
    ! is not uniform across x values. The algorithm itself is more global.
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. 
    subroutine sub_ga_MPI(xstar,f,opts)
        use mod_util
        use mod_mympi
        use mod_matlab
        implicit none
        integer, intent(inout) :: xstar(:)
        type(minNdim_opts), intent(in), optional :: opts
        interface 
            function f(x) 
                use mod_types
                implicit none
                integer, dimension(:), intent(in) :: x
                real(rt) :: f
            end function
        end interface
        ! local other
        integer, allocatable, target :: X(:,:),slavexlist(:,:) ! NOTE: these have target attribute and will be pointed at
        real(rt), allocatable, target :: y(:),slaveylist(:) ! NOTE: these have target attribute and will be pointed at
        logical, allocatable :: finishedtask(:),slaveeverassignedlist(:)
        integer, allocatable :: fevaldsn(:),requestlist(:) ! ,lasti(:)
        real(rt) :: tmp
        integer :: nx, N, ix, iter, i, who, itmp(1), prop, ib, iw
        type(minNdim_opts) :: o

        if (present(opts)) o = opts

        call disp_opts(o)

        nx = size(xstar)

        if (mympi_nproc()<=1) stop 'ERROR: sub_ga_MPI: called with only one rank'

        ! N is the pool of genes size
        N = mympi_nproc()-1 ! ensure at least N-1 ranks, o/w they would be doing nothing
        if (o%poolsize/=-1) N = max(N,o%poolsize) ! if specified, ensure at least that poolsize is used
        N = max(N,100) ! ensure at least 100

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave()
            return
        end if

        ! Generate N random points uniformly from [a,b]
        allocate(finishedtask(mympi_nproc()-1))
        allocate(X(nx,N))
        allocate(y(N))
        allocate(fevaldsn(mympi_nproc()-1))
        allocate(requestlist(mympi_nproc()-1))
        allocate(slavexlist(nx,mympi_nproc()-1))
        allocate(slaveylist(mympi_nproc()-1))
        allocate(slaveeverassignedlist(mympi_nproc()-1))

        ! For random results, initialize based on the clock.
        if (o%reproducible==0) then 
            call init_random_seed()
        elseif (o%reproducible==1) then
            ! should draw from some particular stream each time. 
            ! for now though, just use the default seed (the issue 
            ! is this will only be reproducible if f never changes
            ! the number or order of function calls.
        else
            stop 'ERROR: sub_ga_MPI: opts%reproducible should be 0 or 1'
        end if 

        ! Generate initial pool of points
        X(:,1) = xstar
        do i = 2,N
            do ix = 1,nx
                ! call random_number(tmp)
                ! X(ix,i) = merge(1,0,tmp>.5d0)
                X(ix,i) = mod(myirand(),2)
            end do
        end do

        fevaldsn = 0

        finishedtask = .true.

        ! Evaluate at each of the N points in parallel
        write(*,'(A,i4,A)') 'GA: Solving for initial pool of ',N,' points'
        do i = 1,N

            ! Wait for a free worker
            if (o%dbglvl>0) print*,'MASTER: WAITING FOR FREE WORKER'
            who = wait_for_free_worker()

            ! Assign the point
            if (o%dbglvl>0) print*,'MASTER: X(:,i)',X(:,i)
            call assign_xy_to_worker(X(:,i),y(i),who)

            if (o%printmod>0) write(*,'(A,i4,A,i4,A,i4)') 'GA: image ', who, ' solving for point ',i,' of ',N

            ! For statistical purposes, collect who was assigned that feval
            fevaldsn(who) = fevaldsn(who) + 1

        end do

        if (o%dbglvl>1) print*,'waiting for everyone to finish'

        ! Wait for everyone to finish
        do while (any(.not. finishedtask))

            if (o%dbglvl>1) print*,'finishedtask',finishedtask
            if (o%dbglvl>1) print*,'requestlist',requestlist
            do who = 1,mympi_nproc()-1
                if (.not. finishedtask(who)) then
                    finishedtask(who) = mympi_test(requestlist(who)) ! requestlist is not valid after mympi_test returns true
                end if
            end do

        end do

        finishedtask = .true.
        slaveeverassignedlist = .false.

        if (o%dbglvl>1) print*,'done initial point distribution, proceeding to main loop'

        do iter = 1,o%maxiter

            ! Wait for a slave. If there were previously assigned a task, then record its value.
            who = wait_for_free_worker()

            ! If they had a previously assigned point, use its value
            if (slaveeverassignedlist(who)) then

                ! CRS assumption:

                ! Replace the worst point if found a better one
                ! itmp = minloc(y) 
                ! if (slaveylist(who)>y(itmp(1))) then
                !     y(itmp(1)) = slaveylist(who)
                !     x(:,itmp(1)) = slavexlist(:,who)
                ! end if

                ! GA assumption: replace a random parent. If elitist, ensure best is not replaced
                itmp = maxloc(y) 
                prop = itmp(1)
                do while (prop==itmp(1)) ! ensures best is never replaced (elitist)
                    call random_number(tmp)
                    tmp = tmp*N + .5d0 ! U[.5,N+.5]
                    prop = nint(tmp) 
                    if (prop<1 .or. prop>N) prop = itmp(1)
                end do 
                y(prop) = slaveylist(who)
                x(:,prop) = slavexlist(:,who)

            end if

            ! After this point, "who" will be assigned an x
            slaveeverassignedlist(who) = .true.

            ! Sort the current points (descending)
            ! tmpy = -y
            ! call sort2(y,tmpI,tmpy) 
            ! y = -y
            ! X = X(:,tmpI) ! sort X

            ! Get the best and the worst points
            itmp = maxloc(y); ib = itmp(1) ! best
            itmp = minloc(y); iw = itmp(1) ! worst

            ! Display progress
            if (iter==1 .or. (o%printmod>0 .and. mod(iter,o%printmod)==0)) then
                write(*,'(A,i8,2(A,e14.6))') 'GA: iter',iter,'  best ',y(ib),'  worst ',y(iw)
            end if

            ! Check stopping criteria
            if (abs(y(ib)-y(iw))<o%yabstoler .or. iter==o%maxiter) then

                if (abs(y(ib)-y(iw))<o%yabstoler) then
                    write(*,'(A)') 'GA: convergence criteria satisfied'
                else
                    write(*,'(A)') 'GA: max iterations reached'
                end if

                ! Shutdown and distribute xstar
                xstar = X(:,ib)
                call shutdown_slaves_sharebest()

                ! Wait for everyone to finish 
                call mympi_barrier()

                return

            end if

            ! Generate a new gene by selecting two parents
            if (y(iw)<0._rt) stop 'ERROR: GA: function values should always be positive'
            call gen_gene(slavexlist(:,who),X,y)

            ! Request that they solve for x and store it in slavey
            call assign_xy_to_worker(slavexlist(:,who),slaveylist(who),who)

            fevaldsn(who) = fevaldsn(who) + 1

        end do

    contains

        subroutine gen_gene(new,X,y)
            use mod_discreteAR1
            implicit none
            integer, intent(out) :: new(:) ! new gene
            integer, intent(in) :: X(:,:) ! pool (size new x size y)
            real(rt), intent(in) :: y(:) ! fitnesses
            ! local
            real(rt), allocatable :: prob(:)
            real(rt) :: tmp
            integer :: p1,p2,c1,c2,i,j

            allocate(prob(size(y)))

            ! select two parents according to their fitness
            prob = y - minval(y) 
            prob = prob/sum(prob) ! always drops the worst one

            p1 = drawIid(prob,0)

            prob(p1) = 0d0
            prob = prob/sum(prob)

            p2 = drawIid(prob,0)

            ! select double crossover
            ! call random_number(tmp)
            ! tmp = tmp*size(X,1)
            ! tmp = tmp+.5d0
            ! c1 = nint(tmp)
            ! c1 = min(max(1,c1),size(X,1))

            c1 = 1+mod(myirand(),size(X,1))

            ! call random_number(tmp)
            ! tmp = tmp*size(X,1)
            ! tmp = tmp+.5d0
            ! c2 = nint(tmp)
            ! c2 = min(max(1,c2),size(X,1))

            c2 = 1+mod(myirand(),size(X,1))

            if (c1>c2) then
                i = c1
                c1 = c2
                c2 = i
            end if

            if (c2<c1) stop 'Failed to generate crossover points'

            ! do double crossover
            new(1:c1) = X(1:c1,p1)
            new(c1+1:c2) = X(c1+1:c2,p2)
            new(c2+1:size(new)) = X(c2+1:size(new),p1)

            if (size(new)<10000) then
                ! do bit mutation
                do i = 1,size(new)
                    call random_number(tmp)
                    if (tmp<=o%mutationprob) new(i) = mod(new(i)+1,2)
                end do
            else
                ! need to replace o%mutationprob percent bits randomly.
                ! for large sizes, a more efficient way to do this is to choose o%mutationprob*size(new) indices which
                ! should be accurate roughly via a law of large numbers
                do i = 1,max(1,int(size(new)*o%mutationprob)) ! okay if bigger than size(new)
                    j = 1+mod(myirand(),size(new))
                    new(j) = mod(new(j)+1,2)
                end do
            end if

        end subroutine 

        integer function myirand()
            implicit none
            real(8) :: tmp
            integer, parameter :: bigi = huge(0)-1
            call random_number(tmp)
            myirand = int(bigi*tmp)
        end function 

        ! This routine sends the value of x to a slave "who", who will then store the value in y. 
        subroutine assign_xy_to_worker(x,y,who)
            use iso_c_binding, only: c_loc,c_ptr,c_f_pointer ! GETS THE ACTUALY ADDRESS OF Y WHICH WILL PERSIST AFTER CALL TO ASSIGN ENDS
            implicit none
            integer, intent(in), target :: x(:)
            real(rt), intent(out), target :: y
            integer, intent(in) :: who ! who you want assigned
            ! local
            type(c_ptr) :: yptr_c ! pointer in C
            real(rt), pointer :: yptr_f ! pointer in Fortran 
            integer :: flag,ierr

            if (o%dbglvl>1) print*,'MASTER: assign_xy: x',x(1:nx)

            ! Assign it
            flag = 1
            finishedtask(who) = .false.

            if (o%dbglvl>0) print*,'MASTER: SENDING FLAG: ',flag

            call mympi_send(flag,who) ! blocking send (tell them to proceed to execute)

            if (o%dbglvl>1) print*,'MASTER: FINISHED SENDING FLAG'
            if (o%dbglvl>1) print*,'MASTER: who',who
            if (o%dbglvl>1) print*,'MASTER: SENDING X: ',x(1:nx)

            call mympi_send(x,who) ! assign an x value (wait for it to complete)

            if (o%dbglvl>0) print*,'MASTER: IRECVING Y'

            ! In order to be able to do this assignment where the memory location associated with y will be stored even after 
            ! the subroutine is exited, the safest/most reliable way I have found is to send MPI the actual memory address.
            ! This can be accomplished by getting the memory address of y using the F2003 intrinsic "c_loc" and assigning that
            ! to a C pointer. Then, using c_f_pointer, copy that to a Fortran pointer. Then pass that Fortran poitner to MPI. 
            !
            ! This may be a hack (it seems like completely legal Fortran though), but it simplifies bookkeeping substantially.
            ! I.e., it becomes "set it and forget it."

            yptr_c = c_loc(y) ! get the actual address of y using "c_loc" and have yptr_c point to that
            call c_f_pointer(yptr_c,yptr_f) ! get a fortran pointer to that same address
            call MPI_Irecv(yptr_f, 1, mympi_rt, who, who, MPI_COMM_WORLD, requestlist(who), ierr)

        end subroutine 

        ! a routine to shut down all slaves
        subroutine shutdown_slaves_sharebest() 
            implicit none
            ! local
            integer :: i
            logical :: ft

            ! Cycle through and shutdown
            do i = 1,mympi_nproc()-1

                ! Wait for rank to finish if not finished task
                if (.not. finishedtask(i)) then

                    ! Infinite loop, wait for them to finish 
                    do 

                        ! Check whether i has finished
                        ft = mympi_test(requestlist(i))
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            exit
                        end if

                        ! If sleeptime, then sleep
                        if (o%sleeptime>=1) then

                            call sleep(o%sleeptime)

                        else

                            ! Do some busy work so as to not continually check the status (which may cause MPI problems)
                            call short_sleep()

                        end if

                    end do

                end if
                    
                ! Shut down, sending flag = -1 and xstar
                call mympi_send(-1,i) 
                call mympi_send(xstar,i) 

            end do


        end subroutine


        ! a routine returning the index of a free slave
        ! if no worker is available, who is -1
        function wait_for_free_worker() result(who)
            implicit none
            integer :: who
            ! local
            integer :: i
            logical :: ft

            do 

                ! Infinite loop
                do i = 1,mympi_nproc()-1

                    if (finishedtask(i)) then

                        who = i
                        return  

                    else

                        ! Check whether i has finished
                        if (o%dbglvl>1) print*,'MASTER: TESTING MESSAGE'
                        if (o%dbglvl>2) print*,'MASTER: finishedtask',finishedtask
                        if (o%dbglvl>2) print*,'MASTER: REQUESTLIST',REQUESTLIST
                        if (o%dbglvl>2) print*,'MASTER: i: ',i

                        ft = mympi_test(requestlist(i))
                        if (o%dbglvl>1) print*,'MASTER: FINISHED TEST'
                        if (ft) then
                            finishedtask(i) = .true. ! show that they are now free
                            who = i
                            if (o%dbglvl>0) print*,'MASTER: RECEIVED MESSAGE'
                            return
                        end if

                    end if

                end do

                ! If sleeptime, then sleep
                if (o%sleeptime>=1) then

                    call sleep(o%sleeptime)

                else

                    ! Do some busy work so as to not thrash the system
                    call short_sleep()

                end if

            end do

        end function

        ! Do something according to the value of flag(this_image())[1]
        !  1 := evaluate slavex and store in slavey
        ! -1 := exit the slave routine
        subroutine slave()
            implicit none
            integer :: flag
            real(rt) :: slavey
            integer, allocatable :: slavex(:)

            allocate(slavex(nx))

            do 

                ! Wait for a flag 
                if (o%dbglvl>0) print*,'SLAVE: WAITING TO RECEIVE FLAG'
                call mympi_recv(flag,0)
                if (o%dbglvl>1) print*,'SLAVE: flag is ',flag

! call sleep(1)

                ! Wait for an x
                if (o%dbglvl>0) print*,'SLAVE: WAITING TO RECEIVE SLAVEX'
                call mympi_recv(slavex,0)
                if (o%dbglvl>1) print*,'SLAVE: slavex is ',slavex

                ! If flag==1, execute f(x) and store on master.
                if (flag==1) then

                    ! Compute the result
                    slavey = f(slavex)

                    ! Send the result to master tagged by mympi_iproc()
                    if (o%dbglvl>0) print*,'SLAVE: SENDING SLAVEY'
                    call mympi_send(slavey,0,mympi_iproc()) 
                    if (o%dbglvl>0) print*,'SLAVE: FINISHED SENDING SLAVEY'

                end if

                if (flag==-1) then
                    xstar = slavex

                    ! Wait for everyone to finish
                    call mympi_barrier()
                    exit
                end if

            end do

        end subroutine


    end subroutine

    subroutine short_sleep()
        use mod_mympi
        implicit none
        real(8) :: tic,toc,tmp ! leave as double precision
        integer :: uid

        tic = MPI_WTIME()
        toc = tic
        tmp = 0d0
        do while (toc-tic<1d-6)
            toc = MPI_WTIME()
            tmp = tmp + 1d0
        end do

        ! Trick the compiler into not optimizing away the busy work (hopefully)
        uid = 152
        open(uid,file='/lkandsfkasn/makecompilernotoptimize.dat',status='unknown',err=999) ! file should not exist (but could), so this will jump to line 999
        write(uid,*) tmp
        close(uid)
999     tmp = 0d0

    end subroutine 


end module mod_minNdim

module mod_minNdim_test
    use mod_minNdim
    implicit none
contains

    function mysumf(x) result(y)
        use mod_types
        implicit none
        integer, intent(in) :: x(:)
        real(rt) :: y
        y = sum(real(x,rt)) / real(size(x),rt)
    end function 

    subroutine test_GA_MPI()
        implicit none
        integer, allocatable :: x(:)

        allocate(x(10001))
        x = 0
        call sub_ga_MPI(x,mysumf,optimset('maxiter=1000000'))

    end subroutine 

    function testfn_noisyLM(x,nf) result(y)
        use mod_matlab
        implicit none
        real(rt), intent(in) :: x(:)
        integer, intent(in) :: nf
        real(rt) :: y(nf)

        ! y = sin(x) - real(colon(1,size(x)),8)*(1e0_rt-cos(x)) + sum(cos(x)) - size(x) ! + floor(mod(exp(x)-1+.2,1)) + exp(.3*x)-1

!         y = log(abs(x(1)*x(2) + sin(x(1)) - exp(cos(x(2)))))

        ! x = exp(x)
        y = x - sin(x)**2



    end function 
!     subroutine test_parnoisyLM
!         use mod_mympi
!         implicit none
!         real(rt) :: x(2)
! 
!         x = [-5e0_rt,-5e0_rt]
!         call mympi_barrier()
!         call sub_parnoisyLM(x,testfn_noisyLM,size(x),1d-3 + 0e0_rt*x,ndraws=1000,opts=optimset('maxiter',300e0_rt))
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',testfn_noisyLM(x,size(x))
! 
!     end subroutine 
!     subroutine test_noisyLM
!         use mod_mympi
!         implicit none
!         real(rt) :: x(2)
!         if (mympi_iproc()==0) then
!             x = [-5e0_rt,-5e0_rt]
!             call sub_noisyLM(x,testfn_noisyLM,size(x),1d-3 + 0e0_rt*x,opts=optimset('maxiter',300e0_rt))
!             write(*,*) 'x: ',x
!             write(*,*) 'fx: ',testfn_noisyLM(x,size(x))
!         end if
!     end subroutine 

    subroutine test_opt_along_dim
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)

        call sub_opt_along_dim(x,rosenbrock,-1._rt,1._rt,2)

    end subroutine 

    function lm_foo(x,m) 
        use mod_types
        implicit none
        real(rt), dimension(:), intent(in) :: x
        integer, intent(in) :: m
        real(rt) :: lm_foo(m)
        ! local
        integer :: i

        ! If m==size(x), then solution for min SSRS problem is 1,2,3,...
        do i = 1,m
            lm_foo(i) = abs(x(min(i,size(x)))-i)**2
        end do
        print*,'x',x

    end function

    subroutine test_levenbergmarquardt()
        implicit none
        integer, parameter :: m = 3, n = 3
        real(rt) :: xstar(n)
        xstar = 10._rt
        call sub_levenbergmarquardt(xstar,lm_foo,m,optimset('yreltoler=1d-6,xreltoler=1d-6,fcnerr=1d12,printmod=1')) ! fcnerr=1d16 step size is 100%, 1d14 by 10%, 1d12 by 1%
        if (.not. all(abs(xstar-[1._rt,2._rt,3._rt])<1e-6_rt)) stop 'test_levenbergmarquardt: failed'
        print*,'test_levenbergmarquardt: passed'
    end subroutine

    subroutine test_parmin
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_parmin_MPI(x,rosenbrock,opt=0,e=abs(x)*.5e0_rt)
        call sub_parmin_OMP(x,rosenbrock,x*0e0_rt,x*10e0_rt,[.1e0_rt,.1e0_rt,.1e0_rt],optimset('printmod=1'))
        stop 'checking test_parmin'
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 
!     subroutine test_subplex
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_subplex(x,rosenbrock)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)
!     end subroutine 
    subroutine test_simplex
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_simplex(x,rosenbrock,toler_in=1e-6_rt,maxiter_in=10000,printmod_in=1,randomize_in=.false.)
        ! call sub_simplex(x,rosenbrock,optimset('toler=1d-6,maxiter=10000,printmod=15'))
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 
    subroutine test_parallelsimplex
        use mod_mympi
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_parallelsimplex(x,rosenbrock,toler_in=1e-6_rt,maxiter_in=10000,printmod_in=1)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 


    subroutine test_parallelsimanneal
        implicit none
        real(rt) :: x(3),a(3),b(3)
        x = (/.3e0_rt,.1e0_rt,.5e0_rt/)
        a = 0.e0_rt
        b = (/1.2e0_rt,3e0_rt,99e0_rt/)
        call random_seed
        call sub_parallelsimanneal(x,rosenbrock,a_in=a,b_in=b,e_in=1e0_rt*(/.01e0_rt,.01e0_rt,.01e0_rt/),&
                                   printmod_in=1000,maxiter_in=10000,T0_in=.001e0_rt)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 

    subroutine test_simanneal
        implicit none
        real(rt) :: x(3),a(3),b(3)
        x = (/.3e0_rt,.1e0_rt,.5e0_rt/)
        a = 0.e0_rt
        b = (/1.2e0_rt,3e0_rt,99e0_rt/)
        call random_seed
        call sub_simanneal(x,rosenbrock,a_in=a,b_in=b,e_in=1e0_rt*(/.01e0_rt,.01e0_rt,.01e0_rt/),printmod_in=1000,maxiter_in=10000,T0_in=.001e0_rt)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 

    subroutine test_parCRS_MPI
        use mod_mympi
        implicit none
        real(rt) :: x(3),a(3),b(3),e(3)

        a = 0.e0_rt
        b = [1.2e0_rt,3e0_rt,99e0_rt]

        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_parCRS_MPI(x,rosenbrock,a,b,opts=optimset('printmod=5,sleeptime=0,yabstoler=1d-10,maxiter=125'))
        if (mympi_iproc()==0) then
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',rosenbrock(x)
        end if

        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_parCRS_MPI(x,rosenbrock,a,b,opts=optimset('printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=10000'))
        if (mympi_iproc()==0) then
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',rosenbrock(x)
        end if

        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        e = [.01_rt,.01_rt,.01_rt]
        call sub_parCRS_MPI(x,rosenbrock,a,b,e_in=e,opts=optimset('printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=10000,poolsize=2'))
        if (mympi_iproc()==0) then
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',rosenbrock(x)
        end if

    end subroutine 

!     subroutine test_CRS
!         implicit none
!         real(rt) :: x(3),a(3),b(3)
!         x = [.3e0_rt,.1e0_rt,.5e0_rt]
!         a = 0.e0_rt
!         b = [1.2e0_rt,3e0_rt,99e0_rt]
!         call sub_CRS(x,rosenbrock,a,b,optimset('xabstoler=1d-12,printmod=2500,yabstoler=1d-12,madeuparg=200,maxiter=12345'))
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)
!     end subroutine 


    subroutine test_gridsearch
        use mod_matlab, only: linspace
        implicit none
        real(rt) :: x(3),grids(40,3)
        integer :: n(3)

        n = (/19,29,05/)
        grids(1:n(1),1) = linspace(0e0_rt,2e0_rt,n(1))
        grids(1:n(2),2) = linspace(0e0_rt,2e0_rt,n(2))
        grids(1:n(3),3) = linspace(0e0_rt,2e0_rt,n(3))
        call sub_gridsearch(x,rosenbrock,n,grids,printmod_in=1000)

        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)

    end subroutine

    subroutine test_shrinkingcube
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_shrinkingcube(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine 

    subroutine test_parrandom
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_parrandom(x,rosenbrock,printmod_in=100)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine 

    subroutine test_newtonNdim
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_newtonNdim(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

!     subroutine test_compass
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_compass(x,rosenbrock,shrink_in=.5e0_rt,printmod_in=50)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine
! 
!     subroutine test_sequentialmin
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_sequentialmin(x,rosenbrock,shrink_in=.5e0_rt,printmod_in=1)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine 

    subroutine test_gradDescent
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_gradDescent(x,rosenbrock,maxiter_in=10000,xtoler_in=1e-10_rt,toler_in=1e-10_rt,printmod_in=1000)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_stochGradDescent
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_stochGradDescent(x,rosenbrock,maxiter_in=10000,xtoler_in=1e-10_rt,printmod_in=1000,lambda_in=.1e0_rt)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_BFGS
        implicit none
        real(rt) :: x(3)
        x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
        call sub_BFGS(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

!     subroutine test_smolyaksearch
!         implicit none
!         real(rt) :: x(3)
!         x = (/.3e0_rt, .1e0_rt, .5e0_rt /)
!         call sub_smolyaksearch(x,rosenbrock,e_in=.1e0_rt*(/10e0_rt,10e0_rt,10e0_rt/),mu_in=2,maxiter_in=1000,xtoler_in=1d-6)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine

    subroutine test_numGradHess
        use mod_deriv
        use mod_matlab
        implicit none
        real(rt) :: x(3), grad(3), hess(3,3), gradT(3), hessT(3,3), err

        hessT = reshape((/2e0_rt,3e0_rt,0e0_rt,3e0_rt,2e0_rt,3e0_rt,0e0_rt,3e0_rt,2e0_rt/),(/3,3/)) ! Not a function of x

        x = 0e0_rt
        gradT = 0e0_rt
        call sub_numGradHess(x,quadratic,(/1e0_rt,1e0_rt,1e0_rt/),grad,hess)
        err = maxval(abs(gradT-grad))
        err = max(err,maxval(abs(hessT-hess)))
!         print*,'Gradient',grad
!         print*,'Hess'
!         call disp(hess)

        x = 1e0_rt
        gradT = (/5e0_rt,8e0_rt,5e0_rt/)
        call sub_numGradHess(x,quadratic,(/1e0_rt,1e0_rt,1e0_rt/),grad,hess)
        err = max(err,maxval(abs(gradT-grad)))
        err = max(err,maxval(abs(hessT-hess)))
!         print*,'Gradient',grad
!         print*,'Hess'
!         call disp(hess)

        if (err>1d-07) then
            print*,'Error was: ',err
            STOP 'test_numGradHess: failed'
        else
            print*,'test_numGradHess: passed'
        end if

    end subroutine

    function quadratic(x) result(f)
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt) :: f
        !local
        real(rt) :: tmp
        integer :: i

        tmp = sum(x**2)
        do i = 1,size(x)-1
            tmp = tmp + 3e0_rt*x(i)*x(i+1)
        end do
        f = tmp

    end function

    function rosenbrock(x) result(f)
        use mod_util
        implicit none
        real(rt), dimension(1:), intent(in) :: x
        real(rt) :: f
        !local
        integer :: i,n
        real(rt) :: f_tmp, tmp

        call random_number(tmp)
        if (tmp<.0001_rt) then
            tmp = -1._rt
            f = sqrt(tmp)
            return
        end if

        n = size(x)
        f_tmp = 0.e0_rt
        do i = 1,n-1
             f_tmp = f_tmp + 100.e0_rt*(x(i+1)-x(i)**2)**2 + (1.e0_rt-x(i))**2
        end do
        f = f_tmp

!         call sleep(1)

    end function
end module mod_minNdim_test

!     private :: minN_f_template
! 
!     abstract interface
!         function minN_f_template_w_cache(x,f,cache) result(y)
!            use mod_minNdim_w_cache ! only way I could figure out how to declare minN_f_template and minN_f_cache here
!            implicit none
!            real(rt), intent(in) :: x(1:) 
!            procedure(minN_f_template) :: f
!            type(minN_f_cache), intent(inout) :: cache
!            real(rt) :: y
!        end function
!     end interface
! 
! module mod_minNdim_w_cache
! 
!     type minN_f_cache 
!         real(rt), allocatable :: x(:,:)
!         real(rt), allocatable :: y(:)
!         integer :: n,maxn
! !     contains
! !         procedure, pass :: init => init_cache
! !         procedure, pass :: put => put_cache
! !         procedure, pass :: check => check_cache
! !         procedure, pass :: free => empty_cache
!     end type
! 
!     abstract interface
!         function minN_f_template(x) result(y)
!             implicit none
!             real(rt), intent(in) :: x(1:)
!             real(rt) :: y
!         end function 
!     end interface
! 
! contains
! 
!     ! The function g will first search the cache and if not found, will evaluate f
!     function f_without_cache(x,f,cache) result(g)
!         implicit none
!         procedure(minN_f_template) :: f
!         type(minN_f_cache), intent(inout) :: cache ! not used
!         real(rt), intent(in) :: x(1:)
!         real(rt) :: g
! 
!         g = f(x)
! 
!     end function
! 
!     function f_with_cache(x,f,cache) result(g)
!         implicit none
!         procedure(minN_f_template) :: f
!         type(minN_f_cache), intent(inout) :: cache
!         real(rt), intent(in) :: x(1:)
!         real(rt) :: g
!         ! local
!         real(rt) :: h
!         integer :: flag
! 
!         call check_cache(cache,x,h,flag)
! 
!         if (flag == 1) then
!             ! Cached value found
!             g = h
!             return
!         else
!             ! Cached value not found
!             h = f(x)
!             call put_cache(cache,x,h)
!             g = h
!             return
!         end if
! 
!     end function
! 
!     subroutine init_cache(cache,nx,maxn)
!         implicit none
!         type(minN_f_cache), intent(out) :: cache
!         integer, intent(in) :: nx, maxn
! 
!         cache%maxn = maxn
!         cache%n = 0
!         allocate(cache%x(nx,maxn))
!         allocate(cache%y(maxn))
! 
!     end subroutine
!      
!     subroutine free_cache(cache)
!         implicit none
!         type(minN_f_cache), intent(out) :: cache
!         if (allocated(cache%x)) deallocate(cache%x)
!         if (allocated(cache%y)) deallocate(cache%y)
!     end subroutine
! 
!     subroutine put_cache(cache,x,fx)
!         implicit none
!         real(rt), intent(in) :: x(1:),fx
!         type(minN_f_cache), intent(inout) :: cache
! 
!         if (cache%n>=cache%maxn) then
!             print*,'WARNING: cache size exceeded, not caching more values' 
!         else
!             cache%n= cache%n + 1
!             cache%x(:,cache%n) = x
!             cache%y(cache%n) = fx
!         end if
! 
!     end subroutine 
! 
!     subroutine check_cache(cache,x,fx,flag)
!         implicit none
!         type(minN_f_cache), intent(in) :: cache
!         real(rt), intent(in) :: x(1:)
!         real(rt), intent(inout) :: fx ! Overwritten if found in cache
!         integer, intent(out) :: flag
!         ! local
!         integer :: i
! 
!         flag = 0
!         ! Search the cache
!         do i = 1,cache%n
!             if (all(x==cache%x(:,i))) then
!                 flag = 1
!                 fx = cache%y(i)
!                 return
!             end if
!         end do
!         
!     end subroutine 
!     
! end module 
