program main
    use mod_mympi
    use mod_main
    implicit none
    character(50) :: arg
    character(7*50+7) :: arglist,fname
!     character(1) :: y_or_n
    real(8), allocatable :: soln(:,:,:),guess(:,:,:)
    integer :: model,dims(2),opt,maxTime,maxmethod,shocktype,nmultigrid,mgdims(2),i,unitno
    logical :: useStdOut

    call mympi_init()

!     print*,'testing accgegm'; call test_accgegm(); stop

    ! Command line argument list
    if (command_argument_count()/=8) then
        write(*,*) 'Please call the program again with 8 command line arguments specified as follows'
        write(*,*) '#1: the model, 0 for RBC or 1 for Arellano'
        write(*,*) '#2: the number of asset states (nK)'
        write(*,*) '#3: the number of tfp states (nZ)'
        write(*,*) '#4: the assumption about monotonicity and concavity'
        write(*,*) '    -- options 1-3 have no monotonicity, 4-6 simple mono, and 7-9 binary mono'
        write(*,*) '    -- options 1,4,7 have no concavity, 2,5,8 have simple conc, and 3,6,9 have binary conc'
        write(*,*) '#5: shock type'
        write(*,*) '    -- 0 discrete shocks'
        write(*,*) '    -- 1 continuous shocks integrated assuming linear interpolation between shock values'
        write(*,*) '#6: maximization method'
        write(*,*) '    -- 0 grid search '
        write(*,*) '    -- 1 grid search followed by local search'
        write(*,*) '    -- 2 accelerated GEGM'
        write(*,*) '    -- 3 grid search exploiting monotonicity in k and z (RBC only)'
        write(*,*) '#7: upper bound on run time (if exceeded, code exits)'
        write(*,*) '#8: display lots of info to standard out and print graphs (0 false 1 true)'
        STOP
    end if

    ! Get model, problem size, and other info
    arglist = ''
    call get_command_argument(1,arg); arglist = trim(arglist)//trim(arg)//'_'; read(arg,*) model
    call get_command_argument(2,arg); arglist = trim(arglist)//trim(arg)//'_'; read(arg,*) dims(1)
    call get_command_argument(3,arg); arglist = trim(arglist)//trim(arg)//'_'; read(arg,*) dims(2)
    call get_command_argument(4,arg); arglist = trim(arglist)//trim(arg)//'_'; read(arg,*) opt
    call get_command_argument(5,arg); arglist = trim(arglist)//trim(arg)//'_'; read(arg,*) shocktype
    call get_command_argument(6,arg); arglist = trim(arglist)//trim(arg)//'_'; read(arg,*) maxmethod 
    call get_command_argument(7,arg); arglist = trim(arglist)//trim(arg);      read(arg,*) maxTime

    call get_command_argument(8,arg); read(arg,*) i
    useStdOut = i==1

    ! Now call the routine
    nmultigrid=0
    do i = 0,nmultigrid
        mgdims = nint(dims*(real(i+1,8)/(nmultigrid+1)))
        ! mgdims(1) = nint(dims(1)*(real(i+1,8)/(nmultigrid+1)))
        ! mgdims(2) = dims(2)
        call sub_main(model,mgdims,opt,maxTime,useStdOut,shocktype,maxmethod,doextras=i==nmultigrid,guess=guess,soln=soln)
        if (i/=nmultigrid) then
            if (i>0) deallocate(guess)
            allocate(guess(size(soln,1),size(soln,2),size(soln,3)))
            guess = soln
        end if
    end do

!     y_or_n = 'n'
!     write(*,'(A)',advance='no') 'Would you like to save the model solution? y/(n) : '
!     read(*,'(A1)', err=999) y_or_n 
! 999 y_or_n = y_or_n
! 
!     select case (y_or_n)
!         case ('y')
!             fname = 'results/'//trim(arglist)//'.dat'
!             open(newunit=unitno,file=trim(fname),status='unknown',action='write')
!             write(unitno,*) soln
!             close(unitno)
!             write(*,'(A)') 'File has been saved in binary form as '//trim(fname)
!         case default
!             write(*,'(A)') 'File has not been saved.'
!     end select

end program
