! Fortran routine for A Practical Approach to Testing Calibration Strategies
! Grey Gordon, 2016
! You may use this software and modify it provided you leave this notice and do not sell or license it to anyone.
! Proper attributions have been attempted, but if you find one missing, please contact greygordon </AT/> gmail.com
! The main code is mod_crs.
module mod_types_private
    implicit none
    include 'mpif.h'
end module 
module mod_types
    use mod_types_private, only: MPI_REAL,MPI_DOUBLE_PRECISION
    implicit none

    ! Single precision
    ! integer, parameter :: rt = 4 
    ! integer, parameter :: mympi_rt = MPI_REAL 

    ! Double precision
    integer, parameter :: rt = 8
    integer, parameter :: mympi_rt = MPI_DOUBLE_PRECISION 

    private 
    public :: rt, mympi_rt
end module 
! A collection of core routines used by many of the libgrey routines. These are designed to be basic, i.e., they have no dependencies.

! Routines
module mod_core
    use mod_types
    implicit none

    interface disp
        module procedure disp_scal, disp_vec, disp_mat, disp_3d, disp_vec_i, disp_mat_i, &
                         disp_scal_title, disp_vec_title, disp_mat_title, disp_3d_title, &
                         disp_vec_title_i, disp_mat_title_i, disp_mat_title_l, disp_mat_l
    end interface disp

    integer(8), private :: datetime_start(8)


contains

! Define a function continue to prevent the use of 
integer function continue()
    implicit none
    print*,'Don''t use continue, use cycle. Continue it doesn''t do what you think!'
    continue = -1
end function

! Produces a square n by n identity matrix
function eye(n) result(mat)
    implicit none
    integer, intent(in) :: n
    real(rt), dimension(n,n) :: mat    
    integer :: i

    mat = 0._rt
    do i = 1,n
        mat(i,i) = 1._rt
    end do

end function eye

subroutine disp_mat_title_l(title,X)
    implicit none
    character(len=*), intent(in) :: title
    logical, dimension(1:,1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_mat_l(X)
end subroutine 
subroutine disp_mat_l(X)
    implicit none
    logical, dimension(1:,1:), intent(in) :: X    
    integer :: i
    character(len=5) :: dim2
    write(dim2,'(i5)') size(X,2)
    do i = 1,size(X,1)
        write(*,'('//dim2//'L2)') X(i,:)
    end do
end subroutine disp_mat_l

subroutine disp_mat_title_i(title,X)
    implicit none
    character(len=*), intent(in) :: title
    integer, dimension(1:,1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_mat_i(X)
end subroutine 
subroutine disp_mat_i(X)
    implicit none
    integer, dimension(1:,1:), intent(in) :: X    
    integer :: i
    character(len=5) :: dim2
    write(dim2,'(i5)') size(X,2)
    do i = 1,size(X,1)
        write(*,'('//dim2//'i8)') X(i,:)
    end do
end subroutine disp_mat_i

subroutine disp_vec_title_i(title,X)
    implicit none
    character(len=*), intent(in) :: title
    integer, dimension(1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_vec_i(X)
end subroutine
subroutine disp_vec_i(X)
    implicit none
    integer, dimension(1:), intent(in) :: X    
    integer :: i
    do i = 1,size(X,1)
        write(*,'(i8)') X(i)
    end do
end subroutine disp_vec_i


subroutine disp_vec(X,uid)
    implicit none
    real(rt), dimension(:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),1,1,'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),1,1,'',6)
end subroutine 
subroutine disp_mat(X,uid)
    implicit none
    real(rt), dimension(:,:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),1,'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),1,'',6)
end subroutine 
subroutine disp_3d(X,uid)
    implicit none
    real(rt), dimension(:,:,:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),'',6)
end subroutine 
subroutine disp_vec_title(title,X,uid)
    implicit none
    real(rt), dimension(:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),1,1,title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),1,1,title,6)
end subroutine 
subroutine disp_mat_title(title,X,uid)
    implicit none
    real(rt), dimension(:,:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),1,title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),1,title,6)
end subroutine 
subroutine disp_3d_title(title,X,uid)
    implicit none
    real(rt), dimension(:,:,:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),title,6)
end subroutine 

subroutine disp_core(X,n1,n2,n3,title,uid)
    implicit none
    integer, intent(in) :: n1,n2,n3,uid
    character(len=*), intent(in) :: title
    real(rt), intent(in) :: X(n1,n2,*)
    ! local
    integer :: i1,i2,i3

    ! If no title or uid is present, then use title='' and uid=6
    do i3 = 1,n3
        if (n3>1) then
            write(uid,'(A,i4,A)') title//'(:,:,',i3,')'
        else
            if (title/='') write(uid,'(A)') title
        end if
        do i1 = 1,n1
            do i2 = 1,n2
                if (abs(X(i1,i2,i3))<1d5) then
                    write(uid,'(f16.5)',advance='no') X(i1,i2,i3)
                else
                    write(uid,'(es16.7e3)',advance='no') X(i1,i2,i3)
                end if
            end do
            write(uid,*)
        end do
    end do

end subroutine



subroutine disp_scal_title(title,X)
    implicit none
    character(len=*), intent(in) :: title
    real(rt), intent(in) :: X    
    write(*,'(A)') title
    call disp_scal(X)
end subroutine 
subroutine disp_scal(X)
    implicit none
    real(rt), intent(in) :: X    
    write(*,'(f12.5)') X
end subroutine disp_scal

! Acts just like the colon in Matlab
pure function colon(a,b)
    implicit none
    integer, intent(in) :: a,b
    integer, dimension(b-a+1) :: colon
    integer :: i
    do i = a,b
        colon(i-a+1) = i
    end do
end function colon

! 
subroutine bhunt(ilo,xhat,x)
    implicit none
    integer, intent(inout) :: ilo ! on input, guess; on output, lowerbound on bracketing interval
    real(rt), intent(in) :: xhat
    real(rt), dimension(:), intent(in) :: x(:)

    call bhunt_core(ilo,xhat,x,size(x))

end subroutine

subroutine bhunt_core(ilo,xhat,x,nx)
    implicit none
    real(rt), intent(in) :: x(*),xhat
    integer, intent(in) :: nx
    integer, intent(inout) :: ilo
    ! local
    integer :: l,u,step

    l = ilo 
    u = ilo+1
    step = 1

    ! Expand until xhat is in [x(l),x(u)]
    if (xhat>x(u)) then

        ! Expand to the right
        do 

            ! Check whether expansion is possible
            if (u>=nx) then
                ! Cannot expand to the right, so call bsearch and return
                u = nx
                call local_bsearch()
                return
            end if

            ! Check whether bracketed
            if (xhat<=x(u)) then
                ! Have bracketed
                call local_bsearch()
                return
            end if

            ! If still here, have not bracketed, so shift bracket
            step = 2*step
            l = u
            u = u + step

        end do

    elseif (xhat<x(l)) then

        ! Expand to the left
        do 

            ! Check whether expansion is possible
            if (l<=1) then
                ! Cannot expand to the left, so call bsearch and return
                l = 1
                call local_bsearch()
                return
            end if

            ! Check whether bracketed
            if (xhat>=x(l)) then
                ! Have bracketed
                call local_bsearch()
                return
            end if

            ! If still here, have not bracketed, so shift bracket
            step = 2*step
            u = l
            l = l - step

        end do
 
    else

        ! Here, xhat in [x(l),x(u)] and u == l+1, so return (ilo was correct)
        return
        
    end if

contains

    subroutine local_bsearch()
        implicit none
        integer :: m

        do 

            ! Check convergence criterion
            if (u<=l+1) then
                ilo = l
                return
            end if

            ! Otherwise, divide
            m = (u+l)/2 
            if (xhat>=x(m)) then
                l = m
            else
                u = m
            end if


        end do

    end subroutine

end subroutine 

! Use binary search to find the lowerbound on a bracketing interval for 
! Given sorted ascending array, find lowerbound on bracketing interval. Note that
! ilo will always be in 1,...,size(x)-1.
! 
! This is supposed to work much like the C++ standard library where the interface comes from
subroutine bsearch(ilo,xhat,x)
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), intent(in) :: xhat
    integer, intent(out) :: ilo

    call bsearch_core(ilo,xhat,x,1,size(x))
        
end subroutine 

subroutine bsearch_core(ilo,xhat,x,search_lb,search_ub)
    implicit none
    real(rt), intent(in) :: x(*),xhat
    integer, intent(in) :: search_lb,search_ub
    integer, intent(out) :: ilo
    !local
    integer :: ihi,mid
        
    ilo = search_lb
    ihi = search_ub
    mid = (ihi+ilo)/2 ! mid will have ilo<=mid<ihi as long as ihi>ilo. If ihi=ilo, then mid=ilo, so the 
                      ! while loop is not entered and ilo=1 is returned 
    
    do while (mid>ilo) ! mid=ilo means ihi=ilo+1.

        ! Here, we must have ihi>ilo and ilo<=mid<ihi

        ! Bisect. 
        ! If xhat<x(1), then xhat>=x(mid) will never occur and ilo will never be modified.
        ! If xhat>=x(nx), then xhat>=x(mid) will always occur (note mid<ihi). Consequently, 
        !      ilo = mid occurs until mid==ilo, which only occures if ihi==ilo+1.
        if (xhat>=x(mid)) then
            ilo = mid
        else
            ihi = mid
        end if

        mid = (ihi+ilo)/2 

    end do

end subroutine 

! Given sorted (ascending) lists xhat and x, finds ilo as if call bsearch(ilo(i),xhat(i),x) was called, but does so faster
subroutine bsearch_many(ilo,xhat,x)
    implicit none
    integer, intent(out) :: ilo(:)
    real(rt), intent(in) :: xhat(:),x(:)
    ! local
    integer :: n,nx

    n = size(ilo) 
    nx = size(x)

    if (n>=2) then
        call bsearch_core(ilo(1),xhat(1),x,1,nx)
        call bsearch_core(ilo(n),xhat(n),x,ilo(1),nx)

        if (n==2) return
        call bsearch_many_core(ilo,xhat,x,1,n)

    elseif (n==1) then
        call bsearch(ilo(1),xhat(1),x)
    elseif (n<=0) then
        stop 'ERROR: bsearch_many: size(ilo)<=0'
    end if


end subroutine

! Assumes ilo(lb) and ilo(ub) are known, exploits monotonicity
recursive subroutine bsearch_many_core(ilo,xhat,x,lb,ub)
    implicit none
    integer, intent(inout) :: ilo(*)
    real(rt), intent(in) :: xhat(*),x(*)
    integer, intent(in) :: lb,ub
    ! local
    integer :: m

    m = (lb+ub)/2

    call bsearch_core(ilo(m),xhat(m),x,ilo(lb),ilo(ub)+1) ! need to add 1 to the upper bound because ilo always returns the lower part of the bracketing interval

    if (m>lb+1) call bsearch_many_core(ilo,xhat,x,lb,m)
    if (ub>m+1) call bsearch_many_core(ilo,xhat,x,m,ub)


end subroutine 


! linspace(a,b,n) (of course) constructs a grid of n linearly spaced points between
! a and b. If n==1, then the grid is just "a", consistent with Matlab. If n<=0, the routine
! stops with an error.
function linspace(x1,x2,n) result(grid)
    implicit none
    real(rt), intent(in) :: x1,x2
    integer, intent(in) :: n
    real(rt), dimension(1:n) :: grid
    !local
    integer :: i 

    if (n>1) then
        grid(1) = x1
        do i=2,n-1
            grid(i) = x1 + (x2-x1)*(real(i,rt)-1._rt)/(real(n,rt)-1._rt)
        end do   
        grid(n) = x2
    elseif (n==1) then
        grid(1) = x1
    else
        STOP 'linspace: ERROR: grid has size <= 0'
    end if
end function linspace

! Tic and toc for timing routines.
! Because of the odd behavior of the system clock, there is a known bug here that is 
! more likely to be experienced for long run times.
subroutine tic()
    implicit none
    ! call cpu_time(time_start)
    ! call system_clock(count=count_start,count_rate=countrate_start)
    call date_and_time(values=datetime_start)
end subroutine tic
subroutine toc(time_out)
    implicit none
    real(rt), intent(out), optional :: time_out
    integer(8) :: datetime_end(8), ta(8)
    real(rt) :: minutes,seconds
    logical :: addmonth

    call date_and_time(values=datetime_end)

    ! Starting at the start time, go by minutes and increment
    ta = datetime_start
    minutes = 0d0
    do while (minutes<=144000) ! go for up to 100 days (place the max in case there is an error below)
        if (all(ta([1,2,3,5,6])==datetime_end([1,2,3,5,6]))) exit
        minutes = minutes + 1._rt

        ta(6) = ta(6) + 1 ! add a minute
        if (ta(6)>59) then
            ta(6) = 0
            ta(5) = ta(5) + 1 ! add an hour
            if (ta(5)>23) then
                ta(5) = 0 
                ta(3) = ta(3) + 1 ! add a day
                select case (ta(2))
                    case (2)
                        addmonth = ta(3)>28
                    case (4,6,9,11)
                        addmonth = ta(3)>30
                    case default
                        addmonth = ta(3)>31
                end select
                if (addmonth) then
                    ta(3) = 1 ! reset day to 1
                    ta(2) = ta(2) + 1 ! add a month
                    if (ta(2)>12) then
                        ta(1) = ta(1) + 1 ! add a year
                    end if
                end if
            end if
        end if
    end do

    ! Now have the measure down to minutes, get seconds and milliseconds
    seconds = datetime_end(7) - datetime_start(7) + (datetime_end(8) - datetime_start(8))/1000d0

    if (present(time_out)) then
        time_out = minutes + seconds/60d0
    else
        write(*,'(A,f14.6,A,f14.6,A,f14.6,A,f14.6,A)') 'Elapsed time: ',&
            (minutes+seconds/60d0)/60d0,' (hr) or ',&
            minutes+seconds/60d0,' (min) or ',&
            minutes*60d0 + seconds,' (sec) or ',&
            minutes*60d3 + seconds*1d3,' (ms)'
    end if

    
end subroutine toc


end module mod_core

module mod_core_test
    use mod_core
    implicit none
contains

    subroutine test_bhunt()
        use mod_types
        implicit none
        real(rt), allocatable :: x(:)
        real(rt) :: xhat
        integer :: i,ilo
        
        allocate(x(1000))
        
        x(1) = 0._rt
        do i = 2,size(x)
            call random_number(xhat)
            x(i) = x(i-1) + xhat
        end do

        x = 3._rt*(x-x(1))/(x(size(x))-x(1)) - 1._rt! translate to [-1,2]

        ilo = size(x)/2
        do i = 1,100000
            call random_number(xhat)
            xhat = 4._rt*xhat - 1.5_rt ! translate to [-1.5,2.5]
            ! write(*,'(A,i5)',advance='no') 'iloguess',ilo
            call bhunt(ilo,xhat,x) ! guess is just whatever previous value was
            ! call bsearch(ilo,xhat,x) ! guess is just whatever previous value was
            ! print*,'xhat',xhat,'maxval(x)',maxval(x),'ilo',ilo
            if (ilo>1 .and. ilo<size(x)-1 .and. (xhat<x(ilo) .or. xhat>x(ilo+1))) stop 'ERROR: test_bhunt: failed 1'
            if (ilo==1 .and. xhat>x(2)) stop 'ERROR: test_bhunt: failed 2'
            if (ilo==size(x)-1 .and. xhat<x(size(x)-1)) stop 'ERROR: test_bhunt: failed 3'
        end do

        ! With debug flags on, bhunt takes 16.3 seconds in the lines above. If replaced with bsearch, takes 14.7 seconds. Not much slower. 10^8 evaluations (much of that may be random number generation). When "call random_number(xhat)" is commented, so bhunt has a perfect guess, takes 3.8 seconds (26% of the time relative to bsearch).

        print*,'test_bhunt: passed'

    end subroutine

    subroutine test_bsearch_many()
        use mod_types
        implicit none
        integer, allocatable :: ilo(:),ilo2(:)
        real(rt), allocatable :: xhat(:),x(:)
        real(rt) :: tmp
        integer, parameter :: n = 1000, nx = 570
        real(rt), parameter :: xlb = -3._rt, xub = 4._rt
        real(rt), parameter :: xhatlb = -6._rt, xhatub = 5._rt
        integer :: i,iter

        allocate(x(nx))
        allocate(xhat(n))
        allocate(ilo(n))
        allocate(ilo2(n))

        ! Check that the position agrees with bsearch
        do iter = 1,1000
            call randomgrids()
            call bsearch_many(ilo,xhat,x)
            do i = 1,n
                call bsearch(ilo2(i),xhat(i),x)
            end do
            if (any(ilo/=ilo2)) stop 'test_bsearch_many: failed test 1'
        end do

        print*,'test_bsearch_many: passed'

    contains

        subroutine randomgrids()
            implicit none

            ! Create increasing random lists
            xhat(1) = 0._rt
            do i = 2,n
                call random_number(tmp)
                xhat(i) = xhat(i-1) + tmp
            end do
            xhat = xhat*(xhatub-xhatlb) + xhatlb
            
            x(1) = 0._rt
            do i = 2,nx
                call random_number(tmp)
                x(i) = x(i-1) + tmp
            end do
            x = x*(xub-xlb) + xlb
        end subroutine 

    end subroutine

end module mod_core_test
! Collection of utility programs
! -> isNaN and isInvalidNumber
! -> sleep
! -> getcmdarg_? : gets a requested command argument (the type must be specified by "?", d for real(rt), i for integer, s for string)
! -> *_delim_arg : process a delimited string
! -> conv_i_to_s : convert integer to string
module mod_util
    use mod_types
    implicit none
    private :: getcmdarg_core

    type timer 
        real(8) :: refpt
        logical :: ispaused = .false.
        real(8) :: total = 0d0
    contains
        procedure, pass :: stop => timer_stop
        procedure, pass :: start => timer_start
        procedure, pass :: elapsed => timer_elapsed
    end type

    type type_seed
        integer, allocatable :: saved_seed(:)
    contains
        procedure, pass :: save => seed_save
        procedure, pass :: revert => seed_revert
        procedure, pass :: simpleset => seed_simpleset
        procedure, pass :: randomize => seed_randomize
    end type

contains
    
    subroutine assert(tf,msg,val,ival)
        implicit none
        logical, intent(in) :: tf
        character(len=*), intent(in) :: msg
        real(rt), intent(in), optional :: val
        logical, intent(in), optional :: ival
        if (.not. tf) then
            if (present(val)) then
                write(0,*) 'ERROR: '//msg,val
                stop
            elseif (present(ival)) then
                write(0,*) 'ERROR: '//msg,ival
                stop 
            else
                write(0,*) 'ERROR: '//msg
                stop
            end if
        end if
    end subroutine

    ! https://gcc.gnu.org/onlinedocs/gfortran/RANDOM_005fSEED.html with slight modifications
    subroutine seed_randomize(this)
        use iso_fortran_env, only: int64
        implicit none
        class(type_seed), intent(in) :: this
        ! local
        integer, allocatable :: s(:)
        integer :: i, n, un, istat, dt(8) !, pid
        integer(int64) :: t

        call random_seed(size = n)
        allocate(s(n))
        ! First try if the OS provides a random number generator
        open(newunit=un, file="/dev/urandom", access="stream", &
             form="unformatted", action="read", status="old", iostat=istat)
        if (istat == 0) then
           read(un) s
           close(un)
        else
           ! Fallback to XOR:ing the current time and pid. The PID is
           ! useful in case one launches multiple instances of the same
           ! program in parallel.
           call system_clock(t)
           if (t == 0) then
              call date_and_time(values=dt)
              t = (dt(1) - 1970) * 365_int64 * 24 * 60 * 60 * 1000 &
                   + dt(2) * 31_int64 * 24 * 60 * 60 * 1000 &
                   + dt(3) * 24_int64 * 60 * 60 * 1000 &
                   + dt(5) * 60 * 60 * 1000 &
                   + dt(6) * 60 * 1000 + dt(7) * 1000 &
                   + dt(8)
           end if
!            pid = getpid()
!            t = ieor(t, int(pid, kind(t)))
           do i = 1, n
              s(i) = lcg(t)
           end do
        end if
        call random_seed(put=s)

    contains
        ! This simple PRNG might not be good enough for real work, but is
        ! sufficient for seeding a better PRNG.
        function lcg(s)
          integer :: lcg
          integer(int64) :: s
          if (s == 0) then
             s = 104729
          else
             s = mod(s, 4294967296_int64)
          end if
          s = mod(s * 279470273_int64, 4294967291_int64)
          lcg = int(mod(s, int(huge(0), int64)), kind(0))
        end function lcg
    end subroutine 


    ! Saves seed
    subroutine seed_save(this)
        implicit none
        class(type_seed), intent(inout) :: this
        integer :: lenseed

        if (allocated(this%saved_seed)) deallocate(this%saved_seed)

        call random_seed(size=lenseed)

        allocate(this%saved_seed(lenseed)) 

        call random_seed(get=this%saved_seed)

    end subroutine 

    ! Reverts to a saved seet
    subroutine seed_revert(this)
        implicit none
        class(type_seed), intent(in) :: this
        if (allocated(this%saved_seed)) then
            call random_seed(put=this%saved_seed)
        else
            write(0,*) 'WARNING: seed_revert: revert was called, but no seed had been saved, doing nothing.'
        end if
    end subroutine 

    ! Sets the seed based just on one integer value
    subroutine seed_simpleset(this,i)
        implicit none
        class(type_seed), intent(in) :: this
        integer, intent(in) :: i
        ! local
        integer :: iseed,lenseed
        integer, allocatable :: seed(:)
        call random_seed(size=lenseed)
        allocate(seed(lenseed))
        seed(1) = i     
        do iseed = 2,lenseed
            seed(iseed) = mod(seed(iseed-1)*17+7919,(huge(seed(1))-10000)/20) ! 7919 is prime
        end do
        call random_seed(put=seed)
    end subroutine 
    

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Compute the time
!     subroutine timer_init(this)
!         use omp_lib
!         implicit none
!         class(timer) :: this
!         this%refpt = omp_get_wtime()
!         this%ispaused = .false.
!         this%total = 0d0
!     end subroutine 
    subroutine timer_stop(this)
        use omp_lib
        implicit none
        class(timer) :: this
        this%total = this%total + (omp_get_wtime()-this%refpt)
        this%ispaused = .true.
    end subroutine
    subroutine timer_start(this)
        use omp_lib
        implicit none
        class(timer) :: this
        this%refpt = omp_get_wtime()
        this%ispaused = .false.
    end subroutine
    real(8) function timer_elapsed(this)
        use omp_lib
        implicit none
        class(timer) :: this

        if (this%ispaused) then
            timer_elapsed = this%total 
        else
            timer_elapsed = this%total + (omp_get_wtime()-this%refpt)
        end if

    end function

    subroutine test_util()
        use ieee_arithmetic
        implicit none
        real(rt) :: x
        type(type_seed) :: randstate
        real(rt) :: t,t2,t3,t4
!         character(len=1000) :: str

        ! http://www.ibm.com/support/knowledgecenter/SSGH4D_14.1.0/com.ibm.xlf141.aix.doc/language_ref/isnan.html
        ! NOTE: fftrap must be disabled or taking sqrt(x) will crash.
        x = -1._rt
        if (ieee_support_datatype(x)) then
            if (ieee_support_sqrt(x)) then 
                if (ieee_support_nan(x)) then
                    call assert(ieee_is_nan(sqrt(x)),'ERROR: test_util: ieee_is_nan does not correctly identify sqrt(-1) as being nan')
                end if
            end if
        end if
        call assert(isNaN(sqrt(x)),'ERROR: test_util: ieee_is_nan does not correctly identify sqrt(-1) as being nan')

!         if (.not. isInvalidNumber(log(x))) STOP 'ERROR: isInvalidNumber failed test 1'
        if (isInvalidNumber(3e0_rt)) STOP 'ERROR: isInvalidNumber failed test 2'

        ! str = printx([-tiny(0d0),-huge(0d0)])
        ! print*,'printx',trim(str)

        call randstate%randomize()
        call randstate%save()
        call random_number(t)

        call randstate%simpleset(120480)
        call random_number(t2)
        call assert(t2/=t,'t2 should not equal t')

        call randstate%simpleset(120480)
        call random_number(t3)
        call assert(t2==t3,'t2 must equal t3')

        call randstate%revert()
        call random_number(t4)
        call assert(t==t4,'t must equal t4')

        print*,'test_util: passed'

    end subroutine 

    ! NOTE: this may not work depending on compiler options. E.g., if -ffast-math is enabled, then IEEE arithmetic rules
    ! are ignored.
!     function isNaN(x) result(tf)
!         use ieee_arithmetic, only: ieee_is_nan ! F2003 intrinsic
!         implicit none
!         real(rt) :: x
!         logical :: tf
! 
!         ! NaN is not equal to anything, not even itself
!         tf = ieee_is_nan(x)
! 
!     end function 
    elemental function isNaN(x) result(tf)
        implicit none
        real(rt), intent(in) :: x 
        logical :: tf

        ! NaN is not equal to anything, not even itself
        tf = x/=x

    end function 

    ! This works even with ffast math
    elemental function isInvalidNumber(x) result(tf)
        implicit none
        real(rt), intent(in) :: x
        logical :: tf

        ! This should evaluate to false if x is a normal number (not infinite, not NaN)
        tf = .not. ( x>=-huge(0d0) .and. x<=huge(0d0))

    end function 


    ! http://stackoverflow.com/questions/6931846/sleep-in-fortran
    subroutine sleep(sec)
        use, intrinsic :: iso_c_binding, only: c_int
        implicit none
        integer, intent(in) :: sec
        integer (c_int) :: sec_c
        integer (c_int) :: err

        interface
            function callcsleep(seconds)  bind ( C, name="sleep" )
                use, intrinsic :: iso_c_binding, only: c_int
                import
                implicit none
                integer (c_int) :: callcsleep
                integer (c_int), value :: seconds
            end function 
        end interface

        sec_c = int(sec,c_int)

        err = callcsleep(sec_c)

    end subroutine

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! DISPLAY A DOUBLE PRECISION VECTOR IN A NICE FORMAT
    function printx(x) result(strout)
        implicit none
        integer, parameter :: stride = 17
        real(rt), intent(in) :: x(:)
        character(len=1+stride*size(x)) :: strout
        ! local
        character(len=1+stride*size(x)) :: str
        integer :: ix

        if (1+stride*size(x)>len(str)) then
            str = 'size(x) is too large, cannot display'
            return 
        end if

        str = '' ! blank it out
        write(str(1:1),'(A)') '['
        do ix = 1,size(x)-1
            write(str(2+stride*(ix-1):1+stride*ix),'(ES16.8E3,A)') x(ix),','
        end do
        write(str(2+stride*(ix-1):1+stride*ix),'(ES16.8E3,A)') x(size(x)),']'

        strout = str
        do ix = 1,1+stride*size(x)
            if (str(ix:ix)=="E") strout(ix:ix)="D"
        end do


    end function


    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! PARSING
    ! Given a delimited string, such as "xtoler=1d-6,printmod=1" (in this case, the delimiter could be ","), gives the ith
    ! argument (1 would be "xtoler=1d-6" and 2 would be "printmod=1" in this example). There is also a routine to determine
    ! how many arguments there are.
    function count_delim_arg(str,del) result(n)
        implicit none
        character(len=*) :: str
        character(len=1) :: del
        integer :: n
        ! local
        integer :: i,nloc

        nloc = 0
        do i = 1,len(str)
            if (str(i:i)==del) nloc = nloc + 1
        end do
        n = nloc + 1

    end function 
    function get_delim_arg(str,del,i) result(arg)
        implicit none
        character(len=*) :: str
        character(len=1) :: del
        integer, intent(in) :: i
        character(len=len(str)) :: arg
        ! local
        integer :: j,n,ta,tb

        if (i>count_delim_arg(str,del)) then
            arg = 'error_getting_arg'
            return
        end if

        n = 0
        tb = 1 
        do j = 1,len(str)
            if (str(j:j)==del .or. j==len(str)) then
                ta = tb
                tb = j
                n = n + 1
                if (n==i) exit
            end if
        end do

        ! Here, should have str(ta:tb) almost as the argument. 
        ! However, str(ta) and str(tb) may or may not be delimiters.
        if (ta>1) then 
            ta = ta + 1
        end if
        if (tb<len(str)) then
            tb = tb - 1
        end if

        ! Output the string
        arg = '' ! put blank, or else there can be garbage
        arg(1:tb-ta+1) = str(ta:tb)

    end function 
    subroutine test_delim_arg()
        implicit none
        character(len=200) :: str,chk
        character(len=20) :: a,b,c
        integer :: n 

        a = "argone"
        b = "argtwo"
        c = "argthree"
        
        str = trim(a)//";"//trim(b)//";"//trim(c)

        
        n = count_delim_arg(str,";")

        if (n/=3) stop 'test_delim_arg: failed test 1'

        chk = get_delim_arg(str,";",1)
        if (trim(chk)/=trim(a)) then
            print*,'trim(a)  :',trim(a),'len_trim(a)',len_trim(a)
            print*,'trim(chk):',trim(chk),'len_trim(chk)',len_trim(chk)
            print*,'trim(chk)==trim(a)',trim(chk)==trim(a)
            print*,'trim(chk)/=trim(a)',trim(chk)/=trim(a)
            print*,'chk(1:1)==a(1:1)',chk(1:1)==a(1:1)
            print*,'chk(2:2)==a(2:2)',chk(2:2)==a(2:2)
            print*,'chk(3:3)==a(3:3)',chk(3:3)==a(3:3)
            print*,'chk(4:4)==a(4:4)',chk(4:4)==a(4:4)
            print*,'chk(5:5)==a(5:5)',chk(5:5)==a(5:5)
            print*,'chk(6:6)==a(6:6)',chk(6:6)==a(6:6)
            print*,'chk(7:7)==a(7:7)',chk(7:7)==a(7:7)
            stop 'test_delim_arg: failed test 2a'
        end if
        chk = get_delim_arg(str,";",2)
        if (trim(chk)/=trim(b)) stop 'test_delim_arg: failed test 2b'
        chk = get_delim_arg(str,";",3)
        if (trim(chk)/=trim(c)) stop 'test_delim_arg: failed test 2c'

        print*,'test_delim_arg: passed'

    end subroutine




    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! GET COMMAND LINE ARGUMENTS
    function getcmdarg_d(i) result(arg)
        implicit none
        integer, intent(in) :: i
        real(rt) :: arg
        ! local
        real(rt) :: xd
        integer :: xi
        character(1024) :: xs

        call getcmdarg_core(i,xd,xi,xs,'d')
        arg = xd
        
    end function 
    function getcmdarg_i(i) result(arg)
        implicit none
        integer, intent(in) :: i
        integer :: arg
        ! local
        real(rt) :: xd
        integer :: xi
        character(1024) :: xs

        call getcmdarg_core(i,xd,xi,xs,'i')
        arg = xi
        
    end function 
    function getcmdarg_s(i) result(arg)
        implicit none
        integer, intent(in) :: i
        character(len=1024) :: arg
        ! local
        real(rt) :: xd
        integer :: xi
        character(1024) :: xs

        call getcmdarg_core(i,xd,xi,xs,'s')
        arg = adjustl(trim(xs))
        
    end function 

    subroutine getcmdarg_core(i,xd,xi,xs,type_expected)
        implicit none
        integer, intent(in) :: i ! which command argument is desired
        real(rt), intent(inout) :: xd ! output (if real)
        integer, intent(inout) :: xi ! output (if integer)
        character(*), intent(inout) :: xs ! output (if string)
        character(1), intent(in) :: type_expected ! d for double, i for integer, s for string
        ! local
        character(len=1024) :: str

        if (i>command_argument_count()) then
            xd = -1d0
            xi = -1
            xs = "-1"
            write(0,'(A,i5,A,i5,A)') 'WARNING: getcmdarg: arg ',i,' couldn''t be obtained, only ',&
                                               command_argument_count(),' arguments present'
            return
        end if

        ! Get the argument as a string
        call get_command_argument(i,str)

        ! Convert it to the specified type
        select case (type_expected)
            case ('d')
                xd = -1d0
                ! read(str,'(d30.16)',err=999) xd
                read(str,*,err=999) xd
            case ('i')
                xi = -1
                read(str,'(i30)',err=999) xi
            case ('s')
                xs = trim(adjustl(str))
        end select

        return

        ! If caught an error, warn about it
999     write(0,*) 'ERROR: mod_util: getcmdarg_core: error while reading command argument ',i,' argument is ',trim(adjustl(str))

    end subroutine 
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    function conv_s_to_i(s) result(i)
        implicit none
        character(*), intent(in) :: s
        integer :: i
        ! local
        integer :: itmp
        i = -1
        read(s,*,err=999) itmp
        i = itmp
        return

999     write(0,*) 'ERROR: could not convert '//s//' to an integer'
    end function 

    function conv_i_to_s(i) result(s)
        implicit none
        integer, intent(in) :: i
        character(len=11) :: s ! largest integer is 2147483647 (10 digits)
        ! local
        integer :: j,k
        character(len=1) :: t

        s = ' '

        j = abs(i)

        if (j==0) then
            s = '0'
            return
        end if

        k = 0
        do while (j>0)
            k = k+1
            write(s(k:k),'(i1)') mod(j,10)
            j = j/10
        end do

        ! Now reverse it
        do j = 1,k/2
            t = s(j:j)
            s(j:j) = s(k-j+1:k-j+1)
            s(k-j+1:k-j+1) = t
        end do

        ! If negative, shift right and add -
        if (i<0) then
            do j = k,1,-1
                s(j+1:j+1) = s(j:j)
            end do
            s(1:1) = '-'
        end if

    end function


end module

! program main
!     use mod_cmdarg
!     implicit none
!     character(1024) :: str
! 
!     call get_command(str)
!     print*,repeat('*',100)
!     print*,'Full command line: ',trim(adjustl(str))
!     print*,repeat('*',100)
!     print*,'command argument #0 (name of program):    ', trim(getcmdarg_s(0))
!     print*,'command argument #1 (expecting double):   ', getcmdarg_d(1)
!     print*,'command argument #2 (expecting integer):  ', getcmdarg_i(2)
!     print*,'command argument #3 (expecting string):   ', trim(getcmdarg_s(3))
!     print*,'command argument #4 (only 3 args, error): ', trim(getcmdarg_s(4))
!     print*,repeat('*',100)
! 
! end program
! Grey Gordon

! ! For the stubs library (which will compile without mpi), include the appropriate file.
! include 'include_mympi.f90'
! ! include 'include_mympi_stubs.f90'

! A module that attempts to simplify the process of using mpi w/o a significant
! computational cost (which of course would defeat the point!)
!
! Most routines, if available, are listed as myXXX where XXX is an mpi command.
! E.g., mympi_allgatherv calls MPI_Allgatherv.
! There are some exceptions, such as mympi_stop which calls MPI_Abort after
! displaying an error message.
module mod_mympi
    use mod_types
    implicit none

    include 'mpif.h'

    logical, private :: mympi_initialized = .false.
    integer, private :: nproc,iproc


    interface mympi_allgatherv
        module procedure mympi_allgatherv_vec, mympi_allgatherv_mat,  mympi_allgatherv_3d,&
                         mympi_allgatherv_4d, mympi_allgatherv_5d, mympi_allgatherv_6d,&
                         mympi_allgatherv_ivec,mympi_allgatherv_imat, mympi_allgatherv_i3d, mympi_allgatherv_i4d, &
                         mympi_allgatherv_l1d, mympi_allgatherv_l2d,mympi_allgatherv_l4d, mympi_allgatherv_l5d
    end interface
    interface mympi_allgatherv_r4
        module procedure mympi_allgatherv_r4vec,mympi_allgatherv_r4mat,mympi_allgatherv_r43d,mympi_allgatherv_r44d
    end interface

    interface mympi_allreduce
        module procedure mympi_allreduce_1d,mympi_allreduce_2d,mympi_allreduce_3d, &
                         mympi_allreduce_i0d,mympi_allreduce_i1d,mympi_allreduce_i2d
    end interface

    interface mympi_bcast
        module procedure mympi_bcast_vec, mympi_bcast_mat, mympi_bcast_3d, mympi_bcast_4d, mympi_bcast_5d, mympi_bcast_6d, &
                         mympi_bcast_ivec, mympi_bcast_imat, mympi_bcast_i3d, mympi_bcast_i4d, mympi_bcast_i5d, mympi_bcast_i6d, &
                         mympi_bcast_logical_vec, mympi_bcast_logical_mat, mympi_bcast_logical_3d, mympi_bcast_logical_4d, &
                         mympi_bcast_logical_5d, mympi_bcast_logical_6d
    end interface

    interface mympi_ibcast
        module procedure mympi_ibcast_vec, &
                         mympi_ibcast_ivec
    end interface

    interface mympi_irecv
        module procedure mympi_irecv_1d, mympi_irecv_0d, mympi_irecv_i1d
    end interface

    interface mympi_recv
        module procedure mympi_recv_2d, mympi_recv_1d, mympi_recv_0d, & 
                         mympi_recv_i2d, mympi_recv_i1d, mympi_recv_i0d
    end interface

    interface mympi_send
        module procedure mympi_send_0d, mympi_send_1d, mympi_send_2d, &
                         mympi_send_i0d, mympi_send_i1d, mympi_send_i2d
    end interface

    private :: mympi_bcast_core
    private :: mympi_allgatherv_core
    private :: mympi_allgatherv_core_logical
    private :: mympi_allgatherv_icore
    private :: checkinit
    private :: cumsum
    
contains

    ! Performs a reduction operation and shares the result with all nodes
    ! [ MPI_MAX] maximum
    ! [ MPI_MIN] minimum
    ! [ MPI_SUM] sum
    ! [ MPI_PROD] product
    ! [ MPI_LAND] logical and
    ! [ MPI_BAND] bit-wise and
    ! [ MPI_LOR] logical or
    ! [ MPI_BOR] bit-wise or
    ! [ MPI_LXOR] logical xor
    ! [ MPI_BXOR] bit-wise xor
    ! [ MPI_MAXLOC] max value and location
    ! [ MPI_MINLOC] min value and location 
    subroutine mympi_allreduce_3d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:,:,:)
        real(rt), intent(out) :: B(:,:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_2d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:,:)
        real(rt), intent(out) :: B(:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_1d(B,A,operation)
        implicit none
        real(rt), intent(in) :: A(:)
        real(rt), intent(out) :: B(:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_d(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_core_d(B,A,nelts,operation)
        implicit none
        real(rt), intent(in) :: A(*)
        real(rt), intent(out) :: B(*)
        integer, intent(in) :: nelts,operation
        ! local
        integer :: ierr

        ! Call the mpi routine
        call MPI_Allreduce(A(1),& ! first element assigned to iproc
                           B(1),& ! where to put the reduction  
                           nelts,& ! total numbber of elements assigned to iproc --- may be zero
                           mympi_rt,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine
    
    subroutine mympi_allreduce_i2d(B,A,operation)
        implicit none
        integer, intent(in) :: A(:,:)
        integer, intent(out) :: B(:,:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_i(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_i1d(B,A,operation)
        implicit none
        integer, intent(in) :: A(:)
        integer, intent(out) :: B(:)
        integer, intent(in) :: operation

        if (size(A)/=size(B)) STOP 'ERROR: mympi_allreduce: input and output do not have the same size'

        call mympi_allreduce_core_i(B,A,size(A),operation)

    end subroutine 
    subroutine mympi_allreduce_i0d(B,A,operation)
        implicit none
        integer, intent(in) :: A
        integer, intent(out) :: B
        integer, intent(in) :: operation
        integer :: ierr

        call MPI_Allreduce(A,& ! first element assigned to iproc
                           B,& ! where to put the reduction  
                           1,& ! total number of elements assigned to iproc --- may be zero
                           MPI_INTEGER,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine 
    subroutine mympi_allreduce_core_i(B,A,nelts,operation)
        implicit none
        integer, intent(in) :: A(*)
        integer, intent(out) :: B(*)
        integer, intent(in) :: nelts,operation
        ! local
        integer :: ierr

        ! Call the mpi routine
        call MPI_Allreduce(A(1),& ! first element assigned to iproc
                           B(1),& ! where to put the reduction  
                           nelts,& ! total numbber of elements assigned to iproc --- may be zero
                           MPI_INTEGER,& ! data type of send elements
                           operation,& ! type of operation
                           MPI_COMM_WORLD,&
                           ierr) 

    end subroutine

function mympi_getprocessorname() result(nomen)
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: nomen
    ! local
    integer :: ierr,lenout
    character(len=MPI_MAX_PROCESSOR_NAME) :: tmp

    
    call MPI_GET_PROCESSOR_NAME(tmp,lenout,ierr)

    nomen = ''
    nomen(1:lenout) = tmp(1:lenout) 

end function 

function get_ierr_msg(ierr) result(msg)
    implicit none
    integer, intent(in) :: ierr
    character(len=50) :: msg

    select case (ierr)
        case (MPI_SUCCESS)
            msg='Successful return code'
        case (MPI_ERR_BUFFER)
            msg='Invalid buffer pointer'
        case (MPI_ERR_COUNT)
            msg='Invalid count argument'
        case (MPI_ERR_TYPE)
            msg='Invalid datatype argument'
        case (MPI_ERR_TAG)
            msg='Invalid tag argument'
        case (MPI_ERR_COMM)
            msg='Invalid communicator'
        case (MPI_ERR_RANK)
            msg='Invalid rank'
        case (MPI_ERR_ROOT)
            msg='Invalid root'
        case (MPI_ERR_GROUP)
            msg='Null group passed to function'
        case (MPI_ERR_OP)
            msg='Invalid operation'
        case (MPI_ERR_TOPOLOGY)
            msg='Invalid topology'
        case (MPI_ERR_DIMS)
            msg='Illegal dimension argument'
        case (MPI_ERR_ARG)
            msg='Invalid argument'
        case (MPI_ERR_UNKNOWN)
            msg='Unknown error'
        case (MPI_ERR_TRUNCATE)
            msg='message truncated on receive'
        case (MPI_ERR_OTHER)
            msg='Other error; use Error_string'
        case (MPI_ERR_INTERN)
            msg='internal error code'
        case (MPI_ERR_IN_STATUS)
            msg='Look in status for error value'
        case (MPI_ERR_PENDING)
            msg='Pending request'
        case (MPI_ERR_REQUEST)
            msg='illegal mpi_request handle'
        case (MPI_ERR_LASTCODE)
            msg='Last error code -- always at end'
    end select

end function

subroutine checkinit()
    implicit none
    if (.not. mympi_initialized) STOP 'ERROR: mpi has not been initialized'
end subroutine 

integer function mympi_iproc()
    implicit none
    call checkinit()
    mympi_iproc = iproc
end function 
integer function mympi_nproc()
    implicit none
    call checkinit()
    mympi_nproc = nproc
end function 

subroutine mympi_barrier()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Barrier(MPI_COMM_WORLD,ierr)

end subroutine 

! Aborts all mpi routines and also displays a message
subroutine mympi_abort()
    implicit none
    call mympi_stop('')
end subroutine 
subroutine mympi_stop(msg)
    implicit none
    character(len=*), intent(in) :: msg
    ! local
    integer :: errorcode
    integer :: ierr

    call checkinit()

    ! Display the error message
    write(0,*) msg

    ! Pass error code to environment
    errorcode = 1 
    
    ! Terminate all processes
    call MPI_Abort(MPI_COMM_WORLD,errorcode,ierr)

end subroutine 

! Non blocking test for a message
! function mympi_iprobe(source,tag) result(tf)
!     implicit none
!     integer, intent(in) :: source
!     integer, intent(in), optional :: tag 
!     logical :: tf
!     ! local
!     integer :: ierr, status(MPI_STATUS_SIZE)
!     logical :: flag
! 
!     if (present(tag)) then
!         call MPI_Iprobe(source, tag, MPI_COMM_WORLD, flag, status, ierr)
!     else
!         call MPI_Iprobe(source, MPI_ANY_TAG, MPI_COMM_WORLD, flag, status, ierr)
!     end if
! 
!     tf = flag
! 
! end function 

! Divides n elts as evenly as possible into ranges [a,b]
! If b<a, then b is set to -1 and a to 0. This is convenient as b-a+1 
! then returns the number of elements for each node regardless of whether 
! the range is 0 or not. Also, since b<a, "do i = a,b" is not executed. 
! NOTE: if used to assign trailing dimensions of an array, 
! as it defines contiguous array section this is very useful
subroutine mympi_assign_range(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a,b
    
    call checkinit()
    call mympi_assign_range_core(a,b,iproc,n)

end subroutine 
! Helper functions
integer function mympi_a(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: b
    call checkinit()
    call mympi_assign_range(mympi_a,b,n)
end function mympi_a
integer function mympi_b(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: a
    call checkinit()
    call mympi_assign_range(a,mympi_b,n)
end function mympi_b
! Gives the combined list
subroutine mympi_assign_range_list(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a(nproc),b(nproc)
    ! local
    integer :: ip

    call checkinit()

    do ip = 0,nproc-1
        call mympi_assign_range_core(a(ip+1),b(ip+1),ip,n)
    end do
end subroutine 

! The core routine for dividing elements
! ip = 0,...,nproc-1
subroutine mympi_assign_range_core(a,b,ip,n)
    implicit none
    integer, intent(in) :: ip,n
    integer, intent(out) :: a,b
    ! local
    integer :: rem,jproc,times

    call checkinit()
    if (n<nproc) then
        ! If here, can assign one element to each node. 
        ! Hence, ip gets the ip+1 element if ip+1<=n
        if (ip<n) then
            a = ip+1
            b = ip+1
        else
            a = 0
            b = -1
        end if
    else
        ! If here, need to assign at least n/nproc to each element and the remainder
        ! across the first elements

        rem = mod(n,nproc)
        times = n/nproc ! >=1

        if (n/=times*nproc + rem) call mympi_stop('ERROR: mympi_assign_range: arithmetic not as expected')

        b = 0
        do jproc = 0,ip
            if (jproc<rem) then
                ! Gets times+1 elements
                a = 1 + b
                b = times + 1 + b
            else
                ! Gets times elements
                a = 1 + b
                b = times + b
            end if
        end do

    end if

end subroutine


subroutine mympi_bcast_logical_6d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_5d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_4d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_3d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_mat(A,root)
    implicit none
    logical, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_vec(A,root)
    implicit none
    logical, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_logical_core(A,n,root)
    implicit none
    integer, intent(in) :: n, root
    logical, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_LOGICAL,root,MPI_COMM_WORLD,ierr)

end subroutine

subroutine mympi_ibcast_ivec(request,A,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(inout) :: A(:)
    integer, intent(in) :: root

    call mympi_ibcast_icore(request,A,size(A),root)

end subroutine
subroutine mympi_ibcast_icore(request,A,n,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(inout) :: A(*)
    integer, intent(in) :: n,root
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Ibcast(A(1:n),n,MPI_INTEGER,root,MPI_COMM_WORLD,request,ierr)

    if (ierr/=MPI_SUCCESS) then
        call mympi_stop('error calling mympi_ibcast_icore')
    end if

end subroutine

subroutine mympi_ibcast_vec(request,A,root)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_ibcast_core(request,A,size(A),root)
end subroutine
subroutine mympi_ibcast_core(request,A,n,root)
    implicit none
    integer, intent(out) :: request
    integer, intent(in) :: n,root
    real(rt), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Ibcast(A(1),n,mympi_rt,root,MPI_COMM_WORLD,request,ierr)

end subroutine

subroutine mympi_bcast_6d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_5d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_4d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_3d(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_mat(A,root)
    implicit none
    real(rt), intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_vec(A,root)
    implicit none
    real(rt), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_core(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    real(rt), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,mympi_rt,root,MPI_COMM_WORLD,ierr)

end subroutine

subroutine mympi_bcast_i6d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i5d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i4d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_i3d(A,root)
    implicit none
    integer, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_imat(A,root)
    implicit none
    integer, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine
subroutine mympi_bcast_ivec(A,root)
    implicit none
    integer, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_icore(A,size(A),root)
end subroutine

subroutine mympi_bcast_icore(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    integer, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_INTEGER,root,MPI_COMM_WORLD,ierr)

end subroutine

function mympi_test(request) result(isComplete)
    implicit none
    integer, intent(in) :: request
    logical :: isComplete
    ! local
    logical :: flag
    integer :: ierr ! , status(MPI_STATUS_SIZE)

    call MPI_Test(request,flag,MPI_STATUS_IGNORE,ierr)

    isComplete = flag

end function

subroutine mympi_irecv_1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_core(request,A,size(A),source,tag)
    else
        ! call mympi_irecv_core(request,A,size(A),source,MPI_ANY_TAG)
        call mympi_irecv_core(request,A,size(A),source,0)
    end if

end subroutine
subroutine mympi_irecv_0d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    if (present(tag)) then
        call mympi_irecv_core(request,tmp,1,source,tag)
    else
        ! call mympi_irecv_core(request,tmp,1,source,MPI_ANY_TAG)
        call mympi_irecv_core(request,tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine


subroutine mympi_irecv_core(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    real(rt), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, mympi_rt, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_irecv_i1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    integer, intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_icore(request,A,size(A),source,tag)
    else
        ! call mympi_irecv_icore(request,A,size(A),source,MPI_ANY_TAG)
        call mympi_irecv_icore(request,A,size(A),source,0)
    end if

end subroutine

subroutine mympi_irecv_icore(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    integer, intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, MPI_INTEGER, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_waitall(array_of_requests)
    implicit none
    integer, intent(in) :: array_of_requests(:)
    ! local
    integer :: ierr

    call MPI_Waitall(size(array_of_requests),array_of_requests(1),MPI_STATUSES_IGNORE,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_waitall: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine

subroutine mympi_wait(rqst)
    implicit none
    integer, intent(in) :: rqst
    ! local
    integer :: ierr

    call MPI_Wait(rqst,MPI_STATUS_IGNORE,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_wait: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine



subroutine mympi_recv_2d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A(:,:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        ! print*,'about to call recv_core'
        ! call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_core(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_1d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        ! print*,'about to call recv_core'
        ! call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_core(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_0d(A,source,tag)
    implicit none
    real(rt), intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    if (present(tag)) then
        call mympi_recv_core(tmp,1,source,tag)
    else
        ! call mympi_recv_core(tmp,1,source,MPI_ANY_TAG)
        call mympi_recv_core(tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine


subroutine mympi_recv_core(A,n,source,tag)
    implicit none
    real(rt), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr !, status(MPI_STATUS_SIZE)

    ! print*,'about to call recv'
    ! print*,'source: ',source
    ! print*,'tag: ',tag,'MPI_ANY_TAG',MPI_ANY_TAG
    ! print*,'MPI_STATUS_IGNORE: ',MPI_STATUS_IGNORE
    ! print*,'n',n
    ! print*,'A(1)',A(1)
    ! print*,'A(2)',A(2)
    ! print*,'A(3)',A(3)
    ! print*,'MPI_COMM_WORLD',MPI_COMM_WORLD
    call MPI_Recv(A(1), n, mympi_rt, source, tag, &
                  MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_recv_i0d(A,source,tag)
    implicit none
    integer, intent(out) :: A
    integer, intent(in) :: source
    integer, intent(in), optional :: tag
    ! local
    integer :: tmp(1)

    if (present(tag)) then
        call mympi_recv_icore(tmp,1,source,tag)
    else
        ! call mympi_recv_icore(tmp,1,source,MPI_ANY_TAG)
        call mympi_recv_icore(tmp,1,source,0)
    end if

    A = tmp(1)

end subroutine
subroutine mympi_recv_i2d(A,source,tag)
    implicit none
    integer, intent(out) :: A(:,:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_icore(A,size(A),source,tag)
    else
        ! call mympi_recv_icore(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_icore(A,size(A),source,0)
    end if

end subroutine
subroutine mympi_recv_i1d(A,source,tag)
    implicit none
    integer, intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_icore(A,size(A),source,tag)
    else
        ! call mympi_recv_icore(A,size(A),source,MPI_ANY_TAG)
        call mympi_recv_icore(A,size(A),source,0)
    end if

end subroutine

subroutine mympi_recv_icore(A,n,source,tag)
    implicit none
    integer, intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr !, status(MPI_STATUS_SIZE)

    call MPI_Recv(A(1), n, MPI_INTEGER, source, tag, &
                  MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 


subroutine mympi_send_2d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A(:,:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        ! print*,'about to call mympi_send_core'
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_1d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        ! print*,'about to call mympi_send_core'
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_0d(A,dest,tag)
    implicit none
    real(rt), intent(in) :: A
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag
    ! local
    real(rt) :: tmp(1)

    tmp(1) = A

    if (present(tag)) then
        call mympi_send_core(tmp,1,dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_core(tmp,1,dest,0)
    end if

end subroutine

subroutine mympi_send_core(A,n,dest,tag)
    implicit none
    real(rt), intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    ! print*,'about to send'
    call MPI_send(A(1), n, mympi_rt, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_send_i2d(A,dest,tag)
    implicit none
    integer, intent(in) :: A(:,:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_icore(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_i1d(A,dest,tag)
    implicit none
    integer, intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_icore(A,size(A),dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(A,size(A),dest,0)
    end if

end subroutine
subroutine mympi_send_i0d(A,dest,tag)
    implicit none
    integer, intent(in) :: A
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag
    ! local
    integer :: tmp(1)

    tmp(1) = A

    if (present(tag)) then
        call mympi_send_icore(tmp,1,dest,tag)
    else
        ! NOTE: MPI_ANY_TAG may only be used in "RECV"s
        call mympi_send_icore(tmp,1,dest,0)
    end if

end subroutine


subroutine mympi_send_icore(A,n,dest,tag)
    implicit none
    integer, intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    call MPI_send(A(1), n, MPI_INTEGER, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 



subroutine mympi_allgatherv_l5d(A)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_l4d(A)
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_l2d(A)
    implicit none
    logical, intent(inout) :: A(:,:)
    call mympi_allgatherv_core_logical(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_l1d(A)
    implicit none
    logical, intent(inout) :: A(:)
    call mympi_allgatherv_core_logical(A,1,size(A,1))
end subroutine 

! This gathers data from all tasks and delivers the combined data to all tasks.
! NOTE: it assumes the work was distributed across the trailing dimension according
!       to mympi_assign_range_list.
subroutine mympi_allgatherv_6d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4)*size(A,5),size(A,6))
end subroutine 
subroutine mympi_allgatherv_5d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_4d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_3d(A)
    implicit none
    real(rt), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_mat(A)
    implicit none
    real(rt), intent(inout) :: A(:,:)
    call mympi_allgatherv_core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_vec(A)
    implicit none
    real(rt), intent(inout) :: A(:)
    call mympi_allgatherv_core(A,1,size(A,1))
end subroutine 
subroutine mympi_allgatherv_core(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    real(rt), intent(inout) :: A(nrow,*)
    ! local
    real(rt), allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()

    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        mympi_rt,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        mympi_rt,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)

end subroutine

! Same as mympi_allgatherv_core but with logical
subroutine mympi_allgatherv_core_logical(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    logical, intent(inout) :: A(nrow,*)
    ! local
    logical, allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_LOGICAL,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_LOGICAL,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)


end subroutine

! Single precision version of the above
subroutine mympi_allgatherv_r44d(A)
    implicit none
    real(4), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_r43d(A)
    implicit none
    real(4), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_r4mat(A)
    implicit none
    real(4), intent(inout) :: A(:,:)
    call mympi_allgatherv_r4core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_r4vec(A)
    implicit none
    real(4), intent(inout) :: A(:)
    call mympi_allgatherv_r4core(A,1,size(A))
end subroutine 
subroutine mympi_allgatherv_r4core(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    real(4), intent(inout) :: A(nrow,*)
    ! local
    real(4), allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_REAL4,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_REAL4,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)


end subroutine


! Integer versions of the above
subroutine mympi_allgatherv_i4d(A)
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_i3d(A)
    implicit none
    integer, intent(inout) :: A(:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_imat(A)
    implicit none
    integer, intent(inout) :: A(:,:)
    call mympi_allgatherv_icore(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_ivec(A)
    implicit none
    integer, intent(inout) :: A(:)
    call mympi_allgatherv_icore(A,1,size(A))
end subroutine 
subroutine mympi_allgatherv_icore(A,nrow,ncol)
    implicit none
    integer, intent(in) :: nrow,ncol
    integer, intent(inout) :: A(nrow,*)
    ! local
    integer, allocatable :: B(:,:)
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc,icol

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    allocate(B(nrow,ncol))

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_INTEGER,& ! data type of send elements
                        B(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_INTEGER,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

    ! Now all has been received in B, so copy out to A (this is a restriction of MPI)
    do icol = 1,ncol
        A(1:nrow,icol) = B(:,icol)
    end do
    deallocate(B)



end subroutine

subroutine mympi_init()
    implicit none
    integer :: ierr,flag

    ! Test whether intialized already, and, if not, initialize
    call MPI_initialized(flag,ierr)
    if (flag==0) then 
        call MPI_init(ierr)

        ! Get the number of processors
        call MPI_Comm_size(MPI_COMM_WORLD,nproc,ierr)

        ! Get the processor id
        call MPI_Comm_rank(MPI_COMM_WORLD,iproc,ierr)

        ! Initialize mpi
        mympi_initialized = .true.

    end if

end subroutine
subroutine mympi_finalize()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Finalize(ierr)

end subroutine 

subroutine mympi_hello_world()
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: procname
    integer :: lenprocname,ierr

    call checkinit()

    ! Get the processor name
    call MPI_Get_processor_name(procname,lenprocname,ierr)

    ! Display
    write(*,'(3A,i0,A,i0,A)') 'Hello from ',trim(procname),'. I am process ',iproc,' of ',nproc,'.'

end subroutine 

function cumsum(x_in) result(x_out)
    implicit none
    integer, dimension(1:), intent(in) :: x_in
    integer, dimension(1:size(x_in)) :: x_out
    integer :: i,accum

    accum = 0
    do i = 1,size(x_in)
        accum = accum + x_in(i)
        x_out(i) = accum
    end do

end function 


end module mod_mympi

module mod_mympi_test
    use mod_mympi
    implicit none
contains
    
    ! mympi_init and mympi_finalize should be called before and after this routine
    subroutine test_mympi
        implicit none
        call test_mympi_core(1,7)
        call test_mympi_core(2,7)
        call test_mympi_core(3,7)
        call test_mympi_core(4,7)
        call test_mympi_core(5,7)
        call test_mympi_core(11,7)
        call test_mympi_core(24,7)
        call test_mympi_core(31,7)
        call test_mympi_core(1017,7)
        if (mympi_iproc()==0) print*,'test_mympi: passed'
    end subroutine
    subroutine test_mympi_core(ncol,ndim1)
        implicit none
        integer, intent(in) :: ncol, ndim1 
        real(rt), allocatable :: x_d_1d(:),x_d_2d(:,:),x_d_3d(:,:,:)
        real(4), allocatable :: x_f_1d(:),x_f_2d(:,:)
        integer, allocatable :: x_i_1d(:),x_i_2d(:,:)
        logical, allocatable :: x_l_1d(:),x_l_2d(:,:)
        real(rt), allocatable :: y_d_1d(:),y_d_2d(:,:),y_d_3d(:,:,:)
        real(4), allocatable :: y_f_1d(:),y_f_2d(:,:)
        integer, allocatable :: y_i_1d(:),y_i_2d(:,:)
        logical, allocatable :: y_l_1d(:),y_l_2d(:,:)
        integer, allocatable :: lo(:),hi(:),nelts(:),rqstlst(:)
        integer :: a,b,itmp,nproc,icol,cnt,itmp2
        real(rt) :: tmp

        nproc = mympi_nproc()
        allocate(lo(nproc),hi(nproc),nelts(nproc))
        allocate(rqstlst(ncol))

        ! Make space
        allocate(x_d_1d(ncol),x_d_2d(ndim1,ncol),x_d_3d(ndim1,ncol,3)) 
        allocate(x_f_1d(ncol),x_f_2d(ndim1,ncol)) 
        allocate(x_i_1d(ncol),x_i_2d(ndim1,ncol)) 
        allocate(x_l_1d(ncol),x_l_2d(ndim1,ncol)) 
        allocate(y_d_1d(ncol),y_d_2d(ndim1,ncol),y_d_3d(ndim1,ncol,3)) 
        allocate(y_f_1d(ncol),y_f_2d(ndim1,ncol)) 
        allocate(y_i_1d(ncol),y_i_2d(ndim1,ncol)) 
        allocate(y_l_1d(ncol),y_l_2d(ndim1,ncol)) 
        
        ! Hello world
!         call mympi_hello_world()

        ! Check list assignments
        call mympi_assign_range_list(lo,hi,ncol)

        ! Check list assignments
        nelts = hi - lo + 1
        if (any(nelts<0)) call mympi_stop('ERROR: List assignment has negative number of elements')
        if (maxval(nelts)>minval(nelts)+1) call mympi_stop('ERROR: List assignment was uneven')

        call mympi_assign_range(a,b,ncol)
        if (mympi_a(ncol)/=a) call mympi_stop('ERROR: in mympi_a')
        if (mympi_b(ncol)/=b) call mympi_stop('ERROR: in mympi_b')

        !!! Check that ever element was assigned to exactly one rank
        do icol = 1,ncol
            cnt = 0
            do itmp = 1,nproc
                if (icol>=lo(itmp) .and. icol<=hi(itmp)) cnt = cnt + 1
            end do
            if (cnt/=1) call mympi_stop('ERROR: failed to assign complete range or assigned multiply')
        end do

        ! Test barrier
        call mympi_barrier() ! prettier i/o

        ! Test mympi_allgatherv
        x_d_1d = -1e0_rt; x_d_2d = -1e0_rt
        x_f_1d = -1e0; x_f_2d = -1e0
        x_i_1d = -1  ; x_i_2d = -1
        x_l_1d = -1>0; x_l_2d = -1>0

        !!! Have each rank write its rank to x
        do icol = mympi_a(ncol),mympi_b(ncol)
            x_d_1d(  icol) = real(mympi_iproc(),rt)
            x_d_2d(:,icol) = real(mympi_iproc(),rt)
            x_f_1d(  icol) = real(mympi_iproc(),4)
            x_f_2d(:,icol) = real(mympi_iproc(),4)
            x_i_1d(  icol) = mympi_iproc()  
            x_i_2d(:,icol) = mympi_iproc()
            x_l_1d(  icol) = .true. 
            x_l_2d(:,icol) = .true.
        end do

        !!! Gather
        call mympi_allgatherv(x_d_1d)
        call mympi_allgatherv(x_d_2d)
        call mympi_allgatherv_r4(x_f_1d)
        call mympi_allgatherv_r4(x_f_2d)
        call mympi_allgatherv(x_i_1d)
        call mympi_allgatherv(x_i_2d)
        call mympi_allgatherv(x_l_1d)
        call mympi_allgatherv(x_l_2d)

        !!! Test that it is as it should be
        do icol = 1,ncol
            ! Search for who this column was assigned to
            do itmp = 1,nproc
                if (icol>=lo(itmp) .and. icol<=hi(itmp)) exit
            end do
            if (itmp<=nproc) then ! found an assignment
                itmp = itmp - 1
!                 print*,'icol',icol,'x_d_1d(icol)',x_d_1d(icol),'itmp',itmp
                if (   (x_d_1d(  icol) /= real(itmp,rt))) call mympi_stop('ERROR: failed test mympi_allgatherv d_1d')
                if (any(x_d_2d(:,icol) /= real(itmp,rt))) call mympi_stop('ERROR: failed test mympi_allgatherv d_2d')
                if (   (x_f_1d(  icol) /= real(itmp,4))) call mympi_stop('ERROR: failed test mympi_allgatherv f_1d')
                if (any(x_f_2d(:,icol) /= real(itmp,4))) call mympi_stop('ERROR: failed test mympi_allgatherv f_2d')
                if (   (x_i_1d(  icol) /= itmp))         call mympi_stop('ERROR: failed test mympi_allgatherv i_1d')
                if (any(x_i_2d(:,icol) /= itmp))         call mympi_stop('ERROR: failed test mympi_allgatherv i_2d')
                if (   (x_l_1d(  icol) .neqv. .true. ))      call mympi_stop('ERROR: failed test mympi_allgatherv l_1d')
                if (any(x_l_2d(:,icol) .neqv. .true.))       call mympi_stop('ERROR: failed test mympi_allgatherv l_2d')
            end if
        end do

        !!! Wait and display pass
        call mympi_barrier()

        ! Test reduction

        !!! Have each rank write its rank to x
        x_d_1d = real(mympi_iproc(),rt)
        x_d_2d = real(mympi_iproc(),rt)
        x_d_3d = real(mympi_iproc(),rt)

        !!! Reduce
        call random_number(y_d_1d)
        call random_number(y_d_2d)
        call random_number(y_d_3d)

        call mympi_allreduce(y_d_1d,x_d_1d,MPI_SUM)
        call mympi_allreduce(y_d_2d,x_d_2d,MPI_SUM)
        call mympi_allreduce(y_d_3d,x_d_3d,MPI_SUM)

        
        cnt = 0
        do itmp2 = 0,mympi_nproc()-1
            cnt = cnt + itmp2 + 2

        end do

        itmp = -1000000000
        call mympi_allreduce(itmp,2+mympi_iproc(),MPI_SUM)
        if (itmp/=cnt) call mympi_stop('ERROR: failed test mympi_allreduce i_0d')

        call mympi_barrier()
        stop 'checking'

        !!! Test that it is as it should be
        tmp = 0e0_rt
        do itmp = 0,nproc-1
            tmp = tmp + real(itmp,rt)
        end do
        x_d_1d = tmp
        x_d_2d = tmp
        x_d_3d = tmp

        if (any(y_d_1d/=x_d_1d)) call mympi_stop('ERROR: failed test mympi_allreduce d_1d')
        if (any(y_d_2d/=x_d_2d)) call mympi_stop('ERROR: failed test mympi_allreduce d_2d')
        if (any(y_d_3d/=x_d_3d)) call mympi_stop('ERROR: failed test mympi_allreduce d_3d')

        ! Check mympi_ibcast
        x_i_2d = 0
        if (mympi_iproc()==0) then
            do icol = 1,ncol
                x_i_2d(:,icol) = icol
            end do
        end if
        do icol = 1,ncol
            call mympi_ibcast(rqstlst(icol),x_i_2d(:,icol),0)
        end do
        call mympi_waitall(rqstlst) 
        do icol = 1,ncol
            if (any(x_i_2d(:,icol)/=icol)) call mympi_stop('ERROR: failed test mympi_ibcast')
        end do


    end subroutine


end module mod_mympi_test

! Grey Gordon, 2016
! 
! A master/slave helper routine using MPI. The slaves should call "slave(fcn,nx)" and then use the type-bound procedures in type_gang and type_slave 
! to request the slaves compute fcn(y,x) and access the value when done. 
! 
! See "test_mpislave" for a usage example; sorry for the lack of documentation.
!
module mod_mpislave
    use mod_types
    use mod_mympi
    implicit none

!     integer, parameter :: dbglevel = 1

    type type_slave 
        real(rt), pointer :: x(:) => null()
        real(rt), pointer :: y(:) => null()
        integer :: ny ! size of y (inferred automatically)
        integer :: rank
        integer :: requestid
        logical :: busy = .false.
        logical :: ever_assigned = .false.
    contains
        procedure, pass :: assign_x
        procedure, pass :: check_busy
        procedure, pass :: shutdown => shutdown_slave
    end type 

    type type_gang
        type(type_slave), dimension(:), allocatable :: slaves
        integer :: sleeptime = 0
    contains
        procedure, pass :: init => init_gang
        procedure, pass :: shutdown => shutdown_gang
        procedure, pass :: wait_for_free
    end type

contains

    subroutine test_mpislave()
        use mod_util
        implicit none
        type(type_gang) :: gang
        integer :: i,who
        integer, parameter :: nx=2,sleeptime=0
        logical, parameter :: dbg = .false.
        real(rt) :: x(nx),piest
        type(type_seed) :: randstate

        if (mympi_nproc()==1) then
            call mympi_stop('ERROR: test_mpislave: at least 2 ranks are required')
        end if

        call randstate%randomize()

        if (mympi_iproc()/=0) then 
            call slave(fcn,nx)
            return
        end if

        if (dbg) write(*,'(A)',advance='no') 'initializing gang ...'
        call gang%init(nx,sleeptime)
        if (dbg) write(*,'(A)') ' done.'

        piest = 3._rt

        do i = 2,100000

            if (dbg) write(*,'(A)',advance='no') 'waiting for free ...'
            who = gang%wait_for_free()
            if (dbg) write(*,'(A)') ' done.'
            
            if (gang%slaves(who)%ever_assigned) then
                if (dbg) write(*,'(A)',advance='no') 'getting y value ...'
                piest = gang%slaves(who)%y(1)/i + piest*real(i-1,rt)/i
                if (dbg) write(*,'(A)') ' done.'
            end if

            if (dbg) write(*,'(A)',advance='no') 'assigning x...'
            call random_number(x)
            call gang%slaves(who)%assign_x(x)
            if (dbg) write(*,'(A)') ' done.'

        end do

        if (dbg) write(*,'(A)',advance='no') 'shutting down ...'
        call gang%shutdown()
        if (dbg) write(*,'(A)') ' done.'

        if (abs(piest-3.14d0)<.1d0) then
            print*,'test_mpislave: passed'
        else
            stop 'test_mpislave: failed, pi estimate is not close to pi'
        end if

    contains

        subroutine fcn(y,x)
            real(rt), intent(out), allocatable :: y(:)
            real(rt), intent(in) :: x(:)

            ! Determine whether x1,x2 lies insides a circle inscribed in a square, or outside of it. 

            ! Area ratio is pi over 4, so to get estimate of pi, multiply by 4
            allocate(y(1))
            y(1) = merge(4._rt,0._rt,norm2(x)<1._rt) 

        end subroutine

    end subroutine

    ! Do something according to the value of flag(this_image())[1]
    !  1 := evaluate slavex and store in slavey
    ! -1 := exit the slave routine
    subroutine slave(fcn,nx)
        use mod_mympi
        implicit none
        integer, intent(in) :: nx
        interface 
            subroutine fcn(y,x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(out), allocatable :: y ! dimension inferred after fcn is called
                real(rt), dimension(:), intent(in) :: x
            end subroutine
        end interface
        ! local
        integer :: flag
        real(rt), allocatable :: slavex(:),slavey(:)

        allocate(slavex(nx))

        do 

            ! Wait for a flag 
            call mympi_recv(flag,0)

            ! If flag==-1, then shutdown
            if (flag==-1) return

            ! If still here, wait for an x
            call mympi_recv(slavex,0)

            ! If flag==1, execute f(x) and store on master.
            if (flag==1) then

                ! Compute the result
                call fcn(slavey,slavex) 

                ! Note that here, slavey should be allocated. If not, abort! 
                if (.not. allocated(slavey)) call mympi_stop('ERROR: mod_mpislave: slave: fcn must allocate slavey, but was not allocated.')

                ! Send the dimension to master tagged by mympi_iproc()
                call mympi_send(size(slavey),0,tag=mympi_iproc()) 

                ! Send the result to master tagged by mympi_nproc()+mympi_iproc()
                ! call mympi_send(slavey,0,tag=mympi_nproc()+mympi_iproc()) 
                call mympi_send(slavey,0) !,tag=mympi_nproc()+mympi_iproc()) 

            end if

        end do

    end subroutine

    ! a routine to shut down all slaves
    subroutine shutdown_slave(this,sleeptime) 
        use mod_util
        implicit none
        class(type_slave), intent(inout) :: this
        integer, intent(in) :: sleeptime
        ! local
        integer :: j
        character(1), volatile :: abc ! volatile prevents optimization (this is used for a wait loop)

        ! Wait for rank to finish if not finished task
        do while (this%busy) 

            ! Check if still busy (if not, then this%busy will be set to .false.
            call this%check_busy()

            ! If still busy, then sleep or do something else for a little while to not thrash the system
            if (this%busy) then

                ! If sleeptime, then sleep
                if (sleeptime>=1) then

                    call sleep(sleeptime)

                else

                    do j = 1,1000000
                        abc = 'a'
                    end do

                end if


            end if

        end do
            
        ! Once not busy, then send the shut down signal of flag==-1
        call mympi_send(-1,this%rank) 

    end subroutine

    subroutine check_busy(this)
        implicit none
        class(type_slave), intent(inout) :: this
        ! local
        logical :: tf

        if (this%busy) then

            ! Check whether i has finished
            tf = mympi_test(this%requestid)

            if (tf) then


                ! Finished sending ny, so need to allocate y and receive it
                if (associated(this%y)) then
                    if (this%ny/=size(this%y)) then
                        deallocate(this%y)
                        allocate(this%y(this%ny))
                        ! Now y has the correct shape
                    end if
                else
                    allocate(this%y(this%ny))
                end if

                ! Here, this%y is allocated and is size this%ny. Receive it.
                call mympi_recv(this%y,this%rank)

                this%busy = .false.

            end if

        end if

    end subroutine

    ! This routine sends the value of x to a slave "who", who will then store the value in y. 
    subroutine assign_x(this,x) !,y)
        implicit none
        class(type_slave), intent(inout) :: this
        real(rt), intent(in), target :: x(:)
        integer :: ierr

        this%busy = .true.
        this%ever_assigned = .true.

        this%x = x

        call mympi_send(1,this%rank) ! blocking send (tell them to proceed to execute)
        call mympi_send(x,this%rank) ! assign an x value (wait for it to complete)

        ! Request that the value of ny be sent and stored in this%ny (once finished, will indicate completed task)
        call MPI_Irecv(this%ny, 1, MPI_INTEGER, this%rank, this%rank, MPI_COMM_WORLD, this%requestid, ierr)

        ! Once the slave has finished evaluating f(x), it's value will be stored in this%y 

    end subroutine 

    subroutine init_gang(this,nx,sleeptime)
        implicit none
        class(type_gang), intent(inout) :: this
        integer, intent(in) :: nx ! size of x
        integer, intent(in), optional :: sleeptime
        ! local
        integer :: i,nranks

        if (allocated(this%slaves)) deallocate(this%slaves)

        if (present(sleeptime)) this%sleeptime = sleeptime

        nranks = mympi_nproc()

        allocate(this%slaves(nranks-1))
        do i = 1,nranks-1
            if (associated(this%slaves(i)%x)) deallocate(this%slaves(i)%x)
            allocate(this%slaves(i)%x(nx))
            this%slaves(i)%rank = i
        end do

    end subroutine

    ! a routine to shut down all slaves
    subroutine shutdown_gang(this) 
        implicit none
        class(type_gang), intent(inout) :: this
        ! local
        integer :: i

        do i = 1,size(this%slaves)
            call this%slaves(i)%shutdown(this%sleeptime)
        end do
        
    end subroutine

    ! a routine returning the index of a free slave
    function wait_for_free(this) result(who)
        use mod_util
        implicit none
        class(type_gang), intent(inout) :: this
        integer :: who
        ! local
        real(8) :: tic, toc, tmp
        integer :: i

        do 

            do i = 1,size(this%slaves)

                ! Test whether a particular slave is busy
                call this%slaves(i)%check_busy()
                if (.not. this%slaves(i)%busy) then
                    who = this%slaves(i)%rank
                    return
                end if

            end do


            ! If sleeptime, then sleep
            if (this%sleeptime>=1) then

                call sleep(this%sleeptime)

            else

                ! Do some busy work so as to not continually check the status (which may cause MPI problems)
                tic = MPI_WTIME()
                toc = tic
                tmp = 0d0
                do while (toc-tic<1d-6)
                    toc = MPI_WTIME()
                    tmp = tmp + i 
                end do
                
                ! Trick the compiler into not optimizing away the busy work (hopefully)
                open(newunit=i,file='/lkandsfkasn/makecompilernotoptimize.dat',status='unknown',err=999) ! file should not exist (but could), so this will jump to line 999
                write(i,*) tmp
                close(i)
999             tmp = 0d0

            end if

        end do

    end function



end module
! Module containing sub_crs, the Fortran routine used by 
! "A Practical Approach to Testing Calibration Strategies" (Yongquan Cao and Grey Gordon)
! See sub_crs for the main documentation of the routine. 
! 
! Grey Gordon, 2016. SDG.
module mod_crs
    use mod_types
    use mod_util
    use mod_mympi
    use mod_mpislave
    implicit none

    ! Options for minNdim 
    type type_crs_opts 
        ! Commonly used ones
        integer :: printmod = 1, maxiter = 10000
        real(rt) :: xabstoler = 1e-6_rt !, xreltoler = 1e-6_rt
        real(rt) :: yabstoler = 1e-10_rt, yreltoler = 1e-10_rt
        ! Others (maybe only a few algorithms use them)
        ! For these, -1 means there is no default specified
        integer :: algorithm = 1
        integer :: poolsize = -1
        integer :: sleeptime = -1 
        integer :: displaybestx = 1 ! 1==means true (the default)
    end type 

    private
    public :: sub_crs, test_crs

contains

    ! Controlled random search --- MPI version
    ! 
    ! CRS attempts to solve min f(x) s.t. x in [a,b] (where all are in Rn)
    ! or simply evaluates f at an arbitrary number of random x draws from [a,b]
    ! 
    ! When minimizing, it first generates a random pool of points (the number 
    ! is controlled by the poolsize option in opts). Then, it tries to improve
    ! this pool by taking the worst point and trying to get rid of it. This is
    ! done by reflecting it through a simplex of points (like in Nelder-Mead).
    ! 
    ! This simplex is chosen by randomly selecting n+1 points (whose function 
    ! values are known) with replacement. The centroid of the simplex is then
    ! used for reflecting the worst point. 
    !
    ! If the reflected point improves on the worst point (and is in a,b), it is
    ! accepted. O/w, the process repeats.
    ! 
    ! This is modified for parallel purposes by attempting to simultaneously get 
    ! rid of several of the worst points simultaneously. 
    !
    ! The optional argument sleeptime is useful when there are only a few ranks. 
    ! It specifies how long the master should sleep if no free workers are found. This
    ! keeps processing cycles free for those ranks actually doing work. If using 
    ! sleeptime>0, the number of ranks should typically be one more than the physical 
    ! number of cores.
    !
    ! If e_in is present, then random draws will not be taken uniformly from [a,b], but rather
    ! from [xstar-e,xstar+e] (component-wise). Combined with a small poolsize (which can be set 
    ! to 0 which will then be overridden), this makes the routine effectively like a parallel 
    ! version of Nelder-Mead.
    !
    ! algorithm
    ! 0           : no optimization : just computes the pool of points and stops. This is used for just testing calibration strategies.  
    ! 1 (default) : crs1 algorithm : purely global with random simplex chosen irrespective of best point and random simplex.
    ! 2           : crs2 algorithm : when random trial point fails, local mutation is used to potentially move towards the best point. 
    ! 
    ! Note: was formerly called sub_parCRS_MPI_v2
    !
    ! INPUT/OUTPUT
    !   xstar : initial guess on input (may not be used depending on algorithm) and final guess on exit
    ! 
    ! INPUT
    !   f       : subroutine f(y,x) with x input and y a vector of outputs. The minimization is done wrt y(1) which will typically be 
    !             the SSRS. However, additional output may be returned which will be saved in a csv file if desired (see "fname" input).
    !   a,b     : lower and upper bound on x space
    !   e_in    : optional, see description above
    !   fname   : optional, gives the path and name where a csv file should be saved (if desired)
    !   xlabels : optional, comma separated list of labels for the x vector (defaults to x1,x2,x3,...)
    !   ylabel  : optional, as above, but for y vector
    !   targets : optional, includes an extra row in the csv file where the x values are listed as . (Stata for missing) and the y values
    !             are listed as targets(1),targets(2),... . The Stata code for "A Practical Approach to Testing Calibration Strategies" 
    !             interprets this row as giving the target moments.
    !   opts    : optional, comma separated list of options. Possible options are xabstoler,yabstoler,yreltoler,printmod,maxiter,poolsize,
    !             sleeptime,displaybestx,algorithm. For instance, opts="printmod=5,poolsize=1000,algorithm=2" will display info every 5th 
    !             time f is computed, use a poolsize of 1000, and use algorithm 2 (described above). 
    ! 
    ! REFERENCE: P. Kaelo and M. M. Ali, "Some variants of the controlled random search algorithm for global optimization," J. Optim. Theory Appl. 130 (2), 253-264 (2006). 
    ! 
    subroutine sub_crs(xstar,f,a,b,e_in,fname,xlabels,ylabels,targets,opts)
        implicit none
        real(rt), intent(inout) :: xstar(:)
        real(rt), intent(in) :: a(:),b(:)
        real(rt), intent(in), optional :: e_in(:), targets(:)
        character(*), intent(in), optional :: fname,xlabels,ylabels,opts
        interface 
            subroutine f(y,x) 
                use mod_types
                implicit none
                real(rt), dimension(:), intent(out), allocatable :: y
                real(rt), dimension(:), intent(in) :: x
            end subroutine
        end interface
        ! local other
        type(type_seed) :: randstate
        real(rt), allocatable :: xtmp(:)
        real(rt), allocatable :: X(:,:),Y(:,:) 
        real(rt), allocatable :: R(:),g(:),smx(:,:)
        real(rt) :: pool_dist
        integer :: nx, N, M, iter, ism, who, itmp(1), flag, effPoolSize, ibest, iworst, smx_iter, uid
        type(type_crs_opts) :: o
        type(type_gang) :: gang

        if (present(opts)) o = optimset(opts)

        if (o%algorithm<0 .or. o%algorithm>2) then
            write(0,*) 'sub_crs: selected algorithm not valid, use 0, 1 or 2; selecting algorithm==1'
            o%algorithm = 1
        end if

        if (mympi_nproc()<=1) then
            write(0,*) 'WARNING: sub_crs: called with only one rank, aborting'
            return
        end if

        nx = size(xstar)
        if (size(a)/=nx) stop 'ERROR: sub_crs: size of a and x do not agree'
        if (size(b)/=nx) stop 'ERROR: sub_crs: size of b and x do not agree'
        if (any(a>=b)) stop 'ERROR: sub_crs: a>=b at least once'

        ! Check that input x is in a,b for all i
        call check_guess_in_range(flag,xstar,a,b)
        if (flag/=0) then
            write(0,*) 'WARNING: sub_crs: x is not in a,b or b<a for some dimension. Will bound and continue'
            xstar = min(b,max(a,xstar))
        end if 

        ! M is the number to simultaneously try to get rid of
        M = mympi_nproc()-1

        ! N is number of trial points (the pool of points)
        ! N = max(2*M, ((100+nx)/M+1)*M, nx+2+M)
        N = max(((100+nx)/M+1)*M, nx+2+M)
        ! N = ((100+nx)/M+1)*M ! roughly evenly divide 100+nx points across the M ranks
        if (o%poolsize/=-1) N = max(nx+2+M,o%poolsize)

        if (N-M<=nx+1) call mympi_stop('ERROR: sub_crs: too few N points set relative to M')

        ! If not master, go to the "slave" routine and wait for a x to evaluate
        if (mympi_iproc()/=0) then
            call slave(f,nx)
            call mympi_bcast(xstar,0) ! receive the best point 
            call mympi_barrier()
            return
        end if

        ! If still here, then you are the master. Create the gang of workers.
        call gang%init(nx,o%sleeptime)

        ! Generate N random points uniformly from [a,b]
        allocate(xtmp(nx))
        allocate(X(nx,N))
        allocate(R(nx))
        allocate(g(nx))
        allocate(smx(nx,nx+1))                    

        ! Randomize for the master
        call randstate%randomize()

        ! Until have reached poolsize, expand
        effPoolSize = 0
        do iter = 1,o%maxiter

            ! Get the best point and the worst point
            if (effPoolSize>1) then
                itmp = minloc(Y(1,1:effPoolSize))
                ibest = itmp(1)
                itmp = maxloc(Y(1,1:effPoolSize))
                iworst = itmp(1)

                ! Report 
                pool_dist = norm2(maxval(X,2)-minval(X,2))
                if (o%printmod>0 .and. mod(iter,o%printmod)==0) then
                    write(*,'(A,i8,3(A,ES14.6),2A)') 'sub_crs: iter ',iter,' pool dist ',pool_dist,' worst ',Y(1,iworst),' best ',Y(1,ibest),'  xbest = ',printx(X(:,ibest))
                    if (effPoolSize==2 .or. mod(iter,100*o%printmod)==0) then
                        call report()
                    end if
                end if

            end if

            ! Check stopping criteria
            if (effPoolSize==N) then
                if (o%algorithm==0) then
                    write(*,'(A)') 'sub_crs: select algorithm was 0 and pool size met, so exiting.'
                    exit
                end if
                if (pool_dist<o%xabstoler) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: xabstoler of ',o%xabstoler,' met.'
                    exit
                end if
                if (abs(Y(1,iworst)-Y(1,ibest))<o%yreltoler*.5*abs(Y(1,iworst)+Y(1,ibest))) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: yreltoler of ',o%yreltoler,' met.'
                    exit
                end if
                if (abs(Y(1,iworst)-Y(1,ibest))<o%yabstoler) then
                    write(*,'(A,ES14.6,A)') 'sub_crs: yabstoler of ',o%yabstoler,' met.'
                    exit
                end if
            end if

            ! Generate a point
            if (effPoolSize<N) then

                ! If pool size is still not large enough, generate a random point.
                ! If local algorithm and the first iteration, include xstar in the pool.
                ! Otherwise, take a draw from uniform distibution over [a,b].
                if (iter==1 .and. o%algorithm==2) then
                    xtmp = xstar
                else
                    call random_number(xtmp)
                    if (.not. present(e_in)) then
                        xtmp = xtmp*a + (1._rt-xtmp)*b
                    else
                        xtmp = xtmp*max(a,xstar-e_in) + (1._rt-xtmp)*min(b,xstar+e_in)
                    end if
                end if

            else

                do smx_iter = 1,100

                    ! If here, at the refinement stage. So take the simplex of points and pick a better one
                    select case (o%algorithm)
                        case (1)
                            ! Choose a truly random simplex
                            do ism = 1,nx+1
                                smx(:,ism) = x(:,randi(1,N)) 
                            end do
                        case (2)
                            do ism = 1,nx+1
                                ! CRS with local mutation where simplex always includes best point
                                if (ism==1) then
                                    smx(:,ism) = x(:,ibest) 
                                else
                                    do 
                                        itmp(1) = randi(1,N) 
                                        if (ibest/=itmp(1) .and. iworst/=itmp(1)) exit
                                    end do
                                    smx(:,ism) = X(:,itmp(1))
                                end if
                            end do
                    end select

                    ! Compute the centroid
                    g = sum(smx,2)/(nx+1)

                    ! Reflect the worst point
                    R = 2._rt*g - X(:,iworst) ! pure reflection

                    ! If reflection point is interior, use it. 
                    ! If this is not satisfied before the max iterations is reached, then R will be used, but bounded
                    if (all(R>=a) .or. any(R<=b)) exit

                end do

                ! If goes through boundary, then bound at boundary
                xtmp = min(max(R,a),b)

            end if

            ! Here, xtmp is the point that should be assigned

            ! Wait for a free worker
            who = gang%wait_for_free()

            ! If ever previously assigned, use that value
            if (gang%slaves(who)%ever_assigned) then

                if (isnan(gang%slaves(who)%y(1))) then
                    write(0,*) 'WARNING: sub_crs: function returned a NaN value for x :',printx(gang%slaves(who)%x),' replacing with huge(0d0)'
                    gang%slaves(who)%y(1) = huge(0d0)
                end if

                if (effPoolSize<N) then

                    ! Add to pool
                    effPoolSize = effPoolSize + 1
                    X(:,effPoolSize) = gang%slaves(who)%x
                    if (.not. allocated(Y)) then

                        ! Now, the size of Y can be inferred, so allocate it.
                        allocate(Y(size(gang%slaves(who)%y),size(X,2)))

                        ! If fname is present, then export the x and y values to file.
                        if (present(fname)) then
                            call opencsv(uid,'replace')
                            call writeheader(uid)
                            close(uid) ! close the file, and reopen it later
                        end if

                    end if
                    Y(:,effPoolSize) = gang%slaves(who)%y

                    if (effPoolSize==N) write(*,'(A)') 'sub_crs: achieved full pool, moving to refinement stage'

                else

                    ! Replace the worst point if found a better one
                    if (gang%slaves(who)%y(1)<Y(1,iworst)) then

                        X(:,iworst) = gang%slaves(who)%x
                        Y(:,iworst) = gang%slaves(who)%y

                    else

                        if (o%algorithm==2) then 
                            ! If CRS2 is the algorithm, then in the case of a failure, move randomly close to the best point
                            ! The new point is a sort of random reflection of the reflection through the best point 
                            call random_number(xtmp) 
                            xtmp = (1._rt+xtmp)*X(:,ibest) - xtmp*gang%slaves(who)%x 
                            xtmp = min(max(xtmp,a),b)
                        end if

                    end if

                end if

                ! Save the value to file
                call appendtocsv(gang%slaves(who)%x,gang%slaves(who)%y)

            end if

            ! Assign them a value
            call gang%slaves(who)%assign_x(xtmp)

        end do

        ! Do a final report
        call report()
        if (iter==o%maxiter+1) &
        write(*,'(A,i8)')     'sub_crs: failed to converge in allotted iterations'
        if (iter<=o%maxiter) &
        write(*,'(A,i8)')     'sub_crs: converged successfully'

        ! Output the best
        xstar = X(:,ibest)

        call gang%shutdown()
        call mympi_bcast(xstar,0) ! receive the best point 
        call mympi_barrier()

    contains
    
        subroutine writeheader(uid)
            implicit none
            integer, intent(in) :: uid
            integer :: ix, iy

            ! Header
            if (present(xlabels)) then
                write(uid,'(A)',advance='no') xlabels ! should be comma separated list
                write(uid,'(A)',advance='no') ','
            else
                do ix = 1,nx
                    write(uid,'(A)',advance='no') 'x'//trim(conv_i_to_s(ix))//','
                end do
            end if

            if (present(ylabels)) then
                write(uid,'(A)',advance='no') ylabels
            else
                do iy = 1,size(Y,1)
                    write(uid,'(A)',advance='no') 'stat'//trim(conv_i_to_s(iy))
                    if (iy<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
            end if
            write(uid,*) 

            ! Row listing targets (if applicable)
            if (present(targets)) then

                if (size(targets)/=size(Y,1)) then
                    write(0,*) 'WARNING: size(targets) is incorrect, does not equal dimension of function''s output variable. Targets will not be written.'
                    return
                end if

                do ix = 1,nx
                    write(uid,'(A2)',advance='no') '.,' ! missing
                end do
                do iy = 1,size(Y,1)
                    write(uid,'(ES24.16E3)',advance='no') targets(iy)
                    if (iy<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) 

            end if

        end subroutine

        ! Computes the mean and variance-covariance matrix of some data x(j,i) where i gives the observation and j gives the variable
        subroutine sub_varcov(mean,varcov,d)
            implicit none
            real(8), intent(in) :: d(:,:)
            real(8), intent(out) :: mean(:),varcov(:,:)
            ! local 
            integer :: i,j,ni,nj,j2

            nj = size(d,1)
            ni = size(d,2)

            if (size(mean)/=nj) stop 'ERROR: sub_varcov: mean has the wrong size'
            if (any(shape(varcov)/=[nj,nj])) stop 'ERROR: sub_varcov: varcov has the wrong size'

            mean = sum(d,2)/ni

            varcov = 0d0
            do i = 1,size(d,2)
                do j = 1,nj
                    do j2 = j,nj
                        varcov(j,j2) = varcov(j,j2) + (d(j,i)-mean(j))*(d(j2,i)-mean(j2))/ni
                    end do
                end do
            end do
            
            ! Now the variance-covariance upper triangular is set
            ! Do the transpose to fill in the lower triangular
            do j = 1,nj
                do j2 = 1,j-1
                    varcov(j,j2) = varcov(j2,j)
                end do
            end do

        end subroutine

        subroutine report()
            implicit none
            real(8), allocatable :: mu(:),sigma(:,:)
            integer :: i

            allocate(mu(size(a)))
            allocate(sigma(size(a),size(a)))

            call sub_varcov(mu,sigma,X(:,1:effPoolSize))


            write(*,'(2A)')   'sub_crs:  ',repeat('*',120)
            write(*,'(A,i3)') 'sub_crs:            algorithm = ',o%algorithm
            if (present(fname)) &
            write(*,'(2A)')   'sub_crs:     x y history file = ',fname
            if (.not. present(fname)) &
            write(*,'(2A)')   'sub_crs:     x y history file = ','(no history file created) '
            write(*,'(A,i8)') 'sub_crs:                 iter = ',iter 
            write(*,'(A,i8)') 'sub_crs:    current pool size = ',effPoolSize
            write(*,'(A,i8)') 'sub_crs:    desired pool size = ',N
            write(*,'(2A)')   'sub_crs:                    a = ',printx(a)
            write(*,'(2A)')   'sub_crs:                    b = ',printx(b)
            if (present(e_in)) &
            write(*,'(2A)')   'sub_crs:                 e_in = ',printx(e_in)
            write(*,'(2A)')   'sub_crs:   element-wise max y = ',printx(maxval(Y(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise min y = ',printx(minval(Y(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise max x = ',printx(maxval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:   element-wise min x = ',printx(minval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:         x pool range = ',printx(maxval(X(:,1:effPoolSize),2)-minval(X(:,1:effPoolSize),2))
            write(*,'(2A)')   'sub_crs:  x range % of space  = ',printx((maxval(X(:,1:effPoolSize),2)-minval(X(:,1:effPoolSize),2))/max(b-a,epsilon(0._rt)))
            write(*,'(2A)')   'sub_crs:          x pool mean = ',printx(mu)
            do i = 1,size(a)
            write(*,'(A,I3,2A)')'sub_crs: x pool varcov(',i,',:) = ',printx(sigma(i,:))
            end do
            write(*,'(2A)')   'sub_crs:                ybest = ',printx(Y(:,ibest))
            write(*,'(2A)')   'sub_crs:                xbest = ',printx(X(:,ibest))
            write(*,'(2A)')   'sub_crs:  ',repeat('*',120)
        end subroutine 

        ! Opens fname for writing, returning the unit id corresponding to it. Catches errors in opening.
        subroutine opencsv(uid,status)
            integer, intent(out) :: uid
            character(len=*), intent(in) :: status
            do 
                ! open(newunit=uid,file=fname,action='write',status=status,position='append',err=1234)
                open(newunit=uid,file=fname,action='write',position='append',err=1234)
                exit ! If no error, escape the loop
1234            write (0,*) 'ERROR: sub_crs: File name '//trim(fname)//' could not be created'
                call mympi_abort()
            end do
        end subroutine 

        ! Appends the x and stats values to the csv file
        subroutine appendtocsv(x,y)
            implicit none
            real(8), intent(in) :: x(:),y(:)
            integer :: i

            if (present(fname)) then
                call opencsv(uid,'old')
                do i = 1,nx
                    write(uid,'(ES24.16E3)',advance='no') x(i)
                    write(uid,'(A)',advance='no') ','
                end do
                do i = 1,size(Y,1)
                    write(uid,'(ES24.16E3)',advance='no') y(i)
                    if (i<size(Y,1)) write(uid,'(A)',advance='no') ','
                end do
                write(uid,*) ! new line
                close(uid)
            end if

        end subroutine


    end subroutine


    ! Generate a random integer in a,...,b
    function randi(a,b) result(i)
        implicit none
        integer, intent(in) :: a,b
        integer :: i
        real(rt) :: tmp

        ! Generate a U[a,b+1-epsilon]
        ! Then floor to get a in 
        call random_number(tmp)
        tmp = tmp*(b+1-a) + a ! U[a,b+1]
        tmp = min(tmp,b+.99999999e0_rt)
        i = floor(tmp)

        if (floor(tmp)<a .or. floor(tmp)>b) STOP 'ERROR: randi: failed to generate i in a,...,b'

    end function

    subroutine rosenbrock(y,x)
        implicit none
        real(rt), dimension(:), intent(in) :: x
        real(rt), dimension(:), intent(out), allocatable :: y
        !local
        integer :: i,n
        real(rt) :: f_tmp, tmp

        allocate(y(1))

        call random_number(tmp)
        if (tmp<.0001_rt) then
            tmp = -1._rt
            y = sqrt(tmp)
            return
        end if

        n = size(x)
        f_tmp = 0.e0_rt
        do i = 1,n-1
             f_tmp = f_tmp + 100.e0_rt*(x(i+1)-x(i)**2)**2 + (1.e0_rt-x(i))**2
        end do
        y = f_tmp

!         call sleep(1)

    end subroutine

    subroutine check_guess_in_range(flagout,x,a,b)
        implicit none
        real(rt), intent(in) :: x(:),a(:),b(:)
        integer, intent(out) :: flagout
        ! local
        integer :: i,n,flag

        n = size(x)
        if (size(a)/=n) stop 'ERROR: check_guess_in_range: a and x do not have same length'
        if (size(b)/=n) stop 'ERROR: check_guess_in_range: b and x do not have same length'

        flag = 0
        do i = 1,n 
        if (x(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: x<a for element',i; flag = merge(-1,-999,flag==0); end if
        if (x(i)>b(i)) then; write(0,*) 'WARNING: check_guess_in_range: x>b for element',i; flag = merge(-2,-999,flag==0); end if
        if (b(i)<a(i)) then; write(0,*) 'WARNING: check_guess_in_range: b<a for element',i; flag = merge(-3,-999,flag==0); end if
        end do
        flagout = flag

    end subroutine 


    subroutine test_crs
        implicit none
        real(rt) :: x(3),a(3),b(3),e(3)

        a = 0.2_rt
        b = [1.2e0_rt,3e0_rt,99e0_rt]
        e = [.01_rt,.01_rt,.01_rt]

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,opts='printmod=100,sleeptime=0,yabstoler=1d-10,maxiter=100000,algorithm=1')

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,algorithm=2')

        print*
        x = [.3e0_rt,.1e0_rt,.5e0_rt]
        call sub_crs(x,rosenbrock,a,b,e_in=e,opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,poolsize=15')

        print*
        call sub_crs(x,rosenbrock,a,b,fname='/tmp/statmat.csv',xlabels='param1,param2,param3',ylabels='SSRS',opts='printmod=1000,sleeptime=0,yabstoler=1d-10,maxiter=100000,poolsize=2')

    end subroutine 

    ! Here, options are specified via a string such as "xtoler=1d6,printmod=10000"
    function optimset(str) result(opts)
        implicit none
        character(len=*), intent(in) :: str
        type(type_crs_opts) :: opts
        ! local
        character(len=len(str)) :: s,arg,val
        type(type_crs_opts) :: def
        integer :: nargs,i

        nargs = count_delim_arg(str,",")

        do i = 1,nargs

            ! Get the substring
            s = ''
            s = trim(get_delim_arg(str,",",i))

            ! Process to get what is left and right of the equal sign
            arg = ''
            val = ''
            arg = get_delim_arg(s,"=",1)
            val = get_delim_arg(s,"=",2)

            if     (trim(arg)=="xabstoler")     then; read(val,'(d30.16)',err=999) opts%xabstoler 
!             elseif (trim(arg)=="xreltoler")     then; read(val,'(d30.16)',err=999) opts%xreltoler 
            elseif (trim(arg)=="yabstoler")     then; read(val,'(d30.16)',err=999) opts%yabstoler 
            elseif (trim(arg)=="yreltoler")     then; read(val,'(d30.16)',err=999) opts%yreltoler 
            elseif (trim(arg)=="printmod")      then; read(val,'(i25)',   err=999) opts%printmod
            elseif (trim(arg)=="maxiter")       then; read(val,'(i25)',   err=999) opts%maxiter
            elseif (trim(arg)=="poolsize")      then; read(val,'(i25)',   err=999) opts%poolsize
            elseif (trim(arg)=="sleeptime")     then; read(val,'(i25)',   err=999) opts%sleeptime
            elseif (trim(arg)=="displaybestx")  then; read(val,'(i25)',   err=999) opts%displaybestx
            elseif (trim(arg)=="algorithm")     then; read(val,'(i25)',   err=999) opts%algorithm
            else
                write(0,*) 'WARNING: optimset: did not find option corresponding to argument ',&
                           trim(arg),'  with value ',trim(val),' and overall input ',trim(str)
            end if

        end do

        return

999     write(0,*) 'ERROR: optimset: error encountered processing ',trim(str),', with arg ',trim(arg),' and val ',trim(val),' falling back to default arguments'
        opts = def
    

    end function



end module
