#include <iostream>
#include <cstdio>
#include <new>
#include <cstdlib>
#include <cmath>
#include <cfloat>
#include <time.h> 
#include <clibgrey.h>


// std::default_random_engine greyrandgen;
// std::uniform_real_distribution<double> greyrandU01(0.0,1.0);
//    return greyrandU01(greyrandgen);

extern "C" void sgemm_( char * ta, char * tb, int * m, int *n, int *k, float *alpha, float *a, int *lda, float *b, int *ldb, float *beta, float *c, int *ldc);
extern "C" void dgemm_( char * ta, char * tb, int * m, int *n, int *k, double *alpha, double *a, int *lda, double *b, int *ldb, double *beta, double *c, int *ldc);

void mtimes(array<double>& c,array<double>& a,array<double>& b, char ta='N', char tb='N', double alpha=1., double beta=0.){
    int na1 = a.n[0], na2 = a.n[1], nb1 = b.n[0], nb2 = b.n[1];
    int m,n,k;

    m = (ta=='N' ? na1 : na2);
    n = (tb=='N' ? nb2 : nb1);
    k = (ta=='N' ? na2 : na1);

    dgemm_(&ta,&tb,&m,&n,&k,&alpha,a.data,&na1,b.data,&nb1,&beta,c.data,&na1);

}


void mtimes(double *c, double *a, double *b, int na1, int na2, int nb1, int nb2, char ta='N', char tb='N', double alpha=1., double beta=0.){
    int m,n,k;

    m = (ta=='N' ? na1 : na2);
    n = (tb=='N' ? nb2 : nb1);
    k = (ta=='N' ? na2 : na1);

    dgemm_(&ta,&tb,&m,&n,&k,&alpha,a,&na1,b,&nb1,&beta,c,&na1);

}

void mtimes(float *c, float *a, float *b, int na1, int na2, int nb1, int nb2, char ta='N', char tb='N', float alpha=1.f, float beta=0.f){
    int m,n,k;

    m = (ta=='N' ? na1 : na2);
    n = (tb=='N' ? nb2 : nb1);
    k = (ta=='N' ? na2 : na1);

    sgemm_(&ta,&tb,&m,&n,&k,&alpha,a,&na1,b,&nb1,&beta,c,&na1);
//    std::cout<<"not doing anything in mtimes"<<std::endl;

}

//namespace testbelldata{
//    extern array<double> gA(), gE(), Fee();
//};
//
//array<double> testbelldata::gA();
//array<double> testbelldata::gE();
//array<double> testbelldata::Fee();
//
//double testbudget(int iAp, int iA, int iE){
//    using namespace testbelldata;
//    return gA[iA] + gE[iE] - .96*gA[iAp];
//}

//void bellman(double *V, double *c, int *api, double *W, double (*budget)(int iap, int ia, int ie), int nAp, int nA, int nE){
template <class T>
void bellman(array<T> &V, array<T> &c, array<int> &api, array<T> &W, array<T> &gA, array<T> &gE){

    double vtmp,ctmp;
    int iE,iA,iAp;
    int nA,nE,nAp;

    nA = V.n[0];
    nE = V.n[1];
    nAp = W.n[0];

    for (iE=0; iE<nE; iE++){
        for (iA=0; iA<nA; iA++){

            V(iA,iE) = -1e100;
            c(iA,iE) = -1.;
            api(iA,iE) = -1;

            for (iAp=0; iAp<nAp; iAp++){

                ctmp = gA[iA] + gE[iE] - .96*gA[iAp];
                if (ctmp>0.) {
                    vtmp = -1./ctmp + W(iAp,iE); // the (i,j) notation is an overloaded call to [i+j*n[0]]
                    if (vtmp>V(iA,iE)) {
                        V(iA,iE) = vtmp;
                        c(iA,iE) = ctmp;
                        api(iA,iE) = iAp;
                    }
                }
            }
        }
    }
    
    return;

}

void testbellman(){
//    using namespace testbelldata;

    int nA = 100, nE = 7, nAp = nA;
    const double beta = .99;
    double sup;
    array<double> V(nA,nE), V0(nA,nE), c(nA,nE), W(nA,nE),Fee(nE,nE), gA(nA), gE(nE);
    array<double> work(nA,nE);
    array<int> api(nA,nE);

    // initialize the global variables

    // set up grids and shocks
    linspace(gA,0.,10.);
    tauchen(gE,Fee,3.,.95,.17);

    disp("gA",gA);
    disp("gE",gE);
    disp("Fee",Fee);

    // set continuation utility to zero
    W.fill(0.);
    V.fill(0.);
    disp("W",W);

    for (int iter=0; iter<1000; iter++){
 
        // compute C0 utility
        mtimes(W,V,Fee,'N','T',beta,0.); // discount factor is applied here

        // save previous value function
        V0 = V;

        // solve the bellman equation
        bellman(V,c,api,W,gA,gE);

        // check the supnorm
        minus(work,V,V0);
        abs(work);
        sup = maxval(work);

        // check the supnorm
        if (iter%100==0) std::cout<<"sup: "<<sup<<std::endl;

    }
    disp("V",V);
    V.stats();


//    greyDataSet dset("/Users/grey/Desktop/checkdset.h5");
//    dset.create();
//    dset.save("V",V);
//    dset.save("c",c);
//    dset.save("api",api);
//    dset.save("gA",gA);
//    dset.save("gE",gE);
//    dset.close();
//
//    gE.data[0] = 0.;
//    disp("gE",gE);
//
//    greyDataSet dset2("/Users/grey/Desktop/checkdset.h5");
//    dset2.read();
//    dset2.load("gE",gE);
//    dset2.close();
//    disp("gE",gE);


}

int main(){
    testbellman();
    testbisect();
    return 0; 
}
