! A collection of core routines used by many of the libgrey routines. These are designed to be basic, i.e., they have no dependencies.

! Routines
module mod_core
    use mod_types
    implicit none

    interface disp
        module procedure disp_scal, disp_vec, disp_mat, disp_3d, disp_vec_i, disp_mat_i, &
                         disp_scal_title, disp_vec_title, disp_mat_title, disp_3d_title, &
                         disp_vec_title_i, disp_mat_title_i, disp_mat_title_l, disp_mat_l
    end interface disp

    ! integer(8), private :: datetime_start(8)
    real, parameter :: INVALIDTIME = -huge(0.)    
    real, private :: tictoc_start = INVALIDTIME ! use default real


contains

! Define a function continue to prevent the use of 
integer function continue()
    implicit none
    print*,'Don''t use continue, use cycle. Continue it doesn''t do what you think!'
    continue = -1
end function

! Produces a square n by n identity matrix
function eye(n) result(mat)
    implicit none
    integer, intent(in) :: n
    real(rt), dimension(n,n) :: mat    
    integer :: i

    mat = 0._rt
    do i = 1,n
        mat(i,i) = 1._rt
    end do

end function eye

subroutine disp_mat_title_l(title,X)
    implicit none
    character(len=*), intent(in) :: title
    logical, dimension(1:,1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_mat_l(X)
end subroutine 
subroutine disp_mat_l(X)
    implicit none
    logical, dimension(1:,1:), intent(in) :: X    
    integer :: i
    character(len=5) :: dim2
    write(dim2,'(i5)') size(X,2)
    do i = 1,size(X,1)
        write(*,'('//dim2//'L2)') X(i,:)
    end do
end subroutine disp_mat_l

subroutine disp_mat_title_i(title,X)
    implicit none
    character(len=*), intent(in) :: title
    integer, dimension(1:,1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_mat_i(X)
end subroutine 
subroutine disp_mat_i(X)
    implicit none
    integer, dimension(1:,1:), intent(in) :: X    
    integer :: i
    character(len=5) :: dim2
    write(dim2,'(i5)') size(X,2)
    do i = 1,size(X,1)
        write(*,'('//dim2//'i8)') X(i,:)
    end do
end subroutine disp_mat_i

subroutine disp_vec_title_i(title,X)
    implicit none
    character(len=*), intent(in) :: title
    integer, dimension(1:), intent(in) :: X    
    write(*,'(A)') title
    call disp_vec_i(X)
end subroutine
subroutine disp_vec_i(X)
    implicit none
    integer, dimension(1:), intent(in) :: X    
    integer :: i
    do i = 1,size(X,1)
        write(*,'(i8)') X(i)
    end do
end subroutine disp_vec_i


subroutine disp_vec(X,uid)
    implicit none
    real(rt), dimension(:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),1,1,'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),1,1,'',6)
end subroutine 
subroutine disp_mat(X,uid)
    implicit none
    real(rt), dimension(:,:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),1,'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),1,'',6)
end subroutine 
subroutine disp_3d(X,uid)
    implicit none
    real(rt), dimension(:,:,:), intent(in) :: X    
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),'',uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),'',6)
end subroutine 
subroutine disp_vec_title(title,X,uid)
    implicit none
    real(rt), dimension(:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),1,1,title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),1,1,title,6)
end subroutine 
subroutine disp_mat_title(title,X,uid)
    implicit none
    real(rt), dimension(:,:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),1,title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),1,title,6)
end subroutine 
subroutine disp_3d_title(title,X,uid)
    implicit none
    real(rt), dimension(:,:,:), intent(in) :: X    
    character(len=*), intent(in) :: title
    integer, intent(in), optional :: uid
    if (     present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),title,uid)
    if (.not.present(uid)) call disp_core(X,size(X,1),size(X,2),size(X,3),title,6)
end subroutine 

subroutine disp_core(X,n1,n2,n3,title,uid)
    implicit none
    integer, intent(in) :: n1,n2,n3,uid
    character(len=*), intent(in) :: title
    real(rt), intent(in) :: X(n1,n2,*)
    ! local
    integer :: i1,i2,i3

    ! If no title or uid is present, then use title='' and uid=6
    do i3 = 1,n3
        if (n3>1) then
            write(uid,'(A,i4,A)') title//'(:,:,',i3,')'
        else
            if (title/='') write(uid,'(A)') title
        end if
        do i1 = 1,n1
            do i2 = 1,n2
                if (abs(X(i1,i2,i3))<1d5) then
                    write(uid,'(f16.5)',advance='no') X(i1,i2,i3)
                else
                    write(uid,'(es16.7e3)',advance='no') X(i1,i2,i3)
                end if
            end do
            write(uid,*)
        end do
    end do

end subroutine



subroutine disp_scal_title(title,X)
    implicit none
    character(len=*), intent(in) :: title
    real(rt), intent(in) :: X    
    write(*,'(A)') title
    call disp_scal(X)
end subroutine 
subroutine disp_scal(X)
    implicit none
    real(rt), intent(in) :: X    
    write(*,'(f12.5)') X
end subroutine disp_scal

! Acts just like the colon in Matlab
pure function colon(a,b)
    implicit none
    integer, intent(in) :: a,b
    integer, dimension(b-a+1) :: colon
    integer :: i
    do i = a,b
        colon(i-a+1) = i
    end do
end function colon

! 
subroutine bhunt(ilo,xhat,x)
    implicit none
    integer, intent(inout) :: ilo ! on input, guess; on output, lowerbound on bracketing interval
    real(rt), intent(in) :: xhat
    real(rt), dimension(:), intent(in) :: x(:)

    call bhunt_core(ilo,xhat,x,size(x))

end subroutine

subroutine bhunt_core(ilo,xhat,x,nx)
    implicit none
    real(rt), intent(in) :: x(*),xhat
    integer, intent(in) :: nx
    integer, intent(inout) :: ilo
    ! local
    integer :: l,u,step

    l = ilo 
    u = ilo+1
    step = 1

    ! Expand until xhat is in [x(l),x(u)]
    if (xhat>x(u)) then

        ! Expand to the right
        do 

            ! Check whether expansion is possible
            if (u>=nx) then
                ! Cannot expand to the right, so call bsearch and return
                u = nx
                call local_bsearch()
                return
            end if

            ! Check whether bracketed
            if (xhat<=x(u)) then
                ! Have bracketed
                call local_bsearch()
                return
            end if

            ! If still here, have not bracketed, so shift bracket
            step = 2*step
            l = u
            u = u + step

        end do

    elseif (xhat<x(l)) then

        ! Expand to the left
        do 

            ! Check whether expansion is possible
            if (l<=1) then
                ! Cannot expand to the left, so call bsearch and return
                l = 1
                call local_bsearch()
                return
            end if

            ! Check whether bracketed
            if (xhat>=x(l)) then
                ! Have bracketed
                call local_bsearch()
                return
            end if

            ! If still here, have not bracketed, so shift bracket
            step = 2*step
            u = l
            l = l - step

        end do
 
    else

        ! Here, xhat in [x(l),x(u)] and u == l+1, so return (ilo was correct)
        return
        
    end if

contains

    subroutine local_bsearch()
        implicit none
        integer :: m

        do 

            ! Check convergence criterion
            if (u<=l+1) then
                ilo = l
                return
            end if

            ! Otherwise, divide
            m = (u+l)/2 
            if (xhat>=x(m)) then
                l = m
            else
                u = m
            end if


        end do

    end subroutine

end subroutine 

! Use binary search to find the lowerbound on a bracketing interval for 
! Given sorted ascending array, find lowerbound on bracketing interval. Note that
! ilo will always be in 1,...,size(x)-1.
! 
! This is supposed to work much like the C++ standard library where the interface comes from
subroutine bsearch(ilo,xhat,x)
    implicit none
    real(rt), dimension(:), intent(in) :: x
    real(rt), intent(in) :: xhat
    integer, intent(out) :: ilo

    call bsearch_core(ilo,xhat,x,1,size(x))
        
end subroutine 

subroutine bsearch_core(ilo,xhat,x,search_lb,search_ub)
    implicit none
    real(rt), intent(in) :: x(*),xhat
    integer, intent(in) :: search_lb,search_ub
    integer, intent(out) :: ilo
    !local
    integer :: ihi,mid
        
    ilo = search_lb
    ihi = search_ub
    mid = (ihi+ilo)/2 ! mid will have ilo<=mid<ihi as long as ihi>ilo. If ihi=ilo, then mid=ilo, so the 
                      ! while loop is not entered and ilo=1 is returned 
    
    do while (mid>ilo) ! mid=ilo means ihi=ilo+1.

        ! Here, we must have ihi>ilo and ilo<=mid<ihi

        ! Bisect. 
        ! If xhat<x(1), then xhat>=x(mid) will never occur and ilo will never be modified.
        ! If xhat>=x(nx), then xhat>=x(mid) will always occur (note mid<ihi). Consequently, 
        !      ilo = mid occurs until mid==ilo, which only occures if ihi==ilo+1.
        if (xhat>=x(mid)) then
            ilo = mid
        else
            ihi = mid
        end if

        mid = (ihi+ilo)/2 

    end do

end subroutine 

! Given sorted (ascending) lists xhat and x, finds ilo as if call bsearch(ilo(i),xhat(i),x) was called, but does so faster
subroutine bsearch_many(ilo,xhat,x)
    implicit none
    integer, intent(out) :: ilo(:)
    real(rt), intent(in) :: xhat(:),x(:)
    ! local
    integer :: n,nx

    n = size(ilo) 
    nx = size(x)

    if (n>=2) then
        call bsearch_core(ilo(1),xhat(1),x,1,nx)
        call bsearch_core(ilo(n),xhat(n),x,ilo(1),nx)

        if (n==2) return
        call bsearch_many_core(ilo,xhat,x,1,n)

    elseif (n==1) then
        call bsearch(ilo(1),xhat(1),x)
    elseif (n<=0) then
        stop 'ERROR: bsearch_many: size(ilo)<=0'
    end if


end subroutine

! Assumes ilo(lb) and ilo(ub) are known, exploits monotonicity
recursive subroutine bsearch_many_core(ilo,xhat,x,lb,ub)
    implicit none
    integer, intent(inout) :: ilo(*)
    real(rt), intent(in) :: xhat(*),x(*)
    integer, intent(in) :: lb,ub
    ! local
    integer :: m

    m = (lb+ub)/2

    call bsearch_core(ilo(m),xhat(m),x,ilo(lb),ilo(ub)+1) ! need to add 1 to the upper bound because ilo always returns the lower part of the bracketing interval

    if (m>lb+1) call bsearch_many_core(ilo,xhat,x,lb,m)
    if (ub>m+1) call bsearch_many_core(ilo,xhat,x,m,ub)


end subroutine 


! linspace(a,b,n) (of course) constructs a grid of n linearly spaced points between
! a and b. If n==1, then the grid is just "a", consistent with Matlab. If n<=0, the routine
! stops with an error.
function linspace(x1,x2,n) result(grid)
    implicit none
    real(rt), intent(in) :: x1,x2
    integer, intent(in) :: n
    real(rt), dimension(1:n) :: grid
    !local
    integer :: i 

    if (n>1) then
        grid(1) = x1
        do i=2,n-1
            grid(i) = x1 + (x2-x1)*(real(i,rt)-1._rt)/(real(n,rt)-1._rt)
        end do   
        grid(n) = x2
    elseif (n==1) then
        grid(1) = x1
    else
        STOP 'linspace: ERROR: grid has size <= 0'
    end if
end function linspace

! Tic and toc for timing routines.
! Because of the odd behavior of the system clock, there is a known bug here that is 
! more likely to be experienced for long run times.
subroutine tic()
    implicit none
    ! call cpu_time(time_start)
    ! call system_clock(count=count_start,count_rate=countrate_start)
    ! call date_and_time(values=datetime_start)
    call cpu_time(tictoc_start)
end subroutine tic
subroutine toc(time_out)
    implicit none
    real(rt), intent(out), optional :: time_out
    ! local
    real :: timenow

    if (tictoc_start==INVALIDTIME) then
        write(0,*) 'WARNING: toc has been called without a corresponding tic'
        if (present(time_out)) time_out = 0.
        return
    else
        call cpu_time(timenow)
        if (present(time_out)) then
            time_out = (timenow - tictoc_start)/60.
        else
            write(*,*) 'Elapsed time (minutes): ',(timenow - tictoc_start)/60.
        end if
    end if

!     integer(8) :: datetime_end(8), ta(8)
!     real(rt) :: minutes,seconds
!     logical :: addmonth
! 
!     call date_and_time(values=datetime_end)
! 
!     ! Starting at the start time, go by minutes and increment
!     ta = datetime_start
!     minutes = 0d0
!     do while (minutes<=144000) ! go for up to 100 days (place the max in case there is an error below)
!         if (all(ta([1,2,3,5,6])==datetime_end([1,2,3,5,6]))) exit
!         minutes = minutes + 1._rt
! 
!         ta(6) = ta(6) + 1 ! add a minute
!         if (ta(6)>59) then
!             ta(6) = 0
!             ta(5) = ta(5) + 1 ! add an hour
!             if (ta(5)>23) then
!                 ta(5) = 0 
!                 ta(3) = ta(3) + 1 ! add a day
!                 select case (ta(2))
!                     case (2)
!                         addmonth = ta(3)>28
!                     case (4,6,9,11)
!                         addmonth = ta(3)>30
!                     case default
!                         addmonth = ta(3)>31
!                 end select
!                 if (addmonth) then
!                     ta(3) = 1 ! reset day to 1
!                     ta(2) = ta(2) + 1 ! add a month
!                     if (ta(2)>12) then
!                         ta(1) = ta(1) + 1 ! add a year
!                     end if
!                 end if
!             end if
!         end if
!     end do
! 
!     ! Now have the measure down to minutes, get seconds and milliseconds
!     seconds = datetime_end(7) - datetime_start(7) + (datetime_end(8) - datetime_start(8))/1000d0
! 
!     if (present(time_out)) then
!         time_out = minutes + seconds/60d0
!     else
!         write(*,'(A,f14.6,A,f14.6,A,f14.6,A,f14.6,A)') 'Elapsed time: ',&
!             (minutes+seconds/60d0)/60d0,' (hr) or ',&
!             minutes+seconds/60d0,' (min) or ',&
!             minutes*60d0 + seconds,' (sec) or ',&
!             minutes*60d3 + seconds*1d3,' (ms)'
!     end if
! 
    
end subroutine toc


end module mod_core

module mod_core_test
    use mod_core
    implicit none
contains

    subroutine test_bhunt()
        use mod_types
        implicit none
        real(rt), allocatable :: x(:)
        real(rt) :: xhat
        integer :: i,ilo
        
        allocate(x(1000))
        
        x(1) = 0._rt
        do i = 2,size(x)
            call random_number(xhat)
            x(i) = x(i-1) + xhat
        end do

        x = 3._rt*(x-x(1))/(x(size(x))-x(1)) - 1._rt! translate to [-1,2]

        ilo = size(x)/2
        do i = 1,100000
            call random_number(xhat)
            xhat = 4._rt*xhat - 1.5_rt ! translate to [-1.5,2.5]
            ! write(*,'(A,i5)',advance='no') 'iloguess',ilo
            call bhunt(ilo,xhat,x) ! guess is just whatever previous value was
            ! call bsearch(ilo,xhat,x) ! guess is just whatever previous value was
            ! print*,'xhat',xhat,'maxval(x)',maxval(x),'ilo',ilo
            if (ilo>1 .and. ilo<size(x)-1 .and. (xhat<x(ilo) .or. xhat>x(ilo+1))) stop 'ERROR: test_bhunt: failed 1'
            if (ilo==1 .and. xhat>x(2)) stop 'ERROR: test_bhunt: failed 2'
            if (ilo==size(x)-1 .and. xhat<x(size(x)-1)) stop 'ERROR: test_bhunt: failed 3'
        end do

        ! With debug flags on, bhunt takes 16.3 seconds in the lines above. If replaced with bsearch, takes 14.7 seconds. Not much slower. 10^8 evaluations (much of that may be random number generation). When "call random_number(xhat)" is commented, so bhunt has a perfect guess, takes 3.8 seconds (26% of the time relative to bsearch).

        print*,'test_bhunt: passed'

    end subroutine

    subroutine test_bsearch_many()
        use mod_types
        implicit none
        integer, allocatable :: ilo(:),ilo2(:)
        real(rt), allocatable :: xhat(:),x(:)
        real(rt) :: tmp
        integer, parameter :: n = 1000, nx = 570
        real(rt), parameter :: xlb = -3._rt, xub = 4._rt
        real(rt), parameter :: xhatlb = -6._rt, xhatub = 5._rt
        integer :: i,iter

        allocate(x(nx))
        allocate(xhat(n))
        allocate(ilo(n))
        allocate(ilo2(n))

        ! Check that the position agrees with bsearch
        do iter = 1,1000
            call randomgrids()
            call bsearch_many(ilo,xhat,x)
            do i = 1,n
                call bsearch(ilo2(i),xhat(i),x)
            end do
            if (any(ilo/=ilo2)) stop 'test_bsearch_many: failed test 1'
        end do

        print*,'test_bsearch_many: passed'

    contains

        subroutine randomgrids()
            implicit none

            ! Create increasing random lists
            xhat(1) = 0._rt
            do i = 2,n
                call random_number(tmp)
                xhat(i) = xhat(i-1) + tmp
            end do
            xhat = xhat*(xhatub-xhatlb) + xhatlb
            
            x(1) = 0._rt
            do i = 2,nx
                call random_number(tmp)
                x(i) = x(i-1) + tmp
            end do
            x = x*(xub-xlb) + xlb
        end subroutine 

    end subroutine

end module mod_core_test
