module mod_debug
    implicit none
    interface sub_debug
        module procedure sub_debug_vec, sub_debug_mat
    end interface
contains
    subroutine sub_debug_vec(X)
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(in) :: X
        
        write(*,*) 'Any NaN:             ', any(isNaN(X))
        write(*,*) 'Any large pos:       ', any(X>1d100)
        write(*,*) 'Any large neg:       ', any(X<-1d100)
        write(*,*) 'Any pos:             ', any(X>0.d0)
        write(*,*) 'Any neg:             ', any(X<0.d0)

        write(*,*) 'All NaN:             ', all(isNaN(X))
        write(*,*) 'All large pos:       ', all(X>1d100)
        write(*,*) 'All large neg:       ', all(X<-1d100)
        write(*,*) 'All pos:             ', all(X>0.d0)
        write(*,*) 'All neg:             ', all(X<0.d0)

        write(*,*) 'Monotone increasing: ', all(diff(X)>=0.d0)
        write(*,*) 'Monotone decreasing: ', all(diff(X)<=0.d0)
        write(*,*) 'Flat:                ', all(diff(X)==0.d0)

        write(*,*) 'A plot of X'
        call plot(X)

    end subroutine sub_debug_vec

    subroutine sub_debug_mat(X)
        use mod_matlab
        implicit none
        real(8), dimension(1:,1:), intent(in) :: X
        
        write(*,*) 'Any NaN:             ', any(isNaN(X))
        write(*,*) 'Any large pos:       ', any(X>1d100)
        write(*,*) 'Any large neg:       ', any(X<-1d100)
        write(*,*) 'Any pos:             ', any(X>0.d0)
        write(*,*) 'Any neg:             ', any(X<0.d0)

        write(*,*) 'All NaN:             ', all(isNaN(X))
        write(*,*) 'All large pos:       ', all(X>1d100)
        write(*,*) 'All large neg:       ', all(X<-1d100)
        write(*,*) 'All pos:             ', all(X>0.d0)
        write(*,*) 'All neg:             ', all(X<0.d0)

        write(*,*) 'Monotone increasing: ', all(X(2:size(X,1),:)-X(1:size(X,1)-1,:)>=0.d0)
        write(*,*) 'Monotone decreasing: ', all(X(2:size(X,1),:)-X(1:size(X,1)-1,:)<=0.d0)
        write(*,*) 'Flat:                ', all(X(2:size(X,1),:)-X(1:size(X,1)-1,:)==0.d0)

        write(*,*) 'A sample plot of X'
        call plot(X(:,(1+size(X,1))/2))

    end subroutine sub_debug_mat

end module mod_debug
