module mod_myhdf5
    use h5lt
    implicit none

    private 

    logical :: fortran_interface_init = .false.
    logical :: isinit = .false.
    logical :: fileisopened = .false.

    character(len=1024) :: workdir = './', fname = ''
    
    integer(HID_T) :: file_id = -1

    public :: myh5setworkdir,myh5openfile,myh5closefile,myh5write,myh5read,test_myhdf5

    interface myh5write
        module procedure h5w_6d,h5w_5d,h5w_4d,h5w_3d,h5w_2d,h5w_1d,h5w_0d,&
                         h5w_6f,h5w_5f,h5w_4f,h5w_3f,h5w_2f,h5w_1f,h5w_0f,&
                         h5w_6i,h5w_5i,h5w_4i,h5w_3i,h5w_2i,h5w_1i,h5w_0i,&
                         h5w_6l,h5w_5l,h5w_4l,h5w_3l,h5w_2l,h5w_1l,h5w_0l
    end interface
    interface myh5read
        module procedure h5r_6d,h5r_5d,h5r_4d,h5r_3d,h5r_2d,h5r_1d,h5r_0d,&
                         h5r_6f,h5r_5f,h5r_4f,h5r_3f,h5r_2f,h5r_1f,h5r_0f,&
                         h5r_6i,h5r_5i,h5r_4i,h5r_3i,h5r_2i,h5r_1i,h5r_0i,&
                         h5r_6l,h5r_5l,h5r_4l,h5r_3l,h5r_2l,h5r_1l,h5r_0l
    end interface


    

contains

    subroutine myh5setworkdir(workdir_in)
        implicit none
        character(len=*), intent(in) :: workdir_in
        integer :: n
        workdir = ''
        workdir = workdir_in

        n = len_trim(workdir)
        if (n>1 .and. workdir(1:1)=='~') then
            write(0,*) 'WARNING: myh5setworkdir: you used ~ to specify the home directory,',&
                       'but you must use the full path. Will write files to "./" instead'
            workdir = './'
            return
        end if
        if (n>1 .and. workdir(n:n)/='/') then
            ! Append 
            workdir(n+1:n+1) = '/'
        end if

    end subroutine 

    subroutine init_finterface_if_unint()
        implicit none
        integer :: err

        if (.not. fortran_interface_init) then
            call h5open_f(err)
            fortran_interface_init = .true.
        end if

    end subroutine 

    subroutine myh5closefile()
        implicit none
        integer :: err

        call init_finterface_if_unint()

        call h5fclose_f(file_id,err)
        file_id = -1
        fname = ''
        
    end subroutine


    subroutine myh5openfile(fname_in,rwopt)
        implicit none
        character(len=*), intent(in) :: fname_in,rwopt
        integer :: err

        call init_finterface_if_unint()

        fname = ''
        fname = fname_in

        if (rwopt=='replace') then
            ! Create it, replacing if it exists
            call h5fcreate_f(trim(workdir)//trim(fname),H5F_ACC_TRUNC_F,file_id,err)
        elseif (rwopt=='append') then
            stop 'at this time, have not figured how to do an append where you could modify existing elements. just use replace.'
            call h5fopen_f(trim(workdir)//trim(fname),H5F_ACC_RDWR_F,file_id,err)
        elseif (rwopt=='readonly') then
            call h5fopen_f(trim(workdir)//trim(fname),H5F_ACC_RDONLY_F,file_id,err)
        end if

        fileisopened = .true.

        if (err/=0) then
            write(0,*) 'ERROR: setfile: failed to open ',trim(workdir)//trim(fname)
        else
            isinit = .true.
        end if

    end subroutine

    ! Internal routine
!     subroutine opendataset()
!         implicit none
!         integer :: err
! 
!         call init_finterface_if_unint()
! 
!         ! Open it
!         if (fileisopened) return
! 
!         call h5fopen_f(trim(workdir)//fname,H5F_ACC_RDWR_F,file_id,err)
! 
!         if (err/=0) then
!             write(0,*) 'ERROR: opendataset: failed to open ',trim(workdir)//trim(fname)
!         else
!             isinit = .true.
!         end if
! 
!     end subroutine 

    ! Internal routine
!     subroutine closedataset()
!         implicit none
!         integer :: err
!         call h5fclose_f(file_id,err)
!         if (err/=0) then
!             write(0,*) 'ERROR: closedataset: failed to open ',trim(workdir)//trim(fname)
!         end if
!     end subroutine

    subroutine h5rc_d(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        real(8), intent(out) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer :: err 
        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTread_dataset_f(file_id,"/"//dset_name, H5T_NATIVE_DOUBLE, x, dims, err)
        deallocate(dims)
    end subroutine 
    subroutine h5rc_f(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        real(4), intent(out) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer :: err 
        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTread_dataset_f(file_id,"/"//dset_name, H5T_NATIVE_REAL, x, dims, err)
        deallocate(dims)
    end subroutine 
    subroutine h5rc_i(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        integer, intent(out) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer :: err 
        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTread_dataset_f(file_id,"/"//dset_name, H5T_NATIVE_INTEGER, x, dims, err)
        deallocate(dims)
    end subroutine 
    subroutine h5rc_l(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        logical, intent(out) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer, allocatable :: xcopy(:)
        integer :: err,i

        allocate(xcopy(product(xshape)))

        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTread_dataset_f(file_id,"/"//dset_name, H5T_NATIVE_INTEGER, xcopy, dims, err)
        deallocate(dims)

        do i = 1,product(xshape)
            x(i) = xcopy(i)==1
        end do

    end subroutine 

    subroutine h5wc_f(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        real(4), intent(in) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer :: err 

        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTmake_dataset_f(file_id,"/"//dset_name, size(xshape), dims, H5T_NATIVE_REAL, x, err)
        deallocate(dims)

    end subroutine 

    subroutine h5wc_d(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        real(8), intent(in) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer :: err 

        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTmake_dataset_f(file_id,"/"//dset_name, size(xshape), dims, H5T_NATIVE_DOUBLE, x, err)
        deallocate(dims)

    end subroutine 
    subroutine h5wc_i(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      
        integer, intent(in) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer :: err 

        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTmake_dataset_f(file_id,"/"//dset_name, size(xshape), dims, H5T_NATIVE_INTEGER, x, err)
        deallocate(dims)

    end subroutine 
!     subroutine h5wc_i2(dset_name,x,xshape)
!         implicit none
!         character(len=*), intent(in) :: dset_name      
!         integer(2), intent(in) :: x(*)
!         integer, intent(in) :: xshape(:)
!         integer(hsize_t), dimension(:), allocatable :: dims
!         integer :: err 
!         integer :: dset_id,dspace_id
! 
!         allocate(dims(size(xshape)))
!         dims = int(xshape,hsize_t)
!         ! call H5LTmake_dataset_f(file_id,"/"//dset_name, size(xshape), dims, H5T_NATIVE_INTEGER, x, err)
! 
!         call h5screate_simple_f(size(xshape), dims, dspace_id, err)
!         call h5dcreate_f(file_id,"/"//dset_name,H5T_NATIVE_INTEGER,dspace_id,dset_id,err)
!         call h5dwrite_f(dset_id,H5T_NATIVE_INTEGER,x,err)
!         call h5fflush_f(dset_id,H5F_SCOPE_GLOBAL_F,err)
!         call h5dclose_f(dset_id,err)
! 
!         deallocate(dims)
! 
!     end subroutine 
    subroutine h5wc_l(dset_name,x,xshape)
        implicit none
        character(len=*), intent(in) :: dset_name      ! name of the dataset 
        logical, intent(in) :: x(*)
        integer, intent(in) :: xshape(:)
        integer(hsize_t), dimension(:), allocatable :: dims
        integer, allocatable :: xcopy(:)
        integer :: i,err 

        allocate(xcopy(product(xshape)))
        do i = 1,product(xshape)
            xcopy(i) = merge(1,0,x(i))
        end do

        allocate(dims(size(xshape)))
        dims = int(xshape,hsize_t)
        call H5LTmake_dataset_f(file_id,"/"//dset_name, size(xshape), dims, H5T_NATIVE_INTEGER, xcopy, err)
        deallocate(dims)

        deallocate(xcopy)

    end subroutine 

    subroutine h5w_6d(d,x); implicit none; character(*) d; real(8), intent(in) :: x(:,:,:,:,:,:); call h5wc_d(d,x,shape(x)); end subroutine
    subroutine h5w_5d(d,x); implicit none; character(*) d; real(8), intent(in) :: x(:,:,:,:,:)  ; call h5wc_d(d,x,shape(x)); end subroutine
    subroutine h5w_4d(d,x); implicit none; character(*) d; real(8), intent(in) :: x(:,:,:,:)    ; call h5wc_d(d,x,shape(x)); end subroutine
    subroutine h5w_3d(d,x); implicit none; character(*) d; real(8), intent(in) :: x(:,:,:)      ; call h5wc_d(d,x,shape(x)); end subroutine
    subroutine h5w_2d(d,x); implicit none; character(*) d; real(8), intent(in) :: x(:,:)        ; call h5wc_d(d,x,shape(x)); end subroutine
    subroutine h5w_1d(d,x); implicit none; character(*) d; real(8), intent(in) :: x(:)          ; call h5wc_d(d,x,shape(x)); end subroutine
    subroutine h5w_0d(d,x); implicit none; character(*) d; real(8), intent(in) :: x             ; call h5wc_d(d,[x],[1]); end subroutine

    subroutine h5w_6f(d,x); implicit none; character(*) d; real(4), intent(in) :: x(:,:,:,:,:,:); call h5wc_f(d,x,shape(x)); end subroutine
    subroutine h5w_5f(d,x); implicit none; character(*) d; real(4), intent(in) :: x(:,:,:,:,:)  ; call h5wc_f(d,x,shape(x)); end subroutine
    subroutine h5w_4f(d,x); implicit none; character(*) d; real(4), intent(in) :: x(:,:,:,:)    ; call h5wc_f(d,x,shape(x)); end subroutine
    subroutine h5w_3f(d,x); implicit none; character(*) d; real(4), intent(in) :: x(:,:,:)      ; call h5wc_f(d,x,shape(x)); end subroutine
    subroutine h5w_2f(d,x); implicit none; character(*) d; real(4), intent(in) :: x(:,:)        ; call h5wc_f(d,x,shape(x)); end subroutine
    subroutine h5w_1f(d,x); implicit none; character(*) d; real(4), intent(in) :: x(:)          ; call h5wc_f(d,x,shape(x)); end subroutine
    subroutine h5w_0f(d,x); implicit none; character(*) d; real(4), intent(in) :: x             ; call h5wc_f(d,[x],[1]); end subroutine

    subroutine h5w_6i(d,x); implicit none; character(*) d; integer, intent(in) :: x(:,:,:,:,:,:); call h5wc_i(d,x,shape(x)); end subroutine
    subroutine h5w_5i(d,x); implicit none; character(*) d; integer, intent(in) :: x(:,:,:,:,:)  ; call h5wc_i(d,x,shape(x)); end subroutine
    subroutine h5w_4i(d,x); implicit none; character(*) d; integer, intent(in) :: x(:,:,:,:)    ; call h5wc_i(d,x,shape(x)); end subroutine
    subroutine h5w_3i(d,x); implicit none; character(*) d; integer, intent(in) :: x(:,:,:)      ; call h5wc_i(d,x,shape(x)); end subroutine
    subroutine h5w_2i(d,x); implicit none; character(*) d; integer, intent(in) :: x(:,:)        ; call h5wc_i(d,x,shape(x)); end subroutine
    subroutine h5w_1i(d,x); implicit none; character(*) d; integer, intent(in) :: x(:)          ; call h5wc_i(d,x,shape(x)); end subroutine
    subroutine h5w_0i(d,x); implicit none; character(*) d; integer, intent(in) :: x             ; call h5wc_i(d,[x],[1]); end subroutine

    subroutine h5w_6l(d,x); implicit none; character(*) d; logical, intent(in) :: x(:,:,:,:,:,:); call h5wc_l(d,x,shape(x)); end subroutine
    subroutine h5w_5l(d,x); implicit none; character(*) d; logical, intent(in) :: x(:,:,:,:,:)  ; call h5wc_l(d,x,shape(x)); end subroutine
    subroutine h5w_4l(d,x); implicit none; character(*) d; logical, intent(in) :: x(:,:,:,:)    ; call h5wc_l(d,x,shape(x)); end subroutine
    subroutine h5w_3l(d,x); implicit none; character(*) d; logical, intent(in) :: x(:,:,:)      ; call h5wc_l(d,x,shape(x)); end subroutine
    subroutine h5w_2l(d,x); implicit none; character(*) d; logical, intent(in) :: x(:,:)        ; call h5wc_l(d,x,shape(x)); end subroutine
    subroutine h5w_1l(d,x); implicit none; character(*) d; logical, intent(in) :: x(:)          ; call h5wc_l(d,x,shape(x)); end subroutine
    subroutine h5w_0l(d,x); implicit none; character(*) d; logical, intent(in) :: x             ; call h5wc_l(d,[x],[1]); end subroutine

    subroutine h5r_6d(d,x); implicit none; character(*) d; real(8), intent(out) :: x(:,:,:,:,:,:); call h5rc_d(d,x,shape(x)); end subroutine
    subroutine h5r_5d(d,x); implicit none; character(*) d; real(8), intent(out) :: x(:,:,:,:,:)  ; call h5rc_d(d,x,shape(x)); end subroutine
    subroutine h5r_4d(d,x); implicit none; character(*) d; real(8), intent(out) :: x(:,:,:,:)    ; call h5rc_d(d,x,shape(x)); end subroutine
    subroutine h5r_3d(d,x); implicit none; character(*) d; real(8), intent(out) :: x(:,:,:)      ; call h5rc_d(d,x,shape(x)); end subroutine
    subroutine h5r_2d(d,x); implicit none; character(*) d; real(8), intent(out) :: x(:,:)        ; call h5rc_d(d,x,shape(x)); end subroutine
    subroutine h5r_1d(d,x); implicit none; character(*) d; real(8), intent(out) :: x(:)          ; call h5rc_d(d,x,shape(x)); end subroutine
    subroutine h5r_0d(d,x); implicit none; character(*) d; real(8), intent(out) :: x; real(8) :: tmp(1); call h5rc_d(d,tmp,[1]); x = tmp(1); end subroutine

    subroutine h5r_6f(d,x); implicit none; character(*) d; real(4), intent(out) :: x(:,:,:,:,:,:); call h5rc_f(d,x,shape(x)); end subroutine
    subroutine h5r_5f(d,x); implicit none; character(*) d; real(4), intent(out) :: x(:,:,:,:,:)  ; call h5rc_f(d,x,shape(x)); end subroutine
    subroutine h5r_4f(d,x); implicit none; character(*) d; real(4), intent(out) :: x(:,:,:,:)    ; call h5rc_f(d,x,shape(x)); end subroutine
    subroutine h5r_3f(d,x); implicit none; character(*) d; real(4), intent(out) :: x(:,:,:)      ; call h5rc_f(d,x,shape(x)); end subroutine
    subroutine h5r_2f(d,x); implicit none; character(*) d; real(4), intent(out) :: x(:,:)        ; call h5rc_f(d,x,shape(x)); end subroutine
    subroutine h5r_1f(d,x); implicit none; character(*) d; real(4), intent(out) :: x(:)          ; call h5rc_f(d,x,shape(x)); end subroutine
    subroutine h5r_0f(d,x); implicit none; character(*) d; real(4), intent(out) :: x; real(4) :: tmp(1); call h5rc_f(d,tmp,[1]); x = tmp(1); end subroutine

    subroutine h5r_6i(d,x); implicit none; character(*) d; integer, intent(out) :: x(:,:,:,:,:,:); call h5rc_i(d,x,shape(x)); end subroutine
    subroutine h5r_5i(d,x); implicit none; character(*) d; integer, intent(out) :: x(:,:,:,:,:)  ; call h5rc_i(d,x,shape(x)); end subroutine
    subroutine h5r_4i(d,x); implicit none; character(*) d; integer, intent(out) :: x(:,:,:,:)    ; call h5rc_i(d,x,shape(x)); end subroutine
    subroutine h5r_3i(d,x); implicit none; character(*) d; integer, intent(out) :: x(:,:,:)      ; call h5rc_i(d,x,shape(x)); end subroutine
    subroutine h5r_2i(d,x); implicit none; character(*) d; integer, intent(out) :: x(:,:)        ; call h5rc_i(d,x,shape(x)); end subroutine
    subroutine h5r_1i(d,x); implicit none; character(*) d; integer, intent(out) :: x(:)          ; call h5rc_i(d,x,shape(x)); end subroutine
    subroutine h5r_0i(d,x); implicit none; character(*) d; integer, intent(out) :: x; integer :: tmp(1); call h5rc_i(d,tmp,[1]); x = tmp(1); end subroutine

    subroutine h5r_6l(d,x); implicit none; character(*) d; logical, intent(out) :: x(:,:,:,:,:,:); call h5rc_l(d,x,shape(x)); end subroutine
    subroutine h5r_5l(d,x); implicit none; character(*) d; logical, intent(out) :: x(:,:,:,:,:)  ; call h5rc_l(d,x,shape(x)); end subroutine
    subroutine h5r_4l(d,x); implicit none; character(*) d; logical, intent(out) :: x(:,:,:,:)    ; call h5rc_l(d,x,shape(x)); end subroutine
    subroutine h5r_3l(d,x); implicit none; character(*) d; logical, intent(out) :: x(:,:,:)      ; call h5rc_l(d,x,shape(x)); end subroutine
    subroutine h5r_2l(d,x); implicit none; character(*) d; logical, intent(out) :: x(:,:)        ; call h5rc_l(d,x,shape(x)); end subroutine
    subroutine h5r_1l(d,x); implicit none; character(*) d; logical, intent(out) :: x(:)          ; call h5rc_l(d,x,shape(x)); end subroutine
    subroutine h5r_0l(d,x); implicit none; character(*) d; logical, intent(out) :: x; logical :: tmp(1); call h5rc_l(d,tmp,[1]); x = tmp(1); end subroutine


!         call h5screate_simple_f(size(xshape), dims, dspace_id, err)
!         call h5dcreate_f(file_id,"/"//dset_name,H5T_NATIVE_DOUBLE,dspace_id,dset_id,err)
!         call h5dwrite_f(dset_id,H5T_NATIVE_DOUBLE,c_loc(xcopy(1)),err)
!         call h5fflush_f(dset_id,H5F_SCOPE_GLOBAL_F,err)
!         call h5dclose_f(dset_id,err)
!         deallocate(xcopy)
    

    subroutine test_myhdf5
        implicit none
        real(8), allocatable :: d1(:,:,:,:,:,:),d2(:,:,:,:,:,:)
        real(4), allocatable :: f1(:,:,:,:,:,:),f2(:,:,:,:,:,:)
        integer, allocatable :: i1(:,:,:,:,:,:),i2(:,:,:,:,:,:)
        logical, allocatable :: l1(:,:,:,:,:,:),l2(:,:,:,:,:,:)

        allocate(d1(2,3,4,5,6,7))
        allocate(d2(2,3,4,5,6,7))
        allocate(f1(2,3,4,5,6,7))
        allocate(f2(2,3,4,5,6,7))
        allocate(i1(2,3,4,5,6,7))
        allocate(i2(2,3,4,5,6,7))
        allocate(l1(2,3,4,5,6,7))
        allocate(l2(2,3,4,5,6,7))

        ! Create some random data
        call random_number(d1)
        d1 = d1 * 100d0
        f1 = real(d1,4)
        i1 = nint(d1)
        l1 = d1>50d0

        ! Set the directory
        call myh5setworkdir('/tmp/')

        ! Create a blank file
        call myh5openfile('myhdf5test.h5','replace')

        ! Write
        call myh5write('d1_0',d1(2,2,2,2,2,2))
        call myh5write('d1_1',d1(:,2,2,2,2,2))
        call myh5write('d1_2',d1(:,:,2,2,2,2))
        call myh5write('d1_3',d1(:,:,:,2,2,2))
        call myh5write('d1_4',d1(:,:,:,:,2,2))
        call myh5write('d1_5',d1(:,:,:,:,:,2))
        call myh5write('d1_6',d1(:,:,:,:,:,:))
        call myh5write('f1_0',f1(2,2,2,2,2,2))
        call myh5write('f1_1',f1(:,2,2,2,2,2))
        call myh5write('f1_2',f1(:,:,2,2,2,2))
        call myh5write('f1_3',f1(:,:,:,2,2,2))
        call myh5write('f1_4',f1(:,:,:,:,2,2))
        call myh5write('f1_5',f1(:,:,:,:,:,2))
        call myh5write('f1_6',f1(:,:,:,:,:,:))
        call myh5write('i1_0',i1(2,2,2,2,2,2))
        call myh5write('i1_1',i1(:,2,2,2,2,2))
        call myh5write('i1_2',i1(:,:,2,2,2,2))
        call myh5write('i1_3',i1(:,:,:,2,2,2))
        call myh5write('i1_4',i1(:,:,:,:,2,2))
        call myh5write('i1_5',i1(:,:,:,:,:,2))
        call myh5write('i1_6',i1(:,:,:,:,:,:))
        call myh5write('l1_0',l1(2,2,2,2,2,2))
        call myh5write('l1_1',l1(:,2,2,2,2,2))
        call myh5write('l1_2',l1(:,:,2,2,2,2))
        call myh5write('l1_3',l1(:,:,:,2,2,2))
        call myh5write('l1_4',l1(:,:,:,:,2,2))
        call myh5write('l1_5',l1(:,:,:,:,:,2))
        call myh5write('l1_6',l1(:,:,:,:,:,:))
        call myh5closefile()

        ! Read
        d2 = -1d0
        f2 = -1.0
        i2 = -1
        l2 = .false.
        call myh5openfile('myhdf5test.h5','readonly')
        call myh5read('d1_0',d2(2,2,2,2,2,2)); if (    d2(2,2,2,2,2,2)/=    d1(2,2,2,2,2,2) ) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_0'; end if 
        call myh5read('d1_1',d2(:,2,2,2,2,2)); if (any(d2(:,2,2,2,2,2)/=    d1(:,2,2,2,2,2))) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_1'; end if 
        call myh5read('d1_2',d2(:,:,2,2,2,2)); if (any(d2(:,:,2,2,2,2)/=    d1(:,:,2,2,2,2))) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_2'; end if 
        call myh5read('d1_3',d2(:,:,:,2,2,2)); if (any(d2(:,:,:,2,2,2)/=    d1(:,:,:,2,2,2))) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_3'; end if 
        call myh5read('d1_4',d2(:,:,:,:,2,2)); if (any(d2(:,:,:,:,2,2)/=    d1(:,:,:,:,2,2))) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_4'; end if 
        call myh5read('d1_5',d2(:,:,:,:,:,2)); if (any(d2(:,:,:,:,:,2)/=    d1(:,:,:,:,:,2))) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_5'; end if 
        call myh5read('d1_6',d2(:,:,:,:,:,:)); if (any(d2(:,:,:,:,:,:)/=    d1(:,:,:,:,:,:))) then; d2 = -1d0; stop 'test_myhdf5: failed test d1_6'; end if 
        call myh5read('f1_0',f2(2,2,2,2,2,2)); if (    f2(2,2,2,2,2,2)/=    f1(2,2,2,2,2,2) ) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_0'; end if 
        call myh5read('f1_1',f2(:,2,2,2,2,2)); if (any(f2(:,2,2,2,2,2)/=    f1(:,2,2,2,2,2))) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_1'; end if 
        call myh5read('f1_2',f2(:,:,2,2,2,2)); if (any(f2(:,:,2,2,2,2)/=    f1(:,:,2,2,2,2))) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_2'; end if 
        call myh5read('f1_3',f2(:,:,:,2,2,2)); if (any(f2(:,:,:,2,2,2)/=    f1(:,:,:,2,2,2))) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_3'; end if 
        call myh5read('f1_4',f2(:,:,:,:,2,2)); if (any(f2(:,:,:,:,2,2)/=    f1(:,:,:,:,2,2))) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_4'; end if 
        call myh5read('f1_5',f2(:,:,:,:,:,2)); if (any(f2(:,:,:,:,:,2)/=    f1(:,:,:,:,:,2))) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_5'; end if 
        call myh5read('f1_6',f2(:,:,:,:,:,:)); if (any(f2(:,:,:,:,:,:)/=    f1(:,:,:,:,:,:))) then; f2 = -1.0; stop 'test_myhdf5: failed test f1_6'; end if 
        call myh5read('i1_0',i2(2,2,2,2,2,2)); if (    i2(2,2,2,2,2,2)/=    i1(2,2,2,2,2,2) ) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_0'; end if 
        call myh5read('i1_1',i2(:,2,2,2,2,2)); if (any(i2(:,2,2,2,2,2)/=    i1(:,2,2,2,2,2))) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_1'; end if 
        call myh5read('i1_2',i2(:,:,2,2,2,2)); if (any(i2(:,:,2,2,2,2)/=    i1(:,:,2,2,2,2))) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_2'; end if 
        call myh5read('i1_3',i2(:,:,:,2,2,2)); if (any(i2(:,:,:,2,2,2)/=    i1(:,:,:,2,2,2))) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_3'; end if 
        call myh5read('i1_4',i2(:,:,:,:,2,2)); if (any(i2(:,:,:,:,2,2)/=    i1(:,:,:,:,2,2))) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_4'; end if 
        call myh5read('i1_5',i2(:,:,:,:,:,2)); if (any(i2(:,:,:,:,:,2)/=    i1(:,:,:,:,:,2))) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_5'; end if 
        call myh5read('i1_6',i2(:,:,:,:,:,:)); if (any(i2(:,:,:,:,:,:)/=    i1(:,:,:,:,:,:))) then; i2 = -1  ; stop 'test_myhdf5: failed test i1_6'; end if 
        call myh5read('l1_0',l2(2,2,2,2,2,2)); if (    l2(2,2,2,2,2,2).neqv.l1(2,2,2,2,2,2) ) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_0'; end if 
        call myh5read('l1_1',l2(:,2,2,2,2,2)); if (any(l2(:,2,2,2,2,2).neqv.l1(:,2,2,2,2,2))) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_1'; end if 
        call myh5read('l1_2',l2(:,:,2,2,2,2)); if (any(l2(:,:,2,2,2,2).neqv.l1(:,:,2,2,2,2))) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_2'; end if 
        call myh5read('l1_3',l2(:,:,:,2,2,2)); if (any(l2(:,:,:,2,2,2).neqv.l1(:,:,:,2,2,2))) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_3'; end if 
        call myh5read('l1_4',l2(:,:,:,:,2,2)); if (any(l2(:,:,:,:,2,2).neqv.l1(:,:,:,:,2,2))) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_4'; end if 
        call myh5read('l1_5',l2(:,:,:,:,:,2)); if (any(l2(:,:,:,:,:,2).neqv.l1(:,:,:,:,:,2))) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_5'; end if 
        call myh5read('l1_6',l2(:,:,:,:,:,:)); if (any(l2(:,:,:,:,:,:).neqv.l1(:,:,:,:,:,:))) then; l2 = -1>0; stop 'test_myhdf5: failed test l1_6'; end if 

        call myh5closefile()

        ! Clean up by replacing / deleting the file
        call myh5openfile('myhdf5test.h5','replace')
        call myh5write('deleteme',0d0)
        call myh5closefile()

        print*,'test_myhdf5: passed'

    end subroutine

end module
