! This module generates R.V.s using a Fortran 2003 interface to C++ in order to 
! call Tina's Random Number Generator (TRNG). Besides being very efficient, one advantage
! of using par_random_number is it should leave the Fortran seed unchanged (unless
! there is something very mysterious going on in TRNG).
module mod_par_random
    implicit none
contains

    ! Calls the c++ routine mytrng_uniform01, a front-end for Tina's Random Number Generator
    ! which generates U[0,1] R.V.s in parallel.
    subroutine par_random_number(x)
        use iso_c_binding
        implicit none
        interface
            function mytrng_uniform01() result(x) bind(C, name="mytrng_uniform01")
                use iso_c_binding, only: c_double
                real(c_double) :: x
            end function
        end interface

        real(c_double), intent(out) :: x

        x = mytrng_uniform01()

    end subroutine 
    subroutine par_random_seed()
        use iso_c_binding
        implicit none
        interface
            subroutine mytrng_setdefaultstate_uniform01() bind(C, name="mytrng_setdefaultstate_uniform01")
                use iso_c_binding
            end subroutine
        end interface

        ! Resets to initial state
        call mytrng_setdefaultstate_uniform01

    end subroutine 

    subroutine test_par_random_number()
        implicit none
        real(8) :: x,mean,var,skew,kurt,xhist(10),xhist2(10)
        integer, parameter :: n = 10000000
        integer :: i

        do i = 1,10
            call par_random_number(x)
            print*,'x',x
            xhist(i) = x
        end do

        call par_random_seed()

        do i = 1,10
            call par_random_number(x)
            print*,'x',x
            xhist2(i) = x
        end do

        if (any(xhist/=xhist2)) STOP 'test_par_random_number: failed test 0'

        mean = 0d0
        var = 0d0
        !$OMP PARALLEL DO DEFAULT(PRIVATE) REDUCTION(+:mean,var)
        do i = 1,n
            call par_random_number(x)
            mean = mean + x/n
            var = var + x*x/n
        end do
        !$OMP END PARALLEL DO

        var = var - mean**2

        skew = 0d0
        kurt = 0d0
        !$OMP PARALLEL DO DEFAULT(PRIVATE) SHARED(mean,var) REDUCTION(+:skew,kurt)
        do i = 1,n
            call par_random_number(x)
            skew = skew + (x-mean)**3/n / var**1.5d0
            kurt = kurt + (x-mean)**4/n / var**2 
        end do
        !$OMP END PARALLEL DO

        kurt = kurt - 3d0

        if (abs(mean-.5d0)>1d-2) then
            print*,'mean',mean
            STOP 'test_par_random_number: failed test 1'
        end if
        if (abs(var-1d0/12d0)>1d-2) STOP 'test_par_random_number: failed test 2'
        if (abs(skew)>1d-2) STOP 'test_par_random_number: failed test 3'
        if (abs(kurt-(-1.2d0))>1d-2) STOP 'test_par_random_number: failed test 4'
        
        print*,'test_par_random_number: passed'

    end subroutine
end module 
