module mod_spline
    use mod_types
    implicit none

    interface ppval
        module procedure ppval_scal,ppval_vec
    end interface
!     interface ppval2
!         module procedure :: ppval2_scal,ppval2_vec
!     end interface
!     interface pchip
!         module procedure :: pchip_getpp,pchip_interp
!     end interface
    interface schumaker_interp
        module procedure schumaker_interp_vec, schumaker_interp_scal
    end interface

    ! One-dimensional spline
    type spline
        integer :: n = -1 ! Number of knots 
        integer :: d = -1 ! d-1 is the order of the polynomial in each interval
        real(rt), allocatable :: k(:) ! Knots 
        real(rt), allocatable :: co(:,:) ! Coefficients 
        character(len=20) :: method ! Method of spline
    !!! I removed this final procedure on 4/5/13 since gfortran doesn't support it yet
!     contains
!         final :: freeSpline
    end type

    ! Two dimensional spline
    type spline2 
        integer :: n1 = -1, n2=-1 ! Number of knots 
        integer :: d = -1 ! d-1 is the order of polynomial in each interval
        real(rt), allocatable :: k1(:),k2(:) ! Knots 
        real(rt), allocatable :: co1(:,:,:) ! Coefficients for each 2nd dim knot
!         real(rt), allocatable :: co2(:,:) ! Coefficients: these are not used currently
        character(len=20) :: method ! Method of spline (for now, assume same each dimension)
    end type

!     interface vardim_eval
!         module procedure :: vardim_eval_vec, vardim_eval_scal
!     end interface

    type, public :: vardimspline
        real(rt), allocatable :: xdata(:)
        real(rt), allocatable :: knots(:),taustar(:),f_taustar(:) ! basically work arrays
        real(rt) :: lb,ub
        integer :: p,nx
    contains 
        procedure, pass :: free => vardim_free
        procedure, pass :: config => vardim_config
        procedure, pass :: neval => vardim_getnumberofevalpts
        procedure, pass :: getieval => vardim_getithevalpt
        procedure, pass :: setieval => vardim_setithevalpt
        procedure, pass :: plot => vardim_plot
        procedure, pass :: test => vardim_test
        procedure, pass :: eval => vardim_eval_vec
        procedure, pass :: eval1 => vardim_eval_scal
        procedure, pass :: disp => vardim_disp
    end type

    ! type, public :: cubicspline
    type :: cubicspline
        real(rt), allocatable :: x(:),coef(:,:),a(:),b(:)
    contains 
        procedure, pass :: construct => cubic_construct
        procedure, pass :: eval_q => cubic_eval_quick
        procedure, pass :: eval_s => cubic_eval_stable
        procedure, pass :: plot => cubic_plot
    end type

    type :: type_pchip
        real(rt), allocatable :: x(:),y(:),slopes(:)
    contains 
        procedure, pass :: construct => pchip_construct
        procedure, pass :: construct_nomono => pchip_construct_nomono
        procedure, pass :: eval => pchip_eval_scal
        procedure, pass :: eval_w_i => pchip_eval_w_i_scal
        procedure, pass :: deriv => pchip_deriv_scal
        procedure, pass :: plot => pchip_plot
!         procedure, pass :: eval_q => cubic_eval_quick
!         procedure, pass :: eval_s => cubic_eval_stable
    end type

    private :: norm

contains

    subroutine pchip_plot(this,derivorder) 
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), allocatable :: x(:),y(:),yp(:),ypp(:)
        integer, optional :: derivorder
        integer :: i,n

        n = 1000
        allocate(x(n),y(n),yp(n),ypp(n))

        x = linspace(this%x(1),this%x(size(this%x)),n)
        do i = 1,n
            ! y(i) = this%eval(x(i))
            call this%deriv(x(i),y(i),yp(i),ypp(i))
        end do

        if (present(derivorder)) then
            select case (derivorder)
                case (1)
                    call plot(x,yp,'Derivatives of pchip')
                    return
                case (2)
                    call plot(x,ypp,'Second derivatives of pchip')
                    return
                case (0) 
                case default
                    write(0,*) 'WARNING: pchip: bad deriv order, plot will nto be made'
            end select
        end if

        ! Default
        call plot(x,y,'Values of pchip')

    end subroutine 

    subroutine pchip_deriv_scal(this,x,y,yp,ypp)
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt), intent(out), optional :: y,yp,ypp
        ! local
        real(rt) :: h,t,h00,h10,h01,h11,d00,d10,d01,d11,s00,s01,s10,s11
        integer :: i

        call bsearch(i,x,this%x)

        h = (this%x(i+1)-this%x(i))
        t = (x-this%x(i))/h

        h00 = (1._rt+2._rt*t)*(1._rt-t)**2
        d00 = -2*(1._rt+2._rt*t)*(1._rt-t) + 2*(1._rt-t)**2 ! derivative of above wrt t
        s00 = -2*(2._rt)*(1._rt-t) + 2*(1._rt+2._rt*t) - 4*(1._rt-t) ! 2nd derivative of above wrt t

        h10 = t*(1._rt-t)**2
        d10 = (1-t)**2 - 2*t*(1._rt-t) ! derivative of above wrt t
        s10 = -2*(1-t) - 2*(1-2*t) ! 2nd derivative of above wrt t

        h01 = t**2*(3._rt-2._rt*t)
        d01 = 2*t*(3._rt-2._rt*t) - 2*t**2 ! derivative of above wrt t
        s01 = 2*(3._rt-2._rt*t) -4*t - 4*t

        h11 = t**2*(t-1._rt)
        d11 = t**2 + 2*t*(t-1) ! derivative of above wrt t
        s11 = 2*t + 2*(2*t-1) ! 2nd derivative of above wrt t

        ! (f(t(x)))' : f'*t' = f'/h
        ! 2nd        : (f'*t')' = f''*t' + f'*t'' = f''/h + 0

        ! Convert to derivative wrt x
        d00 = d00/h
        d10 = d10/h
        d01 = d01/h
        d11 = d11/h

        s00 = s00/h
        s10 = s10/h
        s01 = s01/h
        s11 = s11/h

        if (present(y)) y = this%y(i)*h00 + h*this%slopes(i)*h10 + this%y(i+1)*h01 + h*this%slopes(i+1)*h11
        if (present(yp)) yp = this%y(i)*d00 + h*this%slopes(i)*d10 + this%y(i+1)*d01 + h*this%slopes(i+1)*d11
        if (present(ypp)) ypp = this%y(i)*s00 + h*this%slopes(i)*s10 + this%y(i+1)*s01 + h*this%slopes(i+1)*s11
        
    end subroutine
    
    ! User provides the index, avoiding a call to bsearch
    function pchip_eval_w_i_scal(this,x,i) result(y)
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), intent(in) :: x
        integer, intent(in) :: i
        real(rt) :: y
        ! local
        real(rt) :: h,t,h00,h10,h01,h11

        ! Check that the i is correct
        if (i>=size(this%x)) stop 'WARNING: pchip_eval_w_i: i is greater than size(x)'
        if (x<this%x(i) .or. x>this%x(i+1)) write(0,*) 'WARNING: pchip_eval_w_i: i provided did not bracket x'

        h = (this%x(i+1)-this%x(i))
        t = (x-this%x(i))/h

        h00 = (1._rt+2._rt*t)*(1._rt-t)**2
        h10 = t*(1._rt-t)**2
        h01 = t**2*(3._rt-2._rt*t)
        h11 = t**2*(t-1._rt)

        y = this%y(i)*h00 + h*this%slopes(i)*h10 + this%y(i+1)*h01 + h*this%slopes(i+1)*h11
        
    end function

    function pchip_eval_scal(this,x) result(y)
        use mod_matlab
        implicit none
        class(type_pchip), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        real(rt) :: h,t,h00,h10,h01,h11
        integer :: i

        call bsearch(i,x,this%x)

        h = (this%x(i+1)-this%x(i))
        t = (x-this%x(i))/h

        h00 = (1._rt+2._rt*t)*(1._rt-t)**2
        h10 = t*(1._rt-t)**2
        h01 = t**2*(3._rt-2._rt*t)
        h11 = t**2*(t-1._rt)

        y = this%y(i)*h00 + h*this%slopes(i)*h10 + this%y(i+1)*h01 + h*this%slopes(i+1)*h11
        
    end function

    subroutine pchip_construct_nomono(this,x,y)
        use mod_util
        use mod_matlab
        implicit none
        class(type_pchip), intent(out) :: this
        real(rt), intent(in) :: x(:),y(:)
        ! local
        real(rt), dimension(:), allocatable :: delta, m
        integer :: k,n,sig

        if (allocated(this%x)) deallocate(this%x)
        if (allocated(this%y)) deallocate(this%y)
        if (allocated(this%slopes)) deallocate(this%slopes)

        n = size(x)
        call assert(size(y)==n,'phip_construct')
        
        allocate(this%x(n))
        allocate(this%y(n))
        allocate(this%slopes(n))
        allocate(delta(0:n))
        allocate(m(n))

        this%x = x
        this%y = y

        delta(1:n-1) = (y(2:n)-y(1:n-1))/(x(2:n)-x(1:n-1))
        delta(0) = delta(1)
        delta(n) = delta(n-1)

        m(2:n-1) = .5d0*(delta(1:n-2) + delta(2:n-1))
        m(1) = delta(1)
        m(n) = delta(n-1)

        this%slopes = m

    end subroutine 
    
    subroutine pchip_construct(this,x,y)
        use mod_util
        use mod_matlab
        implicit none
        class(type_pchip), intent(out) :: this
        real(rt), intent(in) :: x(:),y(:)
        ! local
        real(rt), dimension(:), allocatable :: delta, m 
        integer :: k,n,sig

        if (allocated(this%x)) deallocate(this%x)
        if (allocated(this%y)) deallocate(this%y)
        if (allocated(this%slopes)) deallocate(this%slopes)

        n = size(x)
        call assert(size(y)==n,'phip_construct')
        
        allocate(this%x(n))
        allocate(this%y(n))
        allocate(this%slopes(n))
        allocate(delta(0:n))
        allocate(m(n))

        this%x = x
        this%y = y

        delta(1:n-1) = (y(2:n)-y(1:n-1))/(x(2:n)-x(1:n-1))
        delta(0) = delta(1)
        delta(n) = delta(n-1)

        m(2:n-1) = .5d0*(delta(1:n-2) + delta(2:n-1))
        m(1) = delta(1)
        m(n) = delta(n-1)

        ! See page 476, eqn 4.2 of randall l. dougherty, alan edelman, and james m. hyman
        do k = 1,n
            if (delta(k-1)*delta(k)>0._rt) then
                sig = merge(1,-1,delta(k)>0._rt)
            else
                sig = 0
            end if
            select case (sig)
                case (1)
                    m(k) = min(max(0._rt,m(k)),3._rt*min(abs(delta(k-1)),abs(delta(k))))
                case (0)
                    m(k) = 0._rt
                case (-1)
                    m(k) = max(min(0._rt,m(k)),-3._rt*min(abs(delta(k-1)),abs(delta(k))))
            end select
        end do

        this%slopes = m


        return

!         where (delta/=0._rt) 
!             alpha = m(1:n-1)/delta
!             beta = m(2:n)/delta
!         end where

        ! Note beta = F.alpha*F.deta

!         do k = 2,n-1
!             if (delta(i-1)*delta(i)>0._rt) then
!                 
!             m = max(
!         end do
! 
!         do k = 1,n-1
!             if (delta(k)==0._rt .or. delta(k)==0._rt) then
!                 m(k) = 0._rt
!                 m(k+1) = 0._rt
!             elseif (alpha(k)<0._rt) then
!                 m(k) = 0._rt
! !                m(k+1) = 0._rt
!             elseif (k>1 .and. beta(max(k,1))<0._rt) then
!                 m(k) = 0._rt
!                 m(k+1) = 0._rt
!             else
!                 norm = norm2([alpha(k),beta(k)])
!                 if (norm>3._rt) then
!                     tau = 3._rt/norm
!                     m(k) = tau*alpha(k)*delta(k)
!                     m(k+1) = tau*beta(k)*delta(k)
!                 else
!                     m(k) = alpha(k)*delta(k)
!                     m(k+1) = beta(k)*delta(k)
!                 end if
!             end if
!         end do
! 
! 
!         this%slopes = m
! 
!         return
! 
!         i = 1
!         d_n = -huge(0._rt)
! 
!         do i = 1,n
! 
!             d_o = d_n
!             
!             if (i==1) then
!                 d_n = (y(i+1)-y(i))/(x(i+1)-x(i))
!                 d_o = d_n
!             elseif (i==n) then
!                 d_n = d_o
!             else
!                 d_n = (y(i+1)-y(i))/(x(i+1)-x(i))
!             end if
! 
!             if (d_n/=0._rt .and. d_o/=0._rt) then
!                 this%slopes(i) = .5d0*(d_o+d_n) ! average of finite difference slopes
!             else
!                 this%slopes(i) = 0d0 ! must be zero to preserve monotonicity
!             end if
!             
!             if (i<n) then
!                 a = this%slopes(i)/d_o
!                 b = this%slopes(i+1)/d_o
!                 if (a<0._rt .or. b<0._rt) then
!                     ! input data not monotone
!                     this%slopes(i) = 0d0
!                 else
!                     if (a**2+b**2<9) then
!                         this%slopes(i) = 
!                     else
!                         this%slopes(i) = 
! 
!                     end if
!                 end if
!             end if
!         end do
! 
!         do i = 1,n-1
!             a = this%slopes(i)
!             b = this%slopes(i+1)
!         end do
! 
! 
!         do i = 1,n-1
!             if (
! 
!             m_o = m_n
!             m_n = 
!         end do
! 

    end subroutine 
    

    
    subroutine vardim_free(this)
        implicit none
        class(vardimspline), intent(inout) :: this
        
        if (allocated(this%knots)) then
            deallocate(this%xdata)
            deallocate(this%knots)
            deallocate(this%taustar)
            deallocate(this%f_taustar)
            this%p = -1
            this%nx = -1
            this%lb = -1e0_rt
            this%ub = -1e0_rt
        end if

    end subroutine 

    subroutine vardim_config(this,order,xdata)
        implicit none
        class(vardimspline), intent(inout) :: this
        integer, intent(in) :: order
        real(rt), intent(in) :: xdata(:)
        ! local
        integer :: p,nx,j

        if (allocated(this%xdata)) call this%free()

        this%p = order
        this%nx = size(xdata)

        p = this%p
        nx = this%nx

        allocate(this%xdata(nx))
        this%xdata = xdata

        this%lb = xdata(1)
        this%ub = xdata(nx)

        allocate(this%knots(2*p + nx))
        this%knots(1:p) = xdata(1)
        this%knots(p+1:p+nx) = xdata
        this%knots(p+nx+1:2*p+nx) = xdata(nx)

        allocate(this%taustar(nx+p-1))
        do j = 1,nx+p-1
            this%taustar(j) = sum(this%knots(j+1:j+p))/p
        end do

        allocate(this%f_taustar(nx+p-1))

    end subroutine 
    function vardim_getnumberofevalpts(this) result(n)
        implicit none
        class(vardimspline), intent(in) :: this
        integer :: n

        n = size(this%taustar)

    end function 
    function vardim_getithevalpt(this,i) result(x)
        implicit none
        class(vardimspline), intent(in) :: this
        integer, intent(in) :: i
        real(rt) :: x

        x = this%taustar(i)

    end function
    subroutine vardim_setithevalpt(this,i,fx) 
        implicit none
        class(vardimspline), intent(inout) :: this
        integer, intent(in) :: i
        real(rt), intent(in) :: fx

        this%f_taustar(i) = fx


    end subroutine
    function vardim_eval_vec(this,x) result(y) 
        implicit none
        class(vardimspline), intent(in) :: this
        real(rt), intent(in) :: x(:)
        real(rt) :: y(size(x))
        ! local
        integer :: ix
        ! local
!         real(rt), allocatable :: work(:)
!         integer :: j
! 
!         allocate(work(size(x)))
! 
!         work = 0e0_rt
! 
!         do j = 1,this%nx + this%p - 1
!             work = work + this%f_taustar(j)*bspline(x,j,this%p+1,this%knots)
!         end do
! 
!         y = work
        
        do ix = 1,size(x)
            y(ix) = vardim_eval_scal(this,x(ix))
        end do

    end function 
    function vardim_eval_scal(this,x) result(y) 
        implicit none
        class(vardimspline), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        real(rt) :: B(this%p+1)
        integer :: ilo,ihi

        call bsplinefast(B,ilo,ihi,x,this%p+1,this%knots) 
        y = dot_product(this%f_taustar(ilo:ihi),B)

    end function 
        
    subroutine vardim_disp(this)
        implicit none
        class(vardimspline), intent(in) :: this

        print*,'*vardim ************************************************************************'
        if (allocated(this%xdata)) then
            print*,'*vardim xdata',this%xdata
        else
            print*,'*vardim xdata not allocated'
        end if
        if (allocated(this%knots)) then
            print*,'*vardim knots',this%knots
        else
            print*,'*vardim knots not allocated'
        end if
        if (allocated(this%taustar)) then
            print*,'*vardim taustar',this%taustar
        else
            print*,'*vardim taustar not allocated'
        end if
        if (allocated(this%f_taustar)) then
            print*,'*vardim f_taustar',this%f_taustar
        else
            print*,'*vardim f_taustar not allocated'
        end if
        print*,'*vardim p',this%p
        print*,'*vardim nx',this%nx
        print*,'*vardim ************************************************************************'

    end subroutine

    subroutine vardim_plot(this,xopt) 
        use mod_matlab
        implicit none
        class(vardimspline), intent(in) :: this
        real(rt), intent(in), optional :: xopt(:)
        ! local
        real(rt), allocatable :: x(:),y(:)
        integer :: nx

        nx = 1000
        if (present(xopt)) nx = size(xopt)

        allocate(x(nx),y(nx))

        x = linspace(this%lb,this%ub,nx)
        if (present(xopt)) x = xopt

        y = this%eval(x)

!         call this%disp()
!         print*,'x',x
!         print*,'y',y

        call plot(x,y,'vardimspline')

    end subroutine

    subroutine vardim_test(this,fcn,xdata,order) 
        use mod_matlab, only: linspace,plot
        implicit none
        class(vardimspline), intent(inout) :: this
        real(rt), intent(in) :: xdata(:)
        integer, intent(in) :: order
        interface 
            function fcn(x) 
                use mod_types
                implicit none
                real(rt), intent(in) :: x
                real(rt) :: fcn
            end function
        end interface
        ! local
        integer :: i
        real(rt) :: tmp
!         real(rt), allocatable :: x(:),y(:)

        call this%config(order,xdata)

        do i = 1,this%neval()
            tmp = fcn(this%getieval(i)) ! evaluate the function at the ith requested point
            call this%setieval(i,tmp) ! set the ith point
        end do

        call this%disp()

        call this%plot()

!         allocate(x(1000),y(1000))
!         
!         x = linspace(xdata(1),xdata(size(xdata)),1000)
!         y = this%eval(x)
! 
!         call plot(x,y,'test vardimspline')
! 

    end subroutine

    ! Uses a variation diminishing sple to interpolate
!     function interp_vardim(X,Y,XX,order) result(YY)
!         implicit none
!         real(8), intent(in) :: X(:),Y(:),XX(:)
!         integer, intent(in) :: order
!         real(8) :: YY(size(XX))
!         type(vardimspline) :: v
! 
!         call v%config(order,X)
! 
!     end function 

    subroutine cubic_plot(this,x)
        use mod_matlab
        implicit none
        class(cubicspline), intent(in) :: this
        real(rt), intent(in), optional :: x(:)
        ! local
        real(rt), allocatable :: xx(:),yy(:)
        integer :: i,n

        if (present(x)) then
            n = size(x)
            allocate(xx(n))
            xx = x
        else
            n = 100
            allocate(xx(n))
            xx = linspace(this%x(1),this%x(size(this%x)),n)
        end if

        allocate(yy(n))

        do i = 1,n
            yy(i) = this%eval_q(xx(i))
        end do

        call plot(xx,yy,'cubic spline')

    end subroutine

    function cubic_eval_quick(this,x) result(y)
        use mod_matlab
        implicit none
        class(cubicspline), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        integer :: i
        real(rt) :: t

        ! Check whether extrapolation
        if (x<this%x(1) .or. x>this%x(size(this%x))) then
            ! Linear
            y = interp1q(this%x,this%coef(:,1),x)
            return
        end if

        ! Get position in grid
        call bsearch(i,x,this%x)

        ! Evaluate spline using Horner's method
        t = (x - this%x(i))/(this%x(i+1)-this%x(i))
        y = this%coef(i,1) + t*(this%coef(i,2) + t*(this%coef(i,3) + t*this%coef(i,4)))

    end function 

    function cubic_eval_stable(this,x) result(y)
        use mod_matlab
        implicit none
        class(cubicspline), intent(in) :: this
        real(rt), intent(in) :: x
        real(rt) :: y
        ! local
        integer :: i
        real(rt) :: t

        ! Check whether extrapolation
        if (x<this%x(1) .or. x>this%x(size(this%x))) then
            ! Linear
            y = interp1q(this%x,this%coef(:,1),x)
            return
        end if


        ! Get position in grid
        call bsearch(i,x,this%x)

        ! Evaluate spline using Horner's method
        t = (x - this%x(i))/(this%x(i+1)-this%x(i))

        ! https://en.wikipedia.org/wiki/Spline_interpolation
        ! Better behaved numerically because all are convex combinations
        y = (1._rt-t)*this%coef(i,1) + t*this%coef(i+1,1) + t*(1._rt-t)*((1._rt-t)*this%a(i) + t*this%b(i))

    end function 
    
    subroutine cubic_construct(this,x,y)
        use mod_matlab
        implicit none
        class(cubicspline), intent(out) :: this
        real(rt), intent(in) :: x(0:),y(0:) ! note for convenience, am indexing from 0 
        ! local
        real(rt), allocatable :: Al(:),Ad(:),Au(:),D(:),dY(:)
        integer :: i,n

        n = size(x)-1

        if (allocated(this%x)) deallocate(this%x)
        if (allocated(this%coef)) deallocate(this%coef)
        if (allocated(this%a)) deallocate(this%a)
        if (allocated(this%b)) deallocate(this%b)

        if (size(y)/=size(x)) stop 'ERROR: cubic_construct: inputs wrongly sized'
        
        allocate(this%x(n+1))
        this%x = x


        allocate(Al(0:n-1),Ad(0:n),Au(0:n-1),D(0:n),Dy(0:n)) ! note lower bounds

! stop 'about to calc coefs'

        Au(0) = 1d0/(x(1)-x(0))
        Ad(0) = 2d0/(x(1)-x(0))
        Dy(0) = 3d0*(y(1)-y(0))/(x(1)-x(0))**2
        do i = 1,n-1
            Au(i) = 1d0/(x(i+1)-x(i))
            Ad(i) = 2d0*(1d0/(x(i+1)-x(i)) + 1d0/(x(i)-x(i-1)))
            Dy(i) = 3d0*((y(i+1)-y(i))/(x(i+1)-x(i))**2 + (y(i)-y(i-1))/(x(i)-x(i-1))**2)
        end do
        Ad(n) = 2d0/(x(n)-x(n-1))
        Dy(n) = 3d0*(y(n)-y(n-1))/(x(n)-x(n-1))**2

        Al = Au
        
        ! Solves the tridiagonal system here 
        ! http://mathworld.wolfram.com/CubicSpline.html
        D(0:n) = mldivide(Al,Ad,Au,Dy) 

        allocate(this%coef(n+1,4)) ! the last row is empty except for the first column, which has y(size(y))
        allocate(this%a(n))
        allocate(this%b(n))

        do i = 0,n-1
            ! this%coef(1+i,1) = y(i)
            ! this%coef(1+i,2) = D(i)
            ! this%coef(1+i,3) = 3d0*(y(i+1)-y(i)) - 2d0*D(i) - D(i+1)
            ! this%coef(1+i,4) = 2d0*(y(i)-y(i+1)) + D(i) + D(i+1)
            this%a(1+i) = D(i)*(x(i+1)-x(i)) - (y(i+1)-y(i))
            this%b(1+i) = -D(i+1)*(x(i+1)-x(i)) + (y(i+1)-y(i))
! print*,'a',a,'b',b
            this%coef(1+i,1) = y(i)
            this%coef(1+i,2) = y(i+1)-y(i) + this%a(1+i)
            this%coef(1+i,3) = this%b(1+i) - 2d0*this%a(1+i)
            this%coef(1+i,4) = this%a(1+i) - this%b(1+i)
            ! this%coef(1+i,1) = y(i)
            !  this%coef(1+i,2) = D(i)*(x(i+1)-x(i))
            ! this%coef(1+i,3) = 3d0*(y(i+1)-y(i)) + (x(i+1)-x(i))*(-2d0*D(i)-D(i+1))
            ! this%coef(1+i,4) = (x(i+1)-x(i))*(D(i+1)-D(i))
            ! this%coef(1+i,4) = 2d0*(y(i)-y(i+1)) + (x(i+1)-x(i))*(D(i)+D(i+1))
        end do
        this%coef(n+1,1) = y(n)

! call disp('this%coef',this%coef)
!          print*,'this%coef',this%coef
! print*,'g'

 !        stop 'checking coefs'



    end subroutine

    ! For a given value of x and knots t, finds the i such that x is in 
    ! [t(i),t(i+1)) and produces the non-zero B-splines from i-k+1,...,i or order k
    ! at x. These are the only non-zero basis functions for this x.
    ! 
    ! The routine, for convenience, returns ilo:=i-k+1 and ihi:=i which
    ! implicitly give the non-zero B-splines in the vector B(ilo:ihi). B itself
    ! may be much larger. 
    !
    ! The B-splines are given recursively by 
    ! 
    !  B_{i,1}(x) =  1[ x in [x_i,x_{i+1})] 
    !  B_{i,k}(x) = \frac{x-x_i}{x_{i+k-1}-x_i} B_{i,k-1}(x) + \frac{x_{i+k} - x}{x_{i+k} - x_{i+1}} B_{i+1,k-1}(x).
    ! 
    ! To use the 
    subroutine bsplinefast(Bout,ilo,ihi,x,k,t) 
        use mod_matlab
        implicit none
        real(rt), intent(in) :: x,t(:)
        integer, intent(in) :: k
        integer, intent(out) :: ilo,ihi
        real(rt), intent(out), target :: Bout(k)
        ! local
        real(rt), pointer :: Bnew(:), B(:)
        integer :: i,j,p

        call bsearch(i,x,t)

        if (.not. (x>=t(i) .and. x<t(i+1))) then
            Bout = 0e0_rt
            return
        end if

        if (k==1) then
            Bout = 1e0_rt
            return
        end if

        do p = 2,k
            
            B(i-p+2:i) => Bout(k-(p-2):k)
            Bnew(i-p+1:i) => Bout(k-(p-1):k)

            if (p==2) then
                B(i) = 1e0_rt
            end if

            ! Keep order
            do j = i-p+1,i 
!             do j = i,i-p+1,-1
                ! Compute B_{j,p} using B_{j,p-1} and B_{j+1,p-1}
                if (j==i) then
                    ! Combine B_{j,p-1}~=0 with the zero term B_{j+1,p-1}
                    Bnew(j) = (x-t(j))/(t(j+p-1)-t(j))*B(j)
                elseif (j==i-p+1) then
                    ! Combine B_{j+1,p-1}~=0 with the zero term B_{j,p-1}
                    Bnew(j) = (1-(x-t(j+1))/(t(j+1+p-1)-t(j+1)))*B(j+1)
                else
                    ! Combine B_{j,p-1}~=0 with B_{j+1,p-1}~=0
                    Bnew(j) = (x-t(j))/(t(j+p-1)-t(j))*B(j) + &
                              (1e0_rt-(x-t(j+1))/(t(j+1+p-1)-t(j+1)))*B(j+1)
                end if
            end do

        end do

        ilo = i-k+1
        ihi = i


    end subroutine 


    ! Evaluates the ith B-spline of a given order k at data x given knots t.
    ! The B-splines are given recursively by 
    ! 
    !  B_{i,1}(x) =  1[ x in [x_i,x_{i+1})] 
    !  B_{i,k}(x) = \frac{x-x_i}{x_{i+k-1}-x_i} B_{i,k-1}(x) + \frac{x_{i+k} - x}{x_{i+k} - x_{i+1}} B_{i+1,k-1}(x).
    ! 
    ! x should be sorted, as I plan to exploit this in the future.
    !
    recursive function bspline(x,i,k,t) result(B)
        implicit none
        real(rt), intent(in) :: x(:),t(:)
        integer, intent(in) :: i,k
        real(rt) :: B(size(x))
        ! local
        integer :: n,ix,nx

        if (k<=0) STOP 'ERROR: bspline: bad input value for k. Must be >= 1'

        n = size(t) 
        nx = size(x)
        
        if (i < 1 .or. i > n-k) then
            B = 0e0_rt
            return
        end if

        if (k == 1) then

            do ix = 1,nx
                if (x(ix)>=t(i) .and. x(ix)<t(i+1)) then
                    B(ix) = 1e0_rt
                else
                    B(ix) = 0e0_rt
                end if
            end do

        else
            
            if (t(i+k-1)==t(i) .and. t(i+k)==t(i+1)) then
                B = 0e0_rt
            elseif (t(i+k-1)==t(i)) then
                B = bspline(x,i+1,k-1,t)*(t(i+k)-x)/(t(i+k)-t(i+1))
            elseif (t(i+k)==t(i+1)) then
                B = bspline(x,i,k-1,t)*(x - t(i))/(t(i+k-1)-t(i)) 
            else
                B = bspline(x,i,k-1,t)*(x - t(i))/(t(i+k-1)-t(i)) + &
                    bspline(x,i+1,k-1,t)*(t(i+k)-x)/(t(i+k)-t(i+1))
            end if

        end if

    end function


!     ! Constructs the coefficients for tensor product pchip routine.
!     ! Note: this precomputes the coeficients (which are slopes) for only one dimension. 
!     subroutine pchip2(pp,X1,X2,Y)
!         implicit none
!         real(rt), dimension(1:), intent(in) :: X1,X2
!         real(rt), dimension(1:,1:), intent(in) :: Y
!         type(spline2), intent(out) :: pp
!         ! real(rt), dimension(size(X1)-1,4,size(X2)), intent(out) :: coef
!         ! local
!         integer :: i
!         type(spline) :: pp_tmp
! 
!         pp%n1 = size(X1)
!         pp%n2 = size(X2)
!         pp%method = 'pchip'
!         pp%d = 4
! 
!         ! Check sizes
!         if (size(Y,1)/=pp%n1) STOP 'size(Y,1) doesn''t agree with size(X1)'
!         if (size(Y,2)/=pp%n2) STOP 'size(Y,2) doesn''t agree with size(X2)'
! 
!         ! Precompute the coefficients for 1 dimension
!         allocate(pp%k1(pp%n1),pp%co1(pp%n1-1,4,pp%n2),pp%k2(pp%n2))
!         pp%k1 = x1
!         pp%k2 = x2
!         do i = 1,pp%n2
!             call pchip(pp_tmp,X1,Y(:,i))
!             pp%co1(:,:,i)= pp_tmp%co(:,:)
!         end do
! 
!     end subroutine pchip2 
! 
!     ! Evaluates a 2-dimensional piecewise polynomial
!     subroutine ppval2_scal(y,x1,x2,pp)
!         use mod_interp, only: search
!         implicit none
!         type(spline2), intent(in) :: pp 
!         real(rt), intent(in) :: x1,x2
!         real(rt), intent(out) :: y
!         !local
!         real(rt) :: del
!         real(rt), allocatable :: yint1(:)
!         integer :: i2,ilo1,id,d !,ilo2
!         type(spline) :: pp_tmp
! 
! !         if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
!         
!         d = pp%d
! 
!         allocate(yint1(pp%n2))
!             
!         ! Find the location of x1 and x2 in their grids
!         call search(ilo1,x1,pp%k1)
! !         call search(ilo2,x2,pp%k2)
! 
!         ! A spline has been computed for y(.,x2) for each x2 on the grid.
!         ! So evaluate the spline at (xi1,x2) for each xi1, x2, to get 
!         ! yhat(xi1,x2) for each xi1 and x2 on the grid.
! 
!         ! Evaluate the polynomial using Horner's method
!         del = x1 - pp%k1(ilo1)
!         do i2 = 1,pp%n2
!             yint1(i2) = pp%co1(ilo1,d-1,i2) + pp%co1(ilo1,d,i2)*del
!             do id = d-2,1,-1
!                 yint1(i2) = pp%co1(ilo1,id,i2) + yint1(i2)*del
!             end do
!         end do
! 
!         ! Now have yhat(xi1,x2). Construct one last spline for (.,yhat(xi1,.)) and 
!         ! evaluate it. 
!         if (pp%method=='pchip') then
!             ! pchip is mostly a local method, so this could be made more efficient
!             ! Would be better to construct a "pchipint" that would be of use also
!             ! in the one dimensional case, constructing just the local info.
!             call pchip(pp_tmp,pp%k2,yint1)
!             call ppval(y,x2,pp_tmp)
!         else
!             STOP 'ppval2: ERROR: currently, only pchip is an option'
!         end if
! 
!     end subroutine ppval2_scal

!     ! Evaluates a 2-dimensional piecewise polynomial
!     ! This is a "vectorized" version where x1 and x2 are vectors, the pp
!     ! is evaluted at the tensor product of x1 and x2. There is computational savings
!     ! in the tensor approach because, for #x1=#x2=n, there are only n spline 
!     ! constructions rather than n**2 evaluations.
!     subroutine ppval2_vec(y,x1,x2,pp)
!         use mod_interp, only: search
!         implicit none
!         type(spline2), intent(in) :: pp 
!         real(rt), dimension(1:), intent(in) :: x1,x2
!         real(rt), dimension(1:size(x1),1:size(x2)), intent(out) :: y
!         !local
!         real(rt) :: del
!         real(rt), allocatable :: yint1(:)
!         integer :: i2,ilo1,id,d, nx1,nx2, i1 
!         type(spline) :: pp_tmp
! 
! !         if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
!         
!         nx1 = size(x1)
!         nx2 = size(x2)
! 
!         d = pp%d
! 
!         allocate(yint1(pp%n2))
!             
!         ! Loop over the x1 values
!         do i1 = 1,nx1
! 
!             ! Find the location of x1 in the 1st dim grid
!             call search(ilo1,x1(i1),pp%k1)
! 
!             ! A spline has been computed for y(.,x2) for each x2 on the grid.
!             ! So evaluate the spline at (xi1,x2) for each xi1, x2, to get 
!             ! yhat(xi1,x2) for each xi1 and x2 on the grid.
! 
!             ! Evaluate the polynomial using Horner's method
!             del = x1(i1) - pp%k1(ilo1)
!             do i2 = 1,pp%n2
!                 yint1(i2) = pp%co1(ilo1,d-1,i2) + pp%co1(ilo1,d,i2)*del
!                 do id = d-2,1,-1
!                     yint1(i2) = pp%co1(ilo1,id,i2) + yint1(i2)*del
!                 end do
!             end do
! 
!             ! Now have yhat(xi1,x2). Construct one last spline for (.,yhat(xi1,.)) and 
!             ! evaluate it. 
!             if (pp%method=='pchip') then
!                 ! pchip is mostly a local method, so this could be made more efficient
!                 ! Would be better to construct a "pchipint" that would be of use also
!                 ! in the one dimensional case, constructing just the local info.
!                 call pchip(pp_tmp,pp%k2,yint1)
!                 call ppval_vec(y(i1,:),x2,pp_tmp) ! This is the vectorized version
!             else
!                 STOP 'ppval2: ERROR: currently, only pchip is an option'
!             end if
! 
!         end do
! 
!     end subroutine ppval2_vec


    ! freeSpline
    ! Free the spline memory
    subroutine freeSpline(pp) 
        implicit none
        type(spline) :: pp ! this deallocates the information
        if (allocated(pp%k)) deallocate(pp%k)
        if (allocated(pp%co)) deallocate(pp%co)
        pp%n = -1 ! this initializes to the default state 
        pp%d = -1 ! this initializes to the default state  
        pp%method = '' ! clears the method
    end subroutine

    ! checkSpline
    ! 
    !  0 == allocated and dimensions agree
    ! -1 == not allocated
    ! -2 == allocated but dimensions don't agree
    function checkSpline(pp) result(flag)
        implicit none
        type(spline), intent(in) :: pp
        integer :: flag

        if (allocated(pp%k) .and. allocated(pp%co)) then
            if (pp%n==size(pp%k) .and. pp%n-1==size(pp%co,1) &
                .and. pp%d==size(pp%co,2) .and. pp%d>0 .and. pp%n>1) then
                flag = 0
                return
            else
                flag = -2
                return
            end if
        else
            flag = -1
            return
        end if

    end function

    ! Given a piecewise polynomial pp, evaluates to recover values y.
    ! y(x) = c(i,1) + c(i,2)(x-xi) + c(i,3)(x-xi)**2 + ... if x \in [ki,ki+1]
    subroutine ppval_vec(y,x,pp) 
        use mod_interp, only: search
        implicit none
        type(spline), intent(in) :: pp 
        real(rt), intent(in) :: x(1:)
        real(rt), intent(out) :: y(1:size(x))
        !local
        real(rt) :: del
        integer :: ilo,ix,id,nx,d

        if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
        
        d = pp%d
        nx = size(x)
        
        ! Loop over each value of x
        do ix = 1,nx
            
            ! Find the location of x in grid k
            call search(ilo,x(ix),pp%k)
            del = x(ix)-pp%k(ilo)

            ! Evaluate the polynomial using Horner's method
            y(ix) = pp%co(ilo,d-1) + pp%co(ilo,d)*del
            do id = d-2,1,-1
                y(ix) = pp%co(ilo,id) + y(ix)*del
            end do

        end do

    end subroutine ppval_vec

    ! Given input x,k,c where k is knots, c is coefficients, and x is data,
    ! evaluate the spline defined by k,co at x to recover values y.
    ! y(x) = c(i,1) + c(i,2)(x-xi) + c(i,3)(x-xi)**2 + ... if x \in [ki,ki+1]
    subroutine ppval_scal(y,x,pp,ilo_in,ilo_out) 
        use mod_interp, only: search
        implicit none
        type(spline), intent(in) :: pp 
        real(rt), intent(in) :: x
        real(rt), intent(out) :: y
        integer, intent(in), optional :: ilo_in
        integer, intent(out), optional :: ilo_out
        !local
        real(rt) :: del
        integer :: ilo,id,d

        if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
        
        d = pp%d
            
        ! Find the location of x in grid k
        if (present(ilo_in)) then
            ilo = ilo_in
            ! Verify that x is in ilo,ilo+1
            if (x<pp%k(ilo) .or. x>pp%k(ilo+1)) then
                STOP 'ppval: ERROR: x is not in x(ilo,ilo+1)'
            end if
        else
            call search(ilo,x,pp%k)
        end if
        if (present(ilo_out)) then
            ilo_out = ilo
        end if

        ! Evaluate the polynomial using Horner's method
        del = x - pp%k(ilo)
        y = pp%co(ilo,d-1) + pp%co(ilo,d)*del
        do id = d-2,1,-1
            y = pp%co(ilo,id) + y*del
        end do

    end subroutine ppval_scal

    ! Given input x,k,c where k is knots, c is coefficients, and x is data,
    ! evaluate the spline defined by k,co at x to recover values y.
    ! y(x) = c(i,1) + c(i,2)(x-xi) + c(i,3)(x-xi)**2 + ... if x \in [ki,ki+1]
    subroutine ppderiv(yp,ypp,x,pp) 
        use mod_interp, only: search
        implicit none
        real(rt), intent(in) :: x(1:)
        type(spline), intent(in) :: pp
        real(rt), intent(out) :: yp(1:size(x)),ypp(1:size(x))
        !local
        real(rt) :: del
        integer :: ilo,ix,id,nx,d

        if (checkSpline(pp)/=0) STOP 'ppval: ERROR: input spline has an error.'     
        
        d = pp%d
        nx = size(x)
        
        ! Loop over each value of x
        do ix = 1,nx
            
            ! Find the location of x in grid k
            call search(ilo,x(ix),pp%k)
            del = x(ix)-pp%k(ilo)

            ! Evaluate the first derivative using Horner's method
            yp(ix) = real(d-1,rt)*pp%co(ilo,d)
            do id = d-1,2,-1
                yp(ix) = real(id-1,rt)*pp%co(ilo,id) + yp(ix)*del
            end do

            ! Evaluate the second derivative using Horner's method
            ypp(ix) = real(d-2,rt)*real(d-1,rt)*pp%co(ilo,d) ! This evaluates to zero if d is 1 or 2
            do id = d-1,3,-1
                ypp(ix) = real(id-2,rt)*real(id-1,rt)*pp%co(ilo,id) + ypp(ix)*del
            end do

        end do

    end subroutine ppderiv

    function schumaker_interp_vec(x,y,xp) result(yp)
        implicit none
        real(rt), dimension(:), intent(in) :: x,y,xp
        real(rt), dimension(size(xp)) :: yp
        type(spline) :: pp

        call schumaker(pp,x,y)
        call ppval(yp,xp,pp)

    end function
    function schumaker_interp_scal(x,y,xp) result(yp)
        implicit none
        real(rt), dimension(:), intent(in) :: x,y
        real(rt) :: xp
        real(rt) :: yp
        type(spline) :: pp

        call schumaker(pp,x,y)
        call ppval(yp,xp,pp)

    end function

    ! Produce a Schumaker shape-preserving spline 
    ! Input data is either Lagrange data x,y or Hermite data x,y,s where s is slope
    subroutine schumaker(pp,x,y,s)
        use mod_core
        implicit none
        real(rt), dimension(1:), intent(in) :: x,y
        real(rt), dimension(1:), intent(in), optional, target :: s
        type(spline), intent(out) :: pp
        !local
        real(rt), dimension(:), allocatable :: yp,dx,dy,m,L,Lm,dyp
        logical, dimension(:), allocatable :: specialCase
        real(rt) :: xsi,alpha,beta,ypbar,yplo_minus_m,yphi_minus_m,theta
        real(rt), parameter :: eps = epsilon(0e0_rt), sqrteps = sqrt(eps)
        integer :: i,n,j,nk

        pp%method = 'schumaker'
        pp%d = 3
        
        n = size(x)

        ! Check inputs
        if (size(y)/=n) STOP 'error: y and x do not have the same size'
!         if (any(isNaN(y))) then
!             print*,'y',y
!             STOP 'y is NaN'
!         end if

        ! Allocate space for local variables
        allocate(yp(n),dx(n-1),dy(n-1),m(n-1),L(n-1),Lm(n-1),dyp(n-1),specialCase(n-1))

        dx = x(2:n)-x(1:n-1)
        dy = y(2:n)-y(1:n-1)
        m = dy/dx

        ! Construct or copy the spline 
        if (present(s)) then
            ! Copy by assigning pointer
            if (size(s)/=n) STOP 'error: s and x do not have the same size'
            yp = s
        else

            ! Construct the slopes
            ! L = sqrt(dx**2 + dy**2)
            do i = 1,n-1
                L(i) = norm((/dx(i),dy(i)/))
            end do

            Lm = L*m
            do i = 2,n-1
                if ((m(i-1)>0e0_rt .and. m(i)>0e0_rt) .or. (m(i-1)<0e0_rt .and. m(i)<0e0_rt)) then
                    yp(i) = (Lm(i-1)+Lm(i))/(L(i-1)+L(i))
                else
                    yp(i) = 0e0_rt
                end if
            !    print*,(Lm(i-1)+Lm(i))/(L(i-1)+L(i))
            end do
            yp(1) = (3e0_rt*m(1) - yp(2))/2e0_rt
            yp(n) = (3e0_rt*m(n-1) - yp(n-1))/2e0_rt

        end if

        dyp = yp(2:n)-yp(1:n-1)
        specialCase = abs((yp(1:n-1)+yp(2:n))/2e0_rt - m)<4e0_rt*eps

        ! Get the number of knots - 1
        nk = 2*(n-1) - count(specialCase)

        pp%n = nk + 1
        
        ! Allocate space for output arguments
        allocate(pp%k(nk+1),pp%co(nk,3))

        ! Loop over intervals and compute coeffs for each
        j = 1
        do i = 1,n-1
        
            ! Check whether the special case where a quadratic polynomial works is satisfied
!             if (specialCase(i)) then
            if (specialCase(i)) then
                ! If so, store the coeffs and knots
                pp%co(j,1) = y(i)
                pp%co(j,2) = yp(i)
                pp%co(j,3) = dyp(i)/dx(i)/2e0_rt
                pp%k(j) = x(i)
                j = j+1

            else

                yplo_minus_m = yp(i)-m(i)
                yphi_minus_m = yp(i+1)-m(i)

                ! Note yplo_minus_m = yphi_minus_

                ! Test whether we can ensure xsi will be in (x(i),x(i+1))
                ! For this, ened |y'_j-m| \in (0, (|y'_{i+1}-y'_i|)^-1) for j = i,i+1
                

                ! Otherwise in the typical case where a quad poly doesn't work, insert a knot
                ! Three cases which will give the knot
                if (abs(yp(i+1)-yp(i))<epsilon(0e0_rt) .or. &
                    (yplo_minus_m<epsilon(0e0_rt) .and. yphi_minus_m<epsilon(0e0_rt)) .or. &
                    (yplo_minus_m>-epsilon(0e0_rt) .and. yphi_minus_m>-epsilon(0e0_rt))) then
!                 if ((yplo_minus_m<1d2*sqrteps .and. yphi_minus_m<=1d2*sqrteps) .or. (yplo_minus_m>=-1d2*sqrteps .and. yphi_minus_m>=-1d2*sqrteps)) then
                    ! Cannot shape preserve
                    xsi = (x(i)+x(i+1))/2e0_rt
                elseif (abs(yplo_minus_m)>abs(yphi_minus_m)) then
!                     xsibar = x(i) + 2e0_rt*dx(i)*yphi_minus_m(i)/dyp(i) 
!                     xsi = (x(i) + xsibar)/2e0_rt
!                     xsi = x(i) + dx(i)*yphi_minus_m/dyp(i) 

                    ! theta*dx will be the step size. we need to ensure this is not too close to 0 or 1
                    theta = yphi_minus_m/dyp(i)  ! < 0
                    theta = min(max(theta,sqrteps),1e0_rt-sqrteps)
                    xsi = x(i) + dx(i)*theta
                else 
!                     xsibar = x(i+1) + 2e0_rt*dx(i)*yplo_minus_m(i)/dyp(i)
!                     xsi = (x(i+1) + xsibar)/2e0_rt
!                     xsi = x(i+1) + dx(i)*yplo_minus_m/dyp(i)
                    theta = yphi_minus_m/dyp(i)  ! < 0
                    theta = min(max(theta,sqrteps),1e0_rt-sqrteps) 
                    theta = theta-1e0_rt
                    xsi = x(i+1) + dx(i)*theta
                end if

if (xsi<=x(i) .or. xsi>=x(i+1)) then
    print*,'spec case criterion',abs( (yp(i)+yp(i+1))/2e0_rt - m(i) )
    print*,'xsi',xsi
    print*,'x(i:i+1)',x(i:i+1)
    print*,'dyp(i)',dyp(i)
    print*,'dx(i)',dx(i)
    print*,'yphi_minus_m',yphi_minus_m
    print*,'yplo_minus_m',yplo_minus_m
    STOP 'xsi is not between x(i) and x(i+1)'
end if

                ! With the knot, there is a unique quadratic polynomial
                alpha = xsi - x(i)
                beta = x(i+1) - xsi
                ypbar = 2e0_rt*m(i) - (alpha*yp(i) + beta*yp(i+1))/dx(i)
                pp%co(j,1) = y(i)
                pp%co(j,2) = yp(i)

                pp%co(j,3) = (ypbar - yp(i))/(2e0_rt*alpha)
                pp%k(j) = x(i)
                pp%co(j+1,1) = pp%co(j,1) + alpha*(pp%co(j,2) + alpha*pp%co(j,3))
                pp%co(j+1,2) = ypbar
                pp%co(j+1,3) = (yp(i+1)-ypbar)/(2e0_rt*beta)
                pp%k(j+1) = xsi

!                 if (any(pp%co(j+1,:)>huge(0e0_rt) .or. pp%co(j,:)>huge(0e0_rt)) &
!                     .or. any(isNaN(pp%co(j+1,:)) .or. isNaN(pp%co(j,:))) &
!                     .or. max(pp%co(j+1,3),pp%co(j,3))>.5e0_rt*huge(0e0_rt)) then
                if (any(pp%co(j+1,:)>huge(0e0_rt) .or. pp%co(j,:)>huge(0e0_rt)) &
                    .or. max(pp%co(j+1,3),pp%co(j,3))>.5e0_rt*huge(0e0_rt)) then
                    print*,'xsi',xsi
                    print*,'x(i)',x(i),'x(i+1)',x(i+1)
                    print*,'y(i)',y(i),'y(i+1)',y(i+1)
                    print*,'dx(i)',dx(i)
                    print*,'dy(i)',dy(i)
                    print*,'m(i)',m(i)
                    print*,'L(i)',L(i)
                    print*,'c(j,1)',pp%co(j,1),'c(j+1,1)',pp%co(j+1,1)
                    print*,'c(j,2)',pp%co(j,2),'c(j+1,2)',pp%co(j+1,2)
                    print*,'c(j,3)',pp%co(j,3),'c(j+1,3)',pp%co(j+1,3)
                    print*,'yp(i)',yp(i),'yp(i+1)',yp(i+1)
                    print*,'yplo_minus_m',yplo_minus_m,'yphi_minus_m',yphi_minus_m
                    print*,'dx',dx(i)
                    print*,'ypbar',ypbar
                    print*,'alpha',alpha
                    print*,'beta',beta
                    print*,'c(j,1)+alpha(c(j,2)+alpha*c(j,3))',pp%co(j,1) + alpha*(pp%co(j,2) + alpha*pp%co(j,3))
                    print*,'c(j,1),alpha*c(j,2),alpha**2*c(j,3)',pp%co(j,1) , alpha*pp%co(j,2) , alpha**2*pp%co(j,3)
                    print*,'c(j,1)+alpha*c(j,2),alpha**2*c(j,3)',pp%co(j,1) + alpha*pp%co(j,2) , alpha**2*pp%co(j,3)
                    print*,'c(j,1),alpha*c(j,2)+alpha**2*c(j,3)',pp%co(j,1) , alpha*pp%co(j,2) + alpha**2*pp%co(j,3)
                    print*,'c(j,1)+alpha*c(j,2)+alpha**2*c(j,3)',pp%co(j,1) + alpha*pp%co(j,2) + alpha**2*pp%co(j,3)
                    print*,'j+1',j+1,'size(c,1)',size(pp%co,1)
                    STOP 'c is inf'
                end if

                j = j + 2

            end if
        end do

        ! Finalize k by adding the last point
        pp%k(j) = x(n)
       
        ! Deallocate space for local variables
        deallocate(dx,dy,m,L,Lm,dyp,specialCase)

        ! Deallocate space for yp if it was allocated
        if (.not. present(s)) deallocate(yp)

    end subroutine schumaker

    ! Computes the euclidean norm of X
    function norm(X) result(l2)
        implicit none
        real(rt), dimension(1:), intent(in) :: X
        real(rt) :: l2
        
        ! Call the minpack version of enorm
        l2 = sqrt(dot_product(X,X))
!         l2 = enorm(size(X),X)

!     contains
! 
!     DOUBLE PRECISION FUNCTION ENORM(N,X)
!     IMPLICIT NONE
!     INTEGER N
!     DOUBLE PRECISION X(N)
!     !     **********
!     !
!     !     function enorm
!     !
!     !     given an n-vector x, this function calculates the
!     !     euclidean norm of x.
!     !
!     !     the euclidean norm is computed by accumulating the sum of
!     !     squares in three different sums. the sums of squares for the
!     !     small and large components are scaled so that no overflows
!     !     occur. non-destructive underflows are permitted. underflows
!     !     and overflows do not occur in the computation of the unscaled
!     !     sum of squares for the intermediate components.
!     !     the definitions of small, intermediate and large components
!     !     depend on two constants, rdwarf and rgiant. the main
!     !     restrictions on these constants are that rdwarf**2 not
!     !     underflow and rgiant**2 not overflow. the constants
!     !     given here are suitable for every known computer.
!     !
!     !     the function statement is
!     !
!     !       double precision function enorm(n,x)
!     !
!     !     where
!     !
!     !       n is a positive integer input variable.
!     !
!     !       x is an input array of length n.
!     !
!     !     subprograms called
!     !
!     !       fortran-supplied ... dabs,dsqrt
!     !
!     !     argonne national laboratory. minpack project. march 1980.
!     !     burton s. garbow, kenneth e. hillstrom, jorge j. more
!     !
!     !     **********
!     INTEGER i
!     DOUBLE PRECISION agiant , floatn , one , rdwarf , rgiant , s1 , s2 , s3 , xabs , x1max , x3max , zero
!     DATA one , zero , rdwarf , rgiant/1.0D0 , 0.0D0 , 3.834D-20 ,  1.304D19/
!     s1 = zero
!     s2 = zero
!     s3 = zero
!     x1max = zero
!     x3max = zero
!     floatn = N
!     agiant = rgiant/floatn
!     DO i = 1 , N
!         xabs = DABS(X(i))
!         IF ( xabs.GT.rdwarf .AND. xabs.LT.agiant ) THEN
!         !
!         !           sum for intermediate components.
!         !
!             s2 = s2 + xabs**2
!         ELSEIF ( xabs.LE.rdwarf ) THEN
!         !
!         !              sum for small components.
!         !
!             IF ( xabs.LE.x3max ) THEN
!                 IF ( xabs.NE.zero ) s3 = s3 + (xabs/x3max)**2
!             ELSE
!                 s3 = one + s3*(x3max/xabs)**2
!                 x3max = xabs
!             ENDIF
!         !
!         !              sum for large components.
!         !
!         ELSEIF ( xabs.LE.x1max ) THEN
!             s1 = s1 + (xabs/x1max)**2
!         ELSE
!             s1 = one + s1*(x1max/xabs)**2
!             x1max = xabs
!         ENDIF
!     ENDDO
!     !
!     !     calculation of norm.
!     !
!     IF ( s1.NE.zero ) THEN
!         ENORM = x1max*DSQRT(s1+(s2/x1max)/x1max)
!     ELSEIF ( s2.EQ.zero ) THEN
!         ENORM = x3max*DSQRT(s3)
!     ELSE
!         IF ( s2.GE.x3max ) ENORM = DSQRT(s2*(one+(x3max/s2)*(x3max*s3)))
!         IF ( s2.LT.x3max ) ENORM = DSQRT(x3max*((s2/x3max)+(x3max*s3)))
!     ENDIF
!     !
!     !     last card of function enorm.
!     !
!     END function
! 

    end function 


!     ! Piecewise cubic hermite interpolating polynomial
!     ! Preserves montonicity 
!     subroutine pchip_interp(Yi,X,Y,Xi)
!         implicit none
!         real(rt), dimension(1:), intent(in) :: X,Y,Xi
!         real(rt), dimension(1:size(Xi)), intent(out) :: Yi
!         ! local
!         type(spline) :: pp 
! 
!         ! Get pp form of pchip
!         call pchip_getpp(pp,X,Y)
! 
!         ! Eval at Xi
!         call ppval(Yi,Xi,pp)
! 
!     end subroutine 
! 
!     subroutine pchip_getpp(pp,X,Y)
!         implicit none
!         type(spline), intent(out) :: pp ! intent(out) should deallocate all the components
!         real(rt), dimension(1:), intent(in) :: X,Y
!         ! local
!         real(rt), allocatable :: s(:)
!         real(rt) :: h
!         integer :: n,ierr,i
! 
!         n = size(X)
! 
!         pp%method = 'pchip'
!         pp%d = 4
! 
!         if (size(Y)/=n) STOP 'pchip: ERROR: X and Y dims don''t agree'
! 
!         pp%n = n
!         allocate(pp%k(n),pp%co(n-1,4))
!         allocate(s(n))
! 
!         ! Knots are just the X values
!         pp%k = X
! 
!         ! Coefficients: 
!         pp%co(:,1) = Y(1:n-1)
!         call dpchim(n,X,Y,s,1,ierr) ! Gets the slopes at each end point
!         pp%co(:,2) = s(1:n-1)
!         do i = 1,n-1
!             h = x(i+1) - x(i)
!             if (h<=0e0_rt) STOP 'pchip: ERROR: X is not increasing' 
!             pp%co(i,3) = - (s(i+1) + 2e0_rt*pp%co(i,2))/h + 3e0_rt*(Y(i+1)-pp%co(i,1))/(h**2)
!             pp%co(i,4) = (s(i+1) - 2e0_rt*pp%co(i,3)*h - pp%co(i,2))/(3e0_rt*h**2)
!         end do
! 
!         deallocate(s)
!     end subroutine

end module mod_spline

module mod_spline_test 
    use mod_types
    use mod_spline
    implicit none

contains

    real(rt) function f1(x)
        implicit none
        real(rt), intent(in) :: x
        f1 = sin(x)
    end function 
    real(rt) function f2(x)
        implicit none
        real(rt), intent(in) :: x
        f2 = sqrt(x)
    end function 

    subroutine test_cubic()
        use mod_spline
        use mod_matlab
        implicit none
        type(cubicspline) :: a
        real(rt), allocatable :: x(:),y(:)
        real(rt) :: lb,ub
        integer :: i

        lb = 1e-6_rt
        ub = 10d0

        allocate(x(250))
        call random_number(x)
        x = sort(x)
        x = lb + (ub-lb)*x 
        x(1) = lb
        x(size(x)) = ub


        call plot(x,sin(x),'Linear interp')

!         call a%construct([-1._rt,0._rt,3._rt],[.5_rt,0._rt,3._rt])
        call a%construct(x,sin(x))

! stop 'here'
        deallocate(x)
        allocate(x(1000),y(1000))
        x = linspace(lb,ub*1.1_rt,1000)

        call a%plot(x)

        do i = 1,size(x)
            y(i) = a%eval_s(x(i))
        end do

        call plot(x,y,'Cubic Spline, Stable Evaluation')


    end subroutine

    subroutine test_vardim()
        use mod_spline
        implicit none
        type(vardimspline) :: a

        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],1)
        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],2)
        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],3)
        call a%test(f1,[-2e0_rt,-1.5e0_rt,-1e0_rt,-.5e0_rt,1e0_rt,3e0_rt],4)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],1)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],2)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],3)
        call a%test(f2,[0e0_rt,.5e0_rt,1e0_rt,1.5e0_rt,2e0_rt,3e0_rt],4)

    end subroutine

    subroutine test_pchip()
        use mod_util
        use mod_matlab
        implicit none
        ! local
        type(type_pchip) :: p
        real(rt), allocatable :: x(:),y(:),yp(:),ypp(:),y2(:)
        integer :: i,n

        n = 20
        allocate(x(n),y(n),y2(n),yp(n),ypp(n))

        ! Check that it gets the linear function exactly right
        x = linspace(-1._rt,1._rt,n)
        y = 1._rt + x ! + .5_rt*x**2

        call p%construct(x,y)

        do i = 1,n
            call p%deriv(x(i),y2(i),yp(i),ypp(i))
            call assert(abs(y2(i)-y(i))<1d-6,'test_pchip: failed evaluation test')
            call assert(abs(-yp(i)+1._rt)<1d-6,'test_pchip: failed first derivative test')
            call assert(abs(-ypp(i))<1d-6,'test_pchip: failed second derivative test')
        end do

        ! Also check a less trivial function
        x = linspace(-10._rt,100._rt,n)
        y = max(10._rt,x,.02_rt*x**2)

        call p%construct_nomono(x,y)

        call p%plot(derivorder=2)

        print*,'test_pchip: passed'

    end subroutine 

!     subroutine test_spline()
!         use mod_matlab, only: plot,tic,toc,cumsum,linspace
!         use mod_spline
!         implicit none
!         type(spline) :: pp
!         type(spline2) :: pp2
!         real(rt) :: X(6),Y(6),X2(8),V(6,8),Vcheck(6,8)
!         real(rt) :: Xi(100),Yi(100),Ypi(100),Yppi(100),Xi2(100),Vi(100),Vitrue(100)
!         real(rt), allocatable :: tmp2(:,:)
!         integer :: i,j
! 
!         ! Test the interpolation properties of pchip
!         call random_number(Xi) 
!         Xi = cumsum(Xi+1d-6) ! xi must be strictly increasing
!         call random_number(Yi)
!         Yi = 1000e0_rt*Yi + log(Xi)
! 
!         ! Test the interpolation properties of pchip
!         call pchip(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>1d3*epsilon(0e0_rt)) then
!             print*,'maxval(abs(Yi-Ypi))',maxval(abs(Yi-Ypi))
!             STOP 'test_spline: failed test 1'
!         end if
! 
!         ! Same for schumaker
!         call schumaker(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>1d3*epsilon(0e0_rt)) STOP 'test_spline: failed test 2'
! 
!         ! Now test that linear functions can be recovered
!         Yi = -100e0_rt + .05e0_rt*Xi
!         call pchip(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 3'
! 
!         call schumaker(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 4'
! 
!         ! Now test that quadratic functions can be recovered
!         Yi = -100e0_rt + .05e0_rt*Xi - 6e0_rt*Xi**2
!         call pchip(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 5'
! 
!         call schumaker(pp,Xi,Yi)
!         call ppval(Ypi,Xi,pp)
!         if (maxval(abs(Yi-Ypi))>epsilon(0e0_rt)) STOP 'test_spline: failed test 6'
! 
!         ! Now test additional
! 
!         X = (/0e0_rt,.1e0_rt,.5e0_rt,1e0_rt,2e0_rt,10e0_rt/)
!         X2 =(/.0001e0_rt,1e0_rt,2e0_rt,3e0_rt,4e0_rt,7e0_rt,15d0,20d0/)
!         Y = sqrt(X)
! 
! 
!         call pchip(pp,X,Y)
! 
!         Xi = linspace(X(1),X(size(X)),size(Xi))
!         Xi2 = linspace(X2(1),X2(size(X2)),size(Xi))
! 
!         ! Test one dimensional
! !         call sub_pchip_eval(Yi,Xi,co,X,Y)
! !         call plot(Xi,Yi)
! 
! !         call sub_pchip_deriv(Ypi,Yi,Xi,co,X,Y)
! !         call plot(Xi,Yi)
! !         call plot(Xi,Ypi)
! 
! !         call myprint('X',X)
! !         call myprint('Y',Y)
! !         call myprint('Yi',Yi)
! !         call sub_pchip_int(yi,xi,X,Y)
! 
!         call pchip(pp,x,y)
!         call ppval(yi,xi,pp)
! ! call plot(x,y)        
! ! call plot(xi,yi,title='pchip')
!         call ppderiv(ypi,yppi,xi,pp)
! ! call plot(xi,ypi,title='pchip, deriv')
! ! call plot(xi,yppi,title='pchip, second deriv')
!         call schumaker(pp,x,y)
!         call ppval(yi,xi,pp)
! ! call plot(xi,yi,title='schumaker')
!         call ppderiv(ypi,yppi,xi,pp)
! ! call plot(xi,ypi,title='schumaker, deriv')
! ! call plot(xi,yppi,title='schumaker, second deriv')
! ! STOP 'plotting'
! 
!         ! Test two dimensional
!         do j = 1,8
!             do i = 1,6
!                 V(i,j) = sqrt(X(i)) - 10e0_rt*sqrt(X2(j))  - X(i)*X2(j)
!             end do
!         end do
!         Vitrue = sqrt(Xi) - 10e0_rt*sqrt(Xi2) - Xi*Xi2
! 
! 
!         call pchip2(pp2,X,X2,V)
! 
!         ! Check the vectorized version
!         call ppval2(Vcheck,X,X2,pp2)
! 
!         if (maxval(abs(Vcheck-V))>1000e0_rt*epsilon(0e0_rt)) then
!             print*,'maxval(abs(Vcheck-V))',maxval(abs(Vcheck-V))
!             STOP 'test_pchip: ERROR: failed test: ppval2 doesn''t interpolate'
!         end if
! 
!         do i = 1,size(xi2)
!             call ppval2(Vi(i),Xi(i),Xi2(i),pp2)
!         end do
! 
!         ! Can compare this with the matlab "test_pchip2.m" file. Good on 1/6/2013
! !         do i = 1,size(xi2)
! !             print*, Vi(i),Vitrue(i)
! !         end do
! !     STOP 'here'
! 
! 
!         if (maxval(abs(Vi-Vitrue)/abs(Vitrue))>.46e0_rt) then
!             print*,'Max relative error Vi,Vitrue',maxval(abs(Vi-Vitrue)/abs(Vitrue))
!             STOP 'test_pchip: ERROR: failed test1: spline error larger than expected'
!         end if
!         
! ! call plot(Vi)
! ! call plot(Vitrue)
! 
! !         STOP 'checking ppval2'
! 
! 
! ! call tic
! ! do i = 1,10000
! !         call sub_pchip2_coef(co2,X,X2,V)
! ! end do
! ! call toc
! ! do i = 1,10000
! !         call sub_pchip2_eval(Vi,xi,xi2,co2,X,X2,V)
! ! end do
! ! call toc
! !end do 
! !         call myprint('V',V)
! !         call myprint('co2',co2)
! !         call myprint('X2',X2)
! !         call myprint('Xi2',xi2)
! !         call myprint('Xi',xi)
! !         call myprint('Vi',Vi)
! !         call myprint('Vitrue',Vitrue)
! 
!         ! Test extreme values, 1 dim routine
!         call pchip(pp,[-2e0_rt,-1e0_rt,0e0_rt,1e0_rt,2e0_rt],[-1d150,-1e0_rt,0d0,1d0,1d150])
!         Xi = linspace(-2e0_rt,2e0_rt,size(Yi))
!         call ppval(Yi,Xi,pp)
!         call plot(Xi,Yi)
! 
!         call pchip2(pp2,[-2e0_rt,-1e0_rt,0e0_rt],[-2e0_rt,-1e0_rt,0e0_rt],reshape(&
!         (/ -2e0_rt, -10e0_rt, -1d200, &
!            -1e0_rt,  -5e0_rt, -200e0_rt, &
!             0e0_rt,  -1e0_rt,  -50e0_rt/)&
!            ,(/3,3/)))
!         Xi = linspace(-2e0_rt,0e0_rt,size(Xi))
!         Yi = linspace(-2e0_rt,0e0_rt,size(Yi))
!         allocate(tmp2(size(Xi),size(Yi)))
!         call ppval2(tmp2,Xi,Yi,pp2)
!         print*,'maxval(tmp2)',maxval(tmp2)
!         
!         call plot(Xi,tmp2)
! 
!         print*,'Yi:',Yi
!         STOP 'testing spline extreme'
! 
!     end subroutine
end module 
