module mod_table
    use mod_types
    implicit none

    integer, parameter :: TABLE_TYPE_DP = 8
    integer, parameter :: TABLE_TYPE_SP = 4
    integer, parameter :: TABLE_TYPE_INT = 2

    character(len=*), parameter :: TABLE_FMT_DP = 'ES24.16E3'
    character(len=*), parameter :: TABLE_FMT_SP = 'ES15.8E2'
    character(len=*), parameter :: TABLE_FMT_INT = 'I10'

    integer, parameter :: TABLE_NAME_LEN = 256
!     integer, parameter :: TABLE_MAX_RANK = 10

    type row
        integer :: t ! type
        integer :: rank
        integer, dimension(:), allocatable :: dims ! dimensions
        integer(1), dimension(:), allocatable :: d ! data (one byte for each element)
        character(len=TABLE_NAME_LEN) :: who ! name
    end type

    type table
        type(row), dimension(:), allocatable :: r ! rows
        integer :: nrows = 0
!     contains
!         procedure, pass :: add => table_add
    end type

    interface create_row
        procedure :: create_row_1d, create_row_2d
    end interface

!     interface pack_data
!         procedure :: pack_data_r8
!     end interface
!     interface get_type
!         procedure :: get_type_r8 
!     end interface
!     interface create_row
!         procedure :: create_row_r8_d1,create_row_r8_d0
!     end interface

contains

    subroutine test_table()
        implicit none
        type(row) :: r

        r = create_row('test',[.01d0,.2d0,.3d0])

        print*,'.01d0',.01d0
        write(*,'(A,ES23.16E3)') ' .01d0   ',.01d0
        print*,'.01',.01
        write(*,'(A,ES14.8E2)') ' .01   ',.01
        print*,'huge(1)',huge(1)
        write(*,'(A,I10)') ' huge(1)  ',huge(1)

        print*,'r%d',r%d
        print*,'r%t',r%t
        print*,'r%rank',r%rank
        print*,'r%dims',r%dims

        call write_row(6,r)

    end subroutine 

    function get_type(x) result(i)
        implicit none
        class(*), intent(in) :: x
        integer :: i
       
        i = -1
        select type (x)
            type is (real(8))
                i = TABLE_TYPE_DP
            type is (real(4))
                i = TABLE_TYPE_SP
            type is (integer)
                i = TABLE_TYPE_INT
            class default
                i = -1
        end select

    end function 
    
    function create_row_1d(who,x) result(r)
        use mod_types
        implicit none
        type(row) :: r
        character(len=*), intent(in) :: who
        class(*), intent(in) :: x(:) ! unlimited polymorphic type
        ! local
        integer(1) :: proto(1)

        r%who = who
        r%rank = size(shape(x))
        allocate(r%dims(r%rank)); r%dims(1:r%rank) = shape(x)
        r%t = get_type(x(1))
        if (allocated(r%d)) deallocate(r%d)
        allocate(r%d(size(transfer(x,proto))))
        r%d = transfer(x,proto)

    end function

    function create_row_2d(who,x) result(r)
        use mod_types
        implicit none
        type(row) :: r
        character(len=*), intent(in) :: who
        class(*), intent(in) :: x(:,:) ! unlimited polymorphic type
        ! local
        integer(1) :: proto(1)

        r%who = who
        r%rank = size(shape(x))
        allocate(r%dims(r%rank)); r%dims(1:r%rank) = shape(x)
        r%t = get_type(x(1,1))
        if (allocated(r%d)) deallocate(r%d)
        allocate(r%d(size(transfer(x,proto))))
        r%d = transfer(x,proto)

    end function
    
    subroutine write_row(uid,r)
        implicit none
        integer, intent(in) :: uid
        type(row), intent(in) :: r
        ! 
        integer :: i
        character(len=12) :: f
        real(8), allocatable :: r8(:)
        real(4), allocatable :: r4(:)
        integer, allocatable :: i4(:)


        write(uid,'(A,A1,I2,A1)',advance='no') trim(r%who),',',r%rank,','
        do i = 1,r%rank
            write(uid,'(I10,A1)',advance='no') r%dims(i),','
        end do

        select case (r%t)
            case (TABLE_TYPE_DP)
                f = '('//TABLE_FMT_DP//')'
                allocate(r8(product(r%dims)))

                r8 = transfer(r%d,r8(1)) 

                do i = 1,product(r%dims)-1
                    write(uid,f,advance='no') r8(i)
                    write(uid,'(A1)',advance='no') ','
                end do
                write(uid,f,advance='no') r8(size(r8))

            case (TABLE_TYPE_SP)
                f = '('//TABLE_FMT_SP//')'
                allocate(r4(product(r%dims)))
                r4 = transfer(r%d,r4(1)) 

                do i = 1,product(r%dims)-1
                    write(uid,f,advance='no') r4(i)
                    write(uid,'(A1)',advance='no') ','
                end do
                write(uid,f,advance='no') r4(size(r4))

            case (TABLE_TYPE_INT)
                f = '('//TABLE_FMT_INT//')'
                allocate(i4(product(r%dims)))
                i4 = transfer(r%d,i4(1)) 

                do i = 1,product(r%dims)-1
                    write(uid,f,advance='no') i4(i)
                    write(uid,'(A1)',advance='no') ','
                end do
                write(uid,f,advance='no') i4(size(i4))

            case default
                stop 'write_row: type not found'
        end select

        write(uid,*)


    end subroutine

    

!     function get_type_r8(x) result(i); real(8) :: x; integer :: i; i = TABLE_TYPE_DP; end function 
! 
!     function create_row_r8_d1(x) result(r); type(row) :: r; real(8) :: x(:); r = create_row_r8_core(x,size(x)); end function 
!     function create_row_r8_d0(x) result(r); type(row) :: r; real(8) :: x; r = create_row_r8_core([x],1); end function 
!     function create_row_r8_core(x,nx) result(r)
!         use mod_types
!         implicit none
!         type(row) :: r
!         integer(1), dimension(:), allocatable, intent(out) :: dat(:)
!         real(rt), intent(in) :: x(*)
!         integer, intent(in) :: nx
!         ! local
!         integer(1) :: proto
! 
!         r%t = get_type_r8(x(1))
!     
!         if (allocated(r%dat)) deallocate(r%dat)
! 
!         allocate(r%dat(size(transfer(x(1:nx),proto))))
!         r%dat = transfer(x(1:nx),proto)
! 
!     end function

!     subroutine pack_data_r8(dat,x,nx)
!         use mod_types
!         implicit none
!         integer(1), dimension(:), allocatable, intent(out) :: dat(:)
!         real(rt), intent(in) :: x(*)
!         integer, intent(in) :: nx
!         ! local
!         integer(1) :: proto
! 
!         allocate(dat(size(transfer(x(1:nx),proto))))
!         dat = transfer(x(1:nx),proto)
! 
!     end subroutine

!     subroutine table_add_r8(this,who,x)
!         implicit none
!         class(table), intent(inout) :: this
!         character(len=*), intent(in) :: who
!         real(rt), intent(in) :: x(:)
! 
!         call table_add_core(this,who,x,shape(x))
! 
!     end subroutine
! 
!     subroutine table_add_r8_d1(this,who,x)
!         implicit none
!         class(table), intent(inout) :: this
!         character(len=*), intent(in) :: who
!         real(8), dimension(:) :: x
!         type(row) :: r
! 
!         r = create_row(this,x)
! 
! 
! 
!     end subroutine

!     subroutine table_add_core(this,who,dims)
!         implicit none
!         class(table), intent(inout) :: this
!         character(len=*), intent(in) :: who
!         class(*), dimension(*) :: x
!         integer, intent(in) :: dims
! 
!         ! Make more room (if necessary)
!         call table_create_empty_row(this)
! 
!         ! Now start assigning
!         call pack_data(this%r(this%nrows)%d,x,nx)
!         this%r(this%nrows)%t = get_type(x(1))
!         this%r(this%nrows)%who = who
!         
!     end subroutine

    ! Create space for another row
    subroutine table_create_empty_row(this)
        implicit none
        class(table), intent(inout) :: this
        type(table) :: copy
        integer :: i

        if (this%nrows==0) then
            allocate(this%r(1)) 
            this%nrows=1
        else
            if (size(this%r)>this%nrows) then
                this%nrows=this%nrows+1
            else

                ! Double the available size of the table
                copy = this

                deallocate(this%r)
                allocate(this%r(this%nrows*2))

                do i = 1,this%nrows
                    this%r(i) = copy%r(i)
                end do

                this%nrows=this%nrows+1

            end if
        end if


    end subroutine

end module
