function [] = analyze()
    
    addpath('matlib012717','-end');

    setdefaults();
    
    clear global
    global EXTRA DATADIR
    
    DATADIR='runtimes/';
    
    EXTRA.abstract = {};
    EXTRA.intro = {};
    EXTRA.basic= {}; 
    EXTRA.twostate = {};
    EXTRA.nonmono = {};
    EXTRA.contopt = {};
    
    
%     figure_timeeval_mono_only();
%     table_mono_only();
%     table_mono_and_conc();    
%     figure_empirical_On();
%     figure_twostate()
%     table_twostate();
%     table_rbcdef_sorting();    
%     table_contopt();
    
    fprintf(1,'\nADDITIONAL STATISTICS IN PAPER\n');
    print_extra();
    
end 

function [] = figure_twostate()
    
    global DATADIR EXTRA
    
    T = myreadtable([DATADIR 'table_joblist4.csv']);
    T = sortrows(T,{'nZ','nK'});
    
    T.lambda = T.x1;
    T.lambda_alt = T.x2;
    
    T.min_lambda = min(T.lambda,T.lambda_alt);
    T.x1 = [];
    T.x2 = [];
        
    yname = 'On2';
%     yname = 'min_lambda';
    
    T.cnt = T.cnt .* T.nZ; % get the total count (it is measured per capital state, and there are nZ of these)

    T.speedup = T.nK.^2 .* T.nZ ./ T.cnt; % speedup relative to brute force
    T.On2 = T.cnt ./ (T.nK .* T.nZ);

%     Lambda = T(:,{'nZ','nK','min_lambda'});
%     Lambda = unstack(Lambda,'min_lambda','nZ');
%     
%     Lambda
%     Lambda = table2array(Lambda);
%     h = plot(log2(Lambda(:,1)),diag(Lambda(1:8,2:9)));
%     mydefaultplotformat(h);
%     title('$\lambda$ as $n_1=n_2=n''$ grows')
%     ylabel('$\lambda$');
%     xlabel('$\log_2 n$')
%     printfig('lambda_growth')
    
    EXTRA.twostate{1+end} = sprintf('Max of min lambda: %f',max(T.min_lambda));
    EXTRA.twostate{1+end} = sprintf('Min of min lambda: %f',min(T.min_lambda));
    
    T = T(:,{'nZ','nK',yname});
        
    T = unstack(T,yname,'nZ');

%     T
    

    n = T.nK;
    [nKK,nZZ] = ndgrid(n,n);
    T.nK = [];
    
    
    T = table2array(T);
        
%     plot3(T.nK,T.nZ,log(T.cnt.*T.nK.*T.nZ),'.')
%     plot3(T.nK,T.nZ,log(T.minutes),'.')
    xlabel('$n_1=n$')
    ylabel('$n_2$')
    zlabel(yname)
    
    tmp = nKK==nZZ;
    
%     T = log(T);
    f = fontoptions();

    close all
    mynewfigure(struct('aspect',.55,'width',7.5));
    
%     plot(nKK(tmp),T(tmp))
    hold on
    j = 0;
    for i = -3:1:3
        j = j+1;
        
        tmp = nKK==(2.^i)*nZZ;

        nstates = log2(nKK(tmp).*nZZ(tmp));
        
%         log2(nstates)
        
        h(j) = plot(nstates,T(tmp));
        if i>=0
            leg{j} = sprintf('$n_1/n_2 = %.0f$',2.^i);
        else
            leg{j} = sprintf('$n_1/n_2 = 1/%.0f$',2.^-i);
%             if i==-1
%             leg{j} = sprintf('$n_1/n_2 = %.1f$',2.^i);
%         elseif i==-2
%             leg{j} = sprintf('$n_1/n_2 = %.2f$',2.^i);
%         elseif i==-3
%             leg{j} = sprintf('$n_1/n_2 = %.2f$',2.^i);
        end
    end
    mydefaultplotformat(h);
%     axis tight
    h = legend(leg{:},'location','northeast');
    set(h,f{:});
%     xlabel('$n_1=n''$')
    xlabel('Base 2 log of number of states, $\log_2(n_1 n_2)$')
    ylabel('Evaluation count per state')
    title('Evaluation count per state for different $n_1/n_2$ ratios')
    grid on
    hold off

%     set(gca,'XTick',160:160:1280)
    
    axis([9.7 17.5 1 3])
%     axis([100 204800 1 Inf])
    
    printfig('On2behavior_multi')

end

function [] = table_rbcdef_sorting()

    global DATADIR
    
    t = myreadtable([DATADIR 'output.csv']);
    
    % Extract the no cash at hand
    for icash = 0:1
        
        w = t(t.nB>=50 & t.docash==icash,{'nB','vfiiter','timetotal','evalcnt','method'});
        
        w.converged = w.vfiiter<500;
        
        if icash==1
            w.evalcnt = w.evalcnt .* w.nB; % want average per state. in Fortran, is divided by nB*nK*nZ, so need to adjust for cash at hand.
        end
        
        w = unstack(w,{'vfiiter','timetotal','evalcnt','converged'},'method');
        w = sortrows(w);
            
        z = table2struct(w,'ToScalar',true);
        
        fcn = @(x,n) exp(interp1(log(n(~isnan(x))),log(x(~isnan(x))),log(n),'linear','extrap'));
%         pfcn = @(x,n) plot(n(~isnan(x)),x(~isnan(x)),'+',n,fcn(x,n),'-');
                
        names = {'timetotal','evalcnt'};
        short = {'time','eval'};
        for inames = 1:length(names)
            for m = [1 4 7]
                
                oldname = sprintf('%s_x%i',names{inames},m);
                fitname = sprintf('fit_%s_x%i',names{inames},m);
                combname = sprintf('%s%i',short{inames},m);
                estname = sprintf('%s%i_est',short{inames},m);
                                
                z.(estname) = false(size(z.(oldname)));
                
%                 if inames==1
%                     figure(m)
%                     pfcn(z.(oldname),z.nB)
%                 end
                z.(fitname)   = fcn(z.(oldname),z.nB);
                
                
                tmp = z.(oldname);
                z.(estname)(isnan(tmp)) = true;
                tmp(isnan(tmp)) = z.(fitname)(isnan(tmp));
                z.(combname) = tmp;
                
                z = rmfield(z,{oldname,fitname});
                
            end
        end

        z = struct2table(z);

        z.speedup = z.time4./z.time7;

        
        global EXTRA
        if (icash==0)
            iK=find(z.nB==50);
            EXTRA.nonmono{1+end} = ...
                sprintf('For 50 pts, binary mono already %.0f and %.0f faster than simple and brute',...
                z.time4(iK)./z.time7(iK),z.time1(iK)./z.time7(iK));
        else
            EXTRA.nonmono{1+end} = ...
                sprintf('Cash at hand, binary mono time speedup ranges from %.1f to %.1f',...
                min((z.time4./z.time7)),max(z.time4./z.time7));
            EXTRA.nonmono{1+end} = ...
                sprintf('Cash at hand, binary mono eval speedup ranges from %.1f to %.1f',...
                min((z.eval4./z.eval7)),max(z.eval4./z.eval7));
        end

        % Print out the table
        fname = sprintf('tables/runeval%i.tex',icash);
        fid = fopen(fname,'w+');
        for row = 1:size(z,1)

            fprintf(fid,'%10i',z.nB(row));

            % Print out run times
            for m = [1 4 7]
                zval = z.(sprintf('time%i',m))(row);
                tval = z.(sprintf('time%i_est',m))(row);
                cval = z.(sprintf('converged_x%i',m))(row);
                [unit,conv] = getunitconv(zval);
                fprintf(fid,' & ');
                niceprint(fid,zval/conv);
                addasterisk(fid,tval);    
%                 if ~tval
%                     addcross(fid,~cval)
%                 end
                fprintf(fid,' %s',unit);
            end

            % Print out evaluation counts
            for m = [1 4 7]
                zval = z.(sprintf('eval%i',m))(row);
                tval = z.(sprintf('eval%i_est',m))(row);
                fprintf(fid,' & ');
                fprintf(fid,' %10.0f',zval);
                addasterisk(fid,tval);    
            end
            
            % Add speedup
            tval = z.('time4_est')(row) || ...
                   z.('time7_est')(row);
            fprintf(fid,'& %10.1f',z.speedup(row));
            addasterisk(fid,tval);    
            
            fprintf(fid,' \\\\\n');

        end
        fclose(fid);
        
        disp(['The following table was saved to ',fname])
        system(sprintf('cat %s',fname));
        disp(' ')
        
    end


end

function [unit,conv] = getunitconv(x)
    if (x/60<=1) 
        unit = '(s)';
        conv = 1;
    elseif (x/60/60<=1) 
        unit = '(m)';
        conv = 60;
    elseif (x/60/60<=24) 
        unit = '(h)';
        conv = 60*60;
    elseif (x/60/60/24<=365) 
        unit = '(d)';
        conv = 60*60*24;
    else
        unit = '(y)';
        conv = 60*60*24*365;
    end
end

function addasterisk(fid,tf)    
    if tf
        fprintf(fid,'${}^*$');        
    else
        fprintf(fid,'      ');        
    end
end

function [] = niceprint(fid,x)

    if x<=10
        fprintf(fid,'%10.2f',x);
    elseif x<=100
        fprintf(fid,'%10.1f',x);        
    elseif x<=10000
        fprintf(fid,'%10.0f',x); 
    else
        fprintf(fid,'%10.2f',x);        
    end
end

function [] = table_twostate()

    global DATADIR
    
    T = myreadtable([DATADIR 'table_joblist3.csv']);
    T.cnt = T.cnt ./ T.nK;
    
    T.seconds = T.minutes * 60;

    T.quad_rel_max = 100*T.time_quad./T.time_max;

	T = T(:,{'nZ','maxmethod','opt','nK','seconds','cnt','quad_rel_max'});
    T = sortrows(T); 
    
    
    A = T;
    
    T = unstack(T,{'seconds','cnt'},'nK');
    T.incr = T.seconds_x500./T.seconds_x250;

   
    list = {'seconds','cnt'};
    for i = 1:2
        
        tmp = A;        
        tmp = tmp(:,{'nZ','maxmethod','opt','nK',list{i}});
        
        tmp = unstack(tmp,list{i},'maxmethod');
        tmp.([list{i} '_rat']) = tmp.x0./tmp.x3;
        tmp.x0 = [];
        tmp.x3 = [];
        A = join(A,tmp);
        A.([list{i} '_rat'])(A.maxmethod==0) = NaN;
    end
    
    A = sortrows(A,{'nK','nZ','maxmethod','opt'});
    A = A(:,{'nK','nZ','maxmethod','opt','cnt','seconds','cnt_rat','seconds_rat','quad_rel_max'});
    
%     A

    global EXTRA
    EXTRA.twostate{1+end} = ...
        sprintf('Two-state relative to one improvement ranges from %.1f to %.1f',min(A.seconds_rat),max(A.seconds_rat));
    
    
    methodlbl = {'$k$ only','','','$k$ and $z$'};
    optlbl = cell(1,9);
    optlbl(7:9) = {'None','Simple','Binary'};
    fid = fopen('tables/multi_combined.tex','w+');
    for i = 1:size(A,1)
        
        if mod(i-1,6)==0
            fprintf(fid,'& & \\multicolumn{4}{c}{$n_1=n''=%i,\\;n_2=%i$}\\\\',A.nK(i),A.nZ(i));
            fprintf(fid,'\n');
            fprintf(fid,'%s','\cmidrule{3-6}');
            fprintf(fid,'\n');
            fprintf(fid,'%s','Monotonicity & Concavity & Eval & Time &  Eval Speedup & Time Speedup \\');
            fprintf(fid,'%s','\midrule');
            fprintf(fid,'\n');
        end
        
        fprintf(fid,'%-15s & ',methodlbl{1+A.maxmethod(i)});
        fprintf(fid,'%-10s & ',optlbl{A.opt(i)});
        fprintf(fid,'%10.1f &',A.cnt(i));
        fprintf(fid,'%10.2f &',A.seconds(i));
        if (isnan(A.cnt_rat(i)))
            fprintf(fid,'%10s &','$-$');
        else
            fprintf(fid,'%10.1f &',A.cnt_rat(i));
        end
        if (isnan(A.seconds_rat(i)))
            fprintf(fid,'%10s ','$-$');
        else
            fprintf(fid,'%10.1f ',A.seconds_rat(i));
        end
        fprintf(fid,'\\\\\n');
        
        if (mod(i,6)==0)
            fprintf(fid,'\\\\\n');
        end
    end
    fclose(fid);
    !cat tables/multi_combined.tex
    disp('Above table was written to tables/multi_combined.tex')
    
    % Also take the multi and write into separate parts
    fid = fopen('tables/multi_combined.tex','r');
    fid2 = -1;
    i = 0;
    while true
        
        i = i+1;
        l = fgetl(fid);
        if l==-1 
            break
        end
        
        j = mod(i,10)-3;
        if j==1
            if (fid2~=-1) 
                fclose(fid2);
            end
            fid2 = fopen(sprintf('tables/multi_%i.tex',floor(i/10)),'w+');            
        end
        if j>=1 && j<=7
            fprintf(fid2,'%s\n',l);
        end
        
    end
    fclose(fid);
    fclose(fid2);
    

end

function [] = table_contopt()

    global EXTRA DATADIR
    
    T = myreadtable([DATADIR 'table_joblist2.csv']);

    T = sortrows(T,{'maxmethod','opt','nK'});

    T.cnt_per = T.cnt ./ T.nK;
    
    T.maxmethod(T.maxmethod==5) = 0; % 5 is a new addition, so to not change code use 0
        
    T.frac_max = T.time_max ./ T.time_tot;
%     T.frac_max
    
%     T.avgEEE = T.x2 % x2, x3, x4, x5, x6 fall monotonicially
%     T
    
    
    smpl = T.nK==500 & T.opt==4 & T.maxmethod==0;
    tmp = T.frac_max(smpl);
    smpl = T.nK==500 & T.opt==7 & T.maxmethod==0;
    tmp2 = T.frac_max(smpl);
    EXTRA.contopt{1+end} = ...
        sprintf('Max time frac of total for D is %.2f and %.2f for simple and binary',...
        tmp,tmp2);

    smpl = T.nK==500 & T.opt==4 & T.maxmethod==2;
    tmp = T.frac_max(smpl);
    smpl = T.nK==500 & T.opt==7 & T.maxmethod==2;
    tmp2 = T.frac_max(smpl);
    EXTRA.contopt{1+end} = ...
        sprintf('Max time frac of total for CG is %.2f and %.2f for simple and binary',...
        tmp,tmp2);

    
    T = T(:,{'nK','opt','maxmethod','minutes','avgEEE','cnt_per'});
    
%     T
    
    T.time = 60*T.minutes;
    T.minutes = [];
    
    
    T = unstack(T,{'time','avgEEE','cnt_per'},'maxmethod');
    T.time = [T.time_x0,T.time_x2,T.time_x1]; % NOTE: CG is 1, CL is 2, and D is 0, so this ordering is correct
    T.avgEEE = [T.avgEEE_x0,T.avgEEE_x2,T.avgEEE_x1];
    T.cnt_per = [T.cnt_per_x0,T.cnt_per_x2,T.cnt_per_x1];
    T(:,{'time_x0','time_x1','time_x2'}) = [];
    T(:,{'avgEEE_x0','avgEEE_x1','avgEEE_x2'}) = [];
    T(:,{'cnt_per_x0','cnt_per_x1','cnt_per_x2'}) = [];
    
    t = T(T.opt==7,:);
    t2 = T(T.opt==4,:);
    assert(all(t2.nK==t.nK),'Must sort so these have the same order!')
    t.speedup = t2.time./t.time; 
        
    t.binrun = t.time;
    t.time = [];
    t.eeeavg= t.avgEEE;
    t.avgEEE = [];

%     t
    
    t = t(:,{'nK','binrun','speedup','eeeavg','cnt_per'});
%     t
    

    fname = 'tables/contopt.tex';
    fid = fopen(fname,'w+');
    for row = 1:size(t,1)
        fprintf(fid,'%8i ',t.nK(row));
        fprintf(fid,'& %10.2f ',t.binrun(row,:));
        fprintf(fid,'& %10.1f ',t.speedup(row,:));
        fprintf(fid,'& %10.3f ',t.eeeavg(row,:));
%         fprintf(fid,'& %10.2f ',t.spreadcyc(row,:));
        fprintf(fid,'\\\\\n');
    end
    fclose(fid);
    disp(['Saved below table to ',fname])
    system(['cat ',fname]);
    
    fname = 'tables/contopt2.tex';
    fid = fopen(fname,'w+');
    for row = 1:size(t,1)
        fprintf(fid,'%8i ',t.nK(row));
        fprintf(fid,'& %10.2f ',t.binrun(row,:));
        fprintf(fid,'& %10.3f ',t.eeeavg(row,:));
        fprintf(fid,'\\\\\n');
    end
    fclose(fid);
    disp(['Saved below table to ',fname])
    system(['cat ',fname]);    
    
    fname = 'tables/contopt3.tex';
    fid = fopen(fname,'w+');
    for row = 1:size(t,1)
        fprintf(fid,'%8i ',t.nK(row));
        fprintf(fid,'& %10.1f ',t.speedup(row,:));
        fprintf(fid,'& %10.1f ',t.cnt_per(row,:));
        fprintf(fid,'\\\\\n');
    end
    fclose(fid);
    disp(['Saved below table to ',fname])
    system(['cat ',fname]);    
    
    EXTRA.contopt{1+end} = ...
        sprintf('Average cost of CG and CL relative to D is %.1f and %.1f percent respectively',...
        mean(100*(t.binrun(:,3)./t.binrun(:,1)-1)),...
        max(100*(t.binrun(:,2)./t.binrun(:,1)-1))); 
    EXTRA.contopt{1+end} = ...
        sprintf('Average cost of CG relative to CL is %.1f percent, max is %.1f percent',...
        mean(100*(t.binrun(:,3)./t.binrun(:,2)-1)),...
        max(100*(t.binrun(:,3)./t.binrun(:,2)-1))); 
    
end

function [] = table_mono_only()

    global DATADIR 
    
    % Table 2    
    methods = [1;4;7];

    T = myreadtable([DATADIR 'table_joblist1.csv']);
    T = T(T.model==1,:);
    T = sortrows(T,{'opt','nK'});
    T.cnt = T.cnt ./ T.nK;
    T.time = T.minutes;
    T = T(:,{'nK','cnt','time','opt'});
    T = unstack(T,{'cnt','time'},'opt');
    T.cnt = [T.cnt_x1,T.cnt_x4,T.cnt_x7];
    T.time = [T.time_x1,T.time_x4,T.time_x7];    
    T = T(:,{'nK','cnt','time'});
    T.cnt(isnan(T.cnt)) = 0;
    T.time(isnan(T.time)) = 0;

    t = T;
    
    for i = 1:3
        I = t.time(:,i)>0;
        cntest(:,i) = exp(interp1(log(t.nK(I)),log(t.cnt(I,i)),log(t.nK),'linear','extrap'));
        timeest(:,i) = exp(interp1(log(t.nK(I)),log(t.time(I,i)),log(t.nK),'linear','extrap'));
    end
    isest = t.cnt==0;
    nK = t.nK;
    t = table(nK,cntest,timeest,isest);
    
    estimated = 60*t.timeest;
    estimated_cnt = t.cntest;
    isestimated = isest;
    
    % Create a formatted latex table
    fname = 'tables/noconcavecompare.tex';
    fid = fopen(fname,'w+');
    
    for iK = 1:length(nK)
        % Number of grid points
        fprintf(fid,'%10i & ',nK(iK));
        % Run time for each method
        for iM = 1:length(methods) 

            n = estimated(iK,iM);
            if (n>100) 
                fprintf(fid,'%8.0f',n);
            elseif (n>10)
                fprintf(fid,'%8.1f',n);
            else 
                fprintf(fid,'%8.2f',n);
            end        
            
            if (isestimated(iK,iM))
                fprintf(fid,'%s','${}^*$');
            else
%                 fprintf(fid,'%s','\textcolor{white}{${}^*$}');
            end
            fprintf(fid,' & ');
        end
        % Evaluation count for each method
        for iM = 1:length(methods) 

            % Since these are very large numbers, use custom exponent
            % notation: http://www.mathworks.com/matlabcentral/newsreader/view_thread/151859
            
            n = estimated_cnt(iK,iM);
            n_exp= floor(log10(n));
            n_coef = n*10^(-n_exp);
            
            if (n>100)
                fprintf(fid,'%8.0f',n);
            else
                fprintf(fid,'%8.1f',n);
            end
            
            
%             fprintf(fid,'%6.1f(%i)',n_coef,n_exp);
            if (isestimated(iK,iM))
                fprintf(fid,'%s','${}^*$');
            else
%                 fprintf(fid,'%s','\textcolor{white}{${}^*$}');
            end
            fprintf(fid,' & ');
        end
        
        % Ratio of binary mono to simple mono
        tmp = estimated(iK,2)/estimated(iK,3);
        if (tmp<100) 
            fprintf(fid,'%8.1f',tmp); 
        else
            fprintf(fid,'%8.0f',tmp); 
        end
        if (isestimated(iK,2))
            fprintf(fid,'%s','${}^*$');
        end

%         fprintf(fid,' & ',tmp); 
        fprintf(fid,' & '); 
        
        % Ratio of binary mono to simple mono
        tmp = estimated_cnt(iK,2)/estimated_cnt(iK,3);
        if (tmp<100) 
            fprintf(fid,'%8.1f',tmp); 
        else
            fprintf(fid,'%8.0f',tmp); 
        end
        if (isestimated(iK,2))
            fprintf(fid,'%s','${}^*$');
        end        
        
        % New line
        fprintf(fid,'\\\\\n');
    end
    
    fclose(fid);
    
    disp(['Below table written to ',fname])
    system(['cat ' fname]);
%     data = [estimated,estimated_cnt./repmat(nK,[1,length(methods)])]; % runtime for all methods 
    
    global EXTRA
%     iK = find(nK==500);
%     EXTRA.abstract{1+end} = ...
%         sprintf('Binary mono faster than simple by a factor of %.0f for a grid size of 500',estimated(iK,2)/estimated(iK,3));
    iK = find(ismember(nK,[100,500,1000]));
    EXTRA.intro{1+end} = ...
        sprintf('Binary mono faster by %.1f,%.1f,%.1f for grid sizes of 100,500,1000, respectively',estimated(iK,2)./estimated(iK,3));
    iK = [find(nK==100) find(nK==2500)];
    EXTRA.basic{1+end} = ...
        sprintf('Ratio brute to binary at 100, 2500: %.1f, %.0f',estimated(iK,1)./estimated(iK,3));
    EXTRA.basic{1+end} = ...
        sprintf('Ratio simple to binary at 100, 2500: %.1f, %.0f',estimated(iK,2)./estimated(iK,3));
    tmp = nK>=100 & nK<=2500;
    tmp = estimated(tmp,1)./estimated(tmp,2);
    EXTRA.basic{1+end} = ...
        sprintf('Range for ratio of simple to brute: %.1f, %.1f',min(tmp),max(tmp));
        
end

function print_extra()
    
    global EXTRA

    tmp = fieldnames(EXTRA);
    for i=1:length(tmp)
        fprintf(1,'\n========%s===========\n\n',tmp{i});
        fprintf(1,'%s\n',EXTRA.(tmp{i}){:});
    end
    
end

function [] = table_mono_and_conc()

    % Mono and concavity combined
    global DATADIR
    
    T = myreadtable([DATADIR 'table_joblist1.csv']);

    T.cnt = T.cnt ./ T.nK;
    
    T = T(T.model==0 & T.nK>=250 & T.nK<=500,{'model','nK','opt','cnt','minutes'});

    T.mono_opt = NaN(size(T.opt));
    T.mono_opt(T.opt>=1 & T.opt<=3) = 0; % brute mono
    T.mono_opt(T.opt>=4 & T.opt<=6) = 1; % simple mono
    T.mono_opt(T.opt>=7 & T.opt<=9) = 2; % binary mono
    
    T.conc_opt = NaN(size(T.opt));
    T.conc_opt(T.opt==1 | T.opt==4 | T.opt==7) = 0; % brute conc
    T.conc_opt(T.opt==2 | T.opt==5 | T.opt==8) = 1; % simple conc 
    T.conc_opt(T.opt==3 | T.opt==6 | T.opt==9) = 2; % binary conc
    
    T.opt = []; % drop opt
    T.model = [];
    T.time = 60*T.minutes;
    T.minutes = [];

    T = sortrows(T,{'nK','conc_opt','mono_opt'});
    
    T = unstack(T,{'cnt','time'},'nK');
    
    T.timeincr = T.time_x500./T.time_x250;
    
    T = T(:,{'cnt_x250','time_x250','cnt_x500','time_x500','timeincr'});

    data = table2array(T);

    i = 1;
    xlabels{i} = 'None   & None  '; i = i+1;
    xlabels{i} = 'Simple & None  '; i = i+1;
    xlabels{i} = 'Binary & None  '; i = i+1;
    xlabels{i} = 'None   & Simple'; i = i+1;
    xlabels{i} = 'Simple & Simple'; i = i+1;
    xlabels{i} = 'Binary & Simple'; i = i+1;
    xlabels{i} = 'None   & Binary'; i = i+1;
    xlabels{i} = 'Simple & Binary'; i = i+1;
    xlabels{i} = 'Binary & Binary'; 

    fid = fopen('tables/monoconc.tex','w+');
    for r = 1:9
        fprintf(fid,'%s & %10.1f & %10.2f & %10.1f & %10.2f & %10.1f \\\\\n',xlabels{r},data(r,:));
        if (r==3 || r==6)
            fprintf(fid,'%s\n','\midrule');
        end
        
    end
    fclose(fid);
    !cat 'tables/monoconc.tex'
    disp('Above written to tables/monoconc.tex')
    
end

function [y] = pos(x)
    y = x(x>0);
end

function o = plotoptions()
    o = {'LineWidth',2};
end

function [] = figure_timeeval_mono_only()
    
    global DATADIR
    
    f = fontoptions();
    o = plotoptions();

    close all
    mynewfigure();

    T = myreadtable([DATADIR 'table_joblist1.csv']);
    T = T(T.model==1,:);
    T = sortrows(T,{'nK','opt'});
    T = T(:,{'nK','minutes','cnt','opt'});
    
    saveT = T;

    for iter = 1:2
        
        subplot(2,1,iter)
        
        if (iter==1)
            series = 'minutes';
            ttl = 'Convergence time ratios';
            ulim = 230;
        else
            series = 'cnt';
            ttl = 'Evaluation count ratios';
            ulim = 300;
        end
        
        
        T = unstack(saveT,series,'opt');
        

        nK = T.nK(:);

        xx = {nK(~isnan(T.x1)),nK(~isnan(T.x4)),nK(~isnan(T.x7))};
        yy = {T.x1(~isnan(T.x1)),T.x4(~isnan(T.x4)),T.x7(~isnan(T.x7))};

        xx = {xx{3:-1:1}};
        yy = {yy{3:-1:1}};


        % log linearly extrapolate
        X = nK;
        for i = 1:3
            Y(:,i) = exp(interp1(log(xx{i}),log(yy{i}),log(X),'linear','extrap'));
        end

        brutebin = Y(:,3)./Y(:,1);
        simpbin = Y(:,2)./Y(:,1);
        brutesimp = Y(:,3)./Y(:,2);

        fprintf(1,'Brute to binary range: %f to %f\n',min(brutebin(X<=2500)),max(brutebin(X<=2500)))
        fprintf(1,'Simple to binary range: %f to %f\n',min(simpbin(X<=2500)),max(simpbin(X<=2500)))
        fprintf(1,'Brute to simple range: %f to %f\n',min(brutesimp(X<=2500)),max(brutesimp(X<=2500)))

        h = plot(X,brutebin,X,simpbin,X,brutesimp,o{:}); 
        mydefaultplotformat(h);
        title(ttl,f{:}); 
        
        if (iter==1) 
            h = legend('Brute force / binary','Simple / binary','Brute force / simple','location','northwest'); 
            set(h,f{:});
%             set(h,'EdgeColor',[1 1 1]*0);
        else
            xlabel('Grid size ($n$)',f{:});
        end
        
        ylabel('Speedup',f{:})

        axis([-Inf,2500,-Inf,ulim])

        grid on

    %     set(gca,'XTick',nK(nK<=5000))
        set(gca,'XTick',unique([nK;(250:250:2500)']))
        set(gca,'TickLabelInterpreter','latex')
    %     return
    %       set(gca,'XTickLabel',{'100','2500','10000','25000','50000','100000'})
    %     
    %     set(gca,'YTick',0:5:20)

    end
    
    printfig('runtimeevalnoconcave')

end


function [] = figure_empirical_On()
    
    % O(n) properties
    global DATADIR
    
    T = myreadtable([DATADIR 'table_joblist1.csv']);
    T = T(T.model==0,:);
    T = T(ismember(T.opt,[7 5 8 9]),:);
    T = sortrows(T,{'nK','cnt'});
    T.On = T.cnt./T.nK;
    T = T(:,{'nK','On','opt'});
    T = unstack(T,'On','opt');

    close all
%     mynewfigure(struct('aspect',.6));
    mynewfigure(struct('aspect',.55,'width',7.5));
    
    f = fontoptions();
    o = plotoptions();
    
    hold on
    p(1) = plot(T.nK,T.x5,'-o',o{:}); 
    p(2) = plot(T.nK,T.x7,'-',o{:}); 
    p(3) = plot(T.nK,T.x8,'--+',o{:}); 
    p(4) = plot(T.nK,T.x9,':*',o{:}); 
    hold off
        
    mydefaultplotformat(p)
    
    title('Average evaluation count over $n$',f{:}); 
    h = legend('Simple monotonicity, simple concavity',...
               'Binary monotonicity, no concavity',...
               'Binary monotonicity, simple concavity',...
               'Binary monotonicity, binary concavity','location','best');
    set(h,f{:},'Position',[0.504344921112059 0.613303570974442 0.328095872061593 0.150238096146357]);        
    legend('boxoff')
    xlabel('Grid size ($n$)',f{:});    
    ylabel('Evaluations',f{:})    

    set(gca,'XTick',unique([1000,0:2500:10000]))
    set(gca,'TickLabelInterpreter','latex')

    fprintf(1,'Eval per state for bin-bin %6.1f\n',T.x9(T.nK==10000));
    fprintf(1,'Eval per state for sim-sim %6.1f\n',T.x5(T.nK==10000));
    
    printfig('Onbehavior')
    
end

function printfig(name)

    print('-dpdf',['pics/' name],'-fillpage')
    
end


% function [] = figure_time_mono_only()
%     
%     global DATADIR
%     
%     T = myreadtable([DATADIR 'table_joblist1.csv']);
%     T = T(T.model==1,:);
%     T = sortrows(T,{'nK','opt'});
%     T = T(:,{'nK','minutes','opt'});
%     T = unstack(T,'minutes','opt');
%     
%     f = fontoptions();
%     o = plotoptions();
% 
%     close all
%     mynewfigure();
%     
%     nK = T.nK(:);
%     
%     xx = {nK(~isnan(T.x1)),nK(~isnan(T.x4)),nK(~isnan(T.x7))};
%     yy = {T.x1(~isnan(T.x1)),T.x4(~isnan(T.x4)),T.x7(~isnan(T.x7))};
%     
%     xx = {xx{3:-1:1}};
%     yy = {yy{3:-1:1}};
%     
%     
%     % log linearly extrapolate
%     X = nK;
%     for i = 1:3
%         Y(:,i) = exp(interp1(log(xx{i}),log(yy{i}),log(X),'linear','extrap'));
%     end
%     
%     brutebin = Y(:,3)./Y(:,1);
%     simpbin = Y(:,2)./Y(:,1);
%     brutesimp = Y(:,3)./Y(:,2);
%     
%     fprintf(1,'Brute to binary range: %f to %f\n',min(brutebin(X<=2500)),max(brutebin(X<=2500)))
%     fprintf(1,'Simple to binary range: %f to %f\n',min(simpbin(X<=2500)),max(simpbin(X<=2500)))
%     fprintf(1,'Brute to simple range: %f to %f\n',min(brutesimp(X<=2500)),max(brutesimp(X<=2500)))
%     
%     h = plot(X,brutebin,X,simpbin,X,brutesimp,o{:}); 
%     mydefaultplotformat(h);
%     title('Convergence time ratios',f{:}); 
%     h = legend('Brute force / binary','Simple / binary','Brute force / simple','location','northwest'); 
%     set(h,f{:});
% %     legend('boxoff')
%     set(h,'EdgeColor',[1 1 1]);
%     
%     xlabel('Grid size ($n$)',f{:});
%     ylabel('Speedup',f{:})
%     
%     axis([-Inf,2500,-Inf,230])
%     
%     grid on
%     
% %     set(gca,'XTick',nK(nK<=5000))
%     set(gca,'XTick',unique([nK;(250:250:2500)']))
%     set(gca,'TickLabelInterpreter','latex')
% %     return
% %       set(gca,'XTickLabel',{'100','2500','10000','25000','50000','100000'})
% %     
% %     set(gca,'YTick',0:5:20)
%     
%     printfig('runtimenoconcave')
% 
% end

