% Produces a job list that can be run on the supercomputer via PCP
function [] = createjoblist()

fmt = @(model,nK,nZ,opt,maxTime,shocktype,maxtech,multigrid) sprintf('%01i %09i %04i %02i %01i %01i %06i %01i',model,nK,nZ,opt,shocktype,maxtech,maxTime,multigrid); 

for ilist = 1:2

    cnt = 0;
    runtime = 0;

    if (ilist==1) 

        % All discrete joblist (Section 4) 
        fname = '../fortran/joblist1.txt';

        model = [0,1];
        nK = [100;250;500;1000;2500;5000;10000;25000;50000;100000];
        shocktypes = 0; % 0 means discrete shocks, 1 is continuous
        maxtech = 0; % 0 for grid search, 1 for continuous, 2 for GEGM
        multigrids = 0; % number of multigrids (not used currently except in testing)
        nZ = 21;
        maxTime = 60; % one hour max

        model_opts = {(1:9)';[1;4;7]};

    else
        
        % Continuous techniques (Section 5) 
        fname = '../fortran/joblist2.txt';

        model = 1; % only Arellano model
        nK = [100;250;500;1000;2500];
        shocktypes = 0; % 0 means discrete shocks, 1 is continuous
        maxtech = [0;1;2]; % 0 for grid search, 1 for continuous, 2 for continuous alternate (was GEGM)
        multigrids = 0; % number of multigrids (not used currently except in testing)
        nZ = 32; % 32 is used in estimation, 100 was used before
        maxTime = 60*30; % let her run
        model_opts = {[4;7]};

    end

    fid = fopen(fname,'w');

    for iM = 1:length(model)

        opt = model_opts{iM};

        for iS = 1:length(shocktypes)
            for iT = 1:length(maxtech)
                for iK = 1:length(nK)
                    for iopt = 1:length(opt)
                        str = fmt(model(iM),nK(iK),nZ,opt(iopt),maxTime,shocktypes(iS),maxtech(iT),multigrids);
                        fprintf(fid,'%s\n',str);
                        cnt = cnt + 1;
                        runtime = runtime + maxTime;
                    end
                end
            end
        end 
    end

    fclose(fid);

    fprintf(1,['Created a job list of %i jobs without errors. The maximum runtime \n'...
              '(without parallelization) should be roughly %i minutes. On a 32 processor\n'...
              'node with perfect speedup, the max is %i minutes.\n'...
              'See "%s."\n'],cnt,runtime,ceil(runtime/32),fname);
    disp('')

end

end













% % For tables
% model = 0;
% nK = [100;200;500;1000]; % for the graph
% nZ = 21;
% maxTime = 60;
% opt = [1;2;3;4;5;6;7;8;9];
% 
% for iK = 1:length(nK)
%     for iopt = 1:length(opt)
%         str = fmt(model,nK(iK),nZ,opt(iopt),maxTime);
%         fprintf(fid,'%s\n',str);
%         cnt = cnt + 1;
%         runtime = runtime + maxTime;
%     end
% end
