% Computes the integral of f over a,b using adaptive quadrature 
function [E xhist] = adaptQuad(f,a,b)

    % Fun test function
    f = @(x) -100.*(x<1) + 2*x.^2.*(x>=1); 
    a = -10; b= 10;
    disp(['Matlab gives'])
    quad(f,a,b)

    % Setup initial values
    fa = f(a);
    fb = f(b);
    
    tol = 1d-3; % Absolute error over the whole interval, roughly

    % Choose some initial level to guarantee at least that level of
    % accuracy
    n = 5;
    x = linspace(a,b,n)';
    fx = [fa;f(x(2:n-1));fb];
    E = 0;
    xhist = [];
    for i = 1:n-1
        E_on_interval = integrationRule(x(i),x(i+1),fx(i),fx(i+1),f);
        [Etmp xhisttmp] = adaptquad_r(f,x(i),x(i+1),fx(i),fx(i+1),E_on_interval,tol);
        E = E + Etmp;
        xhist = [xhist;x(i);xhisttmp];
    end
    xhist = [xhist;x(n)];
    
%     % Call the recursive routine
%     [E xhist] = adaptquad_r(f,a,b,fa,fb,E0);
    
end

function [E] = integrationRule(a,b,fa,fb,f)
    
    % Use trapezoid rule
    E = (b-a)/2*(fb+fa);
    % Use midpoint rule: This is inefficient! Should cache the midpoint
    % function value.
%     m = (b+a)/2;
%     fm = f(m);
%     E = (b-a)*f(m);
    
end

function [E xhist] = adaptquad_r(f,a,b,fa,fb,E0,tol)

    % Compute an update to the expectation
    m = (b+a)/2;
    fm = f(m);
    El = integrationRule(a,m,fa,fm,f);
    Eh = integrationRule(m,b,fm,fb,f);
    xhist = m;
    
    if (abs(-E0 + El + Eh)<tol)
        E = El + Eh;
    else
        [Elnew xlhist] = adaptquad_r(f,a,m,fa,fm,El,tol/2);
        [Ehnew xhhist] = adaptquad_r(f,m,b,fm,fb,Eh,tol/2);
        E = Elnew + Ehnew;
        xhist = [xlhist;m;xhhist];
        
    end    
    
end