% []=autocorr(x,F,mu)
% Computes the autocorrelation of some discrete process with grid x,
% transition matrix F, and initial distribution mu.
function [auto]=autocorr(x,F,mu)

    nx = numel(x);
    
    % Check that mu is a distribution
    if (abs(sum(mu)-1.d0)>1d-12) 
        error('mu is not a distribution, it doesnt sum to 1')
    end
    if (any(mu<0))  
        error('mu is not a distribution, it has negative elements')
    end
    % Check that F is a transition matrix
    if (any(abs(sum(F,2)-1.d0)>1d-12))
        error('F is not a transition matrix')
    end
    
    
    % Compute mean today
    Ex = x'*mu;
    
    % Compute stdev today
    sigx = sqrt(mu'*(x-Ex).^2);
    
    % Compute dsn tomorrow
    mup =(mu'*F)';
    
    % Compute mean tomorrow
    Exp = x'*mup;
    
    % Compute stdev tomorrow
    sigxp = sqrt(mup'*(x-Exp).^2);

    % Compute autocorrelation
    tmp = NaN(nx,1);
    for i = 1:nx
        tmp(i) = (x(i)-Ex)*(F(i,:)*(x-Exp));
    end
    auto = mu'*tmp/(sigx*sigxp);

    
end