function [xstar fevals] = brent(f,a,b,pmod)

toler = 1d-6;
xtoler = 1d-6;
maxiter = 1000;

sqrteps = sqrt(eps(0));

% I follow the algorithm given in wikipedia
ya = f(a);
yb = f(b);
fevals = 2;

% Check whether should stop
if (abs(ya)<toler)
    xstar = a;
    return
elseif (abs(yb)<toler)
    xstar = b;
    return
elseif (sign(ya)==sign(yb))
    fprintf(1,'ya %f yb %f \n',ya,yb);
    error('Root is not bracketed')
end

% b will be the best point. So ensure this.
if (abs(ya)<abs(yb))
    tmp = b;
    tmp2 = yb;
    b = a;
    yb = ya;
    
    a = tmp;
    ya = tmp2;        
end

% Set c = a and fc = fa;
c = a;
yc = ya;


% Main loop
mflag = 1; 
step = 'Init';
for iter = 1:maxiter
    
    % Precompute some things
    dab = ya-yb;
    dbc = yb-yc;
    dac = ya-yc;

    % Report progress
    if (mod(iter,pmod)==0)
        fprintf(1,'iter %i b %f yb %f abs(b-a) %f fevals %i step %s \n',iter,b,yb,abs(dab),fevals,step);
    end
    
    % Check the stopping criterion
    if abs(b-a)<xtoler
        xstar = (b+a)/2;
        return
    end
        
    % Compute either IQI or secant rule potential update
    tiny = eps(0);
    if (abs(dbc)<tiny || abs(dac)<tiny)
        % Do secant. Note: dab should always be greater than tiny.
        s = b + yb*dab/(yb-ya);
        step = 'Secant';
    else
        % Do IQI
        s = a*yb*yc/dab/dac - b*ya*yc/dab/dbc + c*ya*yb/dac/dbc;
        step = 'IQI';
    end
    
    % Check whether to do bisection or not
    tmp = (3*a+b)/4;
    cdn1 = (s<min(tmp,b) || s>max(tmp,b));
    cdn2 = (mflag && abs(s-b)>=abs(dbc)/2);
    cdn4 = mflag && abs(dbc)<sqrteps ;
    if (iter>1) 
        cdn3 = ~mflag && abs(s-b)>=abs(c-d)/2;
        cdn5 = ~mflag && abs(c-d)<sqrteps ;
    else
        cdn3 = 0;
        cdn5 = 0;
    end
    
    
    if (cdn1 || cdn2 || cdn3 || cdn4 || cdn5) 
        s = (a+b)/2;
        mflag = 1; % Previous iteration was bisection
        step = 'Bisect';
    else
        mflag = 0; % Previous iter was not bisection
    end
    
    % Evaluate at s and check for convergence
    ys = f(s);
    if (abs(ys)<toler) 
        xstar = s;
        return
    end
    fevals = fevals + 1;
    
    % Create new triple
    d = c;
    c = b;    
    if (sign(ya)==sign(ys))
        a = s; % s replaces a in bracket
        ya = ys;
    else
        b = s; % s replaces b in bracket
        yb = ys;
    end
    
end


end