/*=================================================================
 *
 * bsearch.c
 *
 * The calling syntax is:
 *
 *		[i] = bsearch(xhat,x)
 *
 * Grey Gordon 2012 (c)
 * There is no warranty at all with this software.
 * This code was created using the Matlab yprime.c as an example  
 *=================================================================*/
/*#include <math.h>*/
#include "mex.h"

/* Input Arguments */

#define	XHAT_IN	prhs[0]
#define X_IN    prhs[1]

/* Output Arguments */

#define	I_OUT   plhs[0]

#if !defined(MAX)
#define	MAX(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define	MIN(A, B)	((A) < (B) ? (A) : (B))
#endif


static void sub_bsearch(
		   double       i[],
 		   double	xhat[],
                   double       x[],
		   int nx,
                   int nxhat)
{
    int ihat, ilo, ihi, mid;

    for (ihat=0;ihat<nxhat;ihat++){
        if (nx<=1) {
            i[ihat] = 1;
            continue;
        }
        if (xhat[ihat]<x[0]){
            i[ihat] = 1;
        } else if (xhat[ihat]>=x[nx-1]){
            i[ihat] = nx-1;
        } else {
            ilo = 0;
            ihi = nx-1;
            mid = (ilo+ihi)/2;
            while (mid>ilo) {
                if (xhat[ihat]>x[mid]) {
                    ilo = mid;
                } else if (xhat[ihat]<x[mid]) {
                    ihi = mid;
                } else {
                    ilo = mid;
                    i[ihat] = ilo+1;
                    continue;
                }
                mid = (ilo+ihi)/2;
            }
            /*i[ihat] = (double)ilo+1;*/
            i[ihat] = (double)ilo+1;
        }
        
        /*printf("hi i[ihat] %f ilo %i \n",i[ihat],ilo);*/
    }

    return;

}



void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
     
{ 
    double *x,*xhat; 
    double *i; 
    mwSize m,n,nxhat,nx; 
    
    /* Check for proper number of arguments */
    if (nrhs != 2) { 
	mexErrMsgTxt("Two input arguments required."); 
    } else if (nlhs > 1) {
	mexErrMsgTxt("Too many output arguments."); 
    } 
    
    /* Check the dimensions of Y.  Y must be a vector */
    m = mxGetM(X_IN); 
    n = mxGetN(X_IN);
    if (m != 1 && n != 1) { 
	mexErrMsgTxt("x must be a vector."); 
    } 
    nx = m*n;
    
    m = mxGetM(XHAT_IN); 
    n = mxGetN(XHAT_IN);
    if (m != 1 && n != 1) { 
	mexErrMsgTxt("xhat must be a scalar or vector."); 
    } 
    nxhat = m*n;

    /* Create a matrix for the return argument */ 
    I_OUT = mxCreateDoubleMatrix(m, n, mxREAL); 
    
    /* Assign pointers to the various parameters */ 
    i = mxGetPr(I_OUT);
    xhat = mxGetPr(XHAT_IN); 
    x = mxGetPr(X_IN);
        
    /* Do the actual computations in a subroutine */
    sub_bsearch(i,xhat,x,nx,nxhat); 
    return;
    
}


