% [Yp Ypp] = cubicSplineDeriv(X,x,abcd)
% Compute the first and second derivative of the spline given implicitly
% by x,abcd at X.
function [Yp Ypp] = cubicSplineDeriv(X,x,abcd)

% Ensure column vectors
if (~iscolumn(X)) 
    error('X must be a column vector')
end

% Use binary search to find the index the lower brackets X in grid x.
i = bsearch(X,x);

% Calculate distance from left node
del = X-x(i);

% Evaluate using Horner's method
Yp = abcd(i,2) + (2*abcd(i,3) + 3*abcd(i,4).*del).*del;
Ypp = 2*abcd(i,3) + 6*abcd(i,4).*del;

end