% [] = export_fortran_csv(file,A)
% 
% Exports an array A to a csv file 'file.csv' (note: places directory in front of file)
% 
% To export A as something other than double precision, one should first
% convert it to that type. The type is detected automatically.
% 
% Note: Currently, this code will fail if a non-number appears in A, such
% as "infinity" or "NaN".
% 
% Grey Gordon, 2014 
function [] = export_fortran_csv(file,A)

    % See if .csv is at end. If so, remove it
    if (length(file)>4) 
        if strcmp(file(end-3:end),'.csv')
            file = file(1:end-4); % drop last four 
        end
    end

    fid = fopen([file,'.csv'],'w');
    
    inds = find(file(:)=='/'|file(:)=='\');
    if isempty(inds)
        varname = file;
    else
        inds = inds(end);
        varname = file(inds+1:end);
    end
    
    % Print the he header info
    fprintf(fid,'Variable name:,%s\n',varname);
    if (isfloat(A)) 
        fprintf(fid,'Variable type:,%s,%i,%s\n','real(8)',0,'(e25.17e3)');
    elseif (isinteger(A))
        fprintf(fid,'Variable type:,%s,%i,%s\n','integer',1,'(i12)');
    elseif (islogical(A))
        fprintf(fid,'Variable type:,%s,%i,%s\n','logical',2,'(i1)');        
    end
    
    % Ensure column vector if vector
    if (isvector(A))
        A = A(:); 
    end
    
    n = ndims(A); % always >= 2
    if (n==2) 
        if (size(A,2)==1); n = 1; end
        if (size(A,1)==1); n = 0; end
    end
    fprintf(fid,'Variable rank:,%i\n',n);
    tmp = size(A);
    fprintf(fid,'Variable dims:');
    if (n>0) 
        fprintf(fid,',%i',tmp(1:n));
    else
        fprintf(fid,',%i',1);
    end
    fprintf(fid,'\n'); % end of row 4
    for i = 5:9
        fprintf(fid,'\n'); % rows 5-9
    end
    
    % Now print the data
    A = reshape(A,[size(A,1),size(A,2),numel(A)/(size(A,1).*size(A,2))]);
    if (isfloat(A)) 
        fmt = '%25.17E';
    elseif (isinteger(A))
        fmt = '%12i';
    elseif (islogical(A))
        fmt = '%1i';
    end
    
    vecfmt = '';
    for j = 1:size(A,2)-1;
        vecfmt = [vecfmt,fmt,','];
    end
    vecfmt = [vecfmt,fmt];
        
    for k = 1:size(A,3)
        % print out matrix conditional on k
        for i = 1:size(A,1)
            fprintf(fid,[vecfmt,'\n'],A(i,:,k));
        end        
    end
    
    fclose(fid);

    
end
