% [S] = import_fortran(folder)
% 
% Imports all the csv files in a folder and returns a structure having all 
% the variables. 
%
% Note: The fields can easily be saved separately to file via 
% "save('import.mat','-struct','S') where S is the structure returned 
% from import_fortran_csv. Then loading import.mat will have all the
% variables separately in the workspace rather than as a structure.
% 
% Grey Gordon, 2014
function [S] = import_fortran(folder,varargin)
  
    if (folder(end)~='/')
        folder = [folder '/'];
    end

    D = dir([folder,'*.csv']);
    
    maxMB = Inf;
    if (nargin>=2)
        maxMB = varargin{1};
    end

    if (length(D)==0) 
      D
      warning(['No .csv files inside ',folder])
    end
    
    for i = 1:length(D)
        stem = D(i).name;
        stem = stem(1:end-4);

        try 
            S.(stem) = import_fortran_csv([folder,D(i).name],maxMB);
        catch me
            disp(['Failed to import ' D(i).name])
            rethrow(me);
        end
    end

end
