function [] = metropolis()
    
    close all

    nx = 2;

    % Function to be maximized (here, making it noisy)
    % Note: must be positive! So, may need to take transformation, like exp
%     f = @(x) normpdf(x(1),10,1)*normpdf(x(2),-10,1); % product of iid normal
    f = @(x) exp(10*(-sum((x-1).^2)));  % + .01*rand()
    f = @(x) exp(-((1-x(1))^2 + 100*(x(2)-x(1)^2)^2)); % exp(-Rosenbrock)
    
    dobayesian = 1;
    
    if (dobayesian)
        g = f;
        
        prior_ind{1} = @(x1) normpdf(x1,2,.1);
        prior_ind{2} = @(x2) normpdf(x2,2,.1);
        prior = @(x) prior_ind{1}(x(1))*prior_ind{2}(x(2));
        
        f = @(x) prior(x)*g(x);
    end
    
    nburn = 1e4;
    niter = 3e5;
    pmod = 1e5;
    
    nhist = 1e5;
    histx = NaN(nx,nhist);
    histf = NaN(nhist,1);
    
    % Initial state
    x0 = zeros(nx,1);
    f0 = f(x0);
    
    % Proposal distribution
    VAR = eye(nx);
    CHOL = chol(VAR)'; % note: transpose
    propose = @(x,alpha) x + alpha*CHOL*randn(nx,1); % Random-walk proposal;
    
    alpha = 1;
    accrate = 0; % acceptance rate
    for iter = 1:niter
        
        % Draw a proposal
        xprop = propose(x0,alpha);
        fprop = f(xprop);
        
        % Determine whether or accept or reject
        pr_accept = min(1,fprop/f0);
        accept = rand()<=pr_accept;
        
        % If accept, replace x0
        if (accept)
            x0 = xprop;
            f0 = fprop;
        end
        
        % Update the proposal distribution to improve acceptance rate
        accrate = (accept + accrate*(iter-1))/iter;
        if (iter<nburn)
            if (accrate<.25) 
                alpha = .99*alpha;
            elseif (accrate>.35) 
                alpha = 1.01*alpha;
            end
        end
                
        if (mod(iter,pmod)==0)
            fprintf(1,'iter %i f0 %f ',iter,f0);
            fprintf(1,'x0');
            printvec(x0);
            fprintf(1,'accept rate %f ',accrate);
            fprintf(1,'alpha %f\n',alpha);
            figure(1)
            hold on 
%             view(3)
%             plot3(x0(1),x0(2),f0,'o')
            plot(x0(1),x0(2),'.')
            hold off
            drawnow
        end
        
        histx(:,1+mod(iter,nhist)) = x0;
        histf(1+mod(iter,nhist)) = f0;
        
        if (mod(iter,nhist)==0) 
    
            figure(2)
            histogram2(histx(1,:),histx(2,:))
            drawnow

            % Report mean, stdev, mode, and median
            fprintf(1,'Mean x0 last %i states: ',nhist);
            printvec(mean(histx,2));
            fprintf(1,'\n');
            
            fprintf(1,'Stdev x0 last %i states: ',nhist);
            printvec(std(histx,0,2));
            fprintf(1,'\n');

            [cnts,x1bins,x2bins] = histcounts2(histx(1,:),histx(2,:));
            [~,imode] = max(cnts(:));
            [imode1,imode2] = ind2sub(size(cnts),imode);
            fprintf(1,'Mode last %i states: ',nhist);
            printvec([mean(x1bins(imode1:imode1+1)),mean(x2bins(imode2:imode2+1))]);
            fprintf(1,'\n');
            
            fprintf(1,'Median last %i states: ',nhist);
            printvec([median(histx(1,:)),median(histx(2,:))]);
            fprintf(1,'\n');
            
        end
        
        
%         if (mod(iter,100000)==0)
%             close all
%         end
        
    end
    
    if (dobayesian) 
        figure(3)
        for i = 1:nx
            subplot(1,nx,i);
            [cnts,bins] = histcounts(histx(i,:));
            tmp = .5*(bins(2:end)+bins(1:end-1));
            tmp2 = linspace(1.5,2.5,1000)';
            plotyy(tmp,cnts,tmp2,prior_ind{i}(tmp2))
        end
        
    end
    
end

function [] = printvec(x)
    for i = 1:length(x(:))
        fprintf(1,' %f ',x(i));
    end
end