function [] = mygnuplot_plot_core(x1,y1,x2,y2,opt)
    % opt should be scalar struct giving
    %
    % opt.title 
    % opt.xlabel
    % opt.ylabel
    % opt.y2label
    % opt.xaxis : length 2 vector giving lower and upper bound of x domain
    % opt.yaxis : length 2 vector giving lower and upper bound of y domain
    % opt.legend : cell array of legend entries
    % opt.legendcmd : gnuplot command such as set key at 1.2,5
    % opt.linepattern : 
    % opt.saveas : name to save as (must end in .pdf)
    % opt.isplotyy : whether plotyy or not
    % opt.extracmd : cell array of additional gnuplot commands that should be added
    %

    isplotyy = 0;
    if (isfield(opt,'isplotyy')) 
        isplotyy = opt.isplotyy;
    end

    if (isoctave()) 
        f1 = fopen('/tmp/data1.dat','w+');
        dlmwrite(f1,[x1 y1],' ',0,0);
        fclose(f1);
    else
        dlmwrite('/tmp/data1.dat',[x1 y1],' ',0,0);
    end
    
    if isplotyy
        f2 = fopen('/tmp/data2.dat','w+');
        dlmwrite(f2,[x2 y2],' ',0,0);
        fclose(f2);
    end

    fid = fopen('/tmp/gnuplot.cfg','w+');

    fprintf(fid,'set terminal postscript eps colour enhanced "Helvetica" 14\n');
    % fprintf(fid,'set terminal epscairo transparent enhanced fontscale 0.5 size 5.00in,4.00in\n');
    fprintf(fid,'set output "%s"\n','/tmp/plot.eps');
    fprintf(fid,'set border linewidth 1\n');

    if isfield(opt,'xaxis') 
        fprintf(fid,'set xrange [%f : %f]\n',opt.xaxis(1),opt.xaxis(2));
    else
        fprintf(fid,'set xrange [%f : %f]\n',min(x1),max(x1));
    end

    fprintf(fid,'set ytics nomirror \n');

    if isfield(opt,'yaxis')
        fprintf(fid,'set yrange [%f : %f]\n',opt.yaxis(1),opt.yaxis(2));
    else
        fprintf(fid,'set autoscale y\n');
    end 
    if (isfield(opt,'title')); fprintf(fid,'set title "%s"\n',opt.title); end
    if (isfield(opt,'xlabel')); fprintf(fid,'set xlabel "%s"\n',opt.xlabel); end
    if (isfield(opt,'ylabel')); fprintf(fid,'set ylabel "%s"\n',opt.ylabel); end
    if (isfield(opt,'legendcmd')); fprintf(fid,'%s\n',opt.legendcmd); end
    
    linepattern = mygnuplot_defaultlinepattern();
    if (isfield(opt,'linepattern'))
        linepattern = opt.linepattern;
    end 

    ny1 = size(y1,2);

    % plotyy options
    if isplotyy
        fprintf(fid,'set y2tics nomirror \n');
        fprintf(fid,'set autoscale y2\n');
        if (isfield(opt,'y2label')); fprintf(fid,'set y2label "%s"\n',opt.y2label); end
        ny2 = size(y2,2);
    else
        ny2 = 0;
    end

    % legend options
    if (isfield(opt,'legend')) 
        legend = opt.legend;
        fprintf(fid,'%s\n','set key on' );
    else
        for i = 1:ny1+ny2
            legend{i} = 'x';
        end 
        fprintf(fid,'%s\n','set key off' );
    end 

    if isfield(opt,'extracmd') 
        for i = 1:numel(opt.extracmd)
            fprintf(fid,'%s\n',opt.extracmd);
        end
    end 

    % Commands for drawing
    fprintf(fid,'%s\n','plot \');
    mygnuplot_datarule(fid,{legend{1:ny1}},'/tmp/data1.dat','x1y1',linepattern,1);
    if isplotyy
        mygnuplot_datarule(fid,{legend{ny1+1:ny1+ny2}},'/tmp/data2.dat','x1y2',linepattern,ny1+1);
    end

    % Stop editing
    fclose(fid);

    % Do the actual drawing
    delete('/tmp/plot.eps')
    if isfield(opt,'saveas')
        system(sprintf('gnuplot /tmp/gnuplot.cfg && epstopdf /tmp/plot.eps --outfile=%s && open %s',opt.saveas,opt.saveas));
    else
        system('gnuplot /tmp/gnuplot.cfg && epstopdf /tmp/plot.eps --outfile=/tmp/plot.pdf && open /tmp/plot.pdf');
    end
    
% 
%     if isfield(opt,'saveas')
%         system(sprintf('gnuplot /tmp/gnuplot.cfg && open %s',opt.saveas));
%         disp(sprintf('File %s written to file',opt.saveas));
%     else
%         delete('/tmp/plot.eps')
%         system(sprintf('gnuplot /tmp/gnuplot.cfg && open %s','/tmp/plot.eps'));
%     end

end 

function [] = mygnuplot_datarule(fid,leg,datname,axesname,lineopts,linejstart)

    linej = linejstart;

    for i = 1:numel(leg)

        linej = 1+mod(linej-1,length(lineopts));

        lo = lineopts(linej);
        % lo
        
        if strcmp(lo.dashtype,'none')
            fprintf(fid,'"%s" using 1:%i axes %s title "%s" linecolor "%s" linewidth %f pointtype %i pointsize %f with %s, \\\n',datname,i+1,axesname,leg{i},lo.rgb,lo.linewidth,lo.markertype,lo.markersize,lo.linestyle);
        else
            fprintf(fid,'"%s" using 1:%i axes %s title "%s" linecolor "%s" linewidth %f pointtype %i pointsize %f dashtype "%s" with %s, \\\n',datname,i+1,axesname,leg{i},lo.rgb,lo.linewidth,lo.markertype,lo.markersize,lo.dashtype,lo.linestyle);
        end
        
        linej = linej+1;

    end 
    % fprintf(fid,'%s\n','"dataex.dat" using 1:5 title 'Data 4' with linespoints');
end
