#include <mex.h>
#include <stdio.h>
#include <errno.h>
#include "matrix.h"

mexFunction(int nlhs, mxArray *plhs[ ],
int nrhs, const mxArray *prhs[ ]) { 

  double *d;
  int i,n;
  int strlen;
  mwSize buflen;

  FILE *my_stream;
  char *my_filename;

  /* Check for proper number of input and output arguments */
  if (nrhs != 2) { 
    mexErrMsgTxt("Two input arguments required.");
  } 
 
  /* Check for proper input type */
  if (!mxIsChar(prhs[0]))  {
    mexErrMsgTxt("First input argument must be a string.");
  }
  if (!mxIsScalar(prhs[1]))  {
    mexErrMsgTxt("Second input argument must be a scalar.");
  }

  /* Get number of elts to be imported */
  n = (mwSize)mxGetScalar(prhs[1]);

  /* Create an mxArray for the output data */
  plhs[0] = mxCreateDoubleMatrix(n, 1, mxREAL);
  d = mxGetPr(plhs[0]);

  /* Get file name */
  strlen = mxGetN(prhs[0]); /* get size*/
  buflen = strlen*sizeof(mxChar)+1; /* get necessary buffer size */
  my_filename = mxMalloc(buflen); /* allocate space for array */
  mxGetString(prhs[0],my_filename,1+strlen); /* copy filename to string */
  
  /* printf("Importing file %s with %i elements \n",my_filename,n); */

  my_stream = fopen(my_filename, "r");

  if (my_stream == NULL) {
    mexErrMsgTxt("Failed to open file.");
    return;
  }

  for(i=0;i<n;i++) {
    fscanf (my_stream, "%lf",&d[i]);
  }

  /* Close stream */
  fclose (my_stream);

return;

}

