module mod_common_var

use mod_types
implicit none

!!!!!!!!!!!!!!!!!!
! Determine whether using Aiyagari or bankruptcy.
logical, parameter :: is_Aiyagari = .FALSE.
logical, parameter :: is_MeansTested = .FALSE.
logical, parameter :: is_Commit = .FALSE.
    !Note on the debt limit: w*emin*1/(1-rho*qbar)
    !The steady state natural limit is ~1.2148 = 1.164*.0502*1/(1-.9519)
    !The sequential limit obviously varies, but worst observed so far would be ~ 1.1786 = 1.1387*.0502*1/(1-.9515)
    !I set the limit tighter than both of these to ensure that if there's forecast error, that no one gets negative consumption
real(dp), parameter :: Aiyagari_debt_limit_SS = -1.1d0
real(dp), dimension(1:2,1:2), parameter :: Aiyagari_debt_limit = Aiyagari_debt_limit_SS !reshape((/.2d0,.0d0,.2d0,0.d0/),(/2,2/))
real(dp) :: e_mean = 1.d0 !used for means testing

!!!!!!!!!!!!!!!!!!!!!!!
! Lengths
!
integer, parameter :: len_z = 2 ! 2 = #{z=g,z=b}
integer, parameter :: len_h = 2 ! 2 = #{h=0,h=1}
integer, parameter :: len_N = 1 ! # of N points.  If F_e'|e,z,z' = F_e'|e, only need one
integer, parameter :: len_A1 = 1 !2
integer, parameter :: len_A2 = 3 !3
integer, parameter :: len_A3 = 6
integer, parameter :: len_m = len_z*len_N*len_A1*len_A2*len_A3
integer, parameter :: len_T = 150 ! # of periods to simulate
integer, parameter :: len_t0 = 50  ! # of period to throwout (only use len_t0+1...T)
integer :: len_kpp = 1250 !750 !1250 ! # of positive grid points used when calculating distributions (len_k = len_kpp+len_Amm+1)

!!!!!!!!!!!!!!!!!!!!!!!!
! Estimated parameters
!
real(dp), parameter :: discount = .917354d0*.975d0           !CCNR have effective discount rate of .8902
real(dp), parameter :: chi        = .0189d0                    !Percent cost of labor income lost
real(dp), parameter :: trans_11 = 0.01979519441340765740d0 !.02d0
real(dp), parameter :: trans_12 = 1.d0-trans_11
real(dp), parameter :: trans_13 = 0.d0
real(dp), parameter :: trans_21 = 0.00081778249954317654d0 !.001d0
real(dp), parameter :: trans_23 = 1.d0 - 0.00081778249954317654d0 - 0.78068033809607384477d0 !.219d0
real(dp), parameter :: trans_22 = 1.d0-trans_21-trans_23
real(dp), parameter :: trans_33 = 0.96397436291636229289d0
real(dp), parameter :: trans_32 = 1.d0-trans_33
real(dp), parameter :: trans_31 = 0.d0

real(dp), parameter :: phi_SS = 0.46972182750383228989d0 !.47d0
real(dp), dimension(1:len_z,1:len_z), parameter :: phi = phi_SS

!!!!!!!!!!!!!!!!!!!!!!!!
! Multigrid Parameters
! 
integer, parameter :: mg_steps = 1 !2
integer, dimension(1:mg_steps), parameter :: mg_n_1 = 6  !1 !2 !1 !(/2,4/)
integer, dimension(1:mg_steps), parameter :: mg_n_2 = 18  !4 !2 !4 !(/3,8/)
integer, dimension(1:mg_steps), parameter :: mg_n_3 = 40  !24 !3  !24 !(/4,24/)
integer, dimension(1:mg_steps), parameter :: mg_len_e = mg_n_1+mg_n_2+mg_n_3
integer, dimension(1:mg_steps), parameter :: mg_xi1_start = 1   !Gives starting index for xi==1
integer, dimension(1:mg_steps), parameter :: mg_xi1_end = mg_n_1 
integer, dimension(1:mg_steps), parameter :: mg_xi2_start = mg_n_1 + 1
integer, dimension(1:mg_steps), parameter :: mg_xi2_end = mg_n_1 + mg_n_2
integer, dimension(1:mg_steps), parameter :: mg_xi3_start = mg_n_1 + mg_n_2 + 1
integer, dimension(1:mg_steps), parameter :: mg_xi3_end = mg_len_e

!!!!!!!!!!!!!!!!!!!!!!!!
! Bankruptcy Parameters
!
real(dp), parameter :: death = -sqrt(huge(0.d0))  !Value of negative consumption
real(dp), parameter :: nonmonotone_relax = 10    !Allows the policy function to not be perfectly monotone.
                                                !Solution is always verified with standard VFI
real(dp), parameter :: sigma = 2.d0            !Coefficient of consumption risk aversion
real(dp), parameter :: alpha_k = .36d0              !Capital share
real(dp), parameter :: delta = .10d0                !Depreciation rate

real(dp), parameter :: lambda = .9d0          !Prob bad stays on record !1/(1-lamba) is average # periods spent in bankruptcy


!!!!!!!!!!!!!!!!!!!!!!!!
! Shock parameters
!
real(dp), parameter :: rho = .975d0                !Conditional probability of survival

real(dp), parameter :: emin_3_SS = 1d0/19.93674729116d0
real(dp), parameter :: emax_3_SS = 30.46468670327320893421d0/19.93674729116d0
real(dp), parameter :: emin_2_SS = 39.24604493098055257860d0/19.93674729116d0
real(dp), parameter :: emax_2_SS = 116.34074857650554690736d0/19.93674729116d0
real(dp), parameter :: emin_1_SS = 14335.69235206637677038088d0/19.93674729116d0
real(dp), parameter :: emax_1_SS = 21263.86711017461493611336d0/19.93674729116d0

real(dp), dimension(1:len_z,1:len_z), parameter :: &
    emax_3 = reshape((/1.266619d0,emax_3_SS,emax_3_SS,1.634849d0/),(/len_z,len_z/))
real(dp), dimension(1:len_z,1:len_z), parameter :: &
    emax_2 = reshape((/5.151411d0,emax_2_SS,emax_2_SS,6.466954d0/),(/len_z,len_z/))
real(dp), dimension(1:len_z,1:len_z), parameter :: &
    emax_1 = reshape((/1005.090923d0,emax_1_SS,emax_1_SS,1123.313219d0/),(/len_z,len_z/))
real(dp), dimension(1:len_z,1:len_z), parameter :: &
    emin_3 = reshape((/.172967d0,emin_3_SS,emin_3_SS,.000001d0/),(/len_z,len_z/))
real(dp), dimension(1:len_z,1:len_z), parameter :: &
    emin_2 = reshape((/2.289856d0,emin_2_SS,emin_2_SS,1.671917d0/),(/len_z,len_z/))
real(dp), dimension(1:len_z,1:len_z), parameter :: &
    emin_1 = reshape((/747.935167d0,emin_1_SS,emin_1_SS,692.403574d0/),(/len_z,len_z/))

integer :: n_1 = mg_n_1(mg_steps) !10 !2 !3 !100 !3 !  10 !2 !10 !1 !10 !20 !3 ! 1 !4 !12 
integer :: n_2 = mg_n_2(mg_steps) !30 !2 !16 !100 !16 ! 20 !2 !20 !4 !40 !44 !11 !3 !8 !44 
integer :: n_3 = mg_n_3(mg_steps) !70 !2 !40 !100 !40 !26 ! 40 !2 !40 !8 !80 !64 !16 !4 !20 !64 
integer :: xi1_beg = mg_xi1_start(mg_steps)
integer :: xi1_end = mg_xi1_end(mg_steps)
integer :: xi2_beg = mg_xi2_start(mg_steps)
integer :: xi2_end = mg_xi2_end(mg_steps)
integer :: xi3_beg = mg_xi3_start(mg_steps)
integer :: xi3_end = mg_xi3_end(mg_steps)
integer :: len_e = mg_len_e(mg_steps) !# of idiosyncratic shocks, should be mod 8 = 0

real(dp), parameter :: g = 1.02d0 !value of z in good state !cf KMS calibration
real(dp), parameter :: b = 0.98d0 !value of z in bad state
integer, parameter :: g_ind = 1 ! index of z grid associated with the good agg state
integer, parameter :: b_ind = 2 ! index of z grid associated with the bad agg state
real(dp), parameter :: trans_gg = .5d0 !prob of trans from g to g !avg business cycle duration ~ 2yrs
real(dp), parameter :: trans_bb = .5d0 !prob of trans from b to b
real(dp), dimension(1:len_z), parameter :: grid_z = (/g,b/)
real(dp), dimension(1:len_z,1:len_z), parameter :: trans_z = reshape((/trans_gg,1.d0-trans_bb,1.d0-trans_gg,trans_bb/),(/2,2/))

!!!!!!!!!!!!!!!!!!!!!!!!
! Targets
! 
real(dp), parameter :: target_r = 0.1169d0              !Target MPK
real(dp), parameter :: target_qbar = rho/(1.d0 + target_r - delta)
real(dp), parameter :: target_KN = (target_r/alpha_k)**(1.d0/(alpha_k-1.d0)) !Target capital labor ratio
real(dp), parameter :: target_KY = 1.d0/(target_r/alpha_k)   !Target capital output ratio
real(dp), parameter :: target_w = (1.d0-alpha_k)*target_KN**alpha_k !Target MPL
real(dp), parameter :: target_pop_in_debt = 0.036d0     !Target measuare of hhs in ddbt
real(dp), parameter :: target_debt_to_output = 0.0036d0 !Target debt/output ratio
real(dp), parameter :: target_pct_hhs_filing = 0.0029d0 !Target % of hhs filing for bankruptcy
real(dp), parameter :: target_pct_in_bnk = target_pct_hhs_filing/(1.d0-rho*lambda) !Fraction of households who will be in default
real(dp), parameter :: target_income_gini = .61d0
real(dp), parameter :: target_wealth_gini = .80d0

!!!!!!!!!!!!!!!!!!!!!!!!
! Grids 
!
real(dp), parameter :: max_a = 35000.d0*target_w  
real(dp), parameter :: min_a = -7.d0*target_w
real(dp), dimension(1:len_N) :: grid_N
real(dp), dimension(1:len_A1) :: grid_A1
real(dp), dimension(1:len_A2) :: grid_A2
real(dp), dimension(1:len_A3) :: grid_A3

!!!!!!!!!!!!!!!!!!!!!!!
! Loop controls, tolerances
integer :: max_iter_VFI = 100           !Maximum number of iterations to perform during VFI
integer :: print_mod = 10
real(dp), parameter :: toler = 1.d-6
real(dp), parameter :: toler_VFI = 1.d-6
!real(dp), parameter :: toler_newton = 1.d-6

end module mod_common_var