! --------------------------------------------------------------------------
! - This module is used to read and write variables from and into Matlab   -
! - .mat format data file, which require linked with                       -
! - libmat.lib, libmat.dll, libmx.lib, libmx.dll and libut.dll             -
! - All these libraries are supported by Mathworks.inc Matlab 13           -
! -                                                                        -
! - Copyright (c) 2003-2004 by Dr. Chen Xianyao                            -
! - Key Lab. of Marine Science and Numerical Modelling                     -
! - First Institute of Oceanography                                        -
! - E-mail: chenxy@fio.org.cn, xianyaochen@yahoo.com.cn                    -
! -                                                                        -
! - All Rights Reserved.                                                   -
! - Disclosure without explicit written consent from the copyright owner   -
! - does not constitute publication.                                       -
! -                                                                        -
! - Matlab is one of products of Mathworks inc.                            -
! --------------------------------------------------------------------------

!Preprocessor
!replaces mwPointer, mwSize by appropriate integer values.  Make sure <matlabroot>/extern/include is included when compiling
#include <fintrf.h>

module mod_matdata

use mod_types
use mexf90
implicit none

!integer, private, parameter :: mwPointer_kind = 8
mwPointer, private :: mp, pa
integer(kind=4), private :: sta

interface matwrite
  module procedure  Put_Int4_0, Put_Real8_0, Put_Char_0, &
                    Put_Int4_1, Put_Real8_1, Put_Char_1, &
                    Put_Int4_2, Put_Real8_2,             &
                    Put_Int4_3, Put_Real8_3,             &
                    Put_Int4_4, Put_Real8_4             
end interface

interface matread
  module procedure  Get_Int4_0, Get_Real8_0, Get_Char_0, &
                    Get_Int4_1, Get_Real8_1, Get_Char_1, &
                    Get_Int4_2, Get_Real8_2,             &
                    Get_Int4_3, Get_Real8_3,             &
                    Get_Int4_4, Get_Real8_4             
end interface

contains

! --- convert string into lowercase
function lowercase(s1) result(s2)
  character(len=*), intent(in) :: s1
  character(len=len(s1)) :: s2
  integer :: id,ich
  do id = 1,len(s1)
    ich = ichar(s1(id:id))
    if (ich >= 65 .and. ich<=90) then
      s2(id:id) = char(ich+32)
    else
      s2(id:id) = s1(id:id)    
    endif
  enddo
end function lowercase

! --- CheckFileName
function CheckFileName(filename) result(s)
  character(len=*), intent(in) :: filename
  character(len=len(filename)+4) :: s
  integer :: L,Ls
  L = len(filename)
  if (L-3 <=0) then
    Ls = L
  else
    Ls = L-3
  endif
  if (lowercase(filename(Ls:L)) /= '.mat') then
    s = filename//'.mat'
  else	  
    s = filename
  endif
end function CheckFileName

! --- CheckAction
function CheckAction(flnm,actionin) result(TrueAction)
  character(len=*), intent(in) :: flnm
  character(len=1), intent(in) :: actionin
  character(len=1) :: TrueAction
  logical :: ext
  
  TrueAction = lowercase(actionin)
  Inquire(file=flnm,exist=ext)
  if (TrueAction == 'a') then
    if (ext) then
      TrueAction = 'u'; return
    else
      TrueAction = 'w'; return
    endif
  endif
  if (TrueAction == 'r' .and. (.not. ext) ) then
    TrueAction = ''; return
  endif    
end function CheckAction

! --- mwPointer 
subroutine Put_Int4_0(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize :: m,n  
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  m = 1
  n = 1
  pa = mxCreateNumericMatrix(m,n,mxClassIDFromClassName('int32'),0)
  ! --- Copy data to variable
  call mxCopyInteger4toPtr(var,mxGetPr(pa),m*n)
  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)

end subroutine Put_Int4_0

! --- real(kind=8) 
subroutine Put_Real8_0(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize :: m,n
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  m = 1
  n = 1
  pa = mxCreateNumericMatrix(m,n,mxClassIDFromClassName('double'),0)
  ! --- Copy data to variable
  call mxCopyReal8toPtr(var,mxGetPr(pa),m*n)
  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Real8_0

! --- mwPointer, 1-dimensional array
subroutine Put_Int4_1(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize :: m,n
    
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  m = size(var)
  n = 1
  pa = mxCreateNumericMatrix(m,n,mxClassIDFromClassName('int32'),0) !I prefer column vector

  ! --- Copy data to variable
  call mxCopyInteger4toPtr(var,mxGetPr(pa),m*n)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Int4_1

! --- real(kind=8), 1-dimensional array
subroutine Put_Real8_1(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction
  
  mwSize :: m,n

  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  m = size(var)
  n = 1
  pa = mxCreateNumericMatrix(m,n,mxClassIDFromClassName('double'),0)

  ! --- Copy data to variable
  call mxCopyReal8toPtr(var,mxGetPr(pa),m*n)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif

  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Real8_1


! --- mwPointer, 2-dimensional array
subroutine Put_Int4_2(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:,:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize :: m,n
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  m = size(var,1)
  n = size(var,2)
  pa = mxCreateNumericMatrix(m,n,mxClassIDFromClassName('int32'),0)

  ! --- Copy data to variable
  call mxCopyInteger4toPtr(var,mxGetPr(pa),m*n)
  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Int4_2


! --- real(kind=8), 2-dimensional array
subroutine Put_Real8_2(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:,:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction
  
  mwSize :: m,n

  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)

  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)

  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif

  ! --- Create variable
  m = size(var,1)
  n = size(var,2)
  pa = mxCreateNumericMatrix(m,n,mxClassIDFromClassName('double'),0)

  ! --- Copy data to variable
  call mxCopyReal8toPtr(var,mxGetPr(pa),m*n)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Real8_2


! --- mwPointer, 3-dimensional array
subroutine Put_Int4_3(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:,:,:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction
  
  mwSize, dimension(1:3) :: varDims
  mwSize :: nDims, varSize

  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  nDims = 3
  varDims(1) = size(var,1)
  varDims(2) = size(var,2)
  varDims(3) = size(var,3)
  varSize = size(var)
  pa = mxCreateNumericArray(nDims,varDims,mxClassIDFromClassName('int32'),0)

  ! --- Copy data to variable
  call mxCopyInteger4toPtr(var,mxGetPr(pa),varSize)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif

  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Int4_3

! --- real(kind=8), 3-dimensional array
subroutine Put_Real8_3(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:,:,:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize, dimension(1:3) :: varDims
  mwSize :: nDims, varSize
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  nDims = 3
  varDims(1) = size(var,1)
  varDims(2) = size(var,2)
  varDims(3) = size(var,3)
  varSize = size(var)
  pa = mxCreateNumericArray(nDims,varDims,mxClassIDFromClassName('double'),0)

  ! --- Copy data to variable
  call mxCopyReal8toPtr(var,mxGetPr(pa),varSize)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Real8_3

! --- mwPointer, 3-dimensional array
subroutine Put_Int4_4(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:,:,:,:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction
  
  mwSize, dimension(1:4) :: varDims
  mwSize :: nDims, varSize

  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  nDims = 4
  varDims(1) = size(var,1)
  varDims(2) = size(var,2)
  varDims(3) = size(var,3)
  varDims(4) = size(var,4)
  varSize = size(var)
  pa = mxCreateNumericArray(nDims,varDims,mxClassIDFromClassName('int32'),0)

  ! --- Copy data to variable
  call mxCopyInteger4toPtr(var,mxGetPr(pa),varSize)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Int4_4

! --- real(kind=8), 3-dimensional array
subroutine Put_Real8_4(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:,:,:,:), intent(in)  :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize, dimension(1:4) :: varDims
  mwSize :: nDims, varSize
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  nDims = 4
  varDims(1) = size(var,1)
  varDims(2) = size(var,2)
  varDims(3) = size(var,3)
  varDims(4) = size(var,4)
  varSize = size(var)
  pa = mxCreateNumericArray(nDims,varDims,mxClassIDFromClassName('double'),0)

  ! --- Copy data to variable
  call mxCopyReal8toPtr(var,mxGetPr(pa),varSize)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Put_Real8_4

! ---------------------------------- 
! -         Put Character          -
! ----------------------------------

! --- Character 0
subroutine Put_Char_0(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  character(len=*), intent(in) :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Create variable
  pa = mxCreateString(var)
  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)  
end subroutine Put_Char_0


! --- Character 1
subroutine Put_Char_1(filename,varname,var,action)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  character(len=*), dimension(:), intent(in) :: var
  character(len=1), intent(in), optional :: action

  character(len=len(filename)+4) :: flnm
  character(len=1) :: actionin, TrueAction

  mwSize :: varSize  

  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  actionin = 'a'
  if (present(action)) actionin = action
  TrueAction = CheckAction(flnm,actionin)
  ! --- Open .mat file
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif

  ! --- Create character matrix from strings
  varSize = size(var)
  pa = mxCreateCharMatrixFromStrings(varSize,var)

  ! --- Put variable into .mat file with given name
  sta = matPutVariable(mp,varname,pa)
  if (sta /= 0) then
    write(stdout,'("Fail to write data ",a," to .mat file ",a)')varname,trim(flnm)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)  
end subroutine Put_Char_1


! --------------------------------------------------------------------------
! -                                                                        -
! -                          Read Subroutines                              -
! -                                                                        -
! --------------------------------------------------------------------------

! --- mwPointer 
subroutine Get_Int4_0(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToInteger4(mxGetPr(pa),var,1)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Int4_0

! --- real(kind=8) 
subroutine Get_Real8_0(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToReal8(mxGetPr(pa),var,1)
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Real8_0

! --- mwPointer 1-dimension
subroutine Get_Int4_1(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToInteger4(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Int4_1

! --- mwPointer 1-dimension
subroutine Get_Real8_1(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToReal8(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Real8_1

! --- integer(kind=4) 2-dimension
subroutine Get_Int4_2(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:,:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToInteger4(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Int4_2

! --- real(kind=8) 2-dimension
subroutine Get_Real8_2(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:,:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToReal8(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Real8_2

! --- mwPointer 3-dimension
subroutine Get_Int4_3(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:,:,:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToInteger4(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Int4_3

! --- REAL(kind=8) 3-dimension
subroutine Get_Real8_3(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:,:,:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToReal8(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Real8_3

! --- mwPointer 4-dimension
subroutine Get_Int4_4(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  integer(kind=4), dimension(:,:,:,:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToInteger4(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Int4_4

! --- real(kind=8) 4-dimension
subroutine Get_Real8_4(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  real(kind=8), dimension(:,:,:,:), intent(out)  :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    var = 0
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    ! --- Copy pointer to data variable
    call mxCopyPtrToReal8(mxGetPr(pa),var,size(var))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Real8_4

! --------------------------
! -     Get Character      -
! --------------------------
! --- character 0
subroutine Get_Char_0(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  character(len=*), intent(out)  :: var
 
  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction
  
  var = ''
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    return
  endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(flnm),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    !write(stdout,*)'pa',pa,mxGetN(pa)
    ! --- Copy pointer to data variable
    sta = mxGetString(pa,var,mxGetN(pa))
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Char_0

! - character 1
subroutine Get_Char_1(filename,varname,var)
  character(len=*), intent(in) :: filename
  character(len=*), intent(in) :: varname
  character(len=*), dimension(:), intent(out) :: var

  character(len=len(filename)+4) :: flnm
  character(len=1) :: TrueAction

  character(len=len(var)*size(var)) :: varget
  integer :: i,j,rows
  
  var = ''
  ! --- check file name with .mat extension
  flnm = CheckFileName(filename)
  ! --- check file I/O action
  TrueAction = CheckAction(flnm,'r')
  if (TrueAction /= 'r') then
    write(stdout,'("Non-exist .mat file ",a)')flnm
    return
 endif
  
  ! --- Open .mat file for read
  mp = matOpen(trim(filename),TrueAction)
  if (mp == 0) then
    write(stdout,'("Fail to open .mat file ",a)')trim(flnm)
    return
  endif
  ! --- Get variable 
  pa = matGetVariable(mp,varname)
  if (pa == 0) then
    write(stdout,'("Non-exist variable ",a)')varname
  else
    rows = size(var)
    ! --- Copy pointer to data variable
    sta = mxGetString(pa,varget,min(mxGetM(pa)*mxGetN(pa),len(var)*size(var)))
    do j = 1,len(var)
      do i = 1,rows
        var(i) = trim(var(i))//varget( (j-1)*rows+i : (j-1)*rows+i )
      enddo  
    enddo
  endif
  sta = matClose(mp)  
  if (sta /= 0) then    
    write(stdout,'("Fail to close .mat file ",a)')trim(flnm)
  endif
  ! --- DestroyArray, Clean Memory
  call mxDestroyArray(pa)
end subroutine Get_Char_1

end module mod_matdata


!Module containing f90 interfaces to called routines
! module mexf90
!     implicit none
!     interface
!     function mxGetPr(mp)
!         mwPointer :: mxGetPr,mp
!     end function mxGetPr
!     function mxGetM(mp)
!         mwPointer :: mxGetM,mp
!     end function mxGetM
!     function mxGetN(mp)
!         mwPointer :: mxGetN,mp
!     end function mxGetN
!     function matOpen(filename,mode)
!         mwPointer :: matOpen
!         character(len=*) :: filename,mode
!     end function matOpen
!     function matClose(mp)
!         integer(kind=4) :: matClose
!         mwPointer :: mp
!     end function matClose
!     function matPutVariable(mp,name,pa)
!         integer(kind=4) :: matPutVariable
!         mwPointer :: mp,pa
!         character(len=*) :: name
!     end function matPutVariable
!     function matGetVariable(mp,name)
!         mwPointer :: matGetVariable
!         mwPointer :: mp
!         character(len=*) :: name
!     end function matGetVariable
!     function mxCreateNumericArray(ndim,dims,classid,ComplexFlag)
!         mwPointer :: mxCreateNumericArray
!         integer(kind=4) :: classid,ComplexFlag
!         mwSize,intent(in) :: ndim
!         mwSize,dimension(1:ndim),intent(in) :: dims
!     end function mxCreateNumericArray
!     function mxCreateNumericMatrix(m,n,classid,ComplexFlag)
!         mwPointer :: mxCreateNumericMatrix
!         integer(kind=4) :: classid,ComplexFlag
!         mwSize,intent(in) :: m,n
!     end function mxCreateNumericMatrix
!     function mxClassIDFromClassName(classname)
!         integer(kind=4) :: mxClassIDFromClassName
!         character(len=*) :: classname
!     end function mxClassIDFromClassName
!     function mxGetString(mp,str,strlen)
!         integer(kind=4) :: mxGetString
!         mwPointer :: mp
!         character(len=*) :: str
!         mwSize :: strlen
!     end function mxGetString
!     function mxCreateString(str)
!         mwPointer :: mxCreateString
!         character(len=*) :: str
!     end function mxCreateString
!     function mxCreateCharMatrixFromStrings(m,str)
!         mwPointer :: mxCreateCharMatrixFromStrings
!         mwSize :: m
!         character(len=*) :: str(m)
!     end function mxCreateCharMatrixFromStrings
! 
!   end interface
! 
! end module mexf90
