% Uses Newton's method to minimize (however the routine will minimize or
% maximize) 
function [x0 f0 g0 H0] = newtonLineSearch_minN(f,x0)
    
    % Check inputs
    if (~isvector(x0)) 
        error('x0 is not a vector')
    end
    
    % Set default parameters
    maxiter = 1000;
    epsilon = 1d-10;
    delta = 1d-10;
    computeHist = 1;
    
    % Other data setup
    n = length(x0);
    x0 = x0(:);
    if (computeHist)
        x0hist = NaN(n,maxiter);
        f0hist = NaN(maxiter);
    end
    
    % Main routine
    for iter = 1:maxiter

        fprintf(1,'iter %i f(x0) %g \n',iter,f(x0));
%         fprintf(1,'x0 is \n');
%         disp(x0)

        % Numerically compute the gradient, hessian, and evaluate f at x0
        [g0 H0 f0] = numGradHess(f,x0);

        if (computeHist)
            x0hist(:,iter) = x0;
            f0hist(iter) = f0;
        end
        
        % Compute search direction
        d = -(H0\g0);
        
        % Minimize in the lambda direction
        fhat = @(lambda) f(x0+lambda*d);
        lambda = fminbnd(fhat,0,2);
        s = lambda*d;  
        lambda
        
        % Check for stop
        if (s'*s<delta*(1+x0'*x0))
            if (g0'*g0<epsilon*(1+abs(f0)));
                if (ones(1,n)*H0*ones(n,1)<0) 
                    disp('Converged to a maximum')
                else
                    disp('Converged to a minimum')
                end
            else
                disp('Converged but to a nonoptimal point -- gradient not nearly zero but step is small')
            end            
            break
        end        
        
        % Update
        x0 = x0 + s;
        
        if (iter==maxiter)
            fprintf(1,'Sequence did not converge. Norm of step is %g \n',s'*s);
        end
            
    end
    
    if (computeHist)
        plot3(x0hist(1,:),x0hist(2,:),f0hist)
    end
    
    
end