function noisyLM()

x = [-5;-5];
scale = 1d-0;
ndraws = 100;

f = @(x) testf(x);

lsqnonlin(f,x,[],[],optimset('Display','Iter'))

for i = 1:1000
    [x,stepsize,firstOrderOptimality] = noisyLMcore(x,f,scale,ndraws); 
    norm = sqrt(f(x)'*f(x));
    fprintf(1,'iter %i : step norm %f SSRS %f SSRS grad norm %f\n',i,stepsize,norm,firstOrderOptimality);
    if (stepsize<1d-8 && firstOrderOptimality<1d-8) 
        break
    end
end

f(x)
x


[X1 X2] = ndgrid(linspace(-10,10,100)',linspace(-10,10,100)');
Y = NaN(size(X1));
for i = 1:numel(X1)   
    Y(i) = f([X1(i);X2(i)])'*f([X1(i);X2(i)]);
end
surfc(X1,X2,Y)


f([0;0;0])

end

function [x,stepsize,firstOrderOptimality] = noisyLMcore(x,f,scale,ndraws)

nx = length(x);
fx = f(x);
SSRSx = fx'*fx;
nf = length(fx);

if (ndraws<min(nx,nf)) then
    warning('number of draws was less than min(nx,nf), cannot compute a solution');
    return
end

e = scale*randn(nx,ndraws);

Y = NaN(ndraws,nf);
for i = 1:ndraws
    fxtmp = f(x+e(:,i));
    Y(i,:) = fxtmp - fx;
end

X = e';
J = ((X'*X)\(X'*Y))';

gradS = J'*fx; % The gradient of the SSRS
firstOrderOptimality = gradS'*gradS; % Euclidean norm 

nL = max(ndraws,12);
if (nL==12) 
    gL = [0,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1d0,1d1,1d2,1d3,1d4];
else
    gL = [0;logspace(-6,4,nL-1)'];
%     gL
end

SSRSbest = 1d100;
for i = 1:nL
    
    lambda = gL(i);
    s = (J'*J + lambda*eye(nx))\(-J'*fx);
    tmpx = x + s;
    tmpf = f(x + s);
    tmpSSRS = tmpf'*tmpf;
    if (tmpSSRS<SSRSbest)
        xbest = tmpx;
        stepsize = sqrt(s'*s);
        SSRSbest = tmpSSRS;
    end
end

if (SSRSbest<SSRSx) 
    x = xbest;
else
    stepsize = 0;
end



end



function y = testf(x)

    y = sin(x) - (1:length(x))'.*(1-cos(x)) + sum(cos(x)) - length(x);% + floor(mod(exp(x)-1+.2,1)) + exp(.3*x)-1;
%     y = log(1+exp(y));
%     y = A*(1 + x + sin(x));
    
    

end