% [fp fpp] = numderiv(f,x0,varargin)
% Takes the numerical derivative of a function f at points x0. The user may
% provide a step size for the derivative as an optional input argument.
%
function [fp fpp] = numderiv(f,x0,varargin)

if (numel(varargin)==0)
%     h = sqrt(2*eps(1.d0));
    h = 1d-6;
else
    h = varargin{1};
end

f0mh = f(x0-h);
f0ph = f(x0+h);

fp = (f0ph-f0mh)./(2*h);

f0 = f(x0);

fpp = (f0mh - 2*f0 + f0ph)/(h^2);

end

function [] = testnumderiv()

f = @(x) log(x);
fp = @(x) 1./x;
fpp = @(x) -x.^(-2);

x = linspace(.10001,.2,1000)';

gh = [1d-4 1d-6 1d-8 sqrt(2*eps(1.d0))];
nh = numel(gh);
fptrue = fp(x);
fpptrue = fpp(x);
errp = NaN(numel(x),nh);
errpp = errp;
for ih=1:nh
    [fphat fpphat] = numderiv(f,x,gh(ih));
    errp(:,ih) = [fptrue-fphat]./fptrue*100;
    errpp(:,ih) = [fpptrue-fpphat]./fpptrue*100;
end

figure(1)
plot(x,errp) % Percent error 1st deriv
figure(2)
plot(x,errpp) % Percent error 2nd deriv

end