function [] = plot3alt(x,y,z)
% plot3alt
% Given vectors x,y,z of equal length (with z=f(x,y)), interpolates to
% create a meshgrid that can be used in any plot such as surf, mesh, etc.

nx = length(x);
minx = min(x);
maxx = max(x);

ny = length(y);
miny = min(y);
maxy = max(y);

nz = length(y);

%Check inputs
if (~isvector(x) || ~isvector(y) || ~isvector(z))
    error('All inputs must be vectors')
end
if (nx~=ny || nx~=nz)
    error('Dims don''t agree') 
end

xevenspace = linspace(minx,maxx,2*nx)';
yevenspace = linspace(miny,maxy,2*ny)';

[X Y] = meshgrid(xevenspace,yevenspace);

%For each point, find the closest (x,y) in Euclidean distance
Z = NaN(size(X));
for yind = 1:size(X,2)
    for xind=1:size(X,1)
        dist = sqrt((X(xind,yind)-x).^2 + (Y(xind,yind)-y).^2);
        [tmp loc] = min(dist);
        Z(xind,yind) = z(loc);
    end
end

surf(X,Y,Z)

end 