% Contains test routines for 1 dimensional root finding
function [] = testRoot1Dim()

clf
clc

%% Linear function
fprintf(1,'\n======== Linear function ===========\n');
f = @(x) 2+x; %root at x = -2
a = -1000;
b = 1000;
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% Quadratic with two roots
fprintf(1,'\n======== Quadratic function ===========\n');

f = @(x) -10 + 2*x + x.^2;
a = -6;
b = 2;
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% Exp(x)-2
fprintf(1,'\n======== Exponential function ===========\n');
f = @(x) exp(x) - 2;
a = -100;
b = 100;
ezplot(f,[a,b])
checkf(f,a,b)
pause()


%% log(x)-e
fprintf(1,'\n======== Logarithmic function ===========\n');
f = @(x) log(x) - exp(1);
a = .0000001;
b = 10000;
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% random spline 
fprintf(1,'\n======== Random spline ===========\n');
x = [-100,-28,-1,.01,.3,15,200]';
y = randn(size(x));
y(1) = abs(y(1));
y(end) = -abs(y(end)); % Ensure root exists
f = @(XX) spline(x,y,XX);
a = x(1);
b = x(end);
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% random lininterp
fprintf(1,'\n======== Random lin interp ===========\n');
x = [-100,-28,-1,.01,.3,15,200]';
y = randn(size(x));
y(1) = abs(y(1));
y(end) = -abs(y(end)); % Ensure root exists
f = @(XX) interp1(x,y,XX);
a = x(1);
b = x(end);
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% discontinuous function (inf or -inf) at 0
fprintf(1,'\n======== Discontinuous Inf function ===========\n');
f = @(x) 1/(x+1d-10);
a = -1000;
b = 1000;
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% discontinuous function at x=2
fprintf(1,'\n======== Discontinuous function ===========\n');
f = @(x) (x>2)*3 + (x<=2)*-5;
a = -1000;
b = 1000;
ezplot(f,[a,b])
checkf(f,a,b)
pause()

%% random function 
fprintf(1,'\n======== Random function ===========\n');
a = -1000;
b = 1000;
f = @(x) (x==a)*-1 + (x==b)*1 + (x~=a & x~=b).*randn(size(x));
ezplot(f,[a,b])
checkf(f,a,b)
pause()

end

function [] = checkf(f,a,b)
    fprintf(1,'\nBISECT\n');
    [x fevals] = bisect(f,a,b,0);
    fprintf(1,'fevals = %i \n', fevals);
    fprintf(1,'x = %f \n',x)
    fprintf(1,'Error is %f \n',f(x));

%     best = 'bisect';
%     fevals_old = fevals;
    
    fprintf(1,'\nBRENT\n');
    [x fevals] = brent(f,a,b,1);
    fprintf(1,'fevals = %i \n', fevals);
    fprintf(1,'x = %f \n',x)
    fprintf(1,'Error is %f \n',f(x));
    fprintf(1,'\nRIDDERS\n');
    [x fevals] = ridders(f,a,b,1);
    fprintf(1,'fevals = %i \n', fevals);
    fprintf(1,'x = %f \n',x)
    fprintf(1,'Error is %f \n',f(x));

    
    
end