function [] = test_pchip2()

% PCHIP  Piecewise Cubic Hermite Interpolating Polynomial.
%     PP = PCHIP(X,Y) provides the piecewise polynomial form of a certain
%     shape-preserving piecewise cubic Hermite interpolant, to the values
%     Y at the sites X, for later use with PPVAL and the spline utility UNMKPP.
%     X must be a vector.
%     If Y is a vector, then Y(j) is taken as the value to be matched at X(j), 
%     hence Y must be of the same length as X.
%     If Y is a matrix or ND array, then Y(:,...,:,j) is taken as the value to
%     be matched at X(j),  hence the last dimension of Y must equal length(X).
%     
%     YY = PCHIP(X,Y,XX) is the same as YY = PPVAL(PCHIP(X,Y),XX), thus
%     providing, in YY, the values of the interpolant at XX.

f = @(x,y) sqrt(x) - x.*y - 10*sqrt(y); %+ log(x+y)  + -(y-2).^2;

x = [0,.1,.5,1,2,10]';
y = [.0001,1,2,3,4,7,15,20]';

[X Y] = ndgrid(x,y);

V = f(X,Y);

surf(X,Y,V)


% Xi = rand(1000,1); Xi = x(1) + (x(end)-x(1))*Xi;
% Yi = rand(1000,1); Yi = y(1) + (y(end)-y(1))*Yi;
Xi = linspace(x(1),x(end),100)';
Yi = linspace(y(1),y(end),100)';
% [Xi Yi] = ndgrid(Xi,Yi);
Vtrue = f(Xi,Yi);


pp_y = pchip2_coef(x,y,V);
% Vhat = pchip2_eval(x,y,pp_y,Xi(:),Yi(:));
Vhat = pchip2_eval(x,y,pp_y,Xi,Yi);

[Vhat,Vtrue]

max(abs(Vhat-Vtrue)./abs(Vtrue))

x = [-2,-1,0]';
y = x;
[X Y] = ndgrid(x,y);
V = [-2,-1,0;-10,-5,-1;-1d50,-200,-50];
surf(X,Y,V)

pp_y = pchip2_coef(x,y,V);
pp_y{1}.coefs
pp_y{2}.coefs
pp_y{3}.coefs
Xi = linspace(x(1),x(end),30)';
Yi = linspace(y(1),y(end),30)';
[Xi Yi] = ndgrid(Xi,Yi);
Vhat = pchip2_eval(x,y,pp_y,Xi(:),Yi(:));

'max is',max(Vhat(:))
Vhat = reshape(Vhat,size(Xi));

size(Vhat)
plot3(Xi,Yi,Vhat,'.')
% figure(1),surf(Xi,Yi,reshape(Vhat,size(Xi)))
% figure(2),surf(Xi,Yi,Vtrue)
% figure(3),surf(Xi,Yi,reshape(Vhat,size(Xi))-Vtrue)

end

function [pp_y] = pchip2_coef(x,y,V)

    nx = size(V,1);
    ny = size(V,2);
    
    % Condiitonal on y, construct the spline for (x,V(.,iy))
    pp_y = cell(ny,1);
    for iy = 1:ny
        pp_y{iy} = pchip(x,V(:,iy));
    end
    
end

function [Vhat] = pchip2_eval(x,y,pp_y,Xi,Yi) 
    
    nx = length(x);
    ny = length(y);
    ni = length(Xi);
    
    
    % For each y, evaluate the spline interpolant at xi.
    Vint_of_y = NaN(ni,ny);
    for iy = 1:ny
        Vint_of_y(:,iy) = ppval(pp_y{iy},Xi);
    end

    % Now we V(xi,y) for y on the grid and all xi. 
    % Compute the spline for y,V(xi,.) and evaluate V(xi,yi)
    Vhat = NaN(size(Xi));
    for i = 1:ni
        Vhat(i) = pchip(y,Vint_of_y(i,:),Yi(i));
    end
end