% Vectorized bisection method
% There are different ways this can be done. I choose to evaluate the
% function each and every time rather than selecting. This makes it easier 
% on the writer of "f" because doesn't have to deal with varying dimensions. 
function [xstar,fevals] = vecbisect(f,a,b,pmod)

toler = 1d-6;
xtoler = 1d-6;
maxiter = 100;

ya = f(a);
yb = f(b);

% Preallocate
yx = NaN(size(a));
x = NaN(size(a));

% Set up convergence matrix: none have converged yet
conv= zeros(size(a));

xstar = NaN(size(a));
% Check whether should stop
test = abs(ya)<toler;
conv(test) = 1;
xstar(test) = a(test);
test = abs(yb)<toler;
conv(test) = 1;
xstar(test) = b(test);
test = sign(ya)==sign(yb);

notBracketed = test;
if (any(notBracketed))
    asave = a;
    bsave = b;
    yasave = ya;
    ybsave = yb;
    [ya,yb]
    
    warning('A root is not bracketed')
end 

% Main loop
for iter = 1:maxiter
    
    % Check the stopping criterion
    test = abs(b-a)<xtoler;
    xstar(test) = (b(test)+a(test))/2;
    conv = conv | test;
    
    % Midpoint
    x = (b+a)/2;
    yx = f(x);
    
    % Check other stopping criterion
    test = abs(yx)<toler;
    xstar(test) = x(test);
    conv = conv | test;
    
    % Replace a or b depending on sign of yx
    testSign = (sign(yx)==sign(ya));
    a(testSign) = x(testSign);
    ya(testSign) = yx(testSign);
    b(~testSign) = x(~testSign);
    yb(~testSign) = yx(~testSign);

    if (mod(iter,pmod)==0) 
        disp(sprintf('iter %i %% converged %f',iter,sum(conv)/numel(conv)))
    end
    if (all(conv)) 
        if (any(notBracketed)) 
            xstar(notBracketed & abs(yasave)<abs(ybsave)) = asave(notBracketed & abs(yasave)<abs(ybsave));
            xstar(notBracketed & abs(yasave)>=abs(ybsave)) = bsave(notBracketed & abs(yasave)>=abs(ybsave));
        end
        return
    end

end


end