function [T] = myreadtable(fname)

    % Use csvread to get the right shape. Don't use it in general because
    % when a field is missing it assigns 0. 
    tmp = csvread(fname,1,0);
    data = NaN(size(tmp));
        
    % Now, manually process the csv file a line at a time
    fid = fopen(fname,'r');
    varnames = csvline(fid,true);
    
    nrows = 0;
    for irow = 1:size(data,1)
        [vals,line] = csvline(fid,false);
        data(irow,1:length(vals)) = vals;
        
        % Error check
        for j = 1:length(vals)
            if (data(irow,j)~=tmp(irow,j)) 
                line
                data(irow,j)
                tmp(irow,j)
                warning('Error importing')
            end
        end
            
            
    end
    fclose(fid);

    % Check that data and tmp agree where not NaN
%     locs = find(~(abs(tmp - data)>1d-14 | tmp==0)); 
%     plot(100*(data(locs)./tmp(locs)-1))
% 
%     pcterr = NaN(size(tmp));
%     pcterr(tmp~=0) = abs(100*(data(tmp~=0)./tmp(tmp~=0)-1));
%     
%     [pcterr(pcterr>1d-6),data(pcterr>1d-6),tmp(pcterr>1d-6)]
%     
%     hist(100*(data(tmp~=0)./tmp(tmp~=0)-1))
    
%     assert(all(all(abs(tmp - data)<1d-14 | tmp==0)),'Error in importing')

    ncol = size(data,2);

    if (length(varnames)>ncol)
        % Need to drop the extra variable names
        varnames = {varnames{1:ncol}};        
    elseif (length(varnames)<ncol)
        % Need to add some extra variables names
        error('Have not programmed this yet');
    end
    
    T = array2table(data,'VariableNames',varnames);

end

function [out,line] = csvline(fid,isstr)

    line = fgetl(fid);
    
    if ~ischar(line)
        out = [];
        return
    end

    sep = line==',';
    sep = [0,find(sep),length(line)+1];
    
    for i = 1:length(sep)-1
        ta = sep(i)+1;
        tb = sep(i+1)-1;
        
        if (isstr)
            out{i} = line(ta:tb);
        else
            out(i) = str2double(line(ta:tb));
        end
        
    end    
    
end